      subroutine set_momenta(pin, mu_Q, m_t)
      implicit none
      include 'constants.f'
      include 'masses.f'
      include 'scale.f'
      double precision pin(20)
      double precision pout(mxpart,4)
      double precision mu_Q
      double precision m_t

      integer i,j

c      do i = 1, 20
c         print *,i,pin(i)
c      enddo

      do i = 1, 2
         pout(i,4) = -pin((i - 1) * 5 + 1)
         do j = 1, 3
            pout(i,j) = -pin((i - 1) * 5 + j + 1)
         enddo
      enddo
      do i = 3, 4
         pout(i,4) = pin((i - 1) * 5 + 1)
         do j = 1, 3
            pout(i,j) = pin((i - 1) * 5 + j + 1)
         enddo
      enddo

      scale = mu_Q
      mt = m_t

c      do i = 1, 4
c         do j = 1, 4
c            print *,'pout',i,j,pout(i,j)
c         enddo
c      enddo

c Check if setup needs to be called for every point !!!      
c      call setup
      call col_operators(pout)

      return
      end



       double precision function etarap(j,p)
      implicit none
C---returns the value of the pseudorapidity
      include 'constants.f'
      integer j
      double precision p(mxpart,4)
      etarap=dsqrt(p(j,1)**2+p(j,2)**2+p(j,3)**2)
      etarap=(etarap+p(j,3))/(etarap-p(j,3))
      if (etarap .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      etarap=100d0
      else
      etarap=0.5d0*dlog(etarap)
      endif
      return
      end

      double precision function aetarap(j,p)
      implicit none
C---returns the absolute value of the pseudorapidity
      include 'constants.f'
      integer j
      double precision p(mxpart,4)
      aetarap=dsqrt(p(j,1)**2+p(j,2)**2+p(j,3)**2)
      aetarap=(aetarap+p(j,3))/(aetarap-p(j,3))
      if (aetarap .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      aetarap=100d0
      else
      aetarap=0.5d0*abs(dlog(aetarap))
      endif
      return
      end
 
      double precision function yrap(j,p)
      implicit none
C---returns the value of the rapidity
      include 'constants.f'
      integer j
      double precision p(mxpart,4)
      yrap=(p(j,4)+p(j,3))/(p(j,4)-p(j,3))
      if (yrap .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      yrap=100d0
      else
      yrap=0.5d0*dlog(yrap)
      endif
      return
      end

      double precision function ayrap(j,p)
      implicit none
C---returns the absolute value of the rapidity
      include 'constants.f'
      integer j
      double precision p(mxpart,4)
      ayrap=(p(j,4)+p(j,3))/(p(j,4)-p(j,3))
      if (ayrap .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      ayrap=100d0
      else
      ayrap=0.5d0*dabs(dlog(ayrap))
      endif
      return
      end
 
      double precision function pt(j,p)
      implicit none
      include 'constants.f'
      integer j
      double precision p(mxpart,4)
c--- This is the formula for pt
      pt=dsqrt(p(j,1)**2+p(j,2)**2)
c--- This is the formula for Et 
c      pt=dsqrt(p(j,1)**2+p(j,2)**2)
c     . *p(j,4)/dsqrt(p(j,1)**2+p(j,2)**2+p(j,3)**2)
      return
      end

      double precision function pttwo(j,k,p)
      implicit none
      include 'constants.f'
      integer j,k
      double precision p(mxpart,4)
      pttwo=dsqrt((p(j,1)+p(k,1))**2+(p(j,2)+p(k,2))**2)
      return
      end

c--- this is the rapidity of pair j,k
      double precision function yraptwo(j,k,p)
      implicit none
      include 'constants.f'
      integer j,k
      double precision p(mxpart,4)
      yraptwo=(p(j,4)+p(k,4)+p(j,3)+p(k,3))
     .       /(p(j,4)+p(k,4)-p(j,3)-p(k,3))
      if (yraptwo .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      yraptwo=100d0
      else 
      yraptwo=0.5d0*dlog(yraptwo)
      endif
            
      return
      end

c--- this is the pseudo-rapidity
      double precision function etaraptwo(j,k,p)
      implicit none
      include 'constants.f'
      integer j,k
      double precision p(mxpart,4)
      
      etaraptwo=dsqrt((p(j,1)+p(k,1))**2+(p(j,2)+p(k,2))**2
     .               +(p(j,3)+p(k,3))**2)
      if (abs(etaraptwo)-abs(p(j,3)+p(k,3)) .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      etaraptwo=100d0
      else 
      etaraptwo=(etaraptwo+p(j,3)+p(k,3))
     .         /(etaraptwo-p(j,3)-p(k,3))
      etaraptwo=0.5d0*dlog(etaraptwo)
      endif
      
      return
      end


CC    NEW

c--- this is the rapidity of pair j,k,l,m
      double precision function yrapfour(j,k,l,m,p)
      implicit none
      include 'constants.f'
      integer j,k,l,m
      double precision p(mxpart,4)
      yrapfour=(p(j,4)+p(k,4)+p(l,4)+p(m,4)+p(j,3)+p(k,3)+p(l,3)+p(m,3))
     .        /(p(j,4)+p(k,4)+p(l,4)+p(m,4)-p(j,3)-p(k,3)-p(l,3)-p(m,3))
      if (yrapfour .lt. 1d-13) then
C-- set to 100 if this is very close to or less than zero
c-- rapidities of 100 will be rejected by any sensible cuts
      yrapfour=100d0
      else 
      yrapfour=0.5d0*dlog(yrapfour)
      endif
            
      return
      end





c      function from dot.f moved here !!!
      double precision function dot(p,i,j)
      implicit none
      include 'constants.f'
      integer i,j
      double precision p(mxpart,4)
      dot=p(i,4)*p(j,4)-p(i,1)*p(j,1)-p(i,2)*p(j,2)-p(i,3)*p(j,3)
      return
      end


c     function from dotem.f moved here !!!
      subroutine dotem(N,p,s)
      implicit none
      include 'constants.f'
      double precision p(mxpart,4),s(mxpart,mxpart)
      integer j,k,N
c---returns 2*piDpj for massless particles
      do j=1,N
      s(j,j)=0d0
      do k=j+1,N
      s(j,k)=2d0*
     & (p(j,4)*p(k,4)-p(j,1)*p(k,1)-p(j,2)*p(k,2)-p(j,3)*p(k,3))
      s(k,j)=s(j,k)
      enddo
      enddo
      return
      end

      
c     function from myli2.f moved here !!!

      FUNCTION myli2(x)                                                     
        implicit none                                                           
*      !! Dilogarithm for arguments x < = 1.0                             
        real*8 X,Y,T,S,A,PI3,PI6,ZERO,ONE,HALF,MALF,MONE,MTWO  
        real*8 C(0:18),H,ALFA,B0,B1,B2,LI2OLD                 
        real*8 myli2                                             
        integer  i                                                       
                                                                           
        DATA ZERO /0.0d0/, ONE /1.0d0/                               
        DATA HALF /0.5d0/, MALF /-0.5d0/                             
        DATA MONE /-1.0d0/, MTWO /-2.0d0/                            
        DATA PI3 /3.289868133696453d0/, PI6 /1.644934066848226d0/    
                                                                           
        DATA C( 0) / 0.4299669356081370d0/                              
        DATA C( 1) / 0.4097598753307711d0/                              
        DATA C( 2) /-0.0185884366501460d0/                              
        DATA C( 3) / 0.0014575108406227d0/                              
        DATA C( 4) /-0.0001430418444234d0/                              
        DATA C( 5) / 0.0000158841554188d0/                              
        DATA C( 6) /-0.0000019078495939d0/                              
        DATA C( 7) / 0.0000002419518085d0/                              
        DATA C( 8) /-0.0000000319334127d0/                              
        DATA C( 9) / 0.0000000043454506d0/                              
        DATA C(10) /-0.0000000006057848d0/                              
        DATA C(11) / 0.0000000000861210d0/                              
        DATA C(12) /-0.0000000000124433d0/                              
        DATA C(13) / 0.0000000000018226d0/                              
        DATA C(14) /-0.0000000000002701d0/                              
        DATA C(15) / 0.0000000000000404d0/                              
        DATA C(16) /-0.0000000000000061d0/                              
        DATA C(17) / 0.0000000000000009d0/                              
        DATA C(18) /-0.0000000000000001d0/                              
                                                                           
        if(x .gt. 1.00000000001d0) then                                    
          write(6,*)'problems in LI2'
          write(6,*)'x=',x
          x = sqrt(MONE)
c       SK: To avoid stop, set ME2 to nan !!!
c          stop                                               
        elseif(x .gt. 1.0d0) then                                          
          x = 1.d0                                                      
        endif                                                              
        IF(X .EQ. ONE) THEN                                                
         LI2OLD=PI6
         myli2=LI2OLD                                                       
         RETURN                                                            
        ELSE IF(X .EQ. MONE) THEN                                          
         LI2OLD=MALF*PI6
         myli2=LI2OLD                                                  
         RETURN                                                            
        END IF                                                             
        T=-X                                                               
        IF(T .LE. MTWO) THEN                                               
         Y=MONE/(ONE+T)                                                    
         S=ONE                                                             
         A=-PI3+HALF*(LOG(-T)**2-LOG(ONE+ONE/T)**2)                        
        ELSE IF(T .LT. MONE) THEN                                          
         Y=MONE-T                                                          
         S=MONE                                                            
         A=LOG(-T)                                                         
         A=-PI6+A*(A+LOG(ONE+ONE/T))                                       
        ELSE IF(T .LE. MALF) THEN                                          
         Y=(MONE-T)/T                                                      
         S=ONE                                                             
         A=LOG(-T)                                                         
         A=-PI6+A*(MALF*A+LOG(ONE+T))                                      
        ELSE IF(T .LT. ZERO) THEN                                          
         Y=-T/(ONE+T)                                                      
         S=MONE                                                            
         A=HALF*LOG(ONE+T)**2                                              
        ELSE IF(T .LE. ONE) THEN                                           
         Y=T                                                               
         S=ONE                                                             
         A=ZERO                                                            
        ELSE                                                               
         Y=ONE/T                                                           
         S=MONE                                                            
         A=PI6+HALF*LOG(T)**2                                              
        END IF                                                             
                                                                           
        H=Y+Y-ONE                                                          
        ALFA=H+H                                                           
        B1=ZERO                                                            
        B2=ZERO                                                            
        DO  I = 18,0,-1                                                    
          B0=C(I)+ALFA*B1-B2                                               
          B2=B1                                                            
          B1=B0                                                            
        ENDDO                                                              
        LI2OLD=-(S*(B0-H*B2)+A) 
         myli2=LI2OLD
        end     

c     function from myli3new.f used instead !!!
      function myLI3(x)
      implicit none
      double precision myLI3,xlog,x,PI,Z3
      PI=3.14159265358979312D0
      Z3=1.20205690315959429D0
      
      
      if (x.lt.-20) then ! asymptotic expansion, suggested by Mathematica
      xlog=dlog(-x)
      myLI3= -1d0/6*(xlog**3+Pi**2*xlog) +1/x +1d0/8/x**2 +1d0/27/x**3
      elseif (x.lt.0.5d0) then
      xlog=dlog(1d0-x)
      myLI3= -xlog -(3*xlog**2)/8. -(17*xlog**3)/216. -(5*xlog**4)/576
     .   -(7*xlog**5)/54000. +(7*xlog**6)/86400. +19*xlog**7/5556600
     .   -xlog**8/752640 -11*xlog**9/127008000 +11*xlog**10/435456000
      elseif (x.lt.1d0) then
      xlog=dlog(x)
      myLI3=Z3 +(Pi**2*xlog)/6 +(3d0/4-dlog(-xlog)/2)*xlog**2 
     .   -xlog**3/12-xlog**4/288 +xlog**6/86400 -xlog**8/10160640  
      elseif (x.eq.1d0) then
      myLI3=Z3
      else
      write(6,*)'wrong argument of Li3!!' 
      endif
      return
      end   

      
c     old version:
c       function myLI3(x)
c       implicit none
c       double precision myLI3,xlog,x,PI,Z3
c       PI=3.14159265358979312D0
c       Z3=1.20205690315959429D0
c  
c       if (x.lt.0.5d0) then
c       xlog=dlog(1d0-x)
c       myLI3=  -xlog -(3*xlog**2)/8.-(17*xlog**3)/216.-(5*xlog**4)/576
c     #   - (7*xlog**5)/54000. + (7*xlog**6)/86400. + 19*xlog**7/5556600
c     #   - xlog**8/752640-11*xlog**9/127008000+11*xlog**10/435456000
c       elseif (x.lt.1d0) then
c       xlog=dlog(x)
c       myLI3=Z3+(Pi**2*xlog)/6+(3d0/4-dlog(-xlog)/2)*xlog**2 
c     # -xlog**3/12-xlog**4/288+xlog**6/86400-xlog**8/10160640  
c       elseif (x.eq.1d0) then
c         myLI3=1.20205690315959429D0
c       else
c        write(6,*)'wrong argument of Li3!!' 
c       endif
c       return
c       end   


c Do not remove !!! actually used !!!
      subroutine callmyli3(x, result)
      implicit none
      double precision x
      double precision myLI3
      double precision result

      result = myLI3(x)

      return
      end

