#ifndef LOGGER_H
#define LOGGER_H

#include <iostream>
#include <string>

typedef enum {
    LOG_DEBUG_VERBOSE = 0,
    LOG_DEBUG_POINT,
    LOG_DEBUG_STEP,
    LOG_DEBUG,
    LOG_INFO,
    LOG_WARN,
    LOG_ERROR,
    LOG_FATAL
} LogLevel;

class Logger;

class Log {
public:
  Log();
  ~Log();

  static void setLogThreshold(LogLevel level);
  static LogLevel getLogThreshold();

private:
  static LogLevel threshold;
};

class Logger {
public:
  //	Logger()  {};
  Logger(const string & _name) : name(_name) {};
  ostream & operator<<(LogLevel level);
  void newLine(LogLevel level);
  
  inline void called() const {if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){cout << name << ": called" << endline_flush;}}
  inline void finished() const {if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){cout << name << ": finished" << endline_flush;}}
  inline void finished(double & value) const {if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){cout << name << ": finished - return" << setw(23) << setprecision(15) << value << endline_flush;}}

private:
  LogLevel level;
  string name;
};

class NoopStream : public std::ostream {
public:
  template<class T>
  NoopStream & operator << (const T & arg){ return *this; };

  virtual ~NoopStream(){};
};

#endif
