#ifndef PARTICLE_H
#define PARTICLE_H

class particle {
private:
  // Should actually be "protected", but are made "public" to keep exactly the particle access format of MATRIX_v2.0.0 in user-defined code !!! 
  /*
protected:
  double pT;
  double pT2;
  double ET;
  double ET2;
  double eta;
  double rapidity;
  double phi;
  double m;
  double m2;
  */
public:
  particle();
  ~particle();
  particle(fourvector momentum);
  particle(fourvector momentum, double M);
  particle(fourvector momentum, double M, double M2);

  particle operator + (const particle & p2) const;
  particle & operator += (const particle & fv);

  double cosphi();
  double sinphi();
  
  void calculate_all();
  void reset_all();
  void reset_all_but_m();

  friend double R2_eta(particle & fv1, particle & fv2);
  friend double R2_rapidity(particle & fv1, particle & fv2);
  friend double R2_cosheta_cosphi(particle & fv1, particle & fv2);
  friend double R2_coshrapidity_cosphi(particle & fv1, particle & fv2);

  friend double R2_eta(particle & p1, const fourvector & fv2);
  friend double R2_rapidity(particle & p1, const fourvector & fv2);
  friend double R2_cosheta_cosphi(particle & p1, const fourvector & fv2);
  friend double R2_coshrapidity_cosphi(particle & p1, const fourvector & fv2);

  friend double R_eta(particle & fv1, particle & fv2);
  friend double R_rapidity(particle & fv1, particle & fv2);

  friend double R_eta(particle & p1, const fourvector & fv2);
  friend double R_rapidity(particle & p1, const fourvector & fv2);

  friend double R2_eta(particle * fv1, particle * fv2);
  friend double R2_rapidity(particle * fv1, particle * fv2);
  friend double R2_cosheta_cosphi(particle * fv1, particle * fv2);
  friend double R2_coshrapidity_cosphi(particle * fv1, particle * fv2);

  friend double R_eta(particle * fv1, particle * fv2);
  friend double R_rapidity(particle * fv1, particle * fv2);

  friend bool lessBypT(particle & p1, particle & p2);
  friend bool greaterBypT(particle & p1, particle & p2);

  friend std::ostream & operator << (ostream & s, particle & p);

  fourvector momentum;
  
  double pT;
  double pT2;
  double ET;
  double ET2;
  double eta;
  double rapidity;
  double phi;
  double m;
  double m2;
  
  bool is_available_pT;
  bool is_available_pT2;
  bool is_available_ET;
  bool is_available_ET2;
  bool is_available_eta;
  bool is_available_rapidity;
  bool is_available_phi;
  bool is_available_m;
  bool is_available_m2;

  /*
  double pTXX();
  double pT2XX();
  double ETXX();
  double ET2XX();
  double etaXX();
  double rapidityXX();
  double phiXX();
  double mXX();
  double m2XX();
  */

  inline double pTXX(){
    if (!is_available_pT){
      pT = sqrt(pT2XX());
      is_available_pT = true;
    }
    return pT;
  }
  inline double pT2XX(){
    if (!is_available_pT2){
      pT2 = momentum.pT2();
      is_available_pT2 = true;
    }
    return pT2;
  }
  inline double ETXX(){
    if (!is_available_ET){
      ET = sqrt(ET2XX());
      is_available_ET = true;
    }
    return ET;
  }
  inline double ET2XX(){
    if (!is_available_ET2){
      ET2 = momentum.ET2();
      is_available_ET2 = true;
    }
    return ET2;
  }
  inline double mXX(){
    if (!is_available_m){
      m = sqrt(m2XX());
      is_available_m = true;
    }
    return m;
  }
  inline double m2XX(){
    if (!is_available_m2){
      m2 = momentum.m2();
      is_available_m2 = true;
    }
    return m2;
  }
  inline double etaXX(){
    if (pT2XX() == 0.){eta = 0.;}
    else {eta = momentum.eta();}
    is_available_eta = true;
    return eta;
  }
  inline double rapidityXX(){
    if (m2XX() == 0. && pT2XX() == 0.){rapidity = 0.;}
    else {rapidity = momentum.rapidity();}
    is_available_rapidity = true;
    return rapidity;
  }
  inline double phiXX(){
    if (pT2XX() == 0.){phi = 0.;} // better 0 than undefined for now
    else {phi = momentum.phi();}
    is_available_phi = true;
    return phi;
  }

};

bool lessBypT(particle & p1, particle & p2);
bool greaterBypT(particle & p1, particle & p2);

bool XlessBypT(particle * p1, particle * p2);
bool XgreaterBypT(particle * p1, particle * p2);

/*
struct particle_id{
  int identity;
  vector<int> content;
  particle *xparticle;
};
bool idgreaterBypT(particle_id & p1, particle_id & p2);
bool idgreaterByET(particle_id & p1, particle_id & p2);
*/

#endif
