#ifndef OBSERVABLESET_H
#define OBSERVABLESET_H

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <map>
#include <string>
#include <vector>

using namespace std;

#include "randommanager.hpp"
#include "logger.hpp"
#include "ftypes.hpp"
#include "fourvector.hpp"
#include "manager.TSV.hpp"
#include "particle.hpp"
#include "dddistribution.hpp"
#include "xdistribution.hpp"
#include "dipole.set.hpp"
#include "collinear.set.hpp"
#include "ioperator.set.hpp"
#include "phasespace.set.hpp"

class munich;
class amplitude_set;

class observable_set : public manager_TSV{
private:

public:
  observable_set();
  virtual ~observable_set();

  observable_set(inputparameter_set * isi, contribution_set * _csi, model_set * _msi, event_set * _esi, phasespace_set * _psi, user_defined * _user);
  observable_set(int _n_set_TSV);

  void calculate_optimization_modifier();
  void calculate_intermediate_result();
  void calculate_Xsection(long long i, double & Xsection, double & Xsection_delta, double & sum_weights, double & sum_weights2, double & temp_sum_weights, double & temp_sum_weights2);
  void initialization_runtime();
  void determine_runtime();

  // contains only TSV information !!!
  friend std::ostream & operator << (std::ostream &, const observable_set &);

  virtual void moments(){}
  virtual void calculate_dynamic_scale(int i_a){}
  virtual void calculate_dynamic_scale_RA(int i_a){}
  virtual void calculate_dynamic_scale_TSV(int i_a){}

  // observable/observable.initialization.cpp
  void initialization(inputparameter_set * isi, contribution_set * _csi, model_set * _msi, event_set * _esi, phasespace_set * _psi, user_defined * _user);
  void initialization_set_default();
  void initialization_input(inputparameter_set * isi);
  void initialization_after_input();

  //  observable/observable.initialization.run.cpp
  void initialization_complete();
  void initialization_result();
  void initialization_tree();
  void initialization_unit();
  double determine_unit_factor(string unit);
  void initialization_filename();
  void initialization_switches();
  // also here, or only in qTsubtraction_basic ???
  void initialization_qTcut();
  void initialization_integration_parameters();
  void initialization_integration();

  // observable/observable.initialization.scale.cpp
  void initialization_TSV();
  // only CV:
  void initialization_basic_CV();
  void initialization_CV();
  void initialization_scale_fact_CV();
  void initialization_scale_ren_CV();

  // observable/observable.initialization.pdf.cpp
  void initialization_LHAPDF();
  void initialization_pdf(vector<string> & input_pdf, vector<string> & list_pdf);
  void initialization_pdf_pair(vector<string> & list_pdf, vector<vector<string> > & list_pdf_pair);
  void determine_selection_content_pdf();
  void perform_selection_content_pdf();
  void perform_selection_content_pdf_collinear();
  void remove_empty_CA_collinear();
  void perform_selection_content_pdf_list();
  void determine_initial_channel();
  string output_list_combination_pdf();

  // observable/observable.pdf.cpp
  void calculate_pdf_LHAPDF_list_TSV();
  void calculate_pdf_LHAPDF_scale_list_TSV(int i_a, int i_v, int i_m);
  // only CV:
  void calculate_pdf_LHAPDF_list_CV();
  void calculate_pdf_LHAPDF_scale_list_CV(int sd, int ss);
  void calculate_pdf_LHAPDF_CA_collinear_CV();
  void calculate_pdf_LHAPDF_scale_CA_collinear_CV(double & mu_fact, vector<vector<vector<double> > > & all_pdf_factor);
  void calculate_pdf_LHAPDF_CV();
  void calculate_pdf_LHAPDF_RA_CV();
  void calculate_pdf_LHAPDF_scale_CV(double & mu_fact, vector<double> & pdf_factor);

  // observable/observable.distribution.cpp
  void initialization_distribution();
  void readin_file_distribution();
  void readin_file_dddistribution();
  xdistribution get_fake_distribution_from_dddistribution(dddistribution & dddist);
  void determine_extended_distribution();
  string output_distribution_list();
  void get_userinput_extravalue_from_readin(vector<string> & user_variable, vector<string> & user_variable_additional, vector<string> & user_value, vector<string> & readin);
  void determine_distribution_complete();
  void determine_single_distribution();
  void determine_double_distribution();
  void preparation_distribution_TSV();
  void determine_single_distribution_TSV();
  void determine_double_distribution_TSV();

  // observable/observable.output.cpp
  void header_integration(ofstream & out_integration);
  void output_check_value_scale();
  void output_check_running_alpha_S();
  void output_finalization_integration();
  void output_integrand_maximum_psp_LAB();
  void output_integrand_maximum_psp();
  void output_integrand_maximum();
  void perform_integration_step_complete();
  void output_zero_contribution_complete();
  void output_step_integration_TSV();
  void output_step_result_TSV();
  void output_step_moment_TSV();
  void output_step_distribution_TSV();
  void output_step_integration();
  void output_step_integration_weight_optimization();
  void output_step_result();
  void output_step_moment();
  void output_step_time();
  void output_step_execution();
  // Should be obsolete and removed !!!
  void output_step_errorplot();
  void output_distribution_complete();
  void output_distribution();
  void output_dddistribution();
  void output_distribution_CV();
  void output_dddistribution_CV();
  void output_distribution_TSV();
  void output_testpoint_input();
  void output_check_selection_psp();

  // observable/observable.output.testpoint.cpp
  void set_testpoint_psp();
  void set_testpoint_psp_dipoles();
  void set_testpoint_x_pdf();
  void set_testpoint_g_z_coll();
  void set_testpoint_scale_complete();
  void set_testpoint_scale_complete_RA();
  void set_testpoint_scale();
  void set_testpoint_scale_RA(int i_a);
  void set_testpoint_scale_TSV(int i_a);
  string output_momenta();
  string output_scales();
  string output_momentum_conservation_check(int x_a);
  string output_momenta_phasespace(int x_a);
  string output_testpoint_kinematics();
  string output_testpoint_integrand();

  // observable/observable.proceeding.cpp
  bool check_proceeding_in(vector<string> & readin);
  void perform_proceeding_in();
  void perform_proceeding_out();
  void perform_proceeding_check();

  // observable/observable.pspweight.cpp
  void determine_psp_weight_TSV();
  void determine_scale_complete();
  // only CV:
  void determine_scale();
  void determine_scale_CA();
  void determine_scale_RA(int i_a);
  void determine_integrand();
  void determine_integrand_VA();
  void determine_integrand_CA();
  void determine_integrand_RA();
  void determine_psp_weight();
  void determine_psp_weight_RA();

#include "../dipolesubtraction/observable.dipolesubtraction.hpp"
#include "../qTsubtraction/observable.qTsubtraction.hpp"


  // switches for observables and output

  int switch_result;
  int switch_distribution;
  int switch_moment;

  int switch_output_execution;
  int switch_output_integration;
  int switch_output_maxevent;
  int switch_output_comparison;
  int switch_output_gnuplot;
  int switch_output_proceeding;
  int switch_output_result;
  int switch_output_time;
  int switch_output_moment;
  int switch_output_distribution;

  int switch_output_cancellation_check;
  int switch_output_testpoint;
  int switch_output_testpoint_reference_only;
  int switch_output_cutinfo;
  int switch_output_check_selection;
 
  int switch_console_output_runtime;
  int switch_console_output_tau_0;
  int switch_console_output_techcut_RA;
  int switch_console_output_running_alpha_S;
  int switch_console_output_phasespace_issue;
  int switch_console_output_ME2_issue;

 
  // filenames
  
  string filename_result;
  string filename_time;
  vector<string> filename_moment;
  string filename_integration;
  string filename_maxevent;
  string filename_comparison;
  string filename_gnuplot;
  string filename_makegnuplot;
  string filename_proceeding;
  string filename_proceeding_2;
  string filename_execution;
  vector<string> filename_distribution;
  vector<string> filename_dddistribution;

  string filename_distribution_all_CV;

  string filename_integration_CV;
  vector<vector<string> > filename_distribution_CV;
  vector<vector<string> > filename_dddistribution_CV;

  vector<string> filename_integration_TSV;
  vector<string> filename_distribution_TSV;
  vector<string> filename_result_TSV;
  vector<string> filename_moment_TSV;


  // scale variation - CV and general ???

  int switch_CV;
  int variation_mu_fact_CV;
  int variation_mu_ren_CV;
  int n_scales_CV;
  int variation_factor_CV;
  int no_central_scale_CV;
  double central_scale_CV;

  double prefactor_reference;
  double prefactor_CV;

  int dynamic_scale;
  int dynamic_scale_CV;


  vector<double> mu_fact;
  vector<double> mu_ren;
  double alpha_S;

  vector<double> scale_ren_CV;
  vector<double> scale_fact_CV;
  vector<double> rel_scale_factor_CV;

  vector<vector<double> > mu_fact_CV;
  vector<vector<double> > mu_ren_CV;
  vector<vector<double> > alpha_S_CV;
  vector<vector<double> > rel_alpha_S_CV;

  vector<double> rel_scale_factor_ren_CV;
  vector<double> rel_scale_factor_fact_CV;

  vector<string> directory_name_scale_CV;

  int id_scales;

  int map_value_scale_fact;
  vector<int> map_value_scale_fact_CV;
  vector<double> value_mu_fact_central;
  vector<vector<double> > value_mu_fact_rel;
  vector<vector<double> > value_mu_fact;

  int map_value_scale_ren;
  vector<int> map_value_scale_ren_CV;
  vector<double> value_mu_ren_central;
  vector<vector<double> > value_mu_ren_rel;
  vector<vector<double> > value_mu_ren;
  vector<vector<double> > value_alpha_S;
  vector<vector<double> > value_factor_alpha_S;

  // should be removed when ...value.. is completely installed ???

  double mu_central;
  double var_mu_ren;
  double var_mu_fact;
  double var_alpha_S_reference;
  double var_rel_alpha_S;
  vector<double> var_mu_ren_CV;
  vector<double> var_mu_fact_CV;
  vector<double> var_alpha_S_CV;
  vector<double> var_rel_alpha_S_CV;


  // reference scale
  string switch_reference;

  
  // PDF organization
  
  vector<string> pdf_selection;
  vector<string> pdf_restriction;
  vector<string> pdf_exclusion;
  vector<string> pdf_disable;
  vector<vector<vector<int> > > allowed_all_pdf;
  // N_f_active should be avoided - functionality is covered by e.g. pdf_disable !!!
  int N_f_active;

  vector<double> pdf_factor;
  vector<vector<double> > pdf_factor_CV;
  
  vector<vector<vector<vector<vector<double> > > > > CX_value_pdf_factor;
  vector<vector<vector<vector<vector<int> > > > > CX_combination_pdf;

  vector<vector<bool> > active_pdf;
  vector<vector<vector<vector<double> > > > value_list_pdf_factor;
  vector<vector<vector<vector<int> > > > list_combination_pdf;
  vector<vector<int> > list_combination_pdf_emission;

  vector<vector<vector<vector<vector<double> > > > > value_pdf_factor;
  // splitting pdf_factor into PDFs from a and b and a possible charge factor:
  vector<vector<vector<vector<vector<vector<vector<double> > > > > > > value_pdf_factor_a;
  vector<vector<vector<vector<vector<vector<vector<double> > > > > > > value_pdf_factor_b;

  vector<vector<vector<vector<vector<double> > > > > value_list_pdf_factor_TSV;


  vector<vector<vector<vector<vector<double*> > > > > pointer_pdf_factor;
  // splitting pdf_factor into PDFs from a and b and a possible charge factor:
  vector<vector<vector<vector<vector<vector<vector<double*> > > > > > > pointer_pdf_factor_a;
  vector<vector<vector<vector<vector<vector<vector<double*> > > > > > > pointer_pdf_factor_b;

  /*
  // for debugging against Sherpa
  ///  vector<vector<vector<vector<vector<double> > > > > value_pdf_factor_1;
  ///  vector<vector<vector<vector<vector<double> > > > > value_pdf_factor_2;
  vector<vector<vector<vector<vector<vector<double> > > > > > value_pdf_factor_combination_1;
  vector<vector<vector<vector<vector<vector<double> > > > > > value_pdf_factor_combination_2;

  ///  vector<vector<vector<vector<vector<double*> > > > > pointer_pdf_factor_1;
  ///  vector<vector<vector<vector<vector<double*> > > > > pointer_pdf_factor_2;
  vector<vector<vector<vector<vector<vector<double*> > > > > > pointer_pdf_factor_combination_1;
  vector<vector<vector<vector<vector<vector<double*> > > > > > pointer_pdf_factor_combination_2;
  // end
  */
  vector<vector<vector<vector<vector<double*> > > > > pointer_list_pdf_factor_TSV;




  // integrand / factor with matrix elements and subtraction-related factors
  
  vector<double> value_ME2term;
  vector<vector<vector<double> > > value_ME2term_ren;
  vector<vector<vector<vector<double> > > > value_ME2term_fact;
  vector<vector<vector<vector<vector<double*> > > > > pointer_ME2term;

  double integrand;
  vector<double> integrand_CV;
  vector<vector<double> > integrand_qTcut_CV;

  vector<vector<double> > integrand_D;
  vector<vector<vector<double> > > integrand_D_CV;
  vector<vector<vector<vector<double> > > > integrand_D_qTcut_CV;

  //  vector<vector<vector<double> > > CX_value_integrand_TSV_qTcut;
  ///  vector<vector<vector<vector<double> > > > value_integrand_qTcut_TSV;
  vector<vector<vector<vector<vector<vector<double> > > > > > value_integrand_qTcut_TSV;
  vector<vector<vector<vector<vector<vector<double> > > > > > value_pc_integrand_qTcut_TSV;
  vector<vector<vector<vector<double> > > > value_ren_integrand_qTcut_TSV;
  vector<vector<vector<vector<double> > > > value_fact_integrand_qTcut_TSV;
  vector<vector<vector<vector<double> > > > value_fact_pc_integrand_qTcut_TSV;
  //  vector<vector<vector<double> > > integrand_TSV_qTcut;
  vector<vector<vector<vector<double> > > > integrand_qTcut_TSV;
  vector<vector<vector<vector<vector<vector<double> > > > > > ps_integrand_qTcut_TSV;

  vector<vector<vector<double> > > CX_value_integrand;
  vector<vector<vector<vector<double> > > > CX_value_integrand_D;
  vector<vector<double> > CX_sum_value_integrand;
  vector<vector<vector<double> > > CX_sum_value_integrand_D;
  vector<vector<vector<double> > > CX_value_integrand_qTcut;
  vector<vector<vector<vector<double> > > > CX_value_integrand_qTcut_D;

  vector<vector<vector<vector<vector<double> > > > > CX_value_integrand_RF;
  vector<vector<vector<vector<vector<vector<double> > > > > > CX_value_integrand_RF_D;
  vector<vector<vector<vector<double> > > > CX_sum_value_integrand_RF;
  vector<vector<vector<vector<vector<double> > > > > CX_sum_value_integrand_RF_D;
  vector<vector<vector<vector<vector<double> > > > > CX_value_integrand_RF_qTcut;
  vector<vector<vector<vector<vector<vector<double> > > > > > CX_value_integrand_RF_qTcut_D;

  double max_integrand;
  double sigma_normalization;
  double sigma_normalization_deviation;


  // cross section 

  double this_psp_weight;
  double this_psp_weight2;
  vector<vector<double> > this_psp_weight_CV;
  vector<vector<double> > this_psp_weight2_CV;
  double step_sum_weight;
  double step_sum_weight2;
  vector<vector<double> > step_sum_weight_CV;
  vector<vector<double> > step_sum_weight2_CV;
  double full_sum_weight;
  double full_sum_weight2;
  vector<vector<double> > full_sum_weight_CV;
  vector<vector<double> > full_sum_weight2_CV;

  vector<vector<vector<vector<vector<double> > > > > ps_integrand_TSV;
  vector<vector<vector<double> > > integrand_TSV;
  vector<vector<vector<double> > > sum_weight_TSV;
  vector<vector<vector<double> > > sum_weight2_TSV;
  vector<vector<vector<double> > > fullsum_weight_TSV;
  vector<vector<vector<double> > > fullsum_weight2_TSV;
  vector<vector<vector<vector<double> > > > sum_weight_qTcut_TSV;
  vector<vector<vector<vector<double> > > > sum_weight2_qTcut_TSV;
  vector<vector<vector<vector<double> > > > fullsum_weight_qTcut_TSV;
  vector<vector<vector<vector<double> > > > fullsum_weight2_qTcut_TSV;

  double Xsection;
  double Xsection_delta;
  vector<vector<double> > Xsection_CV;
  vector<vector<double> > Xsection_delta_CV;
  vector<vector<vector<vector<double> > > > Xsection_TSV;
  vector<vector<vector<vector<double> > > > Xsection_delta_TSV;


  // moments

  // reorganize completely:
  int n_moments;
  vector<int> moment_symm;
  vector<vector<double> > moment;
  vector<vector<vector<double> > > directed_moment;

  vector<double> this_psp_moment;
  vector<double> this_psp_moment2;
  vector<double> step_sum_moment;
  vector<double> step_sum_moment2;
  vector<double> full_sum_moment;
  vector<double> full_sum_moment2;

  vector<vector<vector<double> > > this_psp_moment_CV;
  vector<vector<vector<double> > > this_psp_moment2_CV;
  vector<vector<vector<double> > > step_sum_moment_CV;
  vector<vector<vector<double> > > step_sum_moment2_CV;
  vector<vector<vector<double> > > full_sum_moment_CV;
  vector<vector<vector<double> > > full_sum_moment2_CV;

  vector<vector<vector<vector<vector<vector<double> > > > > > ps_moment_TSV;
  vector<vector<vector<vector<double> > > > moment_TSV;
  vector<vector<vector<vector<double> > > > sum_moment_TSV;
  vector<vector<vector<vector<double> > > > sum_moment2_TSV;
  vector<vector<vector<vector<double> > > > fullsum_moment_TSV;
  vector<vector<vector<vector<double> > > > fullsum_moment2_TSV;
  vector<vector<vector<vector<vector<double> > > > > sum_moment_qTcut_TSV;
  vector<vector<vector<vector<vector<double> > > > > sum_moment2_qTcut_TSV;
  vector<vector<vector<vector<vector<double> > > > > fullsum_moment_qTcut_TSV;
  vector<vector<vector<vector<vector<double> > > > > fullsum_moment2_qTcut_TSV;

  vector<double> Xsection_moment;
  vector<double> Xsection_moment_delta;
  vector<vector<vector<double> > > Xsection_moment_CV;
  vector<vector<vector<double> > > Xsection_moment_delta_CV;
  vector<vector<vector<vector<vector<double> > > > > Xsection_moment_TSV;
  vector<vector<vector<vector<vector<double> > > > > Xsection_moment_delta_TSV;


  // distributions
  
  // rename at some point:
  vector<xdistribution> dat;
  vector<dddistribution> dddat;
  vector<xdistribution> extended_distribution;

  // not yet initialised !!!
  vector<vector<int> > bin;
  vector<vector<int> > bin_max;
  vector<vector<long long> > bin_counts;

  vector<double> fakeasymfactor; // can most likely be removed !!!
  vector<vector<double> > bin_weight;
  vector<vector<double> > bin_weight2;
  vector<vector<vector<double> > > bin_weight_CV;
  vector<vector<vector<double> > > bin_weight2_CV;

  vector<vector<vector<double> > > bin_count_TSV;
  vector<vector<vector<vector<vector<vector<double> > > > > > bin_weight_TSV;
  vector<vector<vector<vector<vector<vector<double> > > > > > bin_weight2_TSV;

  vector<vector<vector<double> > > change_weight_TSV;
  vector<vector<vector<double> > > change_weight2_TSV;
  vector<vector<vector<double> > > change_weight_2nd_TSV;
  vector<vector<vector<double> > > change_weight2_2nd_TSV;

  vector<vector<vector<vector<vector<double> > > > > change_qTcut_bin_weight_TSV;
  vector<vector<vector<vector<vector<double> > > > > change_qTcut_bin_weight2_TSV;
  vector<vector<vector<vector<vector<double> > > > > change_qTcut_bin_weight_2nd_TSV;
  vector<vector<vector<vector<vector<double> > > > > change_qTcut_bin_weight2_2nd_TSV;

  vector<int> no_min_qTcut_TSV;
  vector<int> no_max_qTcut_TSV;
  vector<int> no_min_qTcut_distribution_TSV;
  vector<int> no_max_qTcut_distribution_TSV;


  // optimization modifier
  
  int switch_optimization_modifier;
  string selection_optimization_modifier;
  double optimization_modifier_exponent;

  double optimization_modifier;


  // enforce integration/optimization step in low-efficiency integrations

  int switch_enforce_step_efficiency;
  int enforce_step_efficiency;
  int switch_enforce_step_time;
  int enforce_step_time;


  // CPU time related variables (could be shifted to munich ???)
  
  clock_t start;
  clock_t end;
  int h;
  int min;
  int sec;
  int time_counter;
  int sec_import;
  clock_t start_step;



  // so far only used for dipolesubtraction (could be generalized and unified with similar variables in qTsubtraction !!!
  // int QT_finalstate_massive_coloured;
  // int QT_finalstate_massive_charged;
  int massive_QCD;
  int massive_QEW;



  // (spin/colour/charge correlated) matrix elements: could eventually be moved to amplitude_set !!!

  double ME2;

  int check_vanishing_ME2_end;
  int flag_vanishing_ME2;
  int n_event_vanishing_ME2;


  // RA
  vector<double> RA_ME2_cc;
  vector<double> RA_ME2_sc;

  // VA
  double VA_b_ME2;
  double VA_V_ME2;
  double VA_X_ME2;
  double VA_I_ME2;
  vector<double> VA_X_ME2_CV;
  vector<vector<double> > VA_ME2_cf;

  // CA
  vector<vector<double> > CA_ME2_cf;
  vector<vector<double> > CA_ME2_cf_fi;

  // QT
  // used for mixed corrections in qTsubtraction
  double VA_V_ME2_QCD;
  double VA_X_ME2_QCD;
  double VA_V_ME2_QEW;
  double VA_X_ME2_QEW;
  
  

  // qTsubtraction

  // channel identifiers (so far only used in for qT subtraction; could be moved to contribution_set !!!)

  int order_alphas_born;
  int initial_channel;

  int initial_gg;
  int initial_aa;
  int initial_qqx;
  int initial_ga;

  int initial_pdf_gg;
  int initial_pdf_aa;
  int initial_pdf_qqx;
  int initial_pdf_gq;
  int initial_pdf_aq;
  int initial_pdf_ga;
  int initial_pdf_rest;

  int initial_pdf_diag;

  int initial_diag;
  int initial_diag_gg;
  int initial_diag_aa;
  int initial_diag_qqx;
  int initial_diag_ga;



  // uncategorized (should be reorganized/removed in parts !!!)
  
  double E_beam;
  double E_CMS;

  int n_ps;
  int n_pc;
  ///  vector<int> relation_pc_ps;
  vector<int> change_cut;
  vector<int> distribution_no_qTcut;
  vector<vector<int> > distribution_no_qTcut_phasespace;


  int n_pz;

  // presumably not used !!!
  int QCD_order;
  
  // should belong to model_set ???
  int N_f;
  
  string LHAPDFname;
  int LHAPDFsubset;


  // to correct alpha_e factors in external photon treatment !!!
  double rescaling_factor_alpha_e;


  // switch between amplitude providers: should be reorganized, e.g.: selection_amplitude = "OpenLoops"/"Recola" or alike; switch_OL/RCL could be set based on its value !!!
  string selection_amplitude;
  int switch_OL;
  int switch_RCL;

  // belongs actually to amplitude_set, but also affects subtraction terms !!!
  int switch_polenorm;


  // Used in observable.pdf.cpp (could be unified and moved to phasespace_set !!!)
  //  vector<vector<double> > xz_pdf;
  //  vector<vector<double> > xz_coll;
  //  vector<vector<double> > xz_factor;
  // similar content (used in calculate_pdf_LHAPDF_CA_collinear_TSV/CV)
  // -> psi->xz_pdf

 

  // pointer to other containers
  
  // should be redundant now !!!
  munich * xmunich;

  contribution_set * csi;
  user_defined * user;
  model_set * msi;
  event_set * esi;
  phasespace_set * psi;
  amplitude_set * asi;


  int int_end;

  int proceeding_counter;
  int proceeding_check_size;

  string path_to_main;


  int temp_n_step;


  // units used in calculation and in output for cross section and distribution results
  
  string unit_calculation;
  string unit_result;
  string unit_distribution;

  double unit_factor_calculation;
  double unit_factor2_calculation;
  double unit_factor_result;
  double unit_factor_distribution;


  // variables relevant for b-space resummation
  
  // Check how the resummation-related variables are connected !!!
  ///  int switch_resum; -> switch_resummation
  int switch_resummation;
  int dynamical_Qres;
  double QT_Qres;
  double QT_Qres_prefactor;
  int switch_dynamic_Qres;
  // different names in psi:
  //  double Qres;
  //  double Qres_prefactor;

};

#endif
