#ifndef MANAGERPSP_H
#define MANAGERPSP_H

#include <string>
#include <vector>

using namespace std;

class manager_psp{
private:
  int x_psp;
  int x_a;

  fourvector * p;
  double * s;
  double * sqrts;

  bool switch_one_particle;
  
  bool available_p;
  bool available_s;
  bool available_sqrts;

  int available_level;
  
  bool fail;
  bool required;
  
  vector<int> i_xb;
  
  bool switch_IS_FS;
  vector<int> i_xb_IS;
  vector<int> i_xb_FS;

  phasespace_set * psi;

public:
  manager_psp();
  manager_psp(int _x_psp, int _x_a, phasespace_set * _psi);
  virtual ~manager_psp();

  void initialization(int _x_psp, int _x_a, phasespace_set * _psi);
  void initialization_integration();

  void calculate_p();
  inline void register_p(){ available_p = true; }
  inline void reset_p(){ available_p = false; }
  inline bool get_available_p(){ return available_p; }
  inline void get_p(){ if (!available_p){ calculate_p(); } }
  inline fourvector use_p(){ if (!available_p){ calculate_p(); } return *p; }
  
  void calculate_s();
  inline void register_s(){ available_s = true; }
  inline void reset_s(){ available_s = false; }
  inline void get_s(){ if (!available_s){ calculate_s(); } }
  inline double use_s(){ if (!available_s){ calculate_s(); } return *s; }
  
  void calculate_sqrts();
  inline void register_sqrts(){ available_sqrts = true; }
  inline void reset_sqrts(){ available_sqrts = false; }
  inline void get_sqrts(){ if (!available_sqrts){ calculate_sqrts(); } }
  inline double use_sqrts(){ if (!available_sqrts){ calculate_sqrts(); } return *sqrts; }
  
  inline void reset(){ available_p = false; if (!switch_one_particle){ available_s = false; available_sqrts = false; } }

  string info();

};

#endif
