#ifndef MAPPINGDECAY_H
#define MAPPINGDECAY_H

#include <string>
#include <vector>

using namespace std;

class mapping_decay{
private:
  int x_d;
  int x_a;
  
  int out;
  int out_1;
  int out_2;

  vector<int> required_xbp;
  vector<int> required_xbs;
  vector<int> required_xbsqrts;

  int switch_pTthreshold;
  double pT2_threshold;
  double pT2_max;
  double costheta_pTcut;
  double exp_pTthreshold_cth;
  
  int switch_mapping_cth;
  double exp_cth;
  
  int switch_rotate_system;

public:
  mapping_decay();
  mapping_decay(int _x_t, int _x_a, int _out, int _out_1, int _out_2, phasespace_set * _psi);
  virtual ~mapping_decay();

  void initialization(int _x_d, int _x_a, string & content, phasespace_set * _psi);
  void initialization(int _x_d, int _x_a, int _out, int _out_1, int _out_2, phasespace_set * _psi);
  void initialization_integration();
  
  void calculate_c();
  void calculate_g();
  void calculate_inv_r();

  void errorhandling_c();
  void errorhandling_g();
  void errorhandling_inv_r();

  inline int get_x_d() const { return x_d; }

  inline vector<int> get_required_xbp() const { return required_xbp; }
  inline vector<int> get_required_xbs() const { return required_xbs; }
  inline vector<int> get_required_xbsqrts() const { return required_xbsqrts; }

  string info() const ;
  string output_extended_info();

  inline bool operator == (const mapping_decay & ch) const {
    if ((x_a == ch.x_a) && (x_d == ch.x_d) && (out == ch.out) && (out_1 == ch.out_1) && (out_2 == ch.out_2)){return true;}
    else {return false;}
  }

  string name;
  string MC_IS_name_d_costheta;
  string MC_IS_name_d_phi;

  double lambda_out_012;
  double sqrt_lambda_out_012;
  fourvector k10;

  double phi;
  double costheta;
 
  double * g_d;
  double g_d_costheta;

  double r1;
  double r2;
  double g_IS_r1;
  double g_IS_r2;

  randomvariable * random_psp_r1;
  randomvariable * random_psp_r2;

  fourvector * p_out;
  fourvector * p_out_1;
  fourvector * p_out_2;

  double * s_out;
  double * s_out_1;
  double * s_out_2;

  double * sqrts_out;

  phasespace_set * psi;
  bool available;
  bool fail;

  int n_fail_c;
  int n_fail_g;
  int n_fail_inv_r;
  
};

#endif
