#ifndef MAPPINGPROPAGATOR_H
#define MAPPINGPROPAGATOR_H

#include <string>
#include <vector>

using namespace std;

class mapping_smin;
class mapping_smax;

class mapping_propagator{
private:
  int x_p;
  int x_a;
  
  int m;

  int out;

  vector<int> required_xbp;
  vector<int> required_xbs;
  vector<int> required_xbsqrts;

  int x_smin;
  int x_smax;
  
  double Gamma;
  double mass;
  double mass2;


  // Breit-Wigner
  double arg1;
  double arg2;
  double delta_y;

  // vanishing width
  double nuexp;
  double mass2_mapping;
  double term_smin_mass2;
  double term_smax_mass2;

  int switch_propagator_type;
  int switch_nuexp;

  mapping_smin * smin;
  mapping_smax * smax;

public:
  mapping_propagator();
  mapping_propagator(int _x_p, int _x_a, int _m, int _out, int _x_smin, int _x_smax, phasespace_set * _psi);
  virtual ~mapping_propagator();

  void initialization(int _x_p, int _x_a, string & content, phasespace_set * _psi);
  void initialization(int _x_p, int _x_a, int _m, int _out, int _x_smin, int _x_smax, phasespace_set * _psi);
  void initialization_integration();

  void calculate_propagator_limits();

  void calculate_c();
  void calculate_g();
  void calculate_inv_r();

  void errorhandling_c();
  void errorhandling_g();
  void errorhandling_inv_r();

  void calculate_Breit_Wigner_limits();

  void calculate_Breit_Wigner_c();
  void calculate_Breit_Wigner_g();
  void calculate_Breit_Wigner_inv_r();

  void errorhandling_Breit_Wigner_c();
  void errorhandling_Breit_Wigner_g();
  void errorhandling_Breit_Wigner_inv_r();

  void calculate_vanishing_width_limits();

  void calculate_vanishing_width_c();
  void calculate_vanishing_width_g();
  void calculate_vanishing_width_inv_r();

  void errorhandling_vanishing_width_c();
  void errorhandling_vanishing_width_g();
  void errorhandling_vanishing_width_inv_r();

  inline int get_x_p() const { return x_p; }
  inline int get_x_smin() const { return x_smin; }
  inline int get_x_smax() const { return x_smax; }

  inline vector<int> get_required_xbp() const { return required_xbp; }
  inline vector<int> get_required_xbs() const { return required_xbs; }
  inline vector<int> get_required_xbsqrts() const { return required_xbsqrts; }

  string info() const ;
  string output_extended_info();

  inline bool operator == (const mapping_propagator & ch) const {
    if ((x_a == ch.x_a) && (x_p == ch.x_p) && (m == ch.m) && (out == ch.out) && (x_smin == ch.x_smin) && (x_smax == ch.x_smax)){return true;}
    else {return false;}
  }


  string name;
  string MC_IS_name_p;
  
  double * g_p;

  double r1;
  double g_IS_r1;

  randomvariable * random_psp_r1;

  double * smin_out;
  double * smax_out;

  double * s_out;
  double * sqrts_out;

  phasespace_set * psi;
  bool available;
  bool fail;

  int n_fail_c;
  int n_fail_g;
  int n_fail_inv_r;
  
};

#endif
