#ifndef MAPPINGTCHANNEL_H
#define MAPPINGTCHANNEL_H

#include <string>
#include <vector>

using namespace std;

class phasespace_set;

class mapping_tchannel{
private:
  int x_t;
  int x_a;
  
  int m;

  int out;
  int in_1;
  int in_2;
  int out_1;
  int out_2;
  int in_1_out_1;

  vector<int> required_xbp;
  vector<int> required_xbs;
  vector<int> required_xbsqrts;

  int switch_pTthreshold;
  
  // parameters for former 'tchannel_opt'
  double smin_out_out_1;
  double pTcut_out_1;
  int switch_opt_one_particle;
  double mass2_out_1;

public:
  mapping_tchannel();
  mapping_tchannel(int _x_t, int _x_a, int _m, int _out, int _in_1, int _in_2, int _out_1, int _out_2, phasespace_set * _psi);
  virtual ~mapping_tchannel();

  void initialization(int _x_t, int _x_a, string & content, phasespace_set * _psi);
  void initialization(int _x_t, int _x_a, int _m, int _out, int _in_1, int _in_2, int _out_1, int _out_2, phasespace_set * _psi);
  void initialization_integration();

  void calculate_propagator_limits();

  void calculate_c();
  void calculate_g();
  void calculate_inv_r();

  void errorhandling_c();
  void errorhandling_g();
  void errorhandling_inv_r();

  void calculate_propagator_c();
  void calculate_propagator_g();
  void calculate_propagator_inv_r();

  void errorhandling_propagator_c();
  void errorhandling_propagator_g();
  void errorhandling_propagator_inv_r();

  inline int get_x_t() const { return x_t; }

  inline vector<int> get_required_xbp() const { return required_xbp; }
  inline vector<int> get_required_xbs() const { return required_xbs; }
  inline vector<int> get_required_xbsqrts() const { return required_xbsqrts; }

  string info() const ;
  string output_extended_info();

  inline bool operator == (const mapping_tchannel & ch) const {
    if ((x_a == ch.x_a) && (x_t == ch.x_t) && (m == ch.m) && (out == ch.out) && (in_1 == ch.in_1) && (in_2 == ch.in_2) && (out_1 == ch.out_1) && (out_2 == ch.out_2)){return true;}
    else {return false;}
  }



  string name;
  string MC_IS_name_t_t;
  string MC_IS_name_t_phi;

  double term1;
  double term2in;
  double term2out;
  double term2;
  double term3;
  double tmin;
  double tmax;
  double tmax_cut;
  double smin_out;
  
  fourvector p10;
  fourvector p20;
  fourvector k10;
  fourvector k20;
  fourvector p1b;

   
  double phi;
  double costheta;

  double * g_t;
  double g_t_propagator;

  double r1;
  double r2;
  double g_IS_r1;
  double g_IS_r2;

  randomvariable * random_psp_r1;
  randomvariable * random_psp_r2;
  
  fourvector * p_out;
  fourvector * p_in_1;
  fourvector * p_in_2;
  fourvector * p_out_1;
  fourvector * p_out_2;
  fourvector * p_in_1_out_1;

  double * s_out;
  double * s_in_1;
  double * s_in_2;
  double * s_out_1;
  double * s_out_2;
  double * s_in_1_out_1;

  double * sqrts_out;

  double mass2_mapping;
  double nuexp;
  
  phasespace_set * psi;
  bool available;
  bool fail;

  int n_fail_c;
  int n_fail_g;
  int n_fail_inv_r;
  
};

#endif
