#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::calculate_ME2_born(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_born");
  logger.called();

  set_phasespacepoint();
  if (!csi->class_contribution_loopinduced){
    ol_evaluate_tree(1, P_OpenLoops, &(osi->value_ME2term[0]));
    osi->ME2 = osi->value_ME2term[0];
    if (!(osi->check_vanishing_ME2_end)){check_vanishing_ME2_born();}
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "ME2_OL  = " << osi->ME2 <<endline_flush; }
  }
  else {
    ol_setparameter_double(renscale, osi->var_mu_ren);
    ol_setparameter_double(fact_uv, one);
    ol_setparameter_double(fact_ir, one);
    ol_evaluate_loop2(1, P_OpenLoops, M2L2, &acc);
    osi->ME2 = M2L2[0];
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2L2[0] = " << M2L2[0] << endline_flush; }
    osi->value_ME2term[0] = M2L2[0];
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "ME2_OL  = " << osi->ME2 <<endline_flush; }
  }

  logger.finished();
}


void amplitude_OpenLoops_set::check_vanishing_ME2_born(){
  static Logger logger("amplitude_OpenLoops_set::check_vanishing_ME2_born");
  logger.called();

  if (osi->ME2 == 0.){osi->flag_vanishing_ME2 = 1;}
  else {
    osi->flag_vanishing_ME2 = 0;
    ol_evaluate_cc(1, P_OpenLoops, &M2L0, M2cc, &ewcc);
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2L0 = " << M2L0 << endline_flush; }
    for (int i_c = 0; i_c < n_cc; i_c++){
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2cc[" << setw(2) << i_c << "] = " << M2cc[i_c] << endline_flush; }
      if (abs(M2cc[i_c]) > 1.e12 * abs(M2L0)){
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2L0 = 0. due to numerical cancellations!" << endline_flush; }
	osi->flag_vanishing_ME2 = 1;
	break;
      }
    }
  }
 
  logger.finished();
}

#endif
