#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::calculate_ME2_RA_QCD(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_RA_QCD");
  logger.called();

  if (esi->cut_ps[0] > -1){
    set_phasespacepoint();
    if (!csi->class_contribution_loopinduced){
      ol_evaluate_tree(csi->dipole[0].process_id, P_OpenLoops, &osi->value_ME2term[0]);
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_R  = " << setw(23) << setprecision(15) << osi->value_ME2term[0] << endline_flush; }
      if (!(osi->check_vanishing_ME2_end)){
	osi->flag_vanishing_ME2 = 0;
	ol_evaluate_cc(1, P_OpenLoops, &M2tree, M2cc_check, &ewcc);
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2tree = " << M2tree << endline_flush; }
	for (int i_c = 0; i_c < n_cc_check; i_c++){
	  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2cc_check[" << setw(2) << i_c << "] = " << M2cc_check[i_c] << endline_flush; }
	  if (abs(M2cc_check[i_c]) > 1.e12 * abs(M2tree)){
	    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2tree = 0. due to numerical cancellations!" << endline_flush; }
	    osi->flag_vanishing_ME2 = 1;
	    break;
	  }
	}
      }
    }
    else {
      ol_setparameter_double(renscale, osi->var_mu_ren);
      ol_setparameter_double(fact_uv, one);
      ol_setparameter_double(fact_ir, one);
      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "csi->dipole[0].process_id = " << csi->dipole[0].process_id << endline_flush; }
      ol_evaluate_loop2(csi->dipole[0].process_id, P_OpenLoops, M2L2, &acc);
      osi->value_ME2term[0] = M2L2[0];
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_R  = " << setw(23) << setprecision(15) << osi->value_ME2term[0] << endline_flush; }
      // No "flag_vanishing_ME2" check for loop-induced processes (unstable ME'2 might be set to 0)!
    }
  }
  else {osi->value_ME2term[0] = 0.;}

  osi->calculate_dipole_QCD();

  logger.finished();
}


void amplitude_OpenLoops_set::calculate_dipole_Acc_QCD(int x_a){
  static Logger logger("amplitude_OpenLoops_set::calculate_dipole_Acc_QCD");
  logger.called();

  set_phasespacepoint_dipole(x_a);

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ for (int i = 0; i < n_momentum_dipole; i++){logger << LOG_DEBUG_VERBOSE << "P_OL[" << x_a << "][" << i << "] = " << P_OL[x_a][i] << endline_flush;} }
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "csi->dipole[" << x_a << "].process_id = " << csi->dipole[x_a].process_id << endline_flush; }

  if (csi->class_contribution_loopinduced){
    ol_evaluate_cc2(csi->dipole[x_a].process_id, P_OL[x_a], &M2tree, M2cc, &ewcc);
  }
  else {
    ol_evaluate_cc(csi->dipole[x_a].process_id, P_OL[x_a], &M2tree, M2cc, &ewcc);
  }
  osi->RA_ME2_cc[x_a] = M2cc[csi->dipole[x_a].no_BLHA_entry];

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    for (int i = 0; i < n_cc; i++){logger << LOG_DEBUG_POINT << "M2cc[" << i << "] = " << M2cc[i] << endline_flush;}
    logger << LOG_DEBUG_POINT << "OpenLoops:     RA_ME2_cc[" << x_a << "] = " << setw(23) << setprecision(15) << osi->RA_ME2_cc[x_a] << endline_flush;
  }
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "csi->dipole[" << x_a << "].no_BLHA_entry = " << csi->dipole[x_a].no_BLHA_entry << endline_flush; }
  
  logger.finished();
}


void amplitude_OpenLoops_set::calculate_dipole_Asc_QCD(int x_a, fourvector & Vtensor){
  static Logger logger("amplitude_OpenLoops_set::calculate_dipole_Asc_QCD");
  logger.called();

  set_phasespacepoint_dipole(x_a);
  P_sc[x_a][0] = Vtensor.x0();
  P_sc[x_a][1] = Vtensor.x1();
  P_sc[x_a][2] = Vtensor.x2();
  P_sc[x_a][3] = Vtensor.x3();

  if (csi->class_contribution_loopinduced){
    ol_evaluate_cc2(csi->dipole[x_a].process_id, P_OL[x_a], &M2tree, M2cc, &ewcc);
  }
  else {
    ol_evaluate_cc(csi->dipole[x_a].process_id, P_OL[x_a], &M2tree, M2cc, &ewcc);
  }
  osi->RA_ME2_cc[x_a] = M2cc[csi->dipole[x_a].no_BLHA_entry];

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:     RA_ME2_cc[" << x_a << "] = " << setw(23) << setprecision(15) << osi->RA_ME2_cc[x_a] << endline_flush; }

  if (csi->class_contribution_loopinduced){
    ol_evaluate_sc2(csi->dipole[x_a].process_id, P_OL[x_a], csi->dipole[x_a].no_A_emitter(), P_sc[x_a], M2sc);
  }
  else {
    ol_evaluate_sc(csi->dipole[x_a].process_id, P_OL[x_a], csi->dipole[x_a].no_A_emitter(), P_sc[x_a], M2sc);
  }
  osi->RA_ME2_sc[x_a] = M2sc[csi->dipole[x_a].no_A_spectator() - 1];
  //  osi->RA_ME2_sc[x_a] = 0.; // -> for OpenLoops-Recola validation, since spin-correlated amplitudes are defined differently

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    logger << LOG_DEBUG_POINT << "OpenLoops:     RA_ME2_sc[" << x_a << "] = " << setw(23) << setprecision(15) << osi->RA_ME2_sc[x_a] << endline_flush;
    logger << LOG_DEBUG_POINT << "Vtensor = " << Vtensor << "   M2sc[" << csi->dipole[x_a].no_A_spectator() - 1 << "] = " << osi->RA_ME2_sc[x_a] << endline_flush;
    for (int i = 0; i < n_sc; i++){logger << LOG_DEBUG_POINT << "M2sc[" << i << "] = " << M2sc[i] << endline_flush;}
  }

  logger.finished();
}

#endif
