#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::calculate_ME2_RA_QEW(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_RA_QEW");
  logger.called();

  if (esi->cut_ps[0] > -1){
    set_phasespacepoint();
    ol_evaluate_tree(csi->dipole[0].process_id, P_OpenLoops, &osi->value_ME2term[0]);
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:     R_ME2 = " << setw(23) << setprecision(15) << osi->value_ME2term[0] << endline_flush; }
  }
  else {osi->value_ME2term[0] = 0.;}

  osi->calculate_dipole_QEW();
    
  logger.finished();
}


void amplitude_OpenLoops_set::calculate_dipole_Acc_QEW(int x_a){
  static Logger logger("amplitude_OpenLoops_set::calculate_dipole_Acc_QEW");
  logger.called();

  set_phasespacepoint_dipole(x_a);
  ol_evaluate_tree(csi->dipole[x_a].process_id, P_OL[x_a], &osi->RA_ME2_cc[x_a]);

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:     RA_ME2_cc[" << x_a << "] = " << setw(23) << setprecision(15) << osi->RA_ME2_cc[x_a] << endline_flush; }

  logger.finished();
}


void amplitude_OpenLoops_set::calculate_dipole_Asc_QEW(int x_a, fourvector & Vtensor){
  static Logger logger("amplitude_OpenLoops_set::calculate_dipole_Asc_QEW");
  logger.called();

  set_phasespacepoint_dipole(x_a);
  P_sc[x_a][0] = Vtensor.x0();
  P_sc[x_a][1] = Vtensor.x1();
  P_sc[x_a][2] = Vtensor.x2();
  P_sc[x_a][3] = Vtensor.x3();

  ol_evaluate_tree(csi->dipole[x_a].process_id, P_OL[x_a], &osi->RA_ME2_cc[x_a]);
  ol_evaluate_sc(csi->dipole[x_a].process_id, P_OL[x_a], csi->dipole[x_a].no_A_emitter(), P_sc[x_a], M2sc);
  osi->RA_ME2_sc[x_a] = M2sc[0];

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    logger << LOG_DEBUG_POINT << "OpenLoops:     RA_ME2_cc[" << x_a << "] = " << setw(23) << setprecision(15) << osi->RA_ME2_cc[x_a] << endline_flush; 
    logger << LOG_DEBUG_POINT << "OpenLoops:     RA_ME2_sc[" << x_a << "] = " << setw(23) << setprecision(15) << osi->RA_ME2_sc[x_a] << endline_flush;
  }

  logger.finished();
}

#endif
