#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::set_mu_and_delta_VA(){
  static Logger logger("amplitude_OpenLoops_set::set_mu_and_delta_VA");
  logger.called();

  ol_setparameter_double(pole_uv, osi->VA_DeltaUV);
  ol_setparameter_double(pole_ir1, osi->VA_DeltaIR1);
  ol_setparameter_double(pole_ir2, osi->VA_DeltaIR2);
  ol_setparameter_double(OL_mu_ren, osi->var_mu_ren);
  ol_setparameter_double(OL_mu_reg, mu_reg);
  ol_setparameter_double(fact_uv, one);
  ol_setparameter_double(fact_ir, one);

  logger.finished();
}


void amplitude_OpenLoops_set::set_parameters_for_changed_mu_ren_with_fixed_mu_reg(double & new_mu_ren){
  static Logger logger("amplitude_OpenLoops_set::set_parameters_for_changed_mu_ren_with_fixed_mu_reg");
  logger.called();

  double inv_factor = new_mu_ren / mu_reg;
  ol_setparameter_double(OL_mu_ren, new_mu_ren);
  ol_setparameter_double(OL_mu_reg, new_mu_ren);
  ol_setparameter_double(fact_uv, inv_factor);
  ol_setparameter_double(fact_ir, inv_factor);

  logger.finished();
}


void amplitude_OpenLoops_set::set_testpoint_alpha_S(const double alpha_S){
  static Logger logger("amplitude_OpenLoops_set::set_testpoint_alpha_S");
  logger.called();

  //  ol_setparameter_double(OL_alpha_s, alpha_S);

  logger.finished();
}


void amplitude_OpenLoops_set::set_testpoint_rescale_I_ME2(){
  static Logger logger("amplitude_OpenLoops_set::set_testpoint_rescale_I_ME2");
  logger.called();


  logger.finished();
}


void amplitude_OpenLoops_set::calculate_ME2_ioperator_VA_QCD(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_ioperator_VA_QCD");
  logger.called();

  set_phasespacepoint();

  if (!csi->class_contribution_loopinduced){
    ol_evaluate_tree(process_id, P_OpenLoops, &osi->VA_b_ME2);
    ol_evaluate_cc(process_id, P_OpenLoops, &osi->VA_b_ME2, M2cc, &ewcc);
    for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
      for (int j_a = 0; j_a < (*(osi->VA_ioperator))[i_a].size(); j_a++){
	osi->VA_ME2_cf[i_a][j_a] = M2cc[(*(osi->VA_ioperator))[i_a][j_a].no_BLHA_entry];
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  VA_ME2_cf[" << i_a << "][" << j_a << "] = " << setw(23) << setprecision(15) << osi->VA_ME2_cf[i_a][j_a] << endline_flush; }
      }
    }
  }
  else {
    ol_evaluate_cc2(process_id, P_OpenLoops, &osi->VA_b_ME2, M2cc, &ewcc);
    for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
      for (int j_a = 0; j_a < (*(osi->VA_ioperator))[i_a].size(); j_a++){
	osi->VA_ME2_cf[i_a][j_a] = M2cc[(*(osi->VA_ioperator))[i_a][j_a].no_BLHA_entry];
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  VA_ME2_cf[" << i_a << "][" << j_a << "] = " << setw(23) << setprecision(15) << osi->VA_ME2_cf[i_a][j_a] << endline_flush; }
      }
    }
  }

  osi->calculate_ioperator_QCD();

  for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
    osi->I_ME2_emitter[i_a] = accumulate(osi->VA_I_ME2_cf[i_a].begin(), osi->VA_I_ME2_cf[i_a].end(), 0.);
  }
  osi->VA_I_ME2 = accumulate(osi->I_ME2_emitter.begin(), osi->I_ME2_emitter.end(), 0.);

  logger.finished();
}


void amplitude_OpenLoops_set::calculate_ME2_VA_QCD(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_VA_QCD");
  logger.called();

  if (osi->switch_VI == 0 || osi->switch_VI == 1){

    set_phasespacepoint();
    if (!csi->class_contribution_loopinduced){
      set_mu_reg();
      set_mu_and_delta_VA();

      ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
      osi->VA_V_ME2 = M2L1[0];
      ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2);

      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
	logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_B  = " << setw(23) << setprecision(15) << osi->VA_b_ME2 << endline_flush;
	logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << endline_flush;
      }

      if (osi->switch_output_comparison){
	int R2_on = 0;
	ol_setparameter_int(stch("r2_on"), R2_on);
	double V_ME2_D4 = 0.;
	ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
	V_ME2_D4 = M2L1[0];
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_D4 = " << setw(23) << setprecision(15) << V_ME2_D4 << endline_flush; }
	double V_ME2_R2 = osi->VA_V_ME2 - V_ME2_D4;
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_R2 = " << setw(23) << setprecision(15) << V_ME2_R2 << endline_flush; }
	R2_on = 1;
	ol_setparameter_int(stch("r2_on"), R2_on);
      }

      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
	logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_V  = " << setw(23) << setprecision(15) << osi->VA_V_ME2 << endline_flush;
	logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_CT = " << setw(23) << setprecision(15) << osi->VA_X_ME2 << endline_flush;
	logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << endline_flush;
      }
      
      if (osi->switch_VI == 0){
	if (osi->VA_DeltaIR1 == 0. && osi->VA_DeltaIR2 == 0.){osi->VA_I_ME2 = IRL1[0];}
	else {osi->VA_I_ME2 = IRL1[0] + osi->VA_DeltaIR1 * IRL1[1] + osi->VA_DeltaIR2 * IRL1[2];}
      }
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_I  = " << setw(23) << setprecision(15) << osi->VA_I_ME2 << endline_flush; }

      if (osi->switch_CV){
	for (int s = 0; s < osi->n_scales_CV; s++){
	  double inv_factor_CV = osi->var_mu_ren_CV[s] / mu_reg;
	  ol_setparameter_double(OL_mu_ren, osi->var_mu_ren_CV[s]);
	  ol_setparameter_double(OL_mu_reg, osi->var_mu_ren_CV[s]);
	  ol_setparameter_double(fact_uv, inv_factor_CV);
	  ol_setparameter_double(fact_ir, inv_factor_CV);
	  ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2_CV[s]);
	  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  VA_X_ME2_CV[" << s << "] = " << setw(23) << setprecision(15) << osi->VA_X_ME2_CV[s] + osi->VA_V_ME2 << endline_flush; }
	}
      }
      if (osi->switch_TSV){
	for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
	  for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
	    double inv_factor_TSV = osi->value_scale_ren[0][i_v][i_r] / mu_reg;
	    ol_setparameter_double(OL_mu_ren, osi->value_scale_ren[0][i_v][i_r]);
	    ol_setparameter_double(OL_mu_reg, osi->value_scale_ren[0][i_v][i_r]);
	    ol_setparameter_double(fact_uv, inv_factor_TSV);
	    ol_setparameter_double(fact_ir, inv_factor_TSV);
	    ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->value_ME2term_ren[0][i_v][i_r]);
	    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  VA_VX_ME2_TSV[" << i_v << "][" << i_r << "] = " << osi->value_ME2term_ren[0][i_v][i_r] + osi->VA_V_ME2 << endline_flush; }
	    osi->value_ME2term_ren[0][i_v][i_r] += osi->VA_V_ME2 + osi->VA_I_ME2;
	  }
	}
      }
    }
    else {
      // Assumes that  calculate_H1gg_QCD_2loop()  delivers 2-loop amplitude  VA_V_ME2  at mu = s^ = mu_Q (switch_off_H1gg = 1  sets the finite part of that amplitude to 0) !!!
      // would changed default for mu_reg (result must be independent) -> check !!!
      //      set_mu_reg();
      //      set_mu_and_delta_VA();
      /*
      ol_setparameter_double(pole_uv, osi->VA_DeltaUV);
      ol_setparameter_double(pole_ir1, osi->VA_DeltaIR1);
      ol_setparameter_double(pole_ir2, osi->VA_DeltaIR2);
      double mu_Q = (esi->p_parton[0][1] + esi->p_parton[0][2]).m();
      ol_setparameter_double(OL_mu_ren, mu_Q);
      ol_setparameter_double(OL_mu_reg, mu_Q);
      ol_setparameter_double(fact_uv, one);
      ol_setparameter_double(fact_ir, one);
      */
      set_delta_and_mu_Q();

      osi->VA_b_ME2 = 0.;
      ol_evaluate_loop2(process_id, P_OpenLoops, M2L2, &acc);

      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "M2L2[0] = " << M2L2[0] << endline_flush; }

      osi->VA_b_ME2 = M2L2[0];
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_L2I = " << setw(23) << setprecision(15) << osi->VA_b_ME2 << endline_flush; }

      if (switch_off_H1gg){
	osi->QT_A0 = 0.;
	osi->QT_A1 = 0.;
	osi->VA_V_ME2 = 0.;
      }
      else {
	calculate_H1gg_QCD_2loop();
      }

      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "osi->VA_V_ME2 = " << osi->VA_V_ME2 << endline_flush; }
      // temporary solution because mu_reg did not exist as a standard parameter -> use mu_reg = mu_Q = s^ in I-operator !!!
      double save_osi_var_mu_ren = osi->var_mu_ren;
      osi->var_mu_ren = mu_Q;
      calculate_ME2_ioperator_VA_QCD();
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_L2II  = " << setw(23) << setprecision(15) << osi->VA_I_ME2 << "   (corrected for var_rel_alpha_S)" << endline_flush; }
      osi->var_mu_ren = save_osi_var_mu_ren;

      // beta0 usually not initialized in CS subtraction !!!
      static double beta0 = (33. - 2 * osi->N_f) / 12;
      // scale variation (from mu_ren = s^ = mu_Q to the usual scales...) !!!
      osi->VA_X_ME2 = -csi->order_alpha_s_born * beta0 * log(pow(mu_Q / osi->var_mu_ren, 2)) * osi->VA_b_ME2 * (osi->alpha_S / pi);
      if (osi->switch_CV){
	for (int i_s = 0; i_s < osi->n_scales_CV; i_s++){
	  osi->VA_X_ME2_CV[i_s] = -csi->order_alpha_s_born * beta0 * log(pow(mu_Q / osi->var_mu_ren_CV[i_s], 2)) * osi->VA_b_ME2 * (osi->alpha_S / pi);
	}
      }

      if (osi->switch_TSV){
	for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
	  for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
	    osi->value_ME2term_ren[0][i_v][i_r] = osi->VA_V_ME2 + osi->VA_I_ME2 - csi->order_alpha_s_born * beta0 * log(pow(mu_Q / osi->value_scale_ren[0][i_v][i_r], 2)) * osi->VA_b_ME2 * (osi->alpha_S / pi);
	  }
	}
      }

      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "   V = " << setprecision(15) << setw(23) << osi->VA_V_ME2 << "   X = " << setprecision(15) << setw(23) << osi->VA_X_ME2 << "   I = " << setprecision(15) << setw(23) << osi->VA_I_ME2 << endline_flush; }
    }
  }
  else if (osi->switch_VI == 2){
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "OL: osi->VA_I_ME2 = " << osi->VA_I_ME2 << endline_flush; }
    //  osi->VA_I_ME2 = 0.;
    //  osi->VA_V_ME2 = 0.;
    //  osi->VA_X_ME2 = 0.;
    //  for (int s = 0; s < osi->n_scales_CV; s++){osi->VA_X_ME2_CV[s] = 0.;}
    //  I-operator evaluation -> osi->VA_I_ME2
    calculate_ME2_ioperator_VA_QCD();
    for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
      for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
        osi->value_ME2term_ren[0][i_v][i_r] = osi->VA_I_ME2;
      }
    }
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "SK: osi->VA_I_ME2 = " << osi->VA_I_ME2 << endline_flush; }
  }
  
  if (osi->VA_V_ME2 == 0. && osi->switch_VI != 2){ failed_evaluation = 1; }
  if (munich_isnan(osi->VA_V_ME2) || munich_isnan(osi->VA_X_ME2) || munich_isnan(osi->VA_I_ME2)){ failed_evaluation = 1; }
  
  logger.finished();
}

#endif
