#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::calculate_ME2_ioperator_VA_QEW(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_ioperator_VA_QEW");
  logger.called();

  static int initialization = 1;
  static vector<double> I_ME2_emitter(osi->VA_ioperator->size());
  if (initialization == 1){
    for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
      osi->VA_ME2_cf[i_a].resize((*(osi->VA_ioperator))[i_a].size());
      osi->VA_I_ME2_cf[i_a].resize((*(osi->VA_ioperator))[i_a].size());
    }
    initialization = 0;
  }

  set_phasespacepoint();
  ol_evaluate_tree(process_id, P_OpenLoops, &osi->VA_b_ME2);

  for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
    for (int j_a = 0; j_a < (*(osi->VA_ioperator))[i_a].size(); j_a++){
      osi->VA_ME2_cf[i_a][j_a] = (*(osi->VA_ioperator))[i_a][j_a].charge_factor() * osi->VA_b_ME2;
      ///      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "(*(osi->VA_ioperator))[" << i_a << "][" << j_a << "].charge_factor() = " << (*(osi->VA_ioperator))[i_a][j_a].charge_factor() << endline_flush; }
    }
  }

  osi->calculate_ioperator_QEW();

  for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
    I_ME2_emitter[i_a] = accumulate(osi->VA_I_ME2_cf[i_a].begin(), osi->VA_I_ME2_cf[i_a].end(), 0.);
    //    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "I_ME2_emitter[" << i_a << "] = " << I_ME2_emitter[i_a] << endline_flush; }

  }
  osi->VA_I_ME2 = accumulate(I_ME2_emitter.begin(), I_ME2_emitter.end(), 0.);
  //  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "osi->VA_I_ME2 = " << osi->VA_I_ME2 << endline_flush; }

  logger.finished();
}


void amplitude_OpenLoops_set::calculate_ME2_VA_QEW(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_VA_QEW");
  logger.called();

  if (osi->switch_VI == 0 || osi->switch_VI == 1){
    set_phasespacepoint();

    set_mu_reg();
    set_mu_and_delta_VA();

    ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
    osi->VA_V_ME2 = M2L1[0];
    ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2);

    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_B  = " << setw(23) << setprecision(15) << osi->VA_b_ME2 << endline_flush;
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << endline_flush;
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_V  = " << setw(23) << setprecision(15) << osi->VA_V_ME2 << endline_flush;
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_CT = " << setw(23) << setprecision(15) << osi->VA_X_ME2 << endline_flush;
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << endline_flush;
    }

    if (osi->switch_VI == 0){
      if (osi->user->switch_map["I_Munich"]){
	calculate_ME2_ioperator_VA_QEW();
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_I  = " << setw(23) << setprecision(15) << osi->VA_I_ME2 << "   (from Munich)" << endline_flush; }
      }
      else {
	if (osi->VA_DeltaIR1 == 0. && osi->VA_DeltaIR2 == 0.){osi->VA_I_ME2 = IRL1[0];}
	else {osi->VA_I_ME2 = IRL1[0] + osi->VA_DeltaIR1 * IRL1[1] + osi->VA_DeltaIR2 * IRL1[2];}
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_I  = " << setw(23) << setprecision(15) << osi->VA_I_ME2 << "   (from OpenLoops)" << endline_flush; }
      }
    }
    else if (osi->switch_VI == 1){
      osi->VA_I_ME2 = 0.;
    }
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_I  = " << setw(23) << setprecision(15) << osi->VA_I_ME2 << endline_flush; }

    if (osi->switch_CV){
      for (int s = 0; s < osi->n_scales_CV; s++){
	double inv_factor_CV = osi->var_mu_ren_CV[s] / mu_reg;
        ol_setparameter_double(OL_mu_ren, osi->var_mu_ren_CV[s]);
	ol_setparameter_double(OL_mu_reg, osi->var_mu_ren_CV[s]);
        ol_setparameter_double(fact_uv, inv_factor_CV);
        ol_setparameter_double(fact_ir, inv_factor_CV);
	ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->VA_X_ME2_CV[s]);
	if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  VA_X_ME2_CV[" << s << "] = " << setw(23) << setprecision(15) << osi->VA_X_ME2_CV[s] + osi->VA_V_ME2 << endline_flush; }
      }
    }
    if (osi->switch_TSV){
      for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
	for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
	  double inv_factor_TSV = osi->value_scale_ren[0][i_v][i_r] / mu_reg;
	  ol_setparameter_double(OL_mu_ren, osi->value_scale_ren[0][i_v][i_r]);
	  ol_setparameter_double(OL_mu_reg, osi->value_scale_ren[0][i_v][i_r]);
	  ol_setparameter_double(fact_uv, inv_factor_TSV);
	  ol_setparameter_double(fact_ir, inv_factor_TSV);
	  ol_evaluate_ct(process_id, P_OpenLoops, &M2L0, &osi->value_ME2term_ren[0][i_v][i_r]);
	  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops:  VA_VX_ME2_TSV[" << i_v << "][" << i_r << "] = " << osi->value_ME2term_ren[0][i_v][i_r] + osi->VA_V_ME2 << endline_flush; }
	  osi->value_ME2term_ren[0][i_v][i_r] += osi->VA_V_ME2 + osi->VA_I_ME2;
	}
      }
    }
  }
  else if (osi->switch_VI == 2){
    logger << LOG_DEBUG_VERBOSE << "OL: osi->VA_I_ME2 = " << osi->VA_I_ME2 << endline_flush;
    //  osi->VA_I_ME2 = 0.;
    //  osi->VA_V_ME2 = 0.;
    //  osi->VA_X_ME2 = 0.;
    //  for (int s = 0; s < osi->n_scales_CV; s++){osi->VA_X_ME2_CV[s] = 0.;}
    //  I-operator evaluation -> osi->VA_I_ME2
    calculate_ME2_ioperator_VA_QEW();
    for (int i_v = 0; i_v < osi->max_dyn_ren + 1; i_v++){
      for (int i_r = 0; i_r < osi->n_scale_dyn_ren[i_v]; i_r++){
        osi->value_ME2term_ren[0][i_v][i_r] = osi->VA_I_ME2;
      }
    }
    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "SK: osi->VA_I_ME2 = " << osi->VA_I_ME2 << endline_flush; }
  }

  if (osi->VA_V_ME2 == 0. && osi->switch_VI != 2){ failed_evaluation = 1; }
  if (munich_isnan(osi->VA_V_ME2) || munich_isnan(osi->VA_X_ME2) || munich_isnan(osi->VA_I_ME2)){ failed_evaluation = 1; }

  logger.finished();
}

#endif
