#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::initialization(observable_set * _osi, contribution_set * _csi, model_set * _msi, event_set * _esi, inputparameter_set * isi){
  static Logger logger("amplitude_OpenLoops_set::initialization (osi, csi, msi, esi, isi)");
  logger.called();

  amplitude_set::initialization(_osi, _csi, _msi, _esi, isi);
  initialization(isi);

  logger.finished();
}


void amplitude_OpenLoops_set::initialization(inputparameter_set * isi){
  static Logger logger("amplitude_OpenLoops_set::initialization (isi)");
  logger.called();

  initialization_N_nondecoupled_from_pdf_set();
  initialization_parameter();

  one = 1;

  n_momentum = 5 * (csi->n_particle + 2);
  // colour correlators are also used to check numerical zeros in tree amplitudes:
  n_cc_check = (csi->n_particle + 2) * (csi->n_particle + 1) / 2;
  M2cc_check = new double[n_cc_check];

  P_OpenLoops = new double[n_momentum];

  if (csi->class_contribution_CS_real && osi->n_ps > 1){
    //  if (csi->class_contribution_CS_real){
    n_momentum_dipole = 5 * (csi->n_particle + 1);

    P_OL.resize(osi->n_ps);
    for (size_t i_a = 1; i_a < osi->n_ps; i_a++){
      P_OL[i_a] = new double[n_momentum_dipole];
    }

    n_cc = (csi->n_particle + 1) * (csi->n_particle) / 2;
    M2cc = new double[n_cc];

    n_sc = (csi->n_particle + 1);
    M2sc = new double[n_sc];

    P_sc.resize(osi->n_ps);
    for (size_t i_a = 1; i_a < osi->n_ps; i_a++){
      P_sc[i_a] = new double[4];
    }
  }
  else {
    n_cc = (csi->n_particle + 2) * (csi->n_particle + 1) / 2;
    M2cc = new double[n_cc];

    if (csi->type_contribution == "CT2" || csi->type_contribution == "PT2"){
      M2loopcc = new double[n_cc];
    }
  }

  M2L1 = new double[3];
  IRL1 = new double[3];
  M2L2 = new double[5];
  IRL2 = new double[5];
  for (int i_e = 0; i_e < 3; i_e++){M2L1[i_e] = 0.; IRL1[i_e] = 0.;}
  for (int i_e = 0; i_e < 5; i_e++){M2L2[i_e] = 0.; IRL2[i_e] = 0.;}

  allowed_Delta_IRneqUV = 1;
  string set_OL_model = "";
  for (int i_o = 0; i_o < OL_parameter.size(); i_o++){
    if (OL_parameter[i_o] == "model"){set_OL_model = OL_value[i_o];}
  }
  if (set_OL_model == "heft"){allowed_Delta_IRneqUV = 0;}

  initialization_process_general();
  initialization_process();
  initialization_process_extra();
  initialization_process_integration_check();

  ol_start();

  logger.finished();
}


void amplitude_OpenLoops_set::initialization_set_default(){
  static Logger logger("amplitude_OpenLoops_set::initialization_set_default");
  logger.called();

  amplitude_set::initialization_set_default();

  OL_ct_on = stch("ct_on");
  renscale = stch("renscale");
  fact_uv = stch("fact_uv");
  fact_ir = stch("fact_ir");
  OL_mu_ren = stch("muren");
  OL_mu_reg = stch("mureg");
  pole_uv = stch("pole_uv");
  pole_ir1 = stch("pole_ir1");
  pole_ir2 = stch("pole_ir2");
  polenorm = stch("polenorm");
  OL_alpha_s = stch("alpha_s");

  logger.finished();
}


void amplitude_OpenLoops_set::initialization_input(inputparameter_set * isi){
  static Logger logger("amplitude_OpenLoops_set::initialization_input");
  logger.called();

  amplitude_set::initialization_input(isi);

  // Set the OL parameters separately for each contribution !!!
  
  for (int i_i = 0; i_i < isi->input_amplitude_OpenLoops.size(); i_i++){
    logger << LOG_DEBUG_VERBOSE << "isi->input_amplitude_OpenLoops[" << i_i << "] = " << isi->input_amplitude_OpenLoops[i_i] << endline_flush;

    if (isi->input_amplitude_OpenLoops[i_i].variable == ""){}
    else {
      if (!isi->check_input_flag(isi->input_amplitude_OpenLoops, i_i)){continue;}

      if (isi->input_amplitude_OpenLoops[i_i].variable == "OL"){
	if (isi->input_amplitude_OpenLoops[i_i].specifier != ""){
	  OL_parameter.push_back(isi->input_amplitude_OpenLoops[i_i].specifier);
	  OL_value.push_back(isi->input_amplitude_OpenLoops[i_i].value);
	}
      }
    }
  }

  logger.finished();
}


void amplitude_OpenLoops_set::initialization_parameter(){
  static Logger logger("amplitude_OpenLoops_set::initialization_parameter");
  logger.called();

  // generic version to replace mass/width initialization by particle name
  for (int i_p = 1; i_p < 26; i_p++){
    if ((i_p > 6 && i_p < 11) || (i_p > 16 && i_p < 21) || i_p == 12 || i_p == 14 || i_p == 16 || i_p == 21 || i_p == 22){continue;}
    stringstream temp_mass_ss;
    temp_mass_ss << "mass(" << i_p << ")";
    logger << LOG_DEBUG << "msi->M[" << i_p << "] = " << msi->M[i_p] << endline_flush;
    munich_ol_setparameter_double(temp_mass_ss.str(), msi->M[i_p], logger);
  }

  //  ol_setparameter_double(stch("yuk(5)"), msi->M[5]);

  for (int i_p = 1; i_p < 26; i_p++){
    if ((i_p > 6 && i_p < 11) || (i_p > 16 && i_p < 21) || i_p == 12 || i_p == 14 || i_p == 16 || i_p == 21 || i_p == 22){continue;}
    stringstream temp_width_ss;
    temp_width_ss << "width(" << i_p << ")";
    logger << LOG_DEBUG << "msi->Gamma[" << i_p << "] = " << msi->Gamma[i_p] << endline_flush;
    munich_ol_setparameter_double(temp_width_ss.str(), msi->Gamma[i_p], logger);
  }

  munich_ol_setparameter_int("n_quarks", N_quarks, logger);
  /*
  int minnf_alphasrun = p_pdf->info().get_entry_as<int>("NumFlavors");
  logger << LOG_DEBUG << "manual:    N_nondecoupled = " << N_nondecoupled << endline_flush;
  logger << LOG_DEBUG << "automatic: minnf_alphasrun = " << minnf_alphasrun << endline_flush;
  assert(N_nondecoupled == minnf_alphasrun);
  */
  
  logger << LOG_DEBUG << "N_nondecoupled = " << N_nondecoupled << endline_flush;
  munich_ol_setparameter_int("nq_nondecoupled", N_nondecoupled, logger); // should be calculated from chosen PDF set

  logger << LOG_DEBUG << "osi->scale_ren = " << osi->scale_ren << endline_flush;
  munich_ol_setparameter_double("renscale", osi->scale_ren, logger);

  munich_ol_setparameter_int("polenorm", osi->switch_polenorm, logger);
  logger << LOG_DEBUG << "osi->switch_polenorm = " << osi->switch_polenorm << endline_flush;
  munich_ol_setparameter_int("verbose", 2, logger);

  logger << LOG_DEBUG << "msi->ew_scheme = " << msi->ew_scheme << endline_flush;
  munich_ol_setparameter_int("ew_scheme", msi->ew_scheme, logger);

  logger << LOG_DEBUG << "msi->use_cms = " << msi->use_cms << endline_flush;
  munich_ol_setparameter_int("use_cms", msi->use_cms, logger);

  logger << LOG_DEBUG << "Gmu = " << msi->G_F << endline_flush;
  munich_ol_setparameter_double("Gmu", msi->G_F, logger);

  logger << LOG_DEBUG << "osi->alpha_S = " << osi->alpha_S << endline_flush;
  munich_ol_setparameter_double("alpha_s", osi->alpha_S, logger);


  // CKM matrix
  if (msi->CKM_matrix != "trivial"){
    munich_ol_setparameter_int("ckmorder", 1, logger);
    munich_ol_setparameter_double("VCKMdu", msi->V_du, logger);
    munich_ol_setparameter_double("VCKMsu", msi->V_su, logger);
    munich_ol_setparameter_double("VCKMbu", msi->V_bu, logger);
    munich_ol_setparameter_double("VCKMdc", msi->V_dc, logger);
    munich_ol_setparameter_double("VCKMsc", msi->V_sc, logger);
    munich_ol_setparameter_double("VCKMbc", msi->V_bc, logger);
    munich_ol_setparameter_double("VCKMdt", msi->V_dt, logger);
    munich_ol_setparameter_double("VCKMst", msi->V_st, logger);
    munich_ol_setparameter_double("VCKMbt", msi->V_bt, logger);
  }
  else {
    munich_ol_setparameter_int("ckmorder", 0, logger);
  }

  int last_switch = 1;
  munich_ol_setparameter_int("last_switch", last_switch, logger);

  int use_coli_cache = 1;
  munich_ol_setparameter_int("use_coli_cache", use_coli_cache, logger);

  int out_symmetry = 1;
  munich_ol_setparameter_int("out_symmetry", out_symmetry, logger);

  int leading_colour = 0;
  munich_ol_setparameter_int("leading_colour", leading_colour, logger);

  int n_log = 1;
  //  int n_log = n_step;
  munich_ol_setparameter_int("stability_log", n_log, logger);

  // ???
  double pole1_UV = 0.;
  munich_ol_setparameter_double("pole_uv", pole1_UV, logger);

  double pole1_IR = 0.;
  munich_ol_setparameter_double("pole_ir1", pole1_IR, logger);

  double pole2_IR = 0.;
  munich_ol_setparameter_double("pole_ir2", pole2_IR, logger);


  int CT_on = 0; // modification of ...ME2_VA and OpenLoops calls needed !!!
  munich_ol_setparameter_int("ct_on", CT_on, logger); // modification of ...ME2_VA and OpenLoops calls needed !!!

  int R2_on = 1;
  munich_ol_setparameter_int("r2_on", R2_on, logger);

  ///  int IR_on = 2; // I-operator added to virtual amplitude
  int IR_on = 1;
  munich_ol_setparameter_int("ir_on", IR_on, logger);

  double fact_uv = 1.;
  munich_ol_setparameter_double("fact_uv", fact_uv, logger);

  double fact_ir = 1.;
  munich_ol_setparameter_double("fact_ir", fact_ir, logger);


  int polecheck = 0;
  //  int polecheck = 2; // sets polecheck = 1 and IR_on = 2
  munich_ol_setparameter_int("polecheck", polecheck, logger);

  // Check if this works correctly !!!
  for (int i_ol = 0; i_ol < OL_parameter.size(); i_ol++){munich_ol_setparameter_string(OL_parameter[i_ol], OL_value[i_ol], logger);}
  /*
  for (int i_ol = 0; i_ol < OL_parameter.size(); i_ol++){
    logger << LOG_DEBUG << "OL_parameter[" << i_ol << "] = " << OL_parameter[i_ol] << " = " << OL_value[i_ol] << endline_flush;

    string temp_parameter = OL_parameter[i_ol];
    string temp_value = OL_value[i_ol];
    char * char_parameter = new char[temp_parameter.size() + 1];
    std::copy(temp_parameter.begin(), temp_parameter.end(), char_parameter);
    char_parameter[temp_parameter.size()] = '\0';
    char * char_value = new char[temp_value.size() + 1];
    std::copy(temp_value.begin(), temp_value.end(), char_value);
    char_value[temp_value.size()] = '\0';
    ol_setparameter_string(char_parameter, char_value);
    delete [] char_parameter;
    delete [] char_value;
  }
  */
  // Old OpenLoops initialization parameters (left at default values now):
  /*
  munich_ol_setparameter_int("flavour_mapping", 0);
  logger << LOG_DEBUG << "msi->ew_renorm_scheme = " << msi->ew_scheme << " (per default equal to ew_scheme)" << endline_flush;
  ol_setparameter_int(stch("ew_renorm_scheme"), msi->ew_scheme);
  int amp_switch = 1;
  ol_setparameter_int(stch("redlib1"), amp_switch);
  int amp_switch_rescue = 7;
  ol_setparameter_int(stch("redlib2"), amp_switch_rescue);
  int xtest;
  ol_getparameter_int(stch("stability_mode"), &xtest);
  logger << LOG_DEBUG << "default: stability_mode = " << xtest << endline_flush;
  int stability_mode = 23;
  ol_setparameter_int(stch("stability_mode"), stability_mode);

  */

  

  if (msi->ew_scheme == 1 || msi->ew_scheme == -1){
    munich_ol_setparameter_double("alpha_QED", msi->alpha_e_Gmu, logger);
  }
  if (msi->ew_scheme == 2){
    munich_ol_setparameter_double("alpha_QED_MZ", msi->alpha_e_MZ, logger);
  }
  // Set for all schemes because of possible external identified photons
  munich_ol_setparameter_double("alpha_QED_0", msi->alpha_e_0, logger);

  /*
  // Refinement required !!!
// It might be useful to have a switch to deactivate the factor in the photon couplings.
  if (msi->ew_scheme == 1 || msi->ew_scheme == 2) && msi->use_adapted_ew_coupling == msi->ew_scheme){
    munich_ol_setparameter_int("onshell_photons_lsz", 0, logger);
  }
  else if (msi->use_adapted_ew_coupling > 0){
    munich_ol_setparameter_int("offshell_photons_lsz", 0, logger);
  }
  else {
    logger << LOG_DEBUG << "OpenLoops default is used for photon treatment." << endline_flush;
  }
  */
  
  ol_start();

  double temp_alpha = 0.;
  munich_ol_getparameter_double("alpha_QED", temp_alpha, logger);
  logger << LOG_DEBUG << "OpenLoops:   msi->alpha_e     = " << setw(23) << setprecision(15) << temp_alpha << "   1 / msi->alpha_e     = " << setw(23) << setprecision(15) << 1. / temp_alpha << endline_flush;
  munich_ol_getparameter_double("alpha_QED_0", temp_alpha, logger);
  logger << LOG_DEBUG << "OpenLoops:   msi->alpha_e_0   = " << setw(23) << setprecision(15) << temp_alpha << "   1 / msi->alpha_e_0   = " << setw(23) << setprecision(15) << 1. / temp_alpha << endline_flush;
  munich_ol_getparameter_double("alpha_QED_MZ", temp_alpha, logger);
  logger << LOG_DEBUG << "OpenLoops:   msi->alpha_e_MZ  = " << setw(23) << setprecision(15) << temp_alpha << "   1 / msi->alpha_e_MZ  = " << setw(23) << setprecision(15) << 1. / temp_alpha << endline_flush;

  logger.finished();
}


void amplitude_OpenLoops_set::initialization_process(){
  static Logger logger("amplitude_OpenLoops_set::initialization_process");
  logger.called();

  logger << LOG_DEBUG << "csi->type_perturbative_order         = " << csi->type_perturbative_order << endline_flush;
  logger << LOG_DEBUG << "csi->type_contribution               = " << csi->type_contribution << endline_flush;
  logger << LOG_DEBUG << "csi->type_correction                 = " << csi->type_correction << endline_flush;
  logger << LOG_DEBUG << "csi->contribution_order_alpha_s      = " << csi->contribution_order_alpha_s << endline_flush;
  logger << LOG_DEBUG << "csi->contribution_order_alpha_e      = " << csi->contribution_order_alpha_e << endline_flush;
  logger << LOG_DEBUG << "csi->contribution_order_interference = " << csi->contribution_order_interference << endline_flush;

  process_id = -1;
  if (csi->int_end){logger << LOG_WARN << "Process does not contribute, so no amplitudes are initialized." << endline_flush; return;}
  
  // Born level tree-induced contributions
  
  if (csi->type_contribution == "born" ||
      csi->type_contribution == "RT"){
    int type_amplitude = 1;
    if (osi->user->string_value[osi->user->string_map["model"]] == "Bornloop"){type_amplitude = 12;}

    if (csi->contribution_order_interference == 0){
      munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
      //      output_parameter("testpoint");
      process_id = register_OL_subprocess(0, type_amplitude);
      if (process_id == -1){
	munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	process_id = register_OL_subprocess(0, type_amplitude);
      }
    }
    else if (csi->contribution_order_interference == 1){
      // Check if this initialization makes sense in general !!!
      munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
      process_id = register_OL_subprocess(0, 1);
      logger << LOG_DEBUG << "process_id (bare born)           = " << process_id << endline_flush;

      munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s + 1, logger); // ???
      process_id = register_OL_subprocess(0, 11);
      logger << LOG_DEBUG << "process_id (colour correlations) = " << process_id << endline_flush;
    }
    else {logger << LOG_ERROR << "csi->contribution_order_interference = " << csi->contribution_order_interference <<  " > 1  needs to be implemented." << endline_flush; exit(1);}
  }

  
  // loop-only contributions (includes also L2V/CT at present) !!! to be improved !!!

  else if (csi->type_contribution == "loop" ||
	   csi->type_contribution == "L2I" ||
	   //	   csi->type_contribution == "L2VA" ||
	   csi->type_contribution == "L2RT" ||
	   csi->type_contribution == "L2VT" ||
	   csi->type_contribution == "L2CT" ||
	   csi->type_contribution == "L2PT"){
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 12);
    logger << LOG_DEBUG << "process_id = " << process_id << endline_flush;
  }

  
  // dipolesubtraction for tree-induced processes
  
  else if (csi->class_contribution_CS_collinear && !csi->class_contribution_loopinduced){
    int type_amplitude = 11; // required to enforce OpenLoops to initialize QCD amplitudes and with the correct colour correlators
    if (osi->user->string_value[osi->user->string_map["model"]] == "Bornloop"){type_amplitude = 12; logger << LOG_WARN << "May not happen: different branch must be taken." << endline_flush;}
    if (csi->type_correction == "QCD"){
      if (csi->contribution_order_interference == 0){
	munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	process_id = register_OL_subprocess(0, type_amplitude);
	if (process_id == -1){
	  logger << LOG_WARN << "Initialization with amptype = 1 (instead of 11)" << endline_flush;
	  process_id = register_OL_subprocess(0, 1);
	}
	process_id_QCD = process_id;
      }
      else if (csi->contribution_order_interference == 1){
	// Use order_ew for standard matrix element, and order_qcd for colour correlators
	munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	process_id = register_OL_subprocess(0, 1); // no colour correlations needed in EW corrections
	logger << LOG_DEBUG << "process_id (bare born)           = " << process_id << endline_flush;

	munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	process_id_QCD = register_OL_subprocess(0, type_amplitude);
	logger << LOG_DEBUG << "process_id_QCD (colour correlations) = " << process_id_QCD << endline_flush;
	//	int process_id_cc = register_OL_subprocess(0, type_amplitude);
	//	logger << LOG_DEBUG << "process_id_cc (colour correlations) = " << process_id_cc << endline_flush;
	/*
	// looks actually wrong !!! Shouldn't this be initialized with order_ew as well ???
	munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	process_id = register_OL_subprocess(0, type_amplitude);
	logger << LOG_DEBUG << "process_id (colour correlations) = " << process_id << endline_flush;
	*/
	/*
	munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	process_id = register_OL_subprocess(0, 1); // no colour correlations needed in EW corrections
	logger << LOG_DEBUG << "process_id (bare born)           = " << process_id << endline_flush;
	*/
      }
      else {logger << LOG_ERROR << "csi->contribution_order_interference = " << csi->contribution_order_interference <<  " > 1  needs to be implemented." << endline_flush; exit(1);}
    }
    else if (csi->type_correction == "QEW"){
      munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
      process_id = register_OL_subprocess(0, 1); // charge correlations are handled internally (so no 11 needed)
    }
    else {logger << LOG_FATAL << "Wrong correction type: " << csi->type_contribution << " - " << csi->type_correction << " does not exist." << endline_flush; exit(1);}
  }

  else if (csi->class_contribution_CS_virtual && !csi->class_contribution_loopinduced){
    if (csi->type_correction == "QCD"){
      munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
      process_id = register_OL_subprocess(0, 11);
    }
    else if (csi->type_correction == "QEW"){
      munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
      process_id = register_OL_subprocess(0, 11);
    }
    else if (csi->type_correction == "MIX"){
      /*
      munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
      process_id = register_OL_subprocess(0, 11);
      if (process_id == -1){
	munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	process_id = register_OL_subprocess(0, 11);
      }
      */
      if (process_id == -1){
	munich_ol_setparameter_int("loop_order_qcd", csi->contribution_order_alpha_s, logger);
	process_id = register_OL_subprocess(0, 11);
	logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id  is initialized via  loop_order_qcd = " << csi->contribution_order_alpha_s << "  and  amptype = 11" << endline_flush;
	logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id = " << process_id << endline_flush;
	logger << LOG_DEBUG << "Mixed virtual amplitude initialized.   process_id = " << process_id << endline_flush;
      }
      if (process_id == -1){
	munich_ol_setparameter_int("loop_order_ew", csi->contribution_order_alpha_e, logger);
	process_id = register_OL_subprocess(0, 11);
	logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id  is initialized via  loop_order_ew = " << csi->contribution_order_alpha_e << "  and  amptype = 11" << endline_flush;
	logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id = " << process_id << endline_flush;
	logger << LOG_DEBUG << "Mixed virtual amplitude initialized.   process_id = " << process_id << endline_flush;
      }
      //      for (int i_a = 0; i_a < osi->VA_ioperator->size(); i_a++){
      bool init_QCD = false;
      bool init_QEW = false;
      int process_id_dipole_QCD = -1;
      int process_id_dipole_QEW = -1;
      for (int i_a = 0; i_a < (*(osi->VA_ioperator)).size(); i_a++){
	//	for (int j_a = 0; j_a < osi->VA_ioperator[i_a].size(); j_a++){
	for (int j_a = 0; j_a < (*(osi->VA_ioperator))[i_a].size(); j_a++){
	  logger << LOG_DEBUG << "osi->VA_ioperator[" << i_a << "][" << j_a << "].name() = " << setw(15) << (*(osi->VA_ioperator))[i_a][j_a].name() << "   type_correction = " << (*(osi->VA_ioperator))[i_a][j_a].type_correction() << "   to be processed..." << endline_flush;
	  if ((*(osi->VA_ioperator))[i_a][j_a].type_correction() == 1){
	    if (!init_QCD){
	      /*
	      if (process_id_dipole_QCD == -1){
		munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
		process_id_dipole_QCD = register_OL_subprocess(0, 1);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD  is initialized via  order_ew = " << csi->contribution_order_alpha_e << "  and  amptype = 1" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD = " << process_id_dipole_QCD << endline_flush;
	      }
	      if (process_id_dipole_QCD == -1){
		munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
		process_id_dipole_QCD = register_OL_subprocess(0, 1);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD  is initialized via  order_qcd = " << csi->contribution_order_alpha_s << "  and  amptype = 1" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD = " << process_id_dipole_QCD << endline_flush;
	      }
	      */
	      if (process_id_dipole_QCD == -1){
		munich_ol_setparameter_int("loop_order_qcd", csi->contribution_order_alpha_s, logger);
		process_id_dipole_QCD = register_OL_subprocess(0, 11);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD  is initialized via  loop_order_qcd = " << csi->contribution_order_alpha_s << "  and  amptype = 11" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD = " << process_id_dipole_QCD << endline_flush;
	      }
	      if (process_id_dipole_QCD == -1){
		munich_ol_setparameter_int("loop_order_ew", csi->contribution_order_alpha_e, logger);
		process_id_dipole_QCD = register_OL_subprocess(0, 11);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD  is initialized via  loop_order_ew = " << csi->contribution_order_alpha_e << "  and  amptype = 11" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QCD = " << process_id_dipole_QCD << endline_flush;
	      }
	      init_QCD = true;
	    }
	    (*(osi->VA_ioperator))[i_a][j_a].process_id = process_id_dipole_QCD;
	  }
	  if ((*(osi->VA_ioperator))[i_a][j_a].type_correction() == 2){
	    if (!init_QEW){
	      /*
	      if (process_id_dipole_QEW == -1){
		munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s - 1, logger);
		process_id_dipole_QEW = register_OL_subprocess(0, 1);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW  is initialized via  order_qcd = " << csi->contribution_order_alpha_s - 1 << "  and  amptype = 1" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW = " << process_id_dipole_QEW << endline_flush;
	      }
	      if (process_id_dipole_QEW == -1){
		munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
		process_id_dipole_QEW = register_OL_subprocess(0, 1);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW  is initialized via  order_ew = " << csi->contribution_order_alpha_e << "  and  amptype = 1" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW = " << process_id_dipole_QEW << endline_flush;
	      }
	      */
	      if (process_id_dipole_QEW == -1){
		munich_ol_setparameter_int("loop_order_qcd", csi->contribution_order_alpha_s, logger);
		process_id_dipole_QEW = register_OL_subprocess(0, 1);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW  is initialized via  loop_order_qcd = " << csi->contribution_order_alpha_s << "  and  amptype = 1" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW = " << process_id_dipole_QEW << endline_flush;
	      }
	      if (process_id_dipole_QEW == -1){
		munich_ol_setparameter_int("loop_order_ew", csi->contribution_order_alpha_e, logger);
		process_id_dipole_QEW = register_OL_subprocess(0, 11);
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW  is initialized via  loop_order_ew = " << csi->contribution_order_alpha_e << "  and  amptype = 11" << endline_flush;
		logger << LOG_INFO << "VA.MIX ( " << csi->contribution_order_alpha_s << " / " << csi->contribution_order_alpha_e << " ):  process_id_dipole_QEW = " << process_id_dipole_QEW << endline_flush;
	      }
	      init_QEW = true;
	    }
	    (*(osi->VA_ioperator))[i_a][j_a].process_id = process_id_dipole_QEW;
	  }
	  logger << LOG_DEBUG << "osi->VA_ioperator[" << i_a << "][" << j_a << "].name() = " << setw(15) << (*(osi->VA_ioperator))[i_a][j_a].name() << "   type_correction = " << (*(osi->VA_ioperator))[i_a][j_a].type_correction() << "   process_id = " << (*(osi->VA_ioperator))[i_a][j_a].process_id << endline_flush;
	}
      }
    }
    else {logger << LOG_FATAL << "Wrong correction type: " << csi->type_contribution << " - " << csi->type_correction << " does not exist." << endline_flush; exit(1);}
  }

  else if (csi->class_contribution_CS_real && !csi->class_contribution_loopinduced){
    int type_amplitude = 1;
    if (osi->user->string_value[osi->user->string_map["model"]] == "Bornloop"){type_amplitude = 12; logger << LOG_WARN << "May not happen: different branch must be taken." << endline_flush;}
    for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
      if (i_a == 0){
	munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	csi->dipole[i_a].process_id = register_OL_subprocess(i_a, type_amplitude);
	logger << LOG_DEBUG << "csi->dipole[0].process_id = " << csi->dipole[0].process_id << endline_flush;
	if (csi->dipole[i_a].process_id == -1){
	  munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	  csi->dipole[i_a].process_id = register_OL_subprocess(i_a, type_amplitude);
	  if (csi->dipole[i_a].process_id != -1){logger << LOG_WARN << "Should never enter here: OpenLoops should not distinguish order_qcd and order_ew initialization of tree-level amplitudes." << endline_flush;}
	}
      }
      else {
	if (csi->type_correction == "QCD"){
	  type_amplitude = 11; // required to enforce OpenLoops to initialize QCD amplitudes and with the correct colour correlators
	  if (csi->contribution_order_interference == 0){
	    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	    csi->dipole[i_a].process_id = register_OL_subprocess(i_a, type_amplitude);
	    if (csi->dipole[i_a].process_id == -1){
	      logger << LOG_WARN << "Initialization with amptype = 1 (instead of 11) for csi->dipole[" << i_a << "] = " << csi->dipole[i_a].name() << endline_flush;
	      csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 1);
	    }
	  }
	  else if (csi->contribution_order_interference == 1){
	    // looks actually wrong !!! Shouldn't this be initialized with order_ew as well ???
	    munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	    logger << LOG_DEBUG << "order_qcd   set to   " << csi->contribution_order_alpha_s << endline_flush;
	    // trying this does no work though:
	    //	    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	    csi->dipole[i_a].process_id = register_OL_subprocess(i_a, type_amplitude); 
	  }
	  else {logger << LOG_ERROR << "csi->contribution_order_interference = " << csi->contribution_order_interference <<  " > 1  needs to be implemented for csi->dipole[" << i_a << "] = " << csi->dipole[i_a].name() << "." << endline_flush; exit(1);}
	}
	else if (csi->type_correction == "QEW"){
	  munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	  csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 1); // charge correlations are handled internally (so no 11 needed)
	}
	else if (csi->type_correction == "MIX"){
	  if (csi->dipole[i_a].type_correction() == 1){
	    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	    csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 11); // temporary, to force the use of correct colour correlations !!!
	  }
	  else if (csi->dipole[i_a].type_correction() == 2){
	    munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	    csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 1); // charge correlations are handled internally (so no 11 needed)
	  }
	  else {logger << LOG_FATAL << "Wrong correction type: " << csi->dipole[i_a].type_correction() << " does not exist." << endline_flush; exit(1);}
	}
      }
      logger << LOG_DEBUG << "dipole[" << i_a << "].process_id  = " << csi->dipole[i_a].process_id << endline_flush;
      if (csi->dipole[i_a].process_id == -1){logger << LOG_FATAL << "Requested amplitudes not available for csi->dipole[" << i_a << "] = " << csi->dipole[i_a].name() << endline_flush;}
    }
    logger.newLine(LOG_DEBUG);
  }

    
  // dipolesubtraction for loop-induced processes
  
  else if (csi->class_contribution_CS_collinear && csi->class_contribution_loopinduced){
    if (csi->type_correction == "QCD"){
      if (csi->contribution_order_interference == 0){
	munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	process_id = register_OL_subprocess(0, 12);
      }
      else {logger << LOG_ERROR << "csi->contribution_order_interference = " << csi->contribution_order_interference <<  " > 0  needs to be implemented." << endline_flush; exit(1);}
    }
    else {logger << LOG_FATAL << "Wrong correction type: " << csi->type_contribution << " - " << csi->type_correction << " does not exist." << endline_flush; exit(1);}
  }
  
  else if (csi->class_contribution_CS_virtual && csi->class_contribution_loopinduced){
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 12);
    logger << LOG_DEBUG << "process_id = " << process_id << endline_flush;
  }
   
  else if (csi->class_contribution_CS_real && csi->class_contribution_loopinduced){
    for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
      if (i_a == 0){
	munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 12);
	logger << LOG_DEBUG << "csi->dipole[0].process_id = " << csi->dipole[0].process_id << endline_flush;
	if (csi->dipole[i_a].process_id == -1){
	  munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	  csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 12);
	}
      }
      else {
	if (csi->type_correction == "QCD"){
	  if (csi->contribution_order_interference == 1){
	    munich_ol_setparameter_int("order_qcd", csi->contribution_order_alpha_s, logger);
	    csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 12);
	  }
	  else {
	    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
	    csi->dipole[i_a].process_id = register_OL_subprocess(i_a, 12);
	  }
	}
	else {logger << LOG_FATAL << "Wrong correction type: " << csi->type_contribution << " - " << csi->type_correction << " does not exist." << endline_flush; exit(1);}
      }
      logger << LOG_DEBUG << "dipole[" << i_a << "].process_id  = " << csi->dipole[i_a].process_id << endline_flush;
      if (csi->dipole[i_a].process_id == -1){logger << LOG_FATAL << "Requested amplitudes not available for csi->dipole[" << i_a << "] = " << csi->dipole[i_a].name() << endline_flush;}
    }
    logger.newLine(LOG_DEBUG);
  }


  // qT subtraction (both tree-induced and loop-induced processes) at 1st order

  else if ((csi->class_contribution_QT_collinear || csi->class_contribution_QT_powercorrection) && csi->class_contribution_order == 1){// not split at present !!!   && !csi->class_contribution_loopinduced){
    //  else if (csi->type_contribution == "CT"){
    int type_amplitude = 1;
    if (osi->user->string_value[osi->user->string_map["model"]] == "Bornloop"){type_amplitude = 12;}
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, type_amplitude);
  }

  else if (csi->class_contribution_QT_virtual && csi->class_contribution_order == 1){// not split at present !!!    && !csi->class_contribution_loopinduced){
    //  else if (csi->type_contribution == "VT"){
    int type_amplitude = 11;
    if (osi->user->string_value[osi->user->string_map["model"]] == "Bornloop"){type_amplitude = 12;}
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, type_amplitude);
  }


  // qT subtraction (both tree-induced and loop-induced processes) at 2nd order

 else if ((csi->class_contribution_QT_collinear || csi->class_contribution_QT_powercorrection) && csi->class_contribution_order == 2){// not split at present !!!    && !csi->class_contribution_loopinduced){
    //  else if (csi->type_contribution == "CT2"){
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }

  // actually identical to (csi->class_contribution_QT_collinear && csi->class_contribution_order == 2) case !!!
  else if (csi->class_contribution_QT_virtual && csi->class_contribution_order == 2){// not split at present !!!     && !csi->class_contribution_loopinduced){
    //  else if (csi->type_contribution == "VT2"){
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }
  else if (csi->class_contribution_loopinduced){
    // collect loop-induced process initializations here ???
  }

  else if (csi->type_contribution == "NLL_LO"){
    //  ol_setparameter_int(stch("polenorm"), 1); // polenorm prescribed by H^(1) construction
    //    ol_setparameter_int(stch("order_ew"), csi->contribution_order_alpha_e);
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }
  else if (csi->type_contribution == "NLL_NLO"){
    //  ol_setparameter_int(stch("polenorm"), 1); // polenorm prescribed by H^(1) construction
    //    ol_setparameter_int(stch("order_ew"), csi->contribution_order_alpha_e);
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }
  else if (csi->type_contribution == "NNLL_LO"){
    //  ol_setparameter_int(stch("polenorm"), 1); // polenorm prescribed by H^(1) construction
    //    ol_setparameter_int(stch("order_ew"), csi->contribution_order_alpha_e);
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }
  else if (csi->type_contribution == "NNLL_NLO"){
    //  ol_setparameter_int(stch("polenorm"), 1); // polenorm prescribed by H^(1) construction
    //    ol_setparameter_int(stch("order_ew"), csi->contribution_order_alpha_e);
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }
  else if (csi->type_contribution == "NNLL_NNLO"){
    //  ol_setparameter_int(stch("polenorm"), 1); // polenorm prescribed by H^(1) construction
    //    ol_setparameter_int(stch("order_ew"), csi->contribution_order_alpha_e);
    munich_ol_setparameter_int("order_ew", csi->contribution_order_alpha_e, logger);
    process_id = register_OL_subprocess(0, 11);
  }

  logger << LOG_DEBUG_VERBOSE << "process_id = " << process_id << " determination finished." << endline_flush;

  if (csi->class_contribution_CS_real){
    process_id = 0;
    for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
      logger << LOG_DEBUG << "csi->dipole[" << i_a << "].process_id = " << csi->dipole[i_a].process_id << endline_flush;
      if (csi->dipole[i_a].process_id == -1){process_id = -1;}
    }
  }


  /*
  logger << LOG_DEBUG << "process_id = " << process_id << endline_flush;
  if (csi->type_contribution != "RA" &&
      csi->type_contribution != "RRA" &&
      csi->type_contribution != "L2RA"){
    if (process_id == -1){logger << LOG_WARN << "Requested amplitudes are not available." << endline_flush;}
  }
  else {
    process_id = 0;
    for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
      logger << LOG_DEBUG << "dipole check: csi->dipole[" << i_a << "].process_id = " << csi->dipole[i_a].process_id << endline_flush;
      if (csi->dipole[i_a].process_id == -1){process_id = -1;}
      // Could be accepted for asi->switch_amplitude_not_required == 0, where non-existing amplitudes are understood as vanishing !!!
      // Possibly remove dipoles in that case !!!
    }
    logger << LOG_DEBUG << "After dipole check: process_id = " << process_id << endline_flush;
  }
  */

  //  ol_start();

  logger.finished();
}


void amplitude_OpenLoops_set::initialization_process_extra(){
  static Logger logger("amplitude_OpenLoops_set::initialization_process_extra");
  logger.called();

  if (csi->type_contribution == "VT" || csi->type_contribution == "VT2" || csi->type_contribution == "CT2"){
    // Do not split off the counterterm in general for all qTsubtraction contributions
    //    ol_setparameter_int(OL_ct_on, 1);
    munich_ol_setparameter_int("CT_on", 1, logger);
    // Should not be required, already initialized as 0
    osi->VA_X_ME2 = 0.;
  }

  logger.finished();
}


void amplitude_OpenLoops_set::initialization_process_integration_check(){
  static Logger logger("amplitude_OpenLoops_set::initialization_process_integration_check");
  logger.called();

  logger << LOG_DEBUG << "process_id = " << process_id << endline_flush;
  if (process_id == -1){
    if (switch_amplitude_not_required){logger << LOG_INFO << "Requested amplitudes have not been found, but are interpreted as vanishing." << endline_flush;}
    else {logger << LOG_INFO << "Requested amplitudes are not available." << endline_flush;}
  }

  if (csi->coll_choice != 0 && (csi->mass_parton[0][1] != 0. || csi->mass_parton[0][2] != 0.)){
    logger << LOG_INFO << "Incoming parton are not allowed for hadron collisions." << endline_flush;
    process_id = -1;
    logger << LOG_DEBUG_VERBOSE << "process_id = " << process_id << endline_flush;
  }

  logger.finished();
}


// specific function for OpenLoops amplitudes
int amplitude_OpenLoops_set::register_OL_subprocess(int i_a, int amptype){
  static Logger logger("amplitude_OpenLoops_set::register_OL_subprocess");
  logger.called();

  // modify photons depending on hadron_type_parton content:
  vector<int> this_type_parton_OL = csi->type_parton[i_a];
  // Select IS photons based on collider type (hadronic IS particles):
  for (int i_p = 1; i_p < 3; i_p++){
    if (this_type_parton_OL[i_p] == 22){
      if (abs(csi->type_hadron[i_p]) == 101){this_type_parton_OL[i_p] = -2002;}
      else if (abs(csi->type_hadron[i_p]) == 22){this_type_parton_OL[i_p] = 2002;}
      else {logger << LOG_ERROR << "Should not happen!" << endline_flush; exit(1);}
    }
  }

  // Select identified FS photons:
  int temp_counter_photon_born = 0;
  for (int i_p = 3; i_p < this_type_parton_OL.size(); i_p++){
    if (this_type_parton_OL[i_p] == 22 && temp_counter_photon_born < csi->n_photon_born){
      temp_counter_photon_born++;
      this_type_parton_OL[i_p] = 2002;
    }
  }

  // Select un-identified FS photons (which can untergo gamma->ffx splittings)
  int temp_counter_jet_born = 0;
  for (int i_p = 3; i_p < this_type_parton_OL.size(); i_p++){
    if (abs(this_type_parton_OL[i_p]) < 7){
      // More precisely: Check partons in jets (top, bottom, etc.) !!!
      temp_counter_jet_born++;
    }
  }
  for (int i_p = 3; i_p < this_type_parton_OL.size(); i_p++){
    if (this_type_parton_OL[i_p] == 22 && temp_counter_jet_born < csi->n_jet_born){
      temp_counter_jet_born++;
      this_type_parton_OL[i_p] = -2002;
    }
  }

  // Remaining photons should be unresolved ones.


  stringstream temp_processname_ss;
  for (int i_p = 1; i_p < this_type_parton_OL.size(); i_p++){
    if (this_type_parton_OL[i_p] == 0){temp_processname_ss << 21;}
    // pdg convention:
    // W+ = +24 !!!
    // W- = -24 !!!
    else if (this_type_parton_OL[i_p] == 24){temp_processname_ss << -24;}
    else if (this_type_parton_OL[i_p] == -24){temp_processname_ss << 24;}

    /*
    /// only off-shell Photonen !!! needs to be controlled elsewhere...
    else if (this_type_parton_OL[i_p] == 22){temp_processname_ss << 22;}
    else if (this_type_parton_OL[i_p] == -22){temp_processname_ss << 22;}
    //    else if (this_type_parton_OL[i_p] == 22){temp_processname_ss << -22;}
    ///
    */
    else {temp_processname_ss << this_type_parton_OL[i_p];}
    if (i_p == 2){temp_processname_ss << " -> ";}
    else if (i_p == this_type_parton_OL.size() - 1){}
    else {temp_processname_ss << " ";}
  }
  string temp_processname = temp_processname_ss.str();

  char * char_processname = new char[temp_processname.size() + 1];
  std::copy(temp_processname.begin(), temp_processname.end(), char_processname);
  char_processname[temp_processname.size()] = '\0';

  logger << LOG_DEBUG << "before registration:   processname = " << setw(20) << char_processname << "   amptype = " << amptype << endline_flush;

  int no_reg = ol_register_process(char_processname, amptype);

  logger << LOG_DEBUG << " after registration:   processname = " << setw(20) << char_processname << "   amptype = " << amptype << "   id = " << no_reg << endline_flush;

  delete [] char_processname;

  if (no_reg == -1 && (msi->CKM_matrix == "trivial" || msi->CKM_matrix == "Cabibbo")){
   stringstream temp_symm_processname_ss;
   vector<int> symm_type_parton = this_type_parton_OL;
    for (int i_p = 1; i_p < symm_type_parton.size(); i_p++){
      if   (symm_type_parton[i_p] == 1){symm_type_parton[i_p] = 3;}
      else if (symm_type_parton[i_p] == 2){symm_type_parton[i_p] = 4;}
      else if (symm_type_parton[i_p] == 3){symm_type_parton[i_p] = 1;}
      else if (symm_type_parton[i_p] == 4){symm_type_parton[i_p] = 2;}
      else if (symm_type_parton[i_p] == -1){symm_type_parton[i_p] = -3;}
      else if (symm_type_parton[i_p] == -2){symm_type_parton[i_p] = -4;}
      else if (symm_type_parton[i_p] == -3){symm_type_parton[i_p] = -1;}
      else if (symm_type_parton[i_p] == -4){symm_type_parton[i_p] = -2;}
    }
    for (int i_p = 1; i_p < symm_type_parton.size(); i_p++){
      if (symm_type_parton[i_p] == 0){temp_symm_processname_ss << 21;}
      // pdf convention:
      // W+ = +24 !!!
      // W- = -24 !!!
      else if (symm_type_parton[i_p] == 24){temp_symm_processname_ss << -24;}
      else if (symm_type_parton[i_p] == -24){temp_symm_processname_ss << 24;}

      // temporary !!! Needs to distinguish between ax (-22) and a (22) !!!
      else if (symm_type_parton[i_p] == 22){temp_symm_processname_ss << -22;}

      else {temp_symm_processname_ss << symm_type_parton[i_p];}
      if (i_p == 2){temp_symm_processname_ss << " -> ";}
      else if (i_p == symm_type_parton.size() - 1){}
      else {temp_symm_processname_ss << " ";}
    }
    string temp_symm_processname = temp_symm_processname_ss.str();
    char * char_symm_processname = new char[temp_symm_processname.size() + 1];
    std::copy(temp_symm_processname.begin(), temp_symm_processname.end(), char_symm_processname);
    char_symm_processname[temp_symm_processname.size()] = '\0';

    logger << LOG_DEBUG << "before registration:   symm_processname = " << setw(20) << char_symm_processname << "   amptype = " << amptype << endline_flush;

    no_reg = ol_register_process(char_symm_processname, amptype);

    logger << LOG_DEBUG << " after registration:   symm_processname = " << setw(20) << char_symm_processname << "   amptype = " << amptype << "   id = " << no_reg << endline_flush;

    delete [] char_symm_processname;
  }

  logger << LOG_DEBUG_VERBOSE << "finished - no_reg = " << no_reg << endline_flush;
  return no_reg;
}


// specific function for OpenLoops amplitudes
void amplitude_OpenLoops_set::generate_testpoint(){
  static Logger logger("amplitude_OpenLoops_set::generate_testpoint");
  logger.called();

  double testpoint_E_CMS = osi->E_CMS;
  if (esi->psi->start_xbsqrts_all[0][esi->psi->xb_max - 4] != 0.){testpoint_E_CMS = esi->psi->start_xbsqrts_all[0][esi->psi->xb_max - 4];}
  else if (esi->psi->sqrtsmin_opt[0][4] < osi->E_CMS / 2){testpoint_E_CMS = osi->E_CMS / 2;}
  else {testpoint_E_CMS = (osi->E_CMS + esi->psi->sqrtsmin_opt[0][4]) / 2;}
  // fails sometimes at present when calling ol_phase_space_point !!!
  static int n_momentum = 5 * (csi->n_particle + 2);
  //  logger << LOG_DEBUG_VERBOSE << "n_momentum = " << n_momentum << endline_flush;
  double *P;
  P = new double[n_momentum];
  //  logger << LOG_DEBUG_VERBOSE << "osi->E_CMS = " << osi->E_CMS << endline_flush;
  ol_phase_space_point(1, testpoint_E_CMS, P);
  for (int i_p = 1; i_p < esi->p_parton[0].size(); i_p++){esi->p_parton[0][i_p] = fourvector(P[5 * (i_p - 1)], P[5 * (i_p - 1) + 1], P[5 * (i_p - 1) + 2], P[5 * (i_p - 1) + 3]);}
  //  logger << LOG_DEBUG_VERBOSE << "esi->p_parton[0][1].pT2() = " << esi->p_parton[0][1].pT2() << endline_flush;
  //  logger << LOG_DEBUG_VERBOSE << "esi->p_parton[0][2].pT2() = " << esi->p_parton[0][2].pT2() << endline_flush;
  // Rotate phase space point if particles 1 and 2 are not parallel to z axis:
  if (esi->p_parton[0][1].pT2() != 0. || esi->p_parton[0][2].pT2() != 0.){
    fourvector rotate_vector = esi->p_parton[0][1];
    for (int i_p = 1; i_p < esi->p_parton[0].size(); i_p++){esi->p_parton[0][i_p] = esi->p_parton[0][i_p].rotation_inverse(rotate_vector);}
    // Avoid numerical noise in incoming momenta:
    for (int i_p = 1; i_p < 3; i_p++){esi->p_parton[0][i_p] = fourvector(esi->p_parton[0][i_p].x0(), 0., 0., esi->p_parton[0][i_p].x3());}
  }
  esi->p_parton[0][0] = esi->p_parton[0][1] + esi->p_parton[0][2];
  esi->psi->x_pdf[0] = pow(testpoint_E_CMS / osi->E_CMS, 2);
  esi->psi->x_pdf[1] = pow(esi->psi->x_pdf[0], 0.6);
  esi->psi->x_pdf[2] = esi->psi->x_pdf[0] / esi->psi->x_pdf[1];
  esi->psi->boost = (esi->psi->x_pdf[1] - esi->psi->x_pdf[2]) / (esi->psi->x_pdf[1] + esi->psi->x_pdf[2]);
  delete [] P;

  if (csi->class_contribution_collinear){
    //        if (esi->p_parton[0][0].x0() != 0.){
    // Add z1 and z2 values for the testpoint
    esi->psi->z_coll[1] = pow(esi->psi->x_pdf[1], 0.4);
    esi->psi->z_coll[2] = pow(esi->psi->x_pdf[2], 0.3);
  }

  logger.finished();
}


void amplitude_OpenLoops_set::output_parameter(string label){
  static Logger logger("amplitude_OpenLoops_set::output_parameter");
  logger.called();

  // To make sure that updated parameters are printed:
  ol_start();
  if (label == ""){
    OLP_PrintParameter(stch("log/olparameters." + csi->subprocess + ".txt"));
  }
  else {
    OLP_PrintParameter(stch("log/olparameters." + csi->subprocess + "." + label + ".txt"));
  }

  logger.finished();
}


void amplitude_OpenLoops_set::munich_ol_setparameter_int(string parameter, int value, Logger & logger){
  //  static Logger logger("amplitude_OpenLoops_set::munich_ol_setparameter_int");
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << " called (amplitude_OpenLoops_set::munich_ol_setparameter_int)" << endline_flush; }

  logger << LOG_DEBUG_POINT << "OpenLoops (int) parameter  " << parameter << "  is set to  " << value << " ." << endline_flush;
  char * temp_ch_parameter = new char[parameter.size() + 1];
  std::copy(parameter.begin(), parameter.end(), temp_ch_parameter);
  temp_ch_parameter[parameter.size()] = '\0';
  ol_setparameter_int(temp_ch_parameter, value);
  delete [] temp_ch_parameter;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "finished (amplitude_OpenLoops_set::munich_ol_setparameter_int)" << endline_flush; }
}

void amplitude_OpenLoops_set::munich_ol_setparameter_double(string parameter, double value, Logger & logger){
  //  static Logger logger("amplitude_OpenLoops_set::munich_ol_setparameter_double");
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << " called (amplitude_OpenLoops_set::munich_ol_setparameter_double)" << endline_flush; }

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops (double) parameter  " << parameter << "  is set to  " << value << " ." << endline_flush; }
  char * temp_ch_parameter = new char[parameter.size() + 1];
  std::copy(parameter.begin(), parameter.end(), temp_ch_parameter);
  temp_ch_parameter[parameter.size()] = '\0';
  ol_setparameter_double(temp_ch_parameter, value);
  delete [] temp_ch_parameter;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "finished (amplitude_OpenLoops_set::munich_ol_setparameter_double)" << endline_flush; }
}

void amplitude_OpenLoops_set::munich_ol_setparameter_string(string parameter, string value, Logger & logger){
  //  static Logger logger("amplitude_OpenLoops_set::munich_ol_setparameter_string");
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << " called (amplitude_OpenLoops_set::munich_ol_setparameter_string)" << endline_flush; }

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops (string) parameter  " << parameter << "  is set to  " << value << " ." << endline_flush; }
  char * temp_ch_parameter = new char[parameter.size() + 1];
  std::copy(parameter.begin(), parameter.end(), temp_ch_parameter);
  temp_ch_parameter[parameter.size()] = '\0';
  char * temp_ch_value = new char[value.size() + 1];
  std::copy(value.begin(), value.end(), temp_ch_value);
  temp_ch_value[value.size()] = '\0';
  ol_setparameter_string(temp_ch_parameter, temp_ch_value);
  delete [] temp_ch_parameter;
  delete [] temp_ch_value;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "finished (amplitude_OpenLoops_set::munich_ol_setparameter_string)" << endline_flush; }
}

void amplitude_OpenLoops_set::munich_ol_getparameter_int(string parameter, int & value, Logger & logger){
  //  static Logger logger("amplitude_OpenLoops_set::munich_ol_getparameter_int");
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << " called (amplitude_OpenLoops_set::munich_ol_getparameter_int)" << endline_flush; }

  char * temp_ch_parameter = new char[parameter.size() + 1];
  std::copy(parameter.begin(), parameter.end(), temp_ch_parameter);
  temp_ch_parameter[parameter.size()] = '\0';
  ol_getparameter_int(temp_ch_parameter, &value);
  delete [] temp_ch_parameter;
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops (int) parameter  " << parameter << "  at value  " << value << "  is extracted." << endline_flush; }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "finished (amplitude_OpenLoops_set::munich_ol_getparameter_int)" << endline_flush; }
}

void amplitude_OpenLoops_set::munich_ol_getparameter_double(string parameter, double & value, Logger & logger){
  //  static Logger logger("amplitude_OpenLoops_set::munich_ol_getparameter_double");
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << " called (amplitude_OpenLoops_set::munich_ol_getparameter_double)" << endline_flush; }

  char * temp_ch_parameter = new char[parameter.size() + 1];
  std::copy(parameter.begin(), parameter.end(), temp_ch_parameter);
  temp_ch_parameter[parameter.size()] = '\0';
  ol_getparameter_double(temp_ch_parameter, &value);
  delete [] temp_ch_parameter;
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops (double) parameter  " << parameter << "  at value  " << value << "  is extracted." << endline_flush; }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "finished (amplitude_OpenLoops_set::munich_ol_getparameter_double)" << endline_flush; }
}

void amplitude_OpenLoops_set::munich_ol_getparameter_string(string parameter, string & value, Logger & logger){
  //  static Logger logger("amplitude_OpenLoops_set::munich_ol_getparameter_string");
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << " called (amplitude_OpenLoops_set::munich_ol_getparameter_string)" << endline_flush; }
  /*
  char * temp_ch_parameter = new char[parameter.size() + 1];
  char * temp_ch_value = new char[128];
  std::copy(parameter.begin(), parameter.end(), temp_ch_parameter);
  temp_ch_parameter[parameter.size()] = '\0';
  ol_getparameter_string(temp_ch_parameter, temp_ch_value);
  delete [] temp_ch_parameter;
  delete [] temp_ch_value;
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << "OpenLoops (double) parameter  " << parameter << "  at value  " << value << "  is extracted." << endline_flush; }
  */
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "finished (amplitude_OpenLoops_set::munich_ol_getparameter_string)" << endline_flush; }
}

#endif
