#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::calculate_ME2_CT2_QCD(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_CT2_QCD");
  logger.called();

  if (osi->QT_finalstate_massive_coloured){logger << LOG_WARN << "Process-specific function needs to be called, but has not been defined!" << endline_flush; exit(1);}

  set_phasespacepoint();
  set_delta_and_mu_Q();
  if (!csi->class_contribution_loopinduced){
    ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
    osi->VA_V_ME2 = M2L1[0];

    if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_B = " << setw(23) << setprecision(15) << osi->VA_b_ME2 << endline_flush;
      logger << LOG_DEBUG_POINT << "OpenLoops:  ME2_VX = " << setw(23) << setprecision(15) << osi->VA_V_ME2 + osi->VA_X_ME2 << endline_flush;
    }

    if (osi->initial_diag || osi->initial_pdf_gq){
      //  osi->VA_V_ME2 = 2 Re < M0 | M1 >  etc.
      //  (alpha_S / pi) normalization of H1
      osi->QT_H1_delta_QCD = (osi->VA_V_ME2 + osi->VA_X_ME2) / osi->VA_b_ME2 / (osi->alpha_S / pi);
      // same translation for osi->QT_H1_delta_QCD as in ppllll24_calculate_H2.
      if (osi->switch_polenorm == 1){
	if (osi->initial_gg){osi->QT_H1_delta_QCD -= pi2_6 * C_A;}
	else if (osi->initial_qqx){osi->QT_H1_delta_QCD -= pi2_6 * C_F;}
	else {logger << LOG_FATAL << "Wrong process specified." << endline_flush; exit(1);}
      }
    }
    else {
      osi->QT_H1_delta_QCD = 0.;
    }
  }
  else {
    logger << LOG_ERROR << "L2CT2 contribution has not been implemented yet." << endline_flush;
    exit(1);
  }

  logger.finished();
}

#endif
