#include "header.hpp"
#ifdef OPENLOOPS

void amplitude_OpenLoops_set::calculate_ME2_VT2_QCD(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_VT2_QCD");
  logger.called();

  if (osi->QT_finalstate_massive_coloured){logger << LOG_WARN << "Check if process-specific function needs to be called instead of this one!" << endline_flush;}

  set_phasespacepoint();
  set_delta_and_mu_Q();
  // should be done only once during initialization !!!
  int polenorm_0 = 0;
  ol_setparameter_int(polenorm, polenorm_0);

  //  ol_evaluate_loop2(1, P_OpenLoops, M2L2, &acc);
  ol_evaluate_full(process_id, P_OpenLoops, &osi->VA_b_ME2, M2L1, IRL1, M2L2, IRL2, &acc);
  osi->VA_V_ME2 = M2L1[0] + osi->VA_X_ME2;
  // ??? One could simply use the sum as usually later... (osi->VA_X_ME2 should be 0 anyway since calculated together with M2L1[0] !!!)

  // The following should only contribute  if (!csi->switch_VTX) !!!
  // double-spin-flip contribution for gg-initiated processes (so far only for QQx production):
  if (osi->initial_gg){
    calculate_ME2_VT2_QCD_H0_doublespinflip();
    calculate_ME2_VT2_QCD_H0_DG();
  }

  if (switch_off_H2 || !osi->initial_diag){
    osi->QT_A0 = osi->VA_b_ME2;
    osi->QT_A1 = osi->VA_V_ME2;
    osi->QT_H1_delta_QCD = osi->QT_A1 / 2 / (osi->alpha_S * inv2pi * osi->QT_A0);
    osi->QT_A2 = 0.;
    osi->QT_H2_delta_QCD = 0.;
    //osi->QT_H2_delta_QCD = osi->QT_A2 / 4 / (pow(osi->alpha_S * inv2pi, 2) * osi->QT_A0);
  }
  else {
    calculate_H2_QCD_2loop();
  }

  logger.finished();
}


void amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_doublespinflip(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_doublespinflip");
  logger.called();

  osi->QT_H0_doublespinflip = 1.;
  double temp_ME2 = 0.;
  double temp_m2schsf = 0.;
  ol_evaluate_schsf(1, P_OpenLoops, &temp_ME2, 1, 2, &temp_m2schsf);
  osi->QT_H0_doublespinflip = temp_m2schsf / temp_ME2;

  logger.finished();
}


void amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_DG(){
  static Logger logger("amplitude_OpenLoops_set::calculate_ME2_VT2_QCD_H0_DG");
  logger.called();

  osi->QT_H0_DG = 0.;

  logger.finished();
}

#endif
