#include "header.hpp"

void amplitude_set::calculate_ME2_CA_QCD(){
  static Logger logger("amplitude_set::calculate_ME2_CA_QCD");
  logger.called();

  logger << LOG_FATAL << "Basic amplitude_set function must be overwritten." << endline_flush; exit(1);

  logger.finished();
}


void amplitude_set::calculate_ME2check_CA_QCD(){
  static Logger logger("amplitude_set::calculate_ME2check_CA_QCD");
  logger.called();

  if (osi->switch_output_comparison && process_id != -1){
    esi->phasespacepoint();
    if (esi->p_parton[0][0].x0() == 0.){generate_testpoint();}
    if (esi->p_parton[0][0].x0() != 0.){
      ofstream out_comparison;
      out_comparison.open(osi->filename_comparison.c_str(), ofstream::out | ofstream::app);

      esi->perform_event_selection();
      osi->set_testpoint_scale_complete();
      if (osi->switch_CV){osi->determine_scale_CA();}
      vector<int> cut_ps_acc_all(csi->n_ps, 0);
      swap(esi->cut_ps, cut_ps_acc_all);

      osi->set_testpoint_g_z_coll();
      calculate_ME2_CA_QCD();
      osi->calculate_collinear_QCD();

      out_comparison << output_amplitude_CA_QCD() << endline;

      osi->psi->set_xz_coll_pdf();
      if (osi->switch_TSV){
	//	osi->calculate_pdf_LHAPDF_CA_collinear_TSV();
	osi->calculate_pdf_LHAPDF_list_TSV();
	out_comparison << osi->output_testpoint_CA() << endline;
	//	out_comparison << osi->output_testpoint_CA_Sherpa_output() << endline;
	out_comparison << osi->output_testpoint_integrand() << endline;
	out_comparison << endline;
      }

      swap(esi->cut_ps, cut_ps_acc_all);
      out_comparison << osi->output_testpoint_kinematics() << endline;
      out_comparison << osi->output_scales() << endline;
      out_comparison.close();
    }
  }
  output_parameter("testpoint");

  logger.finished();
}


string amplitude_set::output_amplitude_CA_QCD(){
  static Logger logger("amplitude_set::output_amplitude_CA_QCD");
  logger.called();

  stringstream out_comparison;
  out_comparison << "Born and colour-correlated matrix elements evaluated at alpha_S(µR = " << setprecision(15) << osi->scale_ren << ") = " << setprecision(15) << osi->alpha_S << "" << endline << endline;
  //  out_comparison << "Born and colour-correlated matrix elements (µR = " << osi->scale_ren << ", " << "alpha_S(µR) = " << setprecision(15) << osi->alpha_S << ")" << endline << endline;
  for (int i_c = 0; i_c < (*osi->CA_collinear).size(); i_c++){
    for (int j_c = 0; j_c < (*osi->CA_collinear)[i_c].size(); j_c++){
      out_comparison << "Collinear " << i_c <<  ", " << j_c << ":   ";
      out_comparison << left << setw(26) << (*osi->CA_collinear)[i_c][j_c].name() << ": " << "ME2_cf[" << i_c << "][" << j_c << "] = " << setw(23) << setprecision(15) << osi->CA_ME2_cf[i_c][j_c];
      if (j_c > 0){out_comparison << "   ratio to Born = " << setw(23) << setprecision(15) << osi->CA_ME2_cf[i_c][j_c] / osi->CA_ME2_cf[i_c][0];}
      out_comparison << endline;
      //"   charge_factor = " << setw(15) << setprecision(8) << (*osi->CA_collinear)[i_c][j_c].charge_factor() << endline;
    }
    out_comparison << endline;
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << out_comparison.str() << endline_flush; }
  logger.finished();
  return out_comparison.str();
}


