#include "header.hpp"

void amplitude_set::calculate_ME2_RA_QEW(){
  static Logger logger("amplitude_set::calculate_ME2_RA_QEW");
  logger.called();

  logger << LOG_FATAL << "Basic amplitude_set function must be overwritten." << endline_flush; exit(1);

  logger.finished();
}


void amplitude_set::calculate_dipole_Acc_QEW(int x_a){
  static Logger logger("amplitude_set::calculate_dipole_Acc_QEW");
  logger.called();

  logger << LOG_FATAL << "Basic amplitude_set function must be overwritten." << endline_flush; exit(1);

  logger.finished();
}


void amplitude_set::calculate_dipole_Asc_QEW(int x_a, fourvector & Vtensor){
  static Logger logger("amplitude_set::calculate_dipole_Asc_QEW");
  logger.called();

  logger << LOG_FATAL << "Basic amplitude_set function must be overwritten." << endline_flush; exit(1);

  logger.finished();
}


void amplitude_set::calculate_ME2check_RA_QEW(phasespace_set * psi){
  static Logger logger("amplitude_set::calculate_ME2check_RA_QEW");
  logger.called();

  if (osi->switch_output_comparison && process_id != -1){
    esi->phasespacepoint();
    if (esi->p_parton[0][0].x0() == 0.){generate_testpoint();}
    if (esi->p_parton[0][0].x0() != 0.){
      ofstream out_comparison;
      out_comparison.open(osi->filename_comparison.c_str(), ofstream::out | ofstream::app);

      osi->set_testpoint_psp_dipoles();
      esi->perform_event_selection();
      osi->set_testpoint_scale_complete_RA();
      vector<int> cut_ps_acc_all(csi->n_ps, 0);
      swap(esi->cut_ps, cut_ps_acc_all);

      calculate_ME2_RA_QEW();

      out_comparison << output_amplitude_RA_QEW() << endline;
      out_comparison << osi->output_testpoint_RA() << endline;

      osi->set_testpoint_x_pdf();
      if (osi->switch_TSV){
	osi->calculate_pdf_LHAPDF_list_TSV();
	//	osi->calculate_pdf_LHAPDF_TSV();
	out_comparison << osi->output_testpoint_integrand() << endline;
	out_comparison << endline;
      }

      swap(esi->cut_ps, cut_ps_acc_all);
      out_comparison << osi->output_testpoint_kinematics() << endline;
      out_comparison << osi->output_scales() << endline;
      out_comparison.close();
    }
  }

  output_parameter("testpoint");

  logger.finished();
}


string amplitude_set::output_amplitude_RA_QEW(){
  static Logger logger("amplitude_set::output_amplitude_RA_QEW");
  logger.called();

  stringstream out_comparison;

  out_comparison << "Real-correction and reduced charge/spin-charge correlated matrix elements evaluated at alpha_S(µR = " << setprecision(15) << osi->scale_ren << ") = " << setprecision(15) << osi->alpha_S << endline << endline;
  //  out_comparison << "Real-correction and reduced charge/spin-charge correlated matrix elements (µR = " << osi->scale_ren << ", " << "alpha_S(µR) = " << setprecision(15) << osi->alpha_S << ")" << endline << endline;

  out_comparison << "Real:        ";
  out_comparison << left << setw(26) << csi->dipole[0].name() << ":   " << "RA_ME2[0]     = " << setw(23) << setprecision(15) << osi->RA_ME2[0] << endline;
  out_comparison << endline;
  for (int i_a = 1; i_a < csi->n_ps; i_a++){
    out_comparison << "Dipole " << setw(2) << right << i_a <<  ":   ";
    out_comparison << left << setw(26) << csi->dipole[i_a].name() << ":   " << "RA_ME2_cc[" << setw(2) << right << i_a << "] = " << left << setw(23) << setprecision(15) << osi->RA_ME2_cc[i_a];
    if (csi->dipole[i_a].type_splitting() < 2){
      out_comparison << "   RA_ME2_sc[" << right << i_a << "] = " << left << setw(23) << setprecision(15) << osi->RA_ME2_sc[i_a];
    }
    out_comparison << endline;
  }
  out_comparison << endline;

   if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << out_comparison.str() << endline_flush; }
  logger.finished();
  return out_comparison.str();
}

