#include "header.hpp"

void amplitude_set::calculate_ME2_VT_QCD(){
  static Logger logger("amplitude_set::calculate_ME2_VT_QCD");
  logger.called();

  logger << LOG_FATAL << "Basic amplitude_set function must be overwritten." << endline_flush; exit(1);

  logger.finished();
}


void amplitude_set::calculate_H1_delta_QCD_from_VX_b(){
  static Logger logger("amplitude_set::calculate_H1_delta_QCD_from_VX_b");
  logger.called();

  // QT_H1_delta_QCD should not be required for gq channel at NLO !!!
  if (osi->initial_diag || (osi->initial_pdf_gq && csi->class_contribution_order == 2)){
  //  if (osi->initial_diag || osi->initial_pdf_gq){
    //  osi->VA_V_ME2 = 2 Re < M0 | M1 >  etc.
    osi->QT_H1_delta_QCD = (osi->VA_V_ME2 + osi->VA_X_ME2) / osi->VA_b_ME2 / (osi->alpha_S / pi);
    //  (alpha_S / pi) normalization of H1
    if (osi->switch_polenorm == 1){
      if (osi->initial_gg){osi->QT_H1_delta_QCD -= pi2_6 * C_A;}
      else if (osi->initial_qqx){osi->QT_H1_delta_QCD -= pi2_6 * C_F;}
      else {logger << LOG_FATAL << "Wrong process specified." << endline_flush; exit(1);}
    }
  }
  else {
    osi->QT_H1_delta_QCD = 0.;
  }

  logger.finished();
}


void amplitude_set::calculate_ME2check_VT_QCD(){
  static Logger logger("amplitude_set::calculate_ME2check_VT_QCD");
  logger.called();

  if (osi->switch_output_comparison && process_id != -1){
    esi->phasespacepoint();
    if (esi->p_parton[0][0].x0() == 0.){generate_testpoint();}
    if (esi->p_parton[0][0].x0() != 0.){
      ofstream out_comparison;
      out_comparison.open(osi->filename_comparison.c_str(), ofstream::out | ofstream::app);

      esi->perform_event_selection();
      osi->set_testpoint_scale_complete();
      vector<int> cut_ps_acc_all(csi->n_ps, 0);
      swap(esi->cut_ps, cut_ps_acc_all);

      osi->VA_delta_flag = 1;
      string sDelta;
      osi->VA_b_ME2 = 0.;
      osi->VA_V_ME2 = 0.;
      osi->VA_X_ME2 = 0.;
      osi->VA_I_ME2 = 0.;
      osi->QT_A0 = 0.;
      osi->QT_A1 = 0.;
      osi->QT_H1_delta_QCD = 0.;
      for (int i_s = 0; i_s < osi->n_scales_CV; i_s++){osi->VA_X_ME2_CV[i_s] = 0.;}
      osi->VA_DeltaUV = 0.;
      osi->VA_DeltaIR1 = 0.;
      osi->VA_DeltaIR2 = 0.;

      calculate_ME2_VT_QCD();
      osi->psi->set_xz_coll_pdf();
      osi->set_testpoint_g_z_coll();
      osi->set_testpoint_psp();

      if (osi->switch_CV){osi->calculate_pdf_LHAPDF_list_CV();}
      if (osi->switch_TSV){osi->calculate_pdf_LHAPDF_list_TSV();}
      osi->determine_integrand_CX_ncollinear_VT_QCD();
      
      out_comparison << output_amplitude_VT_QCD();
      out_comparison << osi->output_ME2_VT_QCD();

      swap(esi->cut_ps, cut_ps_acc_all);
      out_comparison << osi->output_testpoint_kinematics() << endline;
      out_comparison << osi->output_scales() << endline;
      out_comparison.close();
    }
  }
  output_parameter("testpoint");
  
  logger.finished();
}


#define Fcw left << setw(17)
#define FldXV setprecision(15) << setw(23) << left

string amplitude_set::output_amplitude_VT_QCD(){
  static Logger logger("amplitude_set::output_amplitude_VT_QCD");
  logger.called();

  stringstream out_comparison;
  out_comparison << "Matrix elements (µreg = " << mu_reg << ", µR = " << mu_reg << ", " << "alpha_S(µR = " << osi->scale_ren << ") = " << setprecision(15) << osi->alpha_S << ")" << endline;
  out_comparison << endline;
  out_comparison << Fcw << "ME2_born" << " = " << FldXV << osi->VA_b_ME2 << endline;
  out_comparison << Fcw << "ME2_V+X" << " = " << FldXV << (osi->VA_V_ME2 + osi->VA_X_ME2) << endline;
  out_comparison << endline;
  out_comparison << Fcw << "ME2_V" << " = " << FldXV << osi->VA_V_ME2 << endline;
  out_comparison << Fcw << "ME2_X" << " = " << FldXV << osi->VA_X_ME2 << endline;
  out_comparison << endline;
  out_comparison << Fcw << "H1_delta_QCD" << " = " << FldXV << osi->QT_H1_delta_QCD << endline;
  out_comparison << endline;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << out_comparison.str() << endline_flush; }
  logger.finished();
  return out_comparison.str();
}
