#include "header.hpp"

qTsubtraction_basic::qTsubtraction_basic(){}

qTsubtraction_basic::qTsubtraction_basic(inputparameter_set * isi, contribution_set * _csi, event_set * _esi, phasespace_set * _psi){
  static Logger logger("qTsubtraction_basic::initialization_qTcut (isi, csi, esi, psi)");
  logger.called();

  initialization(isi, _csi, _esi, _psi);

  logger.finished();
}


void qTsubtraction_basic::initialization(inputparameter_set * isi, contribution_set * _csi, event_set * _esi, phasespace_set * _psi){
  static Logger logger("qTsubtraction_basic::initialization");
  logger.called();

  csi = _csi;
  esi = _esi;
  psi = _psi;

  initialization_set_default();
  initialization_input(isi);
  initialization_after_input();

  logger.finished();
}


void qTsubtraction_basic::initialization_set_default(){
  static Logger logger("qTsubtraction_basic::initialization_set_default");
  logger.called();

  /////////////////////////////////
  //  qT-subtraction parameters  //
  /////////////////////////////////

  switch_qTcut = 0;
  n_qTcut = 0;
  min_qTcut = 0.;
  step_qTcut = 0.;
  max_qTcut = 0.;
  binning_qTcut = "linear";
  selection_qTcut = "";

  selection_qTcut_distribution = "";
  selection_no_qTcut_distribution = "";
  selection_qTcut_result = "";
  selection_no_qTcut_result = "";
  selection_qTcut_integration = "";
  selection_no_qTcut_integration = "";

  QT_Q = nullvector;
  QT_QT = 0.;
  QT_sqrtQ2 = 0.;

  logger.finished();
}


void qTsubtraction_basic::initialization_input(inputparameter_set * isi){
  static Logger logger("qTsubtraction_basic::initialization_input");
  logger.called();

  for (int i_i = 0; i_i < isi->input_observable.size(); i_i++){
    logger << LOG_DEBUG_VERBOSE << "isi->input_observable[" << i_i << "] = " << isi->input_observable[i_i] << endline_flush;

    if (isi->input_observable[i_i].variable == ""){}

    /////////////////////////////////
    //  qT-subtraction parameters  //
    /////////////////////////////////

    // needed elsewhere ???
    else if (isi->input_observable[i_i].variable == "switch_qTcut"){switch_qTcut = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "n_qTcut"){n_qTcut = atoi(isi->input_observable[i_i].value.c_str());}
    // needed elsewhere ???
    else if (isi->input_observable[i_i].variable == "min_qTcut"){min_qTcut = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "step_qTcut"){step_qTcut = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "max_qTcut"){max_qTcut = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "binning_qTcut"){binning_qTcut = isi->input_observable[i_i].value;}
    else if (isi->input_observable[i_i].variable == "selection_qTcut"){
      if (isi->input_observable[i_i].value == "clear"){selection_qTcut = "";}
      else {selection_qTcut = isi->input_observable[i_i].value;}
    }
    else if (isi->input_observable[i_i].variable == "selection_qTcut_distribution"){
      if (isi->input_observable[i_i].value == "clear"){selection_qTcut_distribution = "";}
      else {selection_qTcut_distribution = isi->input_observable[i_i].value;}
    }
    else if (isi->input_observable[i_i].variable == "selection_no_qTcut_distribution"){
      if (isi->input_observable[i_i].value == "clear"){selection_no_qTcut_distribution = "";}
      else {selection_no_qTcut_distribution = isi->input_observable[i_i].value;}
    }
    else if (isi->input_observable[i_i].variable == "selection_qTcut_result"){
      if (isi->input_observable[i_i].value == "clear"){selection_qTcut_result = "";}
      else {selection_qTcut_result = isi->input_observable[i_i].value;}
    }
    else if (isi->input_observable[i_i].variable == "selection_no_qTcut_result"){
      if (isi->input_observable[i_i].value == "clear"){selection_no_qTcut_result = "";}
      else {selection_no_qTcut_result = isi->input_observable[i_i].value;}
    }
    else if (isi->input_observable[i_i].variable == "selection_qTcut_integration"){
      if (isi->input_observable[i_i].value == "clear"){selection_qTcut_integration = "";}
      else {selection_qTcut_integration = isi->input_observable[i_i].value;}
    }
    else if (isi->input_observable[i_i].variable == "selection_no_qTcut_integration"){
      if (isi->input_observable[i_i].value == "clear"){selection_no_qTcut_integration = "";}
      else {selection_no_qTcut_integration = isi->input_observable[i_i].value;}
    }
  }

  logger.finished();
}


void qTsubtraction_basic::initialization_after_input(){
  static Logger logger("qTsubtraction_basic::initialization_after_input");
  logger.called();

  logger << LOG_DEBUG << setw(40) << "switch_qTcut" << "     =     " << setw(20) << switch_qTcut << endline_flush;

  if (!switch_qTcut){
    n_qTcut = 1;
    min_qTcut = 0;
    step_qTcut = 0;
    max_qTcut = 0;
    value_qTcut = vector<double> (1, 0.);
    logger << LOG_DEBUG << "No qTcut variation applied!" << endline_flush;
    logger << LOG_DEBUG << setw(40) << "switch_qTcut" << "     =     " << setw(20) << switch_qTcut << endline_flush;
    logger << LOG_DEBUG << setw(40) << "n_qTcut" << "     =     " << setw(20) << n_qTcut << endline_flush;
    logger << LOG_DEBUG << setw(40) << "min_qTcut" << "     =     " << setw(20) << min_qTcut << endline_flush;
    logger << LOG_DEBUG << setw(40) << "max_qTcut" << "     =     " << setw(20) << max_qTcut << endline_flush;
    logger << LOG_DEBUG << setw(40) << "step_qTcut" << "     =     " << setw(20) << step_qTcut << endline_flush;
    logger << LOG_DEBUG << setw(40) << "binning_qTcut" << "     =     " << setw(20) << binning_qTcut << endline_flush;
    logger << LOG_DEBUG << setw(40) << "selection_qTcut" << "     =     " << setw(20) << selection_qTcut << endline_flush;
  }
  else {
    if (switch_qTcut && !n_qTcut){logger << LOG_FATAL << setw(40) << "switch_qTcut = " << switch_qTcut << "   n_qTcut = " << n_qTcut << "  is not a valid setting." << endline_flush; exit(1);}

    if (binning_qTcut == "linear"){
      // possible ways to set input parameters:
      // min_qTcut, n_qTcut, step_qTcut, max_qTcut
      if (min_qTcut != 0. && n_qTcut != 0 && step_qTcut != 0. && max_qTcut != 0.){
	double temp_max_qTcut = min_qTcut + (n_qTcut - 1) * step_qTcut;
	if (max_qTcut != temp_max_qTcut){logger << LOG_FATAL << "Inconsistently over-defined qTcut input!" << endline_flush; exit(1);}
      }
      // min_qTcut, n_qTcut, step_qTcut
      else if (min_qTcut != 0. && n_qTcut != 0 && step_qTcut != 0.){
	max_qTcut = min_qTcut + (n_qTcut - 1) * step_qTcut;
      }
      // min_qTcut, n_qTcut, max_qTcut
      else if (min_qTcut != 0. && n_qTcut != 0 && max_qTcut != 0.){
	step_qTcut = (max_qTcut - min_qTcut) / (n_qTcut - 1);
      }
      // min_qTcut, step_qTcut, max_qTcut
      else if (min_qTcut != 0. && step_qTcut != 0. && max_qTcut != 0.){
	n_qTcut = (max_qTcut - min_qTcut) / step_qTcut + 1;
      }
      // n_qTcut, step_qTcut, max_qTcut
      else if (min_qTcut != 0. && n_qTcut != 0 && step_qTcut != 0. && max_qTcut != 0.){
	min_qTcut = max_qTcut - (n_qTcut - 1) * step_qTcut;
      }
      else {
	logger << LOG_FATAL << "Inconsistent qTcut input for linear binning!" << endline_flush; exit(1);
      }
      for (int i_q = 0; i_q < n_qTcut; i_q++){
	value_qTcut.push_back(min_qTcut + i_q * step_qTcut);
      }
    }

    else if (binning_qTcut == "logarithmic"){
      // possible ways to set input parameters:
      // min_qTcut, n_qTcut, max_qTcut
      if (min_qTcut != 0. && n_qTcut != 0 && max_qTcut != 0.){
	step_qTcut = 0.; // no constant step width in logarithmic binning!
      }
      else {
	logger << LOG_FATAL << "Inconsistent qTcut input for logarithmic binning!" << endline_flush; exit(1);
      }
      for (int i_q = 0; i_q < n_qTcut; i_q++){
	value_qTcut.push_back(min_qTcut * exp10(log10(max_qTcut / min_qTcut) * double(i_q) / (n_qTcut - 1)));
      }
      value_qTcut[n_qTcut - 1] = max_qTcut;
    }

    else if (binning_qTcut == "irregular"){
      // possible ways to set input parameters:

      // selection_qTcut
      if (selection_qTcut != ""){
	vector<string> vs_selection(1);
	for (int i_b = 0; i_b < selection_qTcut.size(); i_b++){
	  logger << LOG_DEBUG_VERBOSE << "i_b = " << i_b << "   vs_selection.size() = " << vs_selection.size() << endline_flush;
	  if (selection_qTcut[i_b] == ':'){vs_selection.push_back("");}
	  else if (selection_qTcut[i_b] != ':'){vs_selection[vs_selection.size() - 1].push_back(selection_qTcut[i_b]);}
	}
	n_qTcut = vs_selection.size();
	value_qTcut.resize(n_qTcut);
	for (int i_b = 0; i_b < value_qTcut.size(); i_b++){
	  logger << LOG_DEBUG_VERBOSE << "vs_selection[" << i_b << "] = " << setprecision(8) << setw(15) << vs_selection[i_b] << endline_flush;
	  value_qTcut[i_b] = atof(vs_selection[i_b].c_str());
	}
	min_qTcut = value_qTcut[0];
	max_qTcut = value_qTcut[n_qTcut - 1];
      }
      else {
	logger << LOG_FATAL << "Inconsistent qTcut input for logarithmic binning!" << endline_flush; exit(1);
      }
    }
  }

  logger << LOG_DEBUG << setw(40) << "n_qTcut" << "     =     " << setw(20) << n_qTcut << endline_flush;
  logger << LOG_DEBUG << setw(40) << "min_qTcut" << "     =     " << setw(20) << min_qTcut << endline_flush;
  logger << LOG_DEBUG << setw(40) << "max_qTcut" << "     =     " << setw(20) << max_qTcut << endline_flush;
  logger << LOG_DEBUG << setw(40) << "step_qTcut" << "     =     " << setw(20) << step_qTcut << endline_flush;
  logger << LOG_DEBUG << setw(40) << "binning_qTcut" << "     =     " << setw(20) << binning_qTcut << endline_flush;
  logger << LOG_DEBUG << setw(40) << "selection_qTcut" << "     =     " << setw(20) << selection_qTcut << endline_flush;

  logger << LOG_DEBUG << setw(40) << "value_qTcut.size()" << "     =     " << setw(20) << value_qTcut.size() << endline_flush;
  for (int i_q = 0; i_q < value_qTcut.size(); i_q++){logger << LOG_DEBUG << "value_qTcut[" << setw(3) << i_q << "] = " << value_qTcut[i_q] << endline_flush;}

  logger << LOG_DEBUG << "csi->type_contribution = " << csi->type_contribution << endline_flush;

  if (csi->type_contribution == "" ||
      csi->type_contribution == "all" ||
      csi->type_contribution == "CT" ||
      csi->type_contribution == "CJ" ||
      csi->type_contribution == "RT" ||
      csi->type_contribution == "CT2" ||
      csi->type_contribution == "CJ2" ||
      csi->type_contribution == "RVA" ||
      csi->type_contribution == "RCA" ||
      csi->type_contribution == "RRA" ||
      csi->type_contribution == "L2RT" ||
      csi->type_contribution == "L2CT"){
    active_qTcut = 1;
    output_n_qTcut = n_qTcut;
  }
  else {
    active_qTcut = 0;
    output_n_qTcut = 1;
  }

  logger << LOG_DEBUG << "output_n_qTcut = " << output_n_qTcut << endline_flush;
  logger << LOG_DEBUG << "active_qTcut = " << active_qTcut << endline_flush;
  logger << LOG_DEBUG << "selection_qTcut_distribution =    " << selection_qTcut_distribution << endline_flush;
  logger << LOG_DEBUG << "selection_no_qTcut_distribution = " << selection_no_qTcut_distribution << endline_flush;
  logger << LOG_DEBUG << "selection_qTcut_result =    " << selection_qTcut_result << endline_flush;
  logger << LOG_DEBUG << "selection_no_qTcut_result = " << selection_no_qTcut_result << endline_flush;
  logger << LOG_DEBUG << "selection_qTcut_integration =    " << selection_qTcut_integration << endline_flush;
  logger << LOG_DEBUG << "selection_no_qTcut_integration = " << selection_no_qTcut_integration << endline_flush;

  
  determination_rcut_values("distribution", selection_qTcut_distribution, selection_no_qTcut_distribution, no_qTcut_distribution, value_qTcut_distribution);
  determination_rcut_values("result", selection_qTcut_result, selection_no_qTcut_result, no_qTcut_result, value_qTcut_result);
  determination_rcut_values("integration", selection_qTcut_integration, selection_no_qTcut_integration, no_qTcut_integration, value_qTcut_integration);
  
  if (csi->class_contribution_qTcut_explicit){
    counter_killed_qTcut.resize(n_qTcut, 0);
    counter_acc_qTcut.resize(n_qTcut, 0);
  }

  logger.finished();
}


void qTsubtraction_basic::determination_rcut_values(string this_name, string & this_selection_qTcut, string & this_selection_no_qTcut, vector<int> & this_no_qTcut, vector<double> & this_value_qTcut){
  static Logger logger("qTsubtraction_basic::determination_rcut_values");
  logger.called();

  if (active_qTcut){
    if (this_selection_qTcut == "" && this_selection_no_qTcut == ""){
      this_value_qTcut.resize(1, 0.);
      this_no_qTcut.resize(1, 0);
      logger << LOG_DEBUG << this_name << " output generated for lowest qTcut value." << endline_flush;
    }
    else if (this_selection_qTcut == "all" || this_selection_no_qTcut == "all"){
      this_value_qTcut = value_qTcut;
      this_no_qTcut.resize(n_qTcut);
      for (int i_q = 0; i_q < n_qTcut; i_q++){this_no_qTcut[i_q] = i_q;}
      logger << LOG_DEBUG << this_name << " output generated for all qTcut values." << endline_flush;
    }
    else if (this_selection_qTcut != "" && this_selection_no_qTcut == ""){
      logger << LOG_DEBUG << this_name << " output generated for selected qTcut values." << endline_flush;

      vector<string> vs_selection(1);
      for (int i_q = 0; i_q < this_selection_qTcut.size(); i_q++){
	logger << LOG_DEBUG_VERBOSE << "i_q = " << i_q << "   vs_selection.size() = " << vs_selection.size() << endline_flush;
	if (this_selection_qTcut[i_q] == ':'){vs_selection.push_back("");}
	else if (this_selection_qTcut[i_q] != ':'){vs_selection[vs_selection.size() - 1].push_back(this_selection_qTcut[i_q]);}
      }
      this_no_qTcut.resize(vs_selection.size());
      this_value_qTcut.resize(vs_selection.size());

      for (int i_q = 0; i_q < this_value_qTcut.size(); i_q++){
	logger << LOG_DEBUG_VERBOSE << "vs_selection[" << i_q << "] = " << setprecision(8) << setw(15) << vs_selection[i_q] << endline_flush;
	this_value_qTcut[i_q] = atof(vs_selection[i_q].c_str());
	int flag = n_qTcut;
	for (int i_q = 0; i_q < n_qTcut; i_q++){
	  if (abs(this_value_qTcut[i_q] - value_qTcut[i_q]) < 1.e-12 * value_qTcut[i_q]){flag = i_q; break;}
	}
	if (flag == n_qTcut){this_value_qTcut.erase(this_value_qTcut.begin() + i_q); i_q--;}
	else {this_no_qTcut[i_q] = flag;}
      }
    }

    else if (this_selection_qTcut == "" && this_selection_no_qTcut != ""){
      logger << LOG_DEBUG << this_name << " output generated for selected no_qTcut values." << endline_flush;
      vector<string> vs_selection(1);
      for (int i_q = 0; i_q < this_selection_no_qTcut.size(); i_q++){
	logger << LOG_DEBUG_VERBOSE << "i_q = " << i_q << "   vs_selection.size() = " << vs_selection.size() << endline_flush;
	if (this_selection_no_qTcut[i_q] == ':'){vs_selection.push_back("");}
	else if (this_selection_no_qTcut[i_q] != ':'){vs_selection[vs_selection.size() - 1].push_back(this_selection_no_qTcut[i_q]);}
      }
      this_no_qTcut.resize(vs_selection.size());
      this_value_qTcut.resize(vs_selection.size());

      for (int i_q = 0; i_q < this_no_qTcut.size(); i_q++){
	logger << LOG_DEBUG << "vs_selection[" << i_q << "] = " << setprecision(8) << setw(15) << vs_selection[i_q] << endline_flush;
	this_no_qTcut[i_q] = atoi(vs_selection[i_q].c_str());
	this_value_qTcut[i_q] = value_qTcut[this_no_qTcut[i_q]];
      }
    }

    else if (this_selection_qTcut != "" && this_selection_no_qTcut != ""){
      logger << LOG_FATAL << "Inconsistent qTcut input for " << this_name << "!" << endline_flush; exit(1);
    }

    logger << LOG_DEBUG << this_name << "   this_value_qTcut.size() = " << this_value_qTcut.size() << endline_flush;
    for (int i_q = 0; i_q < this_value_qTcut.size(); i_q++){
      logger << LOG_DEBUG << "value_qTcut_" << this_name << "[" << i_q << "] = " << setw(15) << setprecision(8) << this_value_qTcut[i_q] << " -> " << setw(3) << this_no_qTcut[i_q] << endline_flush;
    }
  }

  else {
    this_value_qTcut.resize(1, 0.);
    this_no_qTcut.resize(1, 0);
    for (int i_q = 0; i_q < this_value_qTcut.size(); i_q++){
      logger << LOG_DEBUG << "value_qTcut_" << this_name << "[" << i_q << "] = " << setw(15) << setprecision(8) << this_value_qTcut[i_q] << " -> " << setw(3) << this_no_qTcut[i_q] << endline_flush;
    }
  }

  logger.finished();
}


void qTsubtraction_basic::output(){
  static Logger logger("qTsubtraction_basic::output");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "&csi - csi = " << csi << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "csi->n_particle_born = " << csi->n_particle_born << endline_flush;

  logger.finished();
}
