#include "header.hpp"

runresumption_set::runresumption_set(){}

runresumption_set::runresumption_set(observable_set & _osi, phasespace_set & _psi, observable_set & _save_osi, phasespace_set & _save_psi){
  static Logger logger("runresumption_set::runresumption_set");
  logger.called();

  osi = &_osi;
  psi = &_psi;

  resumption_osi = &_save_osi;
  resumption_psi = &_save_psi;

  logger.finished();
}


void runresumption_set::perform_proceeding_step(){
  static Logger logger("runresumption_set::perform_proceeding_step");
  logger.called();

  *resumption_osi = *osi;
  *resumption_psi = *psi;

  logger.finished();
}


void runresumption_set::perform_proceeding_in(){
  static Logger logger("runresumption_set::perform_proceeding_in");
  logger.called();

  // psi dependence should be treated via the respective pointer in resumption_osi !!!
  (*osi).perform_proceeding_in();

  logger.finished();
}


void runresumption_set::perform_proceeding_out(){
  static Logger logger("runresumption_set::perform_proceeding_out");
  logger.called();

  // Should depend on value of switch_proceeding when used !!!
  // Check addresses of osi, psi and resumption_osi, resumption_psi !!!
  (*resumption_osi).perform_proceeding_out();

  logger.finished();

}


void runresumption_set::perform_proceeding_check(){
  static Logger logger("runresumption_set::perform_proceeding_check");
  logger.called();

  (*osi).perform_proceeding_check();

  logger.finished();
}


void runresumption_set::perform_iteration_step(){
  static Logger logger("runresumption_set::perform_iteration_step");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){ logger << LOG_DEBUG_POINT << left << "PSP started: " << "   i_gen = " << setw(12) << psi->i_gen << "   i_acc = " << setw(12) << psi->i_acc << "   i_rej = " << setw(12) << psi->i_rej << "   i_nan = " << setw(4) << psi->i_nan << "   i_tec = " << setw(4) << psi->i_tec << endline_flush; }

  // Add a feature to perform an integration/optimization step earlier if a certain time has passed or a certain number of events (relative to n_step) has been generated (low acceptance):
  if (osi->switch_enforce_step_efficiency){
    psi->enforce_step = false;
    //    logger << LOG_INFO << "psi->i_gen_step = " << psi->i_gen_step << " =?= " << psi->n_step * osi->enforce_step_efficiency << " = psi->n_step * osi->enforce_step_efficiency" << endline_flush;
    
    if (psi->i_gen_step == psi->n_step * osi->enforce_step_efficiency){
      psi->enforce_step = true;
      if (Log::getLogThreshold() <= LOG_DEBUG_STEP){
	logger << LOG_DEBUG_STEP << "psi->i_acc_step" << " = " << setw(12) << psi->i_acc_step << "   " << "psi->i_gen_step" << " = " << setw(12) << psi->i_gen_step <<  "   " << "psi->i_step" << " = " << setw(12) << psi->i_step << endline_flush;
	logger << LOG_DEBUG_STEP << "psi->i_acc" << " = " << setw(12) << psi->i_acc << "   " << "psi->i_gen" << " = " << setw(12) << psi->i_gen << endline_flush;
      }
    }
  }
  if (osi->switch_enforce_step_time){
    if (clock() / CLOCKS_PER_SEC - osi->start_step/CLOCKS_PER_SEC >= osi->enforce_step_time){psi->enforce_step = true;}
    // condition on runtime per step
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ 
    logger << LOG_DEBUG_VERBOSE << "*psi->i_step_mode = " << setw(12) << *psi->i_step_mode << "   " << setw(12) << psi->i_gen << " [" << setw(3) << psi->i_nan << "] " << " " << setw(10) << psi->i_acc << " (" << setw(7) << psi->i_tec << ") " << " " << setw(10) << psi->i_rej << " " << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "(*psi->i_step_mode == psi->n_step) = " << (*psi->i_step_mode == psi->n_step) << endline_flush;
  }

  if (*psi->i_step_mode == psi->n_step || psi->enforce_step){

    if (psi->i_acc == 0){
      logger << LOG_FATAL << "Apparently, the complete phase space is cut with the given input." << endline_flush;
      osi->int_end = 1;
      psi->enforce_end_optimization();
    }

    psi->i_step++;

    if (Log::getLogThreshold() <= LOG_DEBUG_STEP){ logger << LOG_DEBUG_STEP << "psi->i_step = " << setw(12) << psi->i_step << "   psi->i_gen_step = " << setw(12) << psi->i_gen_step << "   psi->i_acc_step = " << setw(10) << psi->i_acc_step << "   psi->enforce_step = " << psi->enforce_step << endline_flush; }

    if (osi->switch_output_proceeding == 2){perform_proceeding_out();}
    osi->calculate_intermediate_result();

    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "osi->perform_integration_step_complete" << endline_flush; }
    osi->perform_integration_step_complete();
    if (osi->switch_output_distribution){osi->output_distribution_complete();}

    if (Log::getLogThreshold() <= LOG_DEBUG_STEP){
      if (osi->csi->class_contribution_qTcut_explicit){
	int temp_killed_total = accumulate(osi->esi->qT_basic->counter_killed_qTcut.begin(), osi->esi->qT_basic->counter_killed_qTcut.end(), 0);
	if (temp_killed_total == 0){
	  logger << LOG_DEBUG_STEP << "No phase-space points were killed due to instabilities at " << psi->i_acc << " events." << endline_flush;
	}
	else {
	  logger << LOG_DEBUG_STEP << "Information about killed phase-space points at " << psi->i_acc << " events:" << endline_flush;
	  for (int i_q = 0; i_q < osi->n_qTcut; i_q++){
	    if (osi->esi->qT_basic->counter_acc_qTcut[i_q] > 0){
	      logger << LOG_DEBUG_STEP << "counter_killed_qTcut[" << setw(3) << i_q << "] = " << setw(8) << osi->esi->qT_basic->counter_killed_qTcut[i_q] << "   counter_acc_qTcut[" << setw(3) << i_q << "] = " << setw(8) << osi->esi->qT_basic->counter_acc_qTcut[i_q] << "   ratio = " << double(osi->esi->qT_basic->counter_killed_qTcut[i_q]) / double(osi->esi->qT_basic->counter_acc_qTcut[i_q]) << endline_flush;
	    }
	  }
	}
      }
    }

    psi->step_optimization_complete();
    // to be implemented !!! better distinction for normal (non-grid) runs !!!
    // maybe psi->end_optimization = 0, 1, 2 ???

    psi->result_optimization_complete();

    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "psi->MC_phasespace.end_optimization = " << psi->MC_phasespace.end_optimization << endline_flush; }

    // could become a psi function ???
    if (psi->switch_output_optimization){osi->output_step_integration_weight_optimization();}
    
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "psi->MC_phasespace.end_optimization = " << psi->MC_phasespace.end_optimization << endline_flush; }

    // shifted:    psi->output_optimization_complete();

    if (osi->switch_output_proceeding){perform_proceeding_step();}
    if (osi->switch_output_proceeding == 1){perform_proceeding_out();}

    psi->output_optimization_complete();

    psi->i_gen_step = 0;
    psi->i_acc_step = 0;
    osi->start_step = clock();
    
    if (psi->i_step == psi->n_repetition){osi->int_end = 1;}

    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "*psi->i_step_mode = " << setw(12) << *psi->i_step_mode << "   " << setw(12) << psi->i_gen << " [" << setw(3) << psi->i_nan << "] " << " " << setw(10) << psi->i_acc << " (" << setw(7) << psi->i_tec << ") " << " " << setw(10) << psi->i_rej << " " << endline_flush; }
    if (Log::getLogThreshold() <= LOG_DEBUG_STEP){ logger << LOG_DEBUG_STEP << "psi->i_step = " << setw(12) << psi->i_step << "   psi->i_gen_step = " << setw(12) << psi->i_gen_step << "   psi->i_acc_step = " << setw(10) << psi->i_acc_step << "   psi->enforce_step = " << psi->enforce_step << endline_flush; }

    psi->enforce_step = false;
  }
  
  logger.finished();
}
