#include "header.hpp"

void contribution_set::initialization_complete(){
  static Logger logger("contribution_set::initialization_complete");
  logger.called();

  initialization_contribution_order();
  
  determination_subprocess_psp(0);
  combination_subprocess_psp(0);

  dipole.clear();
  if (type_contribution == "RA" ||
      type_contribution == "L2RA" ||
      type_contribution == "RRA"){
    dipole.push_back(dipole_set("ME2_R", this, 0, 0));
    if (type_correction == "QCD" || type_correction == "MIX"){determination_dipole_QCD();}
    if (type_correction == "QEW" || type_correction == "MIX"){determination_dipole_QEW();}
    n_ps = dipole.size();
  }
  else if (type_contribution == "PT" || type_contribution == "PT2"){
    n_ps = 2;
    type_parton.push_back(type_parton[0]);
    basic_type_parton.push_back(basic_type_parton[0]);
    swap_parton.push_back(swap_parton[0]);
  }
  else {n_ps = 1;}

  fill_charge_parton();
  fill_coloured_parton();
  fill_mass_parton();

  // Remove subprocesses where heavy quarks beyond those present at Born level are generated ???
  // important e.g. for WW production in 4FS to remove WWbb final states
  // relevant also e.g for bbx production where bbbxbx final states couls be inclulded or not
  // -> should become a switch) !!!
  selection_heavy_quark_content();

  if (coll_choice && (mass_parton[0][1] > 0. || mass_parton[0][2] > 0.)){
    logger << LOG_WARN << "Massive incoming partons are not allowed in hadron-collider processes." << endline_flush;
    int_end = 1;
  }


  // for identification of phase space data files:
  if (!switch_to_newscheme){ newscheme_fill_name_particle(); }

  logger << LOG_DEBUG << "type_parton.size() = " << type_parton.size() << endline_flush;
  logger << LOG_DEBUG << "swap_parton.size() = " << swap_parton.size() << endline_flush;
  // Check if possible permutations of incoming/outgoing particles need to be taken into account !!!
  // should be contained in  swap_parton  !!!
  subprocess_name_new_format.resize(type_parton.size());
  for (int i_a = 0; i_a < type_parton.size(); i_a++){
  logger << LOG_DEBUG << "type_parton[" << i_a << "].size() = " << type_parton[i_a].size() << endline_flush;
  logger << LOG_DEBUG << "swap_parton[" << i_a << "].size() = " << swap_parton[i_a].size() << endline_flush;
    stringstream temp_ss;
    for (int i_p = 1; i_p < type_parton[i_a].size(); i_p++){
      logger << LOG_DEBUG << "type_parton[" << i_a << "][" << i_p << "] = " << setw(4) << type_parton[i_a][i_p] << "   swap_parton[" << i_a << "][" << i_p << "] = " << setw(4) << swap_parton[i_a][i_p] << "   type_parton[" << i_a << "][swap_parton[" << i_a << "][" << i_p << "]] = " << setw(4) << type_parton[i_a][swap_parton[i_a][i_p]] << endline_flush;
      if (abs(type_parton[i_a][swap_parton[i_a][i_p]]) > 10 && abs(type_parton[i_a][swap_parton[i_a][i_p]]) < 17){
	temp_ss << name_particle[lepton_exchange[type_parton[i_a][swap_parton[i_a][i_p]]]];
      }
      else {
	temp_ss << name_particle[type_parton[i_a][swap_parton[i_a][i_p]]];
      }
      if (i_p == 2){ temp_ss << "_"; }
    }
    subprocess_name_new_format[i_a] = temp_ss.str();
    logger << LOG_DEBUG << "subprocess_name_new_format[" << i_a << "] = " << subprocess_name_new_format[i_a] << endline_flush;
  }
  logger.newLine(LOG_DEBUG);
 
  if (!switch_to_newscheme){ oldscheme_fill_name_particle(); }
  
  logger << LOG_INFO << "(contribution_set::output_parton_content)" << endline << endline << output_parton_content() << endline_flush;
  
  logger.finished();
}


void contribution_set::fill_charge_parton(){
  static Logger logger("contribution_set::fill_charge_parton");
  logger.called();

  charge_parton.resize(n_ps);
  charge2_parton.resize(n_ps);
  is_charged_parton.resize(n_ps);
  for (int i_a = 0; i_a < n_ps; i_a++){
    charge_parton[i_a].resize(type_parton[i_a].size(), 0.);
    charge2_parton[i_a].resize(type_parton[i_a].size(), 0.);
    is_charged_parton[i_a].resize(type_parton[i_a].size(), false);
    for (int i_p = 1; i_p < type_parton[i_a].size(); i_p++){
      charge_parton[i_a][i_p] = msi->charge_pdg[type_parton[i_a][i_p]];
      charge2_parton[i_a][i_p] = msi->charge2_pdg[type_parton[i_a][i_p]];
      if (charge_parton[i_a][i_p] != 0.){is_charged_parton[i_a][i_p] = true;}
    }
  }
  
  logger.finished();
}


void contribution_set::fill_coloured_parton(){
  static Logger logger("contribution_set::fill_coloured_parton");
  logger.called();

  coloured_parton.resize(n_ps);
  is_coloured_parton.resize(n_ps);
  casimir_parton.resize(n_ps);
  for (int i_a = 0; i_a < n_ps; i_a++){
    coloured_parton[i_a].resize(type_parton[i_a].size(), 0);
    is_coloured_parton[i_a].resize(type_parton[i_a].size(), false);
    casimir_parton[i_a].resize(type_parton[i_a].size(), 0.);
    for (int i_p = 1; i_p < type_parton[i_a].size(); i_p++){
      if (type_parton[i_a][i_p] == 0 || type_parton[i_a][i_p] == 21){coloured_parton[i_a][i_p] = 2;}
      else if (type_parton[i_a][i_p] > 0 && type_parton[i_a][i_p] < 7){coloured_parton[i_a][i_p] = 1;}
      else if (type_parton[i_a][i_p] < 0 && type_parton[i_a][i_p] > -7){coloured_parton[i_a][i_p] = -1;}
      else {coloured_parton[i_a][i_p] = 0;}
      if (coloured_parton[i_a][i_p] != 0){is_coloured_parton[i_a][i_p] = true;}
      casimir_parton[i_a][i_p] = msi->casimir_pdg[type_parton[i_a][i_p]];
    }
  }
  
  logger.finished();
}


void contribution_set::fill_mass_parton(){
  static Logger logger("contribution_set::fill_mass_parton");
  logger.called();

  mass_parton.resize(n_ps);
  mass2_parton.resize(n_ps);
  is_massive_parton.resize(n_ps);
  for (int i_a = 0; i_a < n_ps; i_a++){
    mass_parton[i_a].resize(type_parton[i_a].size(), 0.);
    mass2_parton[i_a].resize(type_parton[i_a].size(), 0.);
    is_massive_parton[i_a].resize(type_parton[i_a].size(), false);
    for (int i_p = 1; i_p < type_parton[i_a].size(); i_p++){
      mass_parton[i_a][i_p] = msi->M_pdg[type_parton[i_a][i_p]];
      mass2_parton[i_a][i_p] = msi->M2_pdg[type_parton[i_a][i_p]];
      if (mass_parton[i_a][i_p] != 0.){is_massive_parton[i_a][i_p] = true;}
    }
  }

  logger.finished();
}


void contribution_set::selection_heavy_quark_content(){
  static Logger logger("contribution_set::selection_heavy_quark_content");
  logger.called();

  for (int i_q = 1; i_q < 7; i_q++){
    if (msi->M_pdg[i_q] == 0.){continue;}
    for (int i_a = 0; i_a < n_ps; i_a++){
      int count_partonic_Q = 0;
      int count_partonic_Qx = 0;
      for (int i_p = 1; i_p < type_parton[i_a].size(); i_p++){
	if (type_parton[i_a][i_p] == i_q){count_partonic_Q++;}
	if (type_parton[i_a][i_p] == -i_q){count_partonic_Qx++;}
      }
      int count_hadronic_Q = 0;
      int count_hadronic_Qx = 0;
      for (int i_p = 1; i_p < type_hadron.size(); i_p++){
	if (type_hadron[i_p] == i_q){count_hadronic_Q++;}
	if (type_hadron[i_p] == -i_q){count_hadronic_Qx++;}
      }
      logger << LOG_DEBUG << "[" << i_q << "]   " << "count_partonic_Q  = " << count_partonic_Q << " =?= " << count_hadronic_Q << " = count_hadronic_Q" << endline_flush;
      logger << LOG_DEBUG << "[" << i_q << "]   " << "count_partonic_Qx = " << count_partonic_Qx << " =?= " << count_hadronic_Qx << " = count_hadronic_Qx" << endline_flush;
      if (count_partonic_Q != count_hadronic_Q || count_partonic_Qx != count_hadronic_Qx){
	logger << LOG_WARN << "Removed due to [" << i_q << "] in subprocess" << " = " << subprocess << ":  count_partonic_Q  = " << count_partonic_Q << " != " << count_hadronic_Q << " = count_hadronic_Q" << " || count_partonic_Qx = " << count_partonic_Qx << " != " << count_hadronic_Qx << " = count_hadronic_Qx" << endline_flush;
	int_end = 1;
      }
    }
  }
    
  logger.finished();
}


void contribution_set::initialization_contribution_order(){
  static Logger logger("contribution_set::initialization_contribution_order");
  logger.called();

  logger << LOG_DEBUG << "The same information is (was ??? ) also contained in observable_set !!!" << endline_flush;

  if (class_contribution_order == 0){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 1 && class_contribution_CS_collinear){
    if (type_correction == "QCD"){
      phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
      phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
      phasespace_order_interference.resize(1, contribution_order_interference);
    }
    else if (type_correction == "QEW"){
      phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
      phasespace_order_alpha_e.resize(1, contribution_order_alpha_e - 1);
      phasespace_order_interference.resize(1, contribution_order_interference);
    }
    else if (type_correction == "MIX"){
      logger << LOG_FATAL << "Contribution CA.MIX is not defined!" << endline_flush;
      exit(1);
    }
  }
  else if (class_contribution_order == 1 && class_contribution_CS_virtual){
    if (type_correction == "QCD"){
      phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
      phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
      phasespace_order_interference.resize(1, contribution_order_interference);
    }
    else if (type_correction == "QEW"){
      phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
      phasespace_order_alpha_e.resize(1, contribution_order_alpha_e - 1);
      phasespace_order_interference.resize(1, contribution_order_interference);
    }
    else if (type_correction == "MIX"){
      // Implement a general procedure to find out if the contribution is QCD to a squared Born overlapping with EW corrections to an interference Born or vice versa,
      // e.g. check for which orders the given subprocess exists, and decide based on chosen couplings !!!
      // So far,  order_alpha_s_born  is determined by counting the maximum order of alpha_s based on external QCD partons (might be sufficient).
      phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
      phasespace_order_alpha_e.resize(1, contribution_order_alpha_e - 1);
      phasespace_order_interference.resize(1, contribution_order_interference);

      logger << LOG_DEBUG << "phasespace_order_alpha_s[0] = " << phasespace_order_alpha_s[0] << endline_flush;
      logger << LOG_DEBUG << "phasespace_order_alpha_e[0] = " << phasespace_order_alpha_e[0] << endline_flush;
      logger << LOG_DEBUG << "phasespace_order_interference[0] = " << phasespace_order_interference[0] << endline_flush;
      logger << LOG_DEBUG << "contribution_order_alpha_s = "<< contribution_order_alpha_s << endline_flush;
      logger << LOG_DEBUG << "contribution_order_alpha_e = "<< contribution_order_alpha_e << endline_flush;
      logger << LOG_DEBUG << "order_alpha_s_born = "<< order_alpha_s_born << endline_flush;
      // meaning of order_alpha_s_born changed for subleading orders !!!
      /*
      if (contribution_order_alpha_s > order_alpha_s_born){
	logger << LOG_DEBUG << "For (contribution_order_alpha_s >= order_alpha_s_born) VA must be a pure QCD correction at NLO. Assignment to VA.MIX needs to be checked." << endline_flush;
      }
      else 
      */
      if (contribution_order_alpha_s == order_alpha_s_born){
	logger << LOG_DEBUG << "For (contribution_order_alpha_s == order_alpha_s_born) VA.MIX combines QEW corrections to leading Born and QCD corrections to next sub-leading interference Born. Phase space from leading Born contribution is used." << endline_flush;
      }
      else if (contribution_order_alpha_s == order_alpha_s_born + 1){
	//      else if (contribution_order_alpha_s == order_alpha_s_born - 1){
	logger << LOG_DEBUG << "For (contribution_order_alpha_s == order_alpha_s_born + 1) VA.MIX combines QCD corrections to sub-leading Born (not interference, but as(sub) = as(lead) - 2) and QEW corrections to sub-leading interference Born. Phase space from subleading Born contribution is used." << endline_flush;
	phasespace_order_alpha_s[0] = contribution_order_alpha_s - 1;
	phasespace_order_alpha_e[0] = contribution_order_alpha_e;
	contribution_order_interference = 0;
      }
      logger << LOG_DEBUG << "phasespace_order_alpha_s[0] = " << phasespace_order_alpha_s[0] << endline_flush;
      logger << LOG_DEBUG << "phasespace_order_alpha_e[0] = " << phasespace_order_alpha_e[0] << endline_flush;
      logger << LOG_DEBUG << "phasespace_order_interference[0] = " << phasespace_order_interference[0] << endline_flush;
      logger << LOG_DEBUG << "Phase-space is taken according to correction wrt. a non-interference Born." << endline_flush;
      //      logger << LOG_DEBUG << "Phase-space is taken according to involved QEW correction! (could be improved later...)" << endline_flush;
    }
  }
  else if (class_contribution_order == 1 && class_contribution_CS_real){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
    logger << LOG_DEBUG << "Information on dipole phase-spaces is supplied after dipole determination..." << endline_flush;
  }

  // qT subtraction at NLO
  else if (class_contribution_order == 1 && class_contribution_QT_virtual){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 1 && class_contribution_QT_collinear){
   phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
   phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
   phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 1 && class_contribution_QT_powercorrection){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 1 && class_contribution_QT_real){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  // NNLO subtraction with CS subtraction for first emission
  else if (class_contribution_order == 2 && class_contribution_CS_virtual){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 2 && class_contribution_CS_collinear){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 2 && class_contribution_CS_real){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
    logger << LOG_DEBUG << "Information on dipole phase-spaces is supplied after dipole determination..." << endline_flush;
  }

  // qT subtraction at NNLO
  else if (class_contribution_order == 2 && class_contribution_QT_virtual){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 2);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 2 && class_contribution_QT_collinear){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 2);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (class_contribution_order == 2 && class_contribution_QT_powercorrection){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 2);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }

  else if (type_contribution == "NLL_LO" || 
	   type_contribution == "NLL_NLO" || 
	   type_contribution == "NNLL_LO" || 
	   type_contribution == "NNLL_NLO"){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 1);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else if (type_contribution == "NNLL_NNLO"){
    phasespace_order_alpha_s.resize(1, contribution_order_alpha_s - 2);
    phasespace_order_alpha_e.resize(1, contribution_order_alpha_e);
    phasespace_order_interference.resize(1, contribution_order_interference);
  }
  else {
    logger << LOG_FATAL << "Contribution " << type_contribution << "." << type_correction << " has not been defined yet!" << endline;
    //    logger << LOG_FATAL << output_parton_content() << endline_flush;
    exit(1);
  }

  logger << LOG_DEBUG << "process_class = " << process_class << endline_flush;
  logger << LOG_DEBUG << "subprocess = " << subprocess << endline_flush;
  logger << LOG_DEBUG << "type_perturbative_order = " << type_perturbative_order << endline_flush;
  logger << LOG_DEBUG << "type_contribution = " << type_contribution << endline_flush;
  logger << LOG_DEBUG << "type_correction = " << type_correction << endline_flush;
  
  logger << LOG_DEBUG << "contribution_order_alpha_s = " << contribution_order_alpha_s << endline_flush;
  logger << LOG_DEBUG << "contribution_order_alpha_e = " << contribution_order_alpha_e << endline_flush;
  logger << LOG_DEBUG << "contribution_order_interference = " << contribution_order_interference << endline_flush;
  logger << LOG_DEBUG << "phasespace_order_alpha_s = " << phasespace_order_alpha_s[0] << endline_flush;
  logger << LOG_DEBUG << "phasespace_order_alpha_e = " << phasespace_order_alpha_e[0] << endline_flush;
  logger << LOG_DEBUG << "phasespace_order_interference = " << phasespace_order_interference[0] << endline_flush;
  
  // strange location !!!
 
  no_process_parton.resize(1);
  swap_parton.resize(1);

  logger.finished();
}


// Somehow incomplete for EW corrections !!!
// Could be extended to really find all singular structures of QCD/QED type (like for dipole_set) !!!
void contribution_set::QCD_selection_phasespace_singularity(){
  static Logger logger("contribution_set::QCD_selection_phasespace_singularity");
  logger.called();

  //  static vector<int> type_parton = psi.csi->type_parton[0];

  singular_region.resize(type_parton[0].size(), vector<double> (type_parton[0].size()));
  singular_region_name.resize(type_parton[0].size(), vector<string> (type_parton[0].size()));

  vector<string> pa_name(type_parton[0].size(), "");
  if (type_parton[0][1] > -10 && type_parton[0][1] < 10){pa_name[1] = "a";}
  if (type_parton[0][2] > -10 && type_parton[0][2] < 10){pa_name[2] = "b";}
  for (int i_p = 3; i_p < pa_name.size(); i_p++){if (type_parton[0][i_p] >= -10 && type_parton[0][i_p] <= 10){pa_name[i_p] = char(105 + i_p - 3);}}
  logger << LOG_DEBUG << "QCD phasespace singularity selection from QCD dipole candidates started." << endline_flush;

  for (int i_p = 0; i_p < type_parton[0].size(); i_p++){
    for (int j_p = i_p + 1; j_p < type_parton[0].size(); j_p++){
      if (j_p < 3){continue;}
      bool new_region = false;

      if (i_p < 3){
	if (type_parton[0][i_p] == type_parton[0][j_p] ||
	    (type_parton[0][i_p] == 0 && abs(type_parton[0][j_p]) < 10) ||
	    (type_parton[0][j_p] == 0 && abs(type_parton[0][i_p]) < 10)){
	  new_region = true;
	}
      }
      else {
	if (type_parton[0][i_p] == type_parton[0][j_p] ||
	    (type_parton[0][i_p] == type_parton[0][j_p] && type_parton[0][i_p] == 0 && abs(type_parton[0][j_p]) < 10) ||
	    (type_parton[0][j_p] == type_parton[0][i_p] && type_parton[0][j_p] == 0 && abs(type_parton[0][i_p]) < 10)){
	  new_region = true;
	}
      }
      if (new_region){
	// i_p < j_p by construction !!!
	singular_region_list.push_back(vector<int> {i_p, j_p});
	singular_region_name[i_p][j_p] = "p" + pa_name[i_p] + "." + "p" + pa_name[j_p];
 	/*
	vector<int> temp_singularity(2);
	temp_singularity[0] = i_p;
	temp_singularity[1] = j_p;
	sort(temp_singularity.begin(), temp_singularity.end());
	singular_region_list.push_back(temp_singularity);
	singular_region_name[temp_singularity[0]][temp_singularity[1]] = "p" + pa_name[temp_singularity[0]] + "." + "p" + pa_name[temp_singularity[1]];
	*/
      }
    }
  }

  for (int i = 0; i < singular_region_list.size(); i++){
    logger << LOG_DEBUG << singular_region_list[i][0] << " " << singular_region_list[i][1] << "   " << singular_region_name[singular_region_list[i][0]][singular_region_list[i][1]] << endline_flush;
  }

  logger.finished();
}


