#include "header.hpp"

void observable_set::determination_collinear_splitting(){
  static Logger logger("observable_set::selection_collinear_splitting");
  logger.called();

  CA_dipole_splitting.resize(4, vector<int> (3, 0));
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int i_em = 0; i_em < 3; i_em++){
      if ((*CA_collinear)[i_c][0].in_collinear()[i_em] == 1){
	CA_dipole_splitting[(*CA_collinear)[i_c][0].type()][i_em] = 1;
      }
    }
  }

  logger.finished();
}


void observable_set::selection_collinear(){
  static Logger logger("observable_set::selection_collinear");
  logger.called();

  for (int i_c = (*CA_collinear).size() - 1; i_c >=0; i_c--){
    for (int j_c = (*CA_collinear)[i_c].size() - 1; j_c >= 0; j_c--){
      logger << LOG_DEBUG << "(*CA_collinear)[" << i_c << "][" << j_c << "].no_emitter() = " << (*CA_collinear)[i_c][j_c].no_emitter() << "   no_spectator() = " << (*CA_collinear)[i_c][j_c].no_spectator() << endline_flush;
      // Possibly: switch_off_CS_ii && switch_off_CS_if && (*CA_collinear)[i_c][j_c].no_spectator() == 0
      if ((*CA_collinear)[i_c][j_c].no_spectator() == 0){continue;}
      // remove final-final contribution:
      if (switch_off_CS_ff && ((*CA_collinear)[i_c][j_c].no_emitter() > 2 && (*CA_collinear)[i_c][j_c].no_spectator() > 2)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_WARN << "SHOULD NOT HAPPEN !!!   kill ff:   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
      // remove final-initial contribution:
      if (switch_off_CS_fi && ((*CA_collinear)[i_c][j_c].no_emitter() > 2 && (*CA_collinear)[i_c][j_c].no_spectator() < 3)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_WARN << "SHOULD NOT HAPPEN !!!   kill fi:   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
      // remove initial-final & final-initial contribution:
      if (switch_off_CS_if && switch_off_CS_fi && ((*CA_collinear)[i_c][j_c].no_emitter() < 3 && (*CA_collinear)[i_c][j_c].no_spectator() > 2)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_DEBUG << "kill if & fi:   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
      // remove initial-initial contribution:
      if (switch_off_CS_ii && ((*CA_collinear)[i_c][j_c].no_emitter() < 3 && (*CA_collinear)[i_c][j_c].no_spectator() < 3)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_DEBUG << "kill ii :   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
    }
    if ((*CA_collinear)[i_c].size() == 0){
      (*CA_collinear).erase((*CA_collinear).begin() + i_c);
      CA_Q2f.erase(CA_Q2f.begin() + i_c);
    }
  }

  logger.finished();
}


void observable_set::fill_CA_combination_pdf(){
  static Logger logger("observable_set::fill_CA_combination_pdf");
  logger.called();

  CA_combination_pdf.resize((*CA_collinear).size(), vector<vector<vector<int> > > (csi->combination_pdf.size()));
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int i_i = 0; i_i < CA_combination_pdf[i_c].size(); i_i++){
      // CA_combination_pdf[i_c] is a vector which contains the usual csi->combination_pdf(3) with 0 -> direction (1, -1), 1 -> parton with x1 (in hadron 1/2 for +1/-1), 2 -> parton with x2 (in hadron 2/1 for +1/-1)
      if (((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10) && ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10)){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	    if (j_q == 0){continue;}
	    vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	    new_temp_combination_pdf[1] = i_q;
	    new_temp_combination_pdf[2] = j_q;
	    CA_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	  }
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[1] = i_q;
	  CA_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10){
	for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	  if (j_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[2] = j_q;
	  CA_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else {
	CA_combination_pdf[i_c][i_i].push_back((*CA_collinear)[i_c][0].all_pdf()[i_i]);
      }
    }
  }

  logger.finished();
}


void observable_set::determine_CA_list_combination_pdf(){
  static Logger logger("observable_set::determine_CA_list_combination_pdf");
  logger.called();

  // direct determination of list_combination_pdf from ncollinear (including list_combination_pdf_emission; no_pdf refers to entry in list_combination_pdf)
  list_combination_pdf.resize((*CA_collinear).size() + 1, vector<vector<vector<int> > > (csi->combination_pdf.size()));
  list_combination_pdf_emission.resize((*CA_collinear).size() + 1, vector<int> (3, 0));

  // Extra entry for endpoint contribution (-> 0 component):
  list_combination_pdf[0].resize(csi->combination_pdf.size(), vector<vector<int> > (1));
  for (int i_i = 0; i_i < csi->combination_pdf.size(); i_i++){list_combination_pdf[0][i_i].push_back(csi->combination_pdf[i_i]);}
  // Alternative: Introduce a (basically empty 0-component for CA_collinear) - to avoid the +1 shift !!!

  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    list_combination_pdf_emission[i_c + 1][(*CA_collinear)[i_c][0].no_emitter()] = 1;
    //    list_combination_pdf_emission[i_c + 1] = (*CA_collinear)[i_c][0].in_collinear();
    //    ncollinear[i_c].no_pdf = i_c;
    for (int i_i = 0; i_i < list_combination_pdf[i_c + 1].size(); i_i++){
      // list_combination_pdf[i_c + 1] contains a vector, which contains the usual osi_csi->combination_pdf(3) with 0 -> direction (1, -1), 1 -> parton with x1 (in hadron 1/2 for +1/-1), 2 -> parton with x2 (in hadron 2/1 for +1/-1)
      if (((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10) && ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10)){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	    if (j_q == 0){continue;}
	    vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	    new_temp_combination_pdf[1] = i_q;
	    new_temp_combination_pdf[2] = j_q;
	    list_combination_pdf[i_c + 1][i_i].push_back(new_temp_combination_pdf);
	  }
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[1] = i_q;
	  list_combination_pdf[i_c + 1][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10){
	for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	  if (j_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[2] = j_q;
	  list_combination_pdf[i_c + 1][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else {
	list_combination_pdf[i_c + 1][i_i].push_back((*CA_collinear)[i_c][0].all_pdf()[i_i]);
      }
    }
    for (int j_c = 0; j_c < (*CA_collinear)[i_c].size(); j_c++){(*CA_collinear)[i_c][j_c].no_pdf = i_c + 1;}
  }

  if (Log::getLogThreshold() <= LOG_DEBUG){
    for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
      for (int j_c = 0; j_c < (*CA_collinear)[i_c].size(); j_c++){
	logger << LOG_DEBUG << "(*CA_collinear)[" << i_c << "][" << j_c << "].no_pdf = " <<(*CA_collinear)[i_c][j_c].no_pdf << endline_flush;
      }
    }
    logger << LOG_DEBUG << "list_combination_pdf determined " << endl << endline_flush;
  }
  
  logger.finished();
}

