#include "header.hpp"

string observable_set::output_testpoint_RA(){
  static Logger logger("observable_set::output_testpoint_RA");
  logger.called();

  stringstream out_comparison;
  out_comparison << "Real-correction matrix element and dipole terms evaluated at alpha_S(µR = " << setprecision(15) << scale_ren << ") = " << setprecision(15) << alpha_S << endline << endline;
  for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
    out_comparison << setw(14) << csi->dipole[i_a].name() << " = " << setw(23) << setprecision(15) << RA_ME2[i_a] << endline;
    if (i_a == 0){out_comparison << endline;}
  }
  out_comparison << endline;
  out_comparison << setw(14) << "ME2_R - sum{D}" << " = " << setw(23) << setprecision(15) << accumulate(RA_ME2.begin(), RA_ME2.end(), 0.) << endline;
  out_comparison << endline;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << out_comparison.str() << endline_flush; }
  logger.finished();
  return out_comparison.str();
}


void observable_set::output_integrand_maximum_RA(){
  static Logger logger("observable_set::output_integrand_maximum_RA");
  logger.called();

  //  logger << LOG_INFO << "integrand = " << integrand << " ?> " << max_integrand << " = max_integrand" << endline_flush;
											
  if (abs(integrand) > max_integrand){
    ofstream out_maxevent;
    out_maxevent.open(filename_maxevent.c_str(), ofstream::out | ofstream::app);
    double max_integrand_signed;
    if (switch_TSV){
      max_integrand_signed = -accumulate(ps_integrand_TSV[no_reference_TSV][no_scale_ren_reference_TSV][no_scale_fact_reference_TSV][0].begin() + 1, ps_integrand_TSV[no_reference_TSV][no_scale_ren_reference_TSV][no_scale_fact_reference_TSV][0].end(), -ps_integrand_TSV[no_reference_TSV][no_scale_ren_reference_TSV][no_scale_fact_reference_TSV][0][0]);
      //      max_integrand = abs(-accumulate(ps_integrand_TSV[x_s][x_r][x_f][0].begin() + 1, ps_integrand_TSV[x_s][x_r][x_f][0].end(), -ps_integrand_TSV[x_s][x_r][x_f][0][0]));
    }
    else {
      max_integrand_signed = -accumulate(var_RA_ME2.begin() + 1, var_RA_ME2.end(), -var_RA_ME2[0]) * psi->ps_factor * rescaling_factor_alpha_e;
    }
    max_integrand = abs(max_integrand_signed);
    out_maxevent << right << setw(10) << psi->i_gen << right << setw(10) << psi->i_acc << "   max_integrand_signed = " << max_integrand_signed << "   " << "max_integrand / sigma_normalization = " << max_integrand / sigma_normalization << endline;
    output_integrand_maximum_psp();

    /*
    for (int ib = 0; ib < esi->p_parton[0].size(); ib++){out_maxevent << "esi->p_parton[0][" << ib << "] = " << esi->p_parton[0][ib] << endline;}

  // write out all relevant pairings to spot badly mapped resonances
  out_maxevent << endline;
  for (int ib1 = 3; ib1 < esi->p_parton[0].size(); ib1++){
    for (int ib2 = ib1+1; ib2 < esi->p_parton[0].size(); ib2++){
      out_maxevent << "esi->p_parton[0][" << ib1 << "]+esi->p_parton[0][" << ib2 << "] = " << esi->p_parton[0][ib1]+esi->p_parton[0][ib2] << ", sqrt(s)= "  << (esi->p_parton[0][ib1]+esi->p_parton[0][ib2]).m() << endline;
    }
  }
  for (int ib1 = 3; ib1 < esi->p_parton[0].size(); ib1++){
    for (int ib2 = ib1+1; ib2 < esi->p_parton[0].size(); ib2++){
      for (int ib3 = ib2+1; ib3 < esi->p_parton[0].size(); ib3++){
      out_maxevent << "esi->p_parton[0][" << ib1 << "]+esi->p_parton[0][" << ib2 << "]+esi->p_parton[0][" << ib3 << "] = " << esi->p_parton[0][ib1]+esi->p_parton[0][ib2]+esi->p_parton[0][ib3] << ", sqrt(s)= "  << (esi->p_parton[0][ib1]+esi->p_parton[0][ib2]+esi->p_parton[0][ib3]).m() << endline;
      }
    }
  }
  out_maxevent << endline;
    */
    int ccount = 0;
    for (int i_a = 1; i_a < esi->cut_ps.size(); i_a++){if (esi->cut_ps[i_a] == -1){ccount++;}}
    out_maxevent << "cut dipoles: " << setw(2) << ccount << endline;
    out_maxevent << left << setw(8) << "" << "   " << right << setw(23) << setprecision(15) << showpoint << "" << "   " << right << setw(23) << setprecision(15) << showpoint << "A_ME2" << "   " << right << setw(23) << setprecision(15) << showpoint << "A_ME2 / R_ME2" << endline; // "   " << right << setw(23) << setprecision(15) << showpoint << "m_top" << "   " << right << setw(23) << setprecision(15) << showpoint << "m_atop" << "   " << endline;
  logger << LOG_DEBUG_VERBOSE << "esi->cut_ps.size() = " << esi->cut_ps.size() << endline_flush;

  for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){
    out_maxevent << left << setw(8) << csi->dipole[i_a].name() << "   " << right << setw(23) << setprecision(15) << showpoint << csi->dipole[i_a].xy << "   " << right << setw(23) << setprecision(15) << showpoint << RA_ME2[i_a] << "   " << right << setw(23) << setprecision(15) << showpoint << RA_ME2[i_a] / RA_ME2[0] << "   " << endline; //right << setw(23) << setprecision(15) << showpoint << (esi->p_parton[i_a][3] + esi->p_parton[i_a][5]).m() << "   " << right << setw(23) << setprecision(15) << showpoint << (esi->p_parton[i_a][4] + esi->p_parton[i_a][6]).m() << endline;
  }
  out_maxevent << endline;
  out_maxevent << left << setw(8) << "Sum(dip) = " << right << setw(23) << setprecision(15) << showpoint << 0 << "   " << right << setw(23) << setprecision(15) << showpoint << accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) << "   " << right << setw(23) << setprecision(15) << showpoint << accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] << "   " << endline; //right << setw(23) << setprecision(15) << showpoint << (d
  //  if (ccount != 0){
    out_maxevent << "Number of cut dipoles: " << ccount << endline;
    //    out_maxevent << left << setw(8) << "" << "   " << right << setw(23) << setprecision(15) << showpoint << "m_bb" << right << setw(23) << setprecision(15) << showpoint << "p_T(b)" << right << setw(23) << setprecision(15) << showpoint << "p_T(b~)" << endline;
    for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){
      for (int ib = 0; ib < esi->p_parton[i_a].size(); ib++){out_maxevent << "esi->p_parton[" << i_a << "][" << ib << "] = " << esi->p_parton[i_a][ib] << endline;}
      //      out_maxevent << left << setw(8) << dipole_name[i_a] << "   " << right << setw(23) << setprecision(15) << showpoint << (esi->p_parton[i_a][7] + esi->p_parton[i_a][8]).m() << right << setw(23) << setprecision(15) << showpoint << esi->p_parton[i_a][7].pT() << right << setw(23) << setprecision(15) << showpoint << (esi->p_parton[i_a][8]).pT() << endline;
    }
    //  }
  //  else {
    for (int i_a = 1; i_a < esi->cut_ps.size(); i_a++){
      if      (csi->dipole[i_a].type_dipole() == 1){
	double y_ij_k = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]);
	double z_i = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]));
	double z_j = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]));
	out_maxevent << setw(2) << i_a << "   y_" << csi->dipole[i_a].no_R_emitter_1() << csi->dipole[i_a].no_R_emitter_2() << "," << csi->dipole[i_a].no_R_spectator() << " = " << setw(23) << setprecision(15) << y_ij_k << "   z_" << csi->dipole[i_a].no_R_emitter_1() << " = " << setw(23) << setprecision(15) << z_i << "      z_" << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << z_j << endline;
	out_maxevent << setw(2) << i_a << "   y_" << csi->dipole[i_a].no_R_emitter_1() << csi->dipole[i_a].no_R_emitter_2() << "," << csi->dipole[i_a].no_R_spectator() << " = " << setw(23) << setprecision(15) << csi->dipole[i_a].xy << endline;//"   z_" << csi->dipole[i_a].no_R_emitter_1() << " = " << setw(23) << setprecision(15) << z_i << "      z_" << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << z_j << endline;
      }
      else if (csi->dipole[i_a].type_dipole() == 2){
	double x_ij_a = (esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] - esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]) / ((esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]) * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]);
	double z_i = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]));
	double z_j = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] * (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]));
	out_maxevent << setw(2) << i_a << "   x_" << csi->dipole[i_a].no_R_emitter_1() << csi->dipole[i_a].no_R_emitter_2() << "," << csi->dipole[i_a].no_R_spectator() << " = " << setw(23) << setprecision(15) << x_ij_a << "   z_" << csi->dipole[i_a].no_R_emitter_1() << " = " << setw(23) << setprecision(15) << z_i << "      z_" << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << z_j << endline;
	out_maxevent << setw(2) << i_a << "   x_" << csi->dipole[i_a].no_R_emitter_1() << csi->dipole[i_a].no_R_emitter_2() << "," << csi->dipole[i_a].no_R_spectator() << " = " << setw(23) << setprecision(15) << csi->dipole[i_a].xy << endline;//"   z_" << csi->dipole[i_a].no_R_emitter_1() << " = " << setw(23) << setprecision(15) << z_i << "      z_" << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << z_j << endline;
      }
      else if (csi->dipole[i_a].type_dipole() == 3){
	double x_ik_a = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] + esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] - esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) / ((esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] + esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()]);
	double u_i = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] + esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]));
	out_maxevent << setw(2) << i_a << "   x_" << csi->dipole[i_a].no_R_emitter_2() << csi->dipole[i_a].no_R_spectator() << "," << csi->dipole[i_a].no_R_emitter_1() << " = " << setw(23) << setprecision(15) << x_ik_a << "   u_" << csi->dipole[i_a].no_R_spectator() << " = " << setw(23) << setprecision(15) << u_i << endline;
	out_maxevent << setw(2) << i_a << "   x_" << csi->dipole[i_a].no_R_emitter_2() << csi->dipole[i_a].no_R_spectator() << "," << csi->dipole[i_a].no_R_emitter_1() << " = " << setw(23) << setprecision(15) << csi->dipole[i_a].xy << endline;//"   u_" << csi->dipole[i_a].no_R_spectator() << " = " << setw(23) << setprecision(15) << u_i << endline;
      }
      else if (csi->dipole[i_a].type_dipole() == 5){
	double x_i_ab = (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()] - esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] - esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]);
	out_maxevent << setw(2) << i_a << "   x_" << csi->dipole[i_a].no_R_spectator() << "," << csi->dipole[i_a].no_R_emitter_1() << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << x_i_ab << "   v_" << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) << endline;
	out_maxevent << setw(2) << i_a << "   x_" << csi->dipole[i_a].no_R_spectator() << "," << csi->dipole[i_a].no_R_emitter_1() << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << csi->dipole[i_a].xy << endline;//"   v_" << csi->dipole[i_a].no_R_emitter_2() << " = " << setw(23) << setprecision(15) << (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_emitter_2()]) / (esi->p_parton[0][csi->dipole[i_a].no_R_emitter_1()] * esi->p_parton[0][csi->dipole[i_a].no_R_spectator()]) << endline;
      }
    }
    //  }
  for (int sr = 0; sr < csi->singular_region_list.size(); sr++){
    int x1 = csi->singular_region_list[sr][0];
    int x2 = csi->singular_region_list[sr][1];
    csi->singular_region[x1][x2] = esi->p_parton[0][x1] * esi->p_parton[0][x2] / psi->xbs_all[0][0];
    out_maxevent << csi->singular_region_name[x1][x2] << "/^s = "<< right << setw(23) << setprecision(15) << showpoint << csi->singular_region[x1][x2] << endline; //"   A/R = " << right << setw(23) << setprecision(15) << showpoint << accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] << endline;
  }
  out_maxevent << right << setw(14) << psi->i_gen << right << setw(12) << psi->i_acc << "   maxI = " << setw(15) << setprecision(8) << -accumulate(var_RA_ME2.begin() + 1, var_RA_ME2.end(), -var_RA_ME2[0]) * psi->ps_factor * rescaling_factor_alpha_e << "   " << "maxI/LO = " << setw(15) << setprecision(8) << max_integrand / sigma_normalization << "   A/R = " << setw(23) << setprecision(15) << accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] << "   n_cut-dipoles: " << setw(2) << ccount << endline;
  out_maxevent << endline << endline << endline;
  //  logger << LOG_DEBUG_POINT << right << setw(14) << psi->i_gen << right << setw(12) << psi->i_acc << "   maxI = " << setw(15) << setprecision(8) << -accumulate(var_RA_ME2.begin() + 1, var_RA_ME2.end(), -var_RA_ME2[0]) * psi->ps_factor * rescaling_factor_alpha_e << "   " << "maxI/LO = " << setw(15) << setprecision(8) << max_integrand / sigma_normalization << "   A/R = " << setw(23) << setprecision(15) << accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] << "   n_cut-dipoles: " << setw(2) << ccount << endline;
  out_maxevent.close();
  }

  logger.finished();
}


void observable_set::output_cancellation_RA(){
  static Logger logger("observable_set::output_cancellation_RA");
  logger.called();

  int flag = 0;
  int y1 = -1;
  int y2 = -1;
  double min_sr = 1.e99;
  for (int sr = 0; sr < csi->singular_region_list.size(); sr++){
    int x1 = csi->singular_region_list[sr][0];
    int x2 = csi->singular_region_list[sr][1];
    csi->singular_region[x1][x2] = esi->p_parton[0][x1] * esi->p_parton[0][x2] / psi->xbs_all[0][0];
    if (csi->singular_region[x1][x2] < psi->cut_technical * 1.e1){
      if (csi->singular_region[x1][x2] < min_sr){min_sr = csi->singular_region[x1][x2]; y1 = x1; y2 = x2;}
      flag++;
    }
  }
  if (flag){
    int ccount = 0;
    for (int i_a = 1; i_a < esi->cut_ps.size(); i_a++){if (esi->cut_ps[i_a] == -1){ccount++;}}
    //    logger << LOG_DEBUG_POINT << "cut dipoles: " << setw(2) << ccount << endline_flush;
    //    logger << LOG_DEBUG_POINT << left << setw(8) << "" << "   " << right << setw(23) << setprecision(15) << showpoint << "" << "   " << right << setw(23) << setprecision(15) << showpoint << "A_ME2" << "   " << right << setw(23) << setprecision(15) << showpoint << "A_ME2 / R_ME2" << endline_flush;
    stringstream temp_ss;
   
    //    temp_ss << right << setw(14) << psi->i_gen << right << setw(12) << psi->i_acc << "   maxI = " << setw(15) << setprecision(8) << -accumulate(var_RA_ME2.begin() + 1, var_RA_ME2.end(), -var_RA_ME2[0]) * psi->ps_factor * rescaling_factor_alpha_e << "   " << "maxI/LO = " << setw(15) << setprecision(8) << max_integrand / sigma_normalization << "   A/R = " << setw(23) << setprecision(15) << accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] << "   n_cut-dipoles: " << setw(2) << ccount << endline;
    temp_ss << right << setw(12) << psi->i_gen;
    temp_ss << right << setw(10) << psi->i_acc;
    temp_ss << "  (";
    temp_ss << right << setw(2) << esi->cut_ps[0] << "/";
    temp_ss << right << setw(2) << ccount;
    temp_ss << "  ";
    temp_ss << csi->singular_region_name[y1][y2] << "/^s = ";
    temp_ss << right << setw(11) << setprecision(4) << showpoint << csi->singular_region[y1][y2];
    temp_ss << ")  int/norm = ";
    temp_ss << setw(15) << setprecision(8) << integrand / sigma_normalization;
    if (accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) != 0. && RA_ME2[0] != 0.){
      temp_ss << "  A/R = ";
      temp_ss << setw(23) << setprecision(15) << -accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0];
      temp_ss << "  A/R-1 = ";
      temp_ss << right << scientific << setw(9) << setprecision(2) << -accumulate(RA_ME2.begin() + 1, RA_ME2.end(), 0.) / RA_ME2[0] - 1.;
      temp_ss << defaultfloat << left;
    }
    //  temp_ss << "   n_cut-dipoles: " << setw(2) << ccount;
    //    if (switch_output_cancellation_check == 1){logger << LOG_INFO << temp_ss.str() << endline_flush;}
    if (switch_output_cancellation_check > 1){
      temp_ss << endline << endline;
      for (int ib = 0; ib < esi->p_parton[0].size(); ib++){temp_ss << "p_parton[" << 0 << "][" << ib << "] = " << esi->p_parton[0][ib] << endline;}
      temp_ss << scientific << endline;
      for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){
	temp_ss << left << setw(8) << csi->dipole[i_a].name();
	if (i_a == 0){
	  if (esi->cut_ps[0] > -1){
	    temp_ss << "   " << right << setw(23) << setprecision(15) << showpoint << -RA_ME2[i_a];
	    temp_ss << "   " << right << setw(23) << "A / ME2_R";
	  }
	  else {temp_ss << setw(26) << "";}
	  temp_ss << "   " << right << setw(23) << "x/y";
	  temp_ss << "   " << right << setw(23) << "z/u/v";
	  temp_ss << "   " << right << setw(23) << "p_em1 x p_em2 / ^s";
	}
	else {
	  if (esi->cut_ps[i_a] > -1){
	    temp_ss << "   " << right << setw(23) << setprecision(15) << showpoint << -RA_ME2[i_a];
	    if (esi->cut_ps[0] > -1){temp_ss << "   " << right << setw(23) << setprecision(15) << showpoint << -RA_ME2[i_a] / RA_ME2[0];}
	  }
	  else {
	    if (esi->cut_ps[0] > -1){temp_ss << setw(52) << "";}
	    else {temp_ss << setw(26) << "";}
	  }
	  temp_ss << "   " << right << showpoint << setw(23) << setprecision(15) << showpoint << csi->dipole[i_a].xy;
	  temp_ss << "   " << right << showpoint << setw(23) << setprecision(15) << showpoint << csi->dipole[i_a].zuv;
	  int x1 = csi->dipole[i_a].no_R_emitter_1();
	  int x2 = csi->dipole[i_a].no_R_emitter_2();
	  if (x1 < x2){temp_ss << "   " << right << showpoint << setw(23) << setprecision(15) << csi->singular_region[x1][x2];}
	  else {temp_ss << "   " << right << showpoint << setw(23) << setprecision(15) << csi->singular_region[x2][x1];}
	}
	temp_ss << endline;
      }
      temp_ss << defaultfloat << noshowpoint << left;
    }
    if (switch_output_cancellation_check == 1){logger << LOG_INFO << temp_ss.str() << endline_flush;}
    //    logger << LOG_INFO << temp_ss.str() << endline_flush;
  }

  logger.finished();
}



