#include "header.hpp"

void observable_set::initialization_CS_splitting_coefficients_QCD(){
  static Logger logger("observable_set::calculate_splitting_coefficients");
  logger.called();

  CS_QCD_gamma_g = (11. / 6.) * C_A - (2. / 3.) * T_R * N_f;
  CS_QCD_gamma_g_ferm = - (2. / 3.) * T_R * N_f;
  CS_QCD_gamma_g_bos = (11. / 6.) * C_A;

  CS_QCD_K_g = (67. / 18. - pi2_6) * C_A - (10. / 9.) * T_R * N_f;
  CS_QCD_K_g_ferm = - (10. / 9.) * T_R * N_f;
  CS_QCD_K_g_bos = (67. / 18. - pi2_6) * C_A;

  CS_QCD_gamma_q = (3. / 2.) * C_F;
  CS_QCD_K_q = (7. / 2. - pi2_6) * C_F;


  // Only K+P terms (CA):

  // P
  // Allow for different conventions in P_qq:
  if (switch_convention_P_qq){CS_QCD_P_qq_delta = C_F * (3. / 2.);}
  else {CS_QCD_P_qq_delta = 0.;}
  CS_QCD_P_gg_delta = (11. / 6. * C_A - 2. / 3. * N_f * T_R);

  // Kbar
  if (switch_convention_plus == 0){
    // original definition of +-distribution:
    CS_QCD_Kbar_qq_delta = C_F * (-(5. - pi2));
  }
  else if (switch_convention_plus == 1){
    // alternative definition of +-distribution:
    CS_QCD_Kbar_qq_delta = C_F * (-(5. - pi2) - pi2_3);
  }
  //  else {logger << LOG_FATAL << "Undefined convention  switch_convention_plus = " << switch_convention_plus  << endline_flush;}

  CS_QCD_Kbar_gg_delta = -(C_A * ((50. / 9.) - pi2) - T_R * N_f * (16. / 9.));

  // Kt
  CS_QCD_Kt_qq_delta = C_F * (-pi2 / 3.);
  CS_QCD_Kt_gg_delta = C_A * (-pi2 / 3.);

  logger.finished();
}


// P
double observable_set::CS_QCD_P_qg(double x){return C_F * (1. + pow(1. - x, 2)) / x;}
double observable_set::CS_QCD_P_gq(double x){return T_R * (pow(x, 2) + pow(1. - x, 2));}

double observable_set::CS_QCD_P_qq_reg(double x){return -C_F * (1. + x);}
// switch_convention_P_qq allow for different conventions via switch: affects all P_qq (apart from P_qq_reg)
double observable_set::CS_QCD_P_qq(double x){
  if (switch_convention_P_qq){return -C_F * (1. + x);}
  else {return 0.;}
}
double observable_set::CS_QCD_P_qq_plus(double x){
  if (switch_convention_P_qq){return C_F * 2. / (1. - x);}
  else {return C_F * (1. + pow(x, 2)) / (1. - x);}
}
double observable_set::CS_QCD_intP_qq_plus(double x){
  if (switch_off_intXplus){return 0.;}
  else {
    if (switch_convention_P_qq){return -C_F * 2 * log(1. - x);}
    else {return C_F * (-.5 * pow(x, 2) - x - 2 * log(1. - x));}
  }
}
//double observable_set::CS_QCD_P_qq_delta(){return CS_QCD_P_qq_delta;}

double observable_set::CS_QCD_P_gg_reg(double x){return 2 * C_A * ((1. - x) / x - 1. + x * (1. - x));}
double observable_set::CS_QCD_P_gg(double x){return 2 * C_A * ((1. - x) / x - 1. + x * (1. - x));}
double observable_set::CS_QCD_P_gg_plus(double x){return 2 * C_A / (1. - x);}
double observable_set::CS_QCD_intP_gg_plus(double x){
  if (switch_off_intXplus){return 0.;}
  else {return -2 * C_A * log(1. - x);}
}
//double observable_set::CS_QCD_P_gg_delta(){return CS_QCD_P_gg_delta;}

// Kbar
double observable_set::CS_QCD_Kbar_qg(double x){return CS_QCD_P_qg(x) * log((1. - x) / x) + C_F * x;}
double observable_set::CS_QCD_Kbar_gq(double x){return CS_QCD_P_gq(x) * log((1. - x) / x) + 2 * T_R * x * (1. - x);}

double observable_set::CS_QCD_Kbar_qq(double x){
  if (switch_convention_plus == 0){// original definition of +-distribution:
    return C_F * (-(1. + x) * log((1. - x) / x) + (1. - x));
  }
  else if (switch_convention_plus == 1){
    return C_F * (-(1. + x) * log((1. - x) / x) + (1. - x) - 2 * log(x) / (1. - x));
  }
  else {exit(1);}
  //  else {logger << LOG_FATAL << "Undefined convention  switch_convention_plus = " << switch_convention_plus  << endline_flush; exit(1);}
}
double observable_set::CS_QCD_Kbar_qq_plus(double x){
  if (switch_convention_plus == 0){
    return C_F * ((2. / (1. - x)) * log((1. - x) / x));
  }
  else if (switch_convention_plus == 1){
    return C_F * ((2. / (1. - x)) * log(1. - x));
  }
  else {exit(1);}
  //  else {logger << LOG_FATAL << "Undefined convention  switch_convention_plus = " << switch_convention_plus  << endline_flush; exit(1);}
}
double observable_set::CS_QCD_intKbar_qq_plus(double x){
  if (switch_off_intXplus){return 0.;}
  else {
    if (switch_convention_plus == 0){
      return C_F * (-pow(log(1. - x), 2) - 2 * gsl_sf_dilog(1. - x) + pi2_3);
    }
    else if (switch_convention_plus == 1){
      return C_F * (-pow(log(1. - x), 2));
    }
    else { exit(1);}
    //    else {logger << LOG_FATAL << "Undefined convention  switch_convention_plus = " << switch_convention_plus  << endline_flush; exit(1);}
  }
}
//double observable_set::CS_QCD_Kbar_qq_delta(){return CS_QCD_Kbar_qq_delta;}

/*
// alternative definition of +-distribution:
//!//double observable_set::CS_QCD_Kbar_qq_plus(double x){return C_F * ((2. / (1. - x)) * log(1. - x));}
//!//double observable_set::CS_QCD_Kbar_qq(double x){return C_F * (-(1. + x) * log((1. - x) / x) + (1. - x) - 2 * log(x) / (1. - x));}
//!//double observable_set::CS_QCD_intKbar_qq_plus(double x){return C_F * (-pow(log(1. - x), 2));}
/////  double observable_set::CS_QCD_intKbar_qq_plus(double x){return 0.;}
*/

double observable_set::CS_QCD_Kbar_gg(double x){return 2 * C_A * ((1. - x) / x - 1. + x * (1. - x)) * log((1. - x) / x);}
double observable_set::CS_QCD_Kbar_gg_plus(double x){return C_A * ((2 / (1. - x)) * log((1. - x) / x));}
double observable_set::CS_QCD_intKbar_gg_plus(double x){
  if (switch_off_intXplus){return 0.;}
  else {
    return C_A * (-pow(log(1. - x), 2) - 2 * gsl_sf_dilog(1. - x) + pi2_3);
  }
}
//double observable_set::CS_QCD_Kbar_gg_delta(){return CS_QCD_Kbar_gg_delta;}

// Kt
double observable_set::CS_QCD_Kt_qg(double x){return CS_QCD_P_qg(x) * log(1. - x);}
double observable_set::CS_QCD_Kt_gq(double x){return CS_QCD_P_gq(x) * log(1. - x);}

double observable_set::CS_QCD_Kt_qq(double x){return CS_QCD_P_qq_reg(x) * log(1. - x);}
double observable_set::CS_QCD_Kt_qq_plus(double x){return C_F * (2 / (1. - x)) * log(1. - x);}
double observable_set::CS_QCD_intKt_qq_plus(double x){
  if (switch_off_intXplus){return 0.;}
  else {
    return -C_F * pow(log(1. - x), 2);
  }
}
//double observable_set::CS_QCD_Kt_qq_delta(){return CS_QCD_Kt_qq_delta;}

double observable_set::CS_QCD_Kt_gg(double x){return CS_QCD_P_gg_reg(x) * log(1. - x);}
double observable_set::CS_QCD_Kt_gg_plus(double x){return C_A * (2 / (1. - x)) * log(1. - x);}
double observable_set::CS_QCD_intKt_gg_plus(double x){
  if (switch_off_intXplus){return 0.;}
  else {
    return -C_A * pow(log(1. - x), 2);
  }
}
//double observable_set::CS_QCD_Kt_gg_delta(){return CS_QCD_Kt_gg_delta;}


