#include "header.hpp"

void phasespace_set::determine_dipole_phasespace_RA(){
  static Logger logger("phasespace_set::determine_dipole_phasespace_RA");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){ logger << LOG_DEBUG_VERBOSE << "xbp_all[" << 0 << "][" << xbi << "] = " << xbp_all[0][xbi] << endline_flush; }
  }

  for (int i_a = 1; i_a < csi->dipole.size(); i_a++){
    //    logger << LOG_DEBUG_VERBOSE << "construct phasespace of dipole no. " << i_a << ": " << csi->dipole[i_a].name() << endline_flush;
    if (csi->dipole[i_a].type_dipole() == 1){
      if (csi->dipole[i_a].massive() == 0){phasespace_ij_k(i_a);}
      else {phasespace_ij_k_massive(i_a);}
    }
    if (csi->dipole[i_a].type_dipole() == 2){
      if (csi->dipole[i_a].massive() == 0){phasespace_ij_a(i_a);}
      else {phasespace_ij_a_massive(i_a);}
    }
    if (csi->dipole[i_a].type_dipole() == 3){phasespace_ai_k(i_a);}
    if (csi->dipole[i_a].type_dipole() == 5){phasespace_ai_b(i_a);}
    //    logger << LOG_DEBUG_VERBOSE << "construct phasespace of dipole no. " << i_a << ": " << csi->dipole[i_a].name() << "   " << csi->dipole[i_a].xy << endline_flush;
  }

  logger.finished();
}


#define p_i xbp_all[0][bp_i]
#define p_j xbp_all[0][bp_j]
#define p_k xbp_all[0][bp_k]
#define pt_ij xbp_all[x_a][bpt_ij]
#define pt_k xbp_all[x_a][bpt_k]
#define y_ij_k csi->dipole[x_a].xy
void phasespace_set::phasespace_ij_k(int x_a){
  static Logger logger("phasespace_ij_k");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();
  y_ij_k = (p_i * p_j) / (p_i * p_j + p_j * p_k + p_k * p_i);
  pt_ij = p_i + p_j - y_ij_k / (1. - y_ij_k) * p_k;
  pt_k = 1. / (1. - y_ij_k) * p_k;

  for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_j){}
    else if (xbi == bp_k){}
    else if (xbi < bp_j){xbp_all[x_a][xbi] = xbp_all[0][xbi];}
    else if (xbi > bp_j){xbp_all[x_a][xbi / 2] = xbp_all[0][xbi];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush;}
  }
  xbp_all[x_a][0] = xbp_all[x_a][1] + xbp_all[x_a][2];
  xbs_all[x_a][0] = xbs_all[0][0];
  xbsqrts_all[x_a][0] = xbsqrts_all[0][0];
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];

  logger.finished();
}
#undef p_i
#undef p_j
#undef p_k
#undef pt_ij
#undef pt_k
#undef y_ij_k


#define p_i xbp_all[0][bp_i]
#define p_j xbp_all[0][bp_j]
#define p_a xbp_all[0][bp_a]
#define pt_ij xbp_all[x_a][bpt_ij]
#define pt_a xbp_all[x_a][bpt_a]
#define x_ij_a csi->dipole[x_a].xy
void phasespace_set::phasespace_ij_a(int x_a){
  static Logger logger("phasespace_ij_a");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_a = csi->dipole[x_a].binary_R_spectator();
  // int bp_b = bp_a % 2 + 1; // not needed here !!!
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  int bpt_a = csi->dipole[x_a].binary_A_spectator(); // always bp_a == bpt_a

  x_ij_a = 1. - (p_i * p_j) / ((p_i + p_j) * p_a);
  pt_ij = p_i + p_j - (1 - x_ij_a) * p_a;
  pt_a = x_ij_a * p_a;

  for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_j){}
    else if (xbi == bp_a){}
    else if (xbi < bp_j){xbp_all[x_a][xbi] = xbp_all[0][xbi];}
    else if (xbi > bp_j){xbp_all[x_a][xbi / 2] = xbp_all[0][xbi];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush;}
  }

  xbp_all[x_a][0] = xbp_all[x_a][1] + xbp_all[x_a][2];
  xbs_all[x_a][0] = x_ij_a * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];

  logger.finished();
}
#undef p_i
#undef p_j
#undef p_a
#undef pt_ij
#undef pt_a
#undef x_ij_a


#define p_a xbp_all[0][bp_a]
#define p_i xbp_all[0][bp_i]
#define p_k xbp_all[0][bp_k]
#define pt_ai xbp_all[x_a][bpt_ai]
#define pt_k xbp_all[x_a][bpt_k]
#define x_ik_a csi->dipole[x_a].xy
void phasespace_set::phasespace_ai_k(int x_a){
  static Logger logger("phasespace_ai_k");
  logger.called();

  int bp_a = csi->dipole[x_a].binary_R_emitter_1();
  int bp_i = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  int bpt_ai = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();

  x_ik_a = 1. - (p_i * p_k) / ((p_i + p_k) * p_a);
  //double u_i = (p_i * p_a) / ((p_i + p_k) * p_a);
  pt_k = p_i + p_k - (1 - x_ik_a) * p_a;
  pt_ai = x_ik_a * p_a;

  for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_k){}
    else if (xbi == bp_a){}
    else if (xbi < bp_i){xbp_all[x_a][xbi] = xbp_all[0][xbi];}
    else if (xbi > bp_i){xbp_all[x_a][xbi / 2] = xbp_all[0][xbi];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush;}
  }

  xbp_all[x_a][0] = xbp_all[x_a][1] + xbp_all[x_a][2];
  xbs_all[x_a][0] = x_ik_a * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];

  logger.finished();
}
#undef p_i
#undef p_k
#undef p_a
#undef pt_ai
#undef pt_k
#undef x_ik_a


#define bpt_ai bp_a
#define bpt_b bp_b
#define p_a xbp_all[0][bp_a]
#define p_i xbp_all[0][bp_i]
#define p_b xbp_all[0][bp_b]
#define pt_ai xbp_all[x_a][bpt_ai]
#define pt_b xbp_all[x_a][bpt_b]
#define x_i_ab csi->dipole[x_a].xy
void phasespace_set::phasespace_ai_b(int x_a){
  static Logger logger("phasespace_ai_b");
  logger.called();

  int bp_a = csi->dipole[x_a].binary_R_emitter_1();
  int bp_i = csi->dipole[x_a].binary_R_emitter_2();
  int bp_b = csi->dipole[x_a].binary_R_spectator();
  //  int bpt_ai = csi->dipole[x_a].binary_A_emitter(); // always bpt_ai == bp_a
  //  int bpt_b = csi->dipole[x_a].binary_A_spectator(); // always bpt_b == bp_b
  x_i_ab = 1. - (p_i * (p_a + p_b)) / (p_a * p_b);
  //  double v_i = (p_i * p_a) / (p_a * p_b);
  pt_ai = x_i_ab * p_a;
  fourvector K = p_a + p_b - p_i;
  fourvector Kt = pt_ai + p_b;
  fourvector KKt = K + Kt;
  double KKt2 = KKt.m2();
  double K2 = K.m2();

  for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_a){}
    else if (xbi == bp_b){xbp_all[x_a][xbi] = xbp_all[0][xbi];}
    else if (xbi == bp_i){}
    else if (xbi < bp_i){xbp_all[x_a][xbi] = xbp_all[0][xbi] - ((2. * (xbp_all[0][xbi] * KKt)) / KKt2) * KKt + (2. * (xbp_all[0][xbi] * K) / K2) * Kt;}
    else if (xbi > bp_i){xbp_all[x_a][xbi / 2] = xbp_all[0][xbi] - ((2. * (xbp_all[0][xbi] * KKt)) / KKt2) * KKt + (2. * (xbp_all[0][xbi] * K) / K2) * Kt;}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush;}
  }
  xbp_all[x_a][0] = xbp_all[x_a][1] + xbp_all[x_a][2];
  xbs_all[x_a][0] = x_i_ab * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];

  logger.finished();
}
#undef p_a
#undef p_i
#undef p_b
#undef pt_ai
#undef pt_b
#undef bpt_ai
#undef bpt_b
#undef x_i_ab


#define p_i xbp_all[0][bp_i]
#define p_j xbp_all[0][bp_j]
#define p_k xbp_all[0][bp_k]
#define pt_ij xbp_all[x_a][bpt_ij]
#define pt_k xbp_all[x_a][bpt_k]
#define m2_i xbs_all[0][bp_i]
#define m2_j xbs_all[0][bp_j]
#define m2_k xbs_all[0][bp_k]
#define m2_ij xbs_all[x_a][bpt_ij]
#define y_ij_k csi->dipole[x_a].xy
void phasespace_set::phasespace_ij_k_massive(int x_a){
  static Logger logger("phasespace_ij_k_massive");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();

  fourvector Q = p_i + p_j + p_k;
  double Q2 = Q.m2();
  double pi_pj = p_i * p_j;

  pt_k = sqrt(lambda(Q2, m2_ij, m2_k) / lambda(Q2, m2_i + m2_j + 2 * pi_pj, m2_k)) * (p_k - ((Q * p_k) / Q2) * Q) + ((Q2 + m2_k - m2_ij) / (2 * Q2)) * Q;
  pt_ij = Q - pt_k;
  // not clear if this belongs here / is valid only for massless case ???
  y_ij_k = (p_i * p_j) / (p_i * p_j + p_j * p_k + p_k * p_i);
  /*
  pt_ij = p_i + p_j - y_ij_k / (1. - y_ij_k) * p_k;
  pt_k = 1. / (1. - y_ij_k) * p_k;
  */

  for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_j){}
    else if (xbi == bp_k){}
    else if (xbi < bp_j){xbp_all[x_a][xbi] = xbp_all[0][xbi];}
    else if (xbi > bp_j){xbp_all[x_a][xbi / 2] = xbp_all[0][xbi];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush;}
  }
  xbp_all[x_a][0] = xbp_all[x_a][1] + xbp_all[x_a][2];
  xbs_all[x_a][0] = xbs_all[0][0];
  xbsqrts_all[x_a][0] = xbsqrts_all[0][0];
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];

  logger.finished();
}
#undef m2_i
#undef m2_j
#undef m2_k
#undef m2_ij
#undef p_i
#undef p_j
#undef p_k
#undef pt_ij
#undef pt_k
#undef y_ij_k


#define p_i xbp_all[0][bp_i]
#define p_j xbp_all[0][bp_j]
#define p_a xbp_all[0][bp_a]
#define pt_ij xbp_all[x_a][bpt_ij]
#define pt_a xbp_all[x_a][bpt_a]
#define m2_i xbs_all[0][bp_i]
#define m2_j xbs_all[0][bp_j]
#define m2_ij xbs_all[x_a][bpt_ij]
#define x_ij_a csi->dipole[x_a].xy
void phasespace_set::phasespace_ij_a_massive(int x_a){
  static Logger logger("phasespace_ij_a_massive");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_a = csi->dipole[x_a].binary_R_spectator();
  // int bp_b = bp_a % 2 + 1; // not needed here !!!
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  int bpt_a = csi->dipole[x_a].binary_A_spectator(); // always bp_a == bpt_a

  double pi_pj = p_i * p_j;
  double pi_pa = p_i * p_a;
  double pj_pa = p_j * p_a;

  double one_minus_x_ij_a = (pi_pj - .5 * (m2_ij - m2_i - m2_j)) / (pi_pa + pj_pa);
  x_ij_a = 1. - one_minus_x_ij_a;
  pt_ij = p_i + p_j - one_minus_x_ij_a * p_a;
  pt_a = x_ij_a * p_a;
  // not clear if this belongs here / is valid only for massless case ???
  /*
  x_ij_a = 1. - (p_i * p_j) / ((p_i + p_j) * p_a);
  pt_ij = p_i + p_j - (1 - x_ij_a) * p_a;
  pt_a = x_ij_a * p_a;
  */
  for (int xbi = 1; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_j){}
    else if (xbi == bp_a){}
    else if (xbi < bp_j){xbp_all[x_a][xbi] = xbp_all[0][xbi];}
    else if (xbi > bp_j){xbp_all[x_a][xbi / 2] = xbp_all[0][xbi];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush;}
  }

  xbp_all[x_a][0] = xbp_all[x_a][1] + xbp_all[x_a][2];
  xbs_all[x_a][0] = x_ij_a * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];

  logger.finished();
}
#undef x_ij_a
#undef m2_i
#undef m2_j
#undef m2_ij
#undef p_i
#undef p_j
#undef p_a
#undef pt_ij
#undef pt_a
#undef x_ij_a
