#include "header.hpp"

void phasespace_set::ac_psp_RA_group_ij_k(int x_a){
  static Logger logger("phasespace_set::ac_psp_RA_group_ij_k");
  logger.called();

  for (int i = 0; i < 3; i++){xbp_all[x_a][i] = xbp_all[0][i];}
  xbs_all[x_a][0] = xbs_all[0][0];
  xbsqrts_all[x_a][0] = xbsqrts_all[0][0];
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];
  ac_psp_dipole(x_a, MC_phasespace.channel - csi->dipole[x_a - 1].sum_channel());
  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();
  fourvector temp_p_ijk = xbp_all[x_a][bpt_ij] + xbp_all[x_a][bpt_k];
  fourvector temp_p_k_boost = xbp_all[x_a][bpt_k].boost(temp_p_ijk);
  double sqrt_p2_ijk = temp_p_ijk.m();
  double y_ij_k = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., 1., exp_ij_k_y, map_technical_x);
  double one_minus_y_ij_k = 1. - y_ij_k;
  double one_minus_z_i = h_propto_pot(random_manager.get_random(x_a, 7, 0), 0., 1., exp_ij_k_z, map_technical_x);
  double z_i = 1. - one_minus_z_i;
  double phi_k_perp = c_phi(random_manager.get_random(x_a, 8, 0));
  double f2 = y_ij_k * z_i / one_minus_z_i;
  double f = sqrt(f2);
  fourvector khat_perp(-f2 * sqrt_p2_ijk, f * sqrt_p2_ijk * cos(phi_k_perp), f * sqrt_p2_ijk * sin(phi_k_perp), -f2 * sqrt_p2_ijk);
  fourvector k_perp = (khat_perp.rotation(temp_p_k_boost)).boost(temp_p_ijk.Pinv());

  xbp_all[0][bp_i] = y_ij_k * (1. + z_i) * xbp_all[x_a][bpt_k] + z_i  * xbp_all[x_a][bpt_ij] + one_minus_z_i * k_perp;
  xbp_all[0][bp_j] = one_minus_z_i * (xbp_all[x_a][bpt_ij] - k_perp) - y_ij_k * z_i * xbp_all[x_a][bpt_k];
  xbp_all[0][bp_k] = one_minus_y_ij_k * xbp_all[x_a][bpt_k];
  for (int xbi = 4; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_j){}
    else if (xbi == bp_k){}
    else if (xbi <  bp_j){xbp_all[0][xbi] = xbp_all[x_a][xbi];}
    else if (xbi >  bp_j){xbp_all[0][xbi] = xbp_all[x_a][xbi / 2];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush; exit(1);}
  }

  logger.finished();
}


void phasespace_set::ag_psp_RA_group_ij_k(int x_a){
  static Logger logger("phasespace_set::ag_psp_RA_group_ij_k");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();
  if (switch_off_new_psp){
    if (xbs_all[0][bp_i + bp_j] == 0.){xbs_all[0][bp_i + bp_j] = (xbp_all[0][bp_i] + xbp_all[0][bp_j]).m2();}
  }
  double p_kp_ij = 2. * (xbp_all[0][bp_k] * (xbp_all[0][bp_i] + xbp_all[0][bp_j]));
  double y_ij_k = 1. / (1. + p_kp_ij / xbs_all[0][bp_i + bp_j]);
  //  double one_minus_y_ij_k = 1. - y_ij_k;
  double z_i = (xbp_all[0][bp_i] * xbp_all[0][bp_k]) / (xbp_all[0][bp_i] * xbp_all[0][bp_k] + xbp_all[0][bp_j] * xbp_all[0][bp_k]);
  csi->dipole[x_a].zuv = z_i;
  double one_minus_z_i = (xbp_all[0][bp_j] * xbp_all[0][bp_k]) / (xbp_all[0][bp_i] * xbp_all[0][bp_k] + xbp_all[0][bp_j] * xbp_all[0][bp_k]);
  double g_alpha = g_propto_pot(y_ij_k, 0., 1., exp_ij_k_y, map_technical_x) * g_propto_pot(one_minus_z_i, 0., 1., exp_ij_k_z, map_technical_x) / (pi * (xbp_all[x_a][bpt_ij] * xbp_all[x_a][bpt_k]) * (1. - y_ij_k));
  if (switch_console_output_phasespace_issue){
    if (g_alpha < 0.){logger << LOG_WARN << "x_a = " << x_a << "   g_alpha = " << g_alpha << endline_flush;}
  }
  if (switch_off_new_psp){
    if (xbs_all[x_a][0] == 0.){xbs_all[x_a][0] = xbs_all[0][0];}
    if (xbsqrts_all[x_a][0] == 0.){xbsqrts_all[x_a][0] = xbsqrts_all[0][0];}
  }
  ag_psp_dipole(x_a, csi->dipole[x_a - 1].sum_channel());
  for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] = MC_phasespace.g_channel[i] * g_alpha;}
  // invert mappings for VAMP implementation
  if (switch_IS_mode_phasespace == 2 || switch_IS_mode_phasespace == 4){
    vector<double> inv_r(3);
    inv_r[0] = inv_propto_pot(y_ij_k, 0., 1., exp_ij_k_y, map_technical_x);
    inv_r[1] = inv_propto_pot(one_minus_z_i, 0., 1., exp_ij_k_z, map_technical_x);
    fourvector k_perp = xbp_all[0][bp_j] + y_ij_k * z_i * xbp_all[x_a][bpt_k];
    k_perp = -(k_perp - xbp_all[x_a][bpt_ij]) / one_minus_z_i;
    fourvector temp_p_ijk = xbp_all[x_a][bpt_ij] + xbp_all[x_a][bpt_k];
    fourvector temp_p_k_boost = xbp_all[x_a][bpt_k].boost(temp_p_ijk);
    k_perp = (k_perp.boost(temp_p_ijk)).rotation_inverse(temp_p_k_boost);
    double phi = atan2(k_perp.x2(), k_perp.x1());
    if (phi < 0){phi += 2 * M_PI;}

    // Use fourvector function for phi ???
    
    inv_r[2] = inv_phi(phi);

    // Here and in the other functions:
    //    if (!(inv_r[2] >= 0) || inv_r[2] > 1){
    if (inv_r[2] <= 0. || inv_r[2] >= 1. || munich_isnan(inv_r[2])){
      if (switch_console_output_phasespace_issue){
	logger << LOG_WARN << "inv_r[2] not in [0; 1] !   r = " << inv_r[2] << "; k_perp.x,y,z = " << k_perp.x1() << ", " << k_perp.x2() << ", " << k_perp.x3() << endline_flush;
      }
      //      if (!(inv_r[2] >= 0)){inv_r[2] = 0.;}
      //      if (inv_r[2] > 1){inv_r[2] = 1.;}
      if (munich_isnan(inv_r[2])){inv_r[2] = map_technical_r;}
      else if (inv_r[2] <= 0){inv_r[2] = map_technical_r;}
      else if (inv_r[2] >= 1){inv_r[2] = 1. - map_technical_r;}
    }
    if (switch_IS_mode_phasespace == 2){
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){
        double g_IS_temp;
        double g_IS = 1.;
        for (int j = 0; j < 3; j++){
          random_psp[(i + 1) * 8 - 3 + j]->get_g_IS(inv_r[j], g_IS_temp);
          g_IS *= g_IS_temp;
        }
        MC_phasespace.g_channel[i] *= g_IS;
      }
    }
    else if (switch_IS_mode_phasespace == 4){
      double g_IS_temp;
      double g_IS = 1.;
      for (int j = 0; j < 3; j++){
        random_psp[map_type_to_list_MC_IS[vector<int> {x_a, 6 + j, 0}]]->get_g_IS(inv_r[j], g_IS_temp);
        g_IS *= g_IS_temp;
      }
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] *= g_IS;}
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){logger << LOG_DEBUG_VERBOSE << "MC_phasespace.g_channel[" << setw(5) << i << "] = " << setprecision(15) << setw(23) << MC_phasespace.g_channel[i] << endline_flush;}
  }

  logger.finished();
}


void phasespace_set::ac_psp_RA_group_ij_a(double & sinx_ij_a_min, int x_a){
  static Logger logger("phasespace_set::ac_psp_RA_group_ij_a");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_a = csi->dipole[x_a].binary_R_spectator();
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  //  int bpt_a = csi->dipole[x_a].binary_A_spectator(); // always bp_a == bpt_a
  int bp_b = bp_a % 2 + 1;

  double one_minus_x_ij_a_min = 1. - sinx_ij_a_min / xbs_all[0][0];
  double x_ij_a = 1.;
  double one_minus_x_ij_a = 0.;

  if (MC_x_dipole[x_a].switch_optimization == -1){
    one_minus_x_ij_a = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., one_minus_x_ij_a_min, exp_ij_a_x, map_technical_x);
    x_ij_a = 1. - one_minus_x_ij_a;
  }
  else {
    // random_MC_tau[1] -> increase number in calculate_intial_tau_IS_x1x2_IS !!! Why is random_MC_tau[1] used ???
    // random numbers missing for MC_x_dipole mappings !!!
    for (int i_c = 0; i_c < MC_x_dipole[x_a].beta.size(); i_c++){if (random_MC_tau[1] <= MC_x_dipole[x_a].beta[i_c]){MC_x_dipole[x_a].channel = i_c; break;}}
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      for (int i_c = 0; i_c < MC_x_dipole[x_a].beta.size(); i_c++){
	logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].beta[" << i_c << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].beta[i_c] << endline_flush;
      }
      logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].channel = " << MC_x_dipole[x_a].channel << endline_flush;
    }
  
    if (MC_x_dipole[x_a].channel == 0){
      one_minus_x_ij_a = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., one_minus_x_ij_a_min, exp_ij_a_x, map_technical_x);
      x_ij_a = 1. - one_minus_x_ij_a;
    }
    else {
      // multi-channel, where channel 0 reproduces the original version, and channels >0 map potential IS resonances...
      if (MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel] < 0){
	double temp_r = random_manager.get_random(x_a, 6, 0);
	x_ij_a = c_propagator_Breit_Wigner(temp_r, -MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel], sinx_ij_a_min, xbs_all[0][0]) / xbs_all[0][0];
	one_minus_x_ij_a = 1. - x_ij_a;
	if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "s_dipole = " << setw(23) << setprecision(15) << sqrt(x_ij_a * xbs_all[0][0]) << "   " << "x_ij_a = " << setw(23) << setprecision(15) << x_ij_a << "   " << "xbs_all[0][0] = " << setw(23) << setprecision(15) << xbs_all[0][0] << endline_flush; }
      }
    }
  }

  xbp_all[x_a][bp_a] = x_ij_a * xbp_all[0][bp_a];
  xbp_all[x_a][bp_b] = xbp_all[0][bp_b];
  xbp_all[x_a][0] = xbp_all[x_a][bp_a] + xbp_all[x_a][bp_b];
  xbs_all[x_a][0] = x_ij_a * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];
  ac_psp_dipole(x_a, MC_phasespace.channel - csi->dipole[x_a - 1].sum_channel());
  fourvector temp_p_ija = xbp_all[x_a][bpt_ij] + xbp_all[x_a][bp_a];
  fourvector temp_p_a_boost = xbp_all[x_a][bp_a].boost(temp_p_ija);
  double sqrt_p2_ija = temp_p_ija.m();
  double one_minus_z_i = h_propto_pot(random_manager.get_random(x_a, 7, 0), 0., 1., exp_ij_a_z, map_technical_x);
  double z_i = 1. - one_minus_z_i;
  double phx_a_perp = c_phi(random_manager.get_random(x_a, 8, 0));
  double f2 = z_i * one_minus_x_ij_a / (one_minus_z_i * x_ij_a);
  double f = sqrt(f2);
  fourvector khat_perp(-f2 * sqrt_p2_ija, f * sqrt_p2_ija * cos(phx_a_perp), f * sqrt_p2_ija * sin(phx_a_perp), -f2 * sqrt_p2_ija);
  fourvector k_perp = (khat_perp.rotation(temp_p_a_boost)).boost(temp_p_ija.Pinv());

  xbp_all[0][bp_i] = one_minus_x_ij_a / x_ij_a * (1. + z_i) * xbp_all[x_a][bp_a] + z_i * xbp_all[x_a][bpt_ij] + one_minus_z_i * k_perp;
  xbp_all[0][bp_j] = one_minus_z_i * (xbp_all[x_a][bpt_ij] - k_perp) - one_minus_x_ij_a / x_ij_a * z_i * xbp_all[x_a][bp_a];
  xbp_all[0][bp_a] = xbp_all[x_a][bp_a] / x_ij_a;
  for (int xbi = 4; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_j){}
    else if (xbi < bp_j){xbp_all[0][xbi] = xbp_all[x_a][xbi];}
    else if (xbi > bp_j){xbp_all[0][xbi] = xbp_all[x_a][xbi / 2];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush; exit(1);}
  }

  logger.finished();
}


void phasespace_set::ag_psp_RA_group_ij_a(double & sinx_ij_a_min, int x_a){
  static Logger logger("phasespace_set::ag_psp_RA_group_ij_a");
  logger.called();

  int bp_i = csi->dipole[x_a].binary_R_emitter_1();
  int bp_j = csi->dipole[x_a].binary_R_emitter_2();
  int bp_a = csi->dipole[x_a].binary_R_spectator();
  // int bp_b = bp_a % 2 + 1; //  not needed !!!
  int bpt_ij = csi->dipole[x_a].binary_A_emitter();
  //  int bpt_a = csi->dipole[x_a].binary_A_spectator(); // always bp_a == bpt_a
  if (switch_off_new_psp){
    if (xbs_all[0][bp_i + bp_j] == 0.){xbs_all[0][bp_i + bp_j] = (xbp_all[0][bp_i] + xbp_all[0][bp_j]).m2();}
  }
  double pa_p12 = 2. * (xbp_all[0][bp_a] * (xbp_all[0][bp_i] + xbp_all[0][bp_j]));

  double one_minus_x_ij_a_min = 1. - sinx_ij_a_min / xbs_all[0][0];
  double one_minus_x_ij_a = xbs_all[0][bp_i + bp_j] / pa_p12;
  double x_ij_a = 1. - one_minus_x_ij_a;

  double g_x_ij_a = 0.;
  if (MC_x_dipole[x_a].switch_optimization == -1){
    g_x_ij_a = g_propto_pot(one_minus_x_ij_a, 0., one_minus_x_ij_a_min, exp_ij_a_x, map_technical_x);
  }
  else {
    double x_ij_a_s_part = x_ij_a * xbs_all[0][0];
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "x_ij_a_s_part = " << x_ij_a_s_part << "   sqrt -> " << sqrt(x_ij_a_s_part) << endline_flush; }
    // multi-channel, where channel 0 reproduces the original version, and channels >0 map potential IS resonances...
    for (int i_c = 0; i_c < MC_x_dipole[x_a].g_channel.size(); i_c++){
      if (i_c == 0){
	MC_x_dipole[x_a].g_channel[i_c] = g_propto_pot(one_minus_x_ij_a, 0., one_minus_x_ij_a_min, exp_ij_a_x, map_technical_x);
      }
      else {
	if (MC_x_dipole_mapping[x_a][i_c] < 0){MC_x_dipole[x_a].g_channel[i_c] = xbs_all[0][0] * g_propagator_Breit_Wigner(x_ij_a_s_part, -MC_x_dipole_mapping[x_a][i_c], sinx_ij_a_min, xbs_all[0][0]);}
	///      else if (tau_MC_map[i_c] > 0){MC_x_dipole[x_a].g_channel[i_c] = s_had * g_propagator_vanishing_width(x_ij_a_s_part, M2[tau_MC_map[i_c]], tau_0_s_had, s_had, nuxs);}
	if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	  logger << LOG_DEBUG_VERBOSE << "sinx_ij_a_min = " << setw(23) << setprecision(15) << sinx_ij_a_min << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "x_ij_a_s_part = " << setw(23) << setprecision(15) << x_ij_a_s_part << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "xbs_all[0][0] = " << setw(23) << setprecision(15) << xbs_all[0][0] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "-MC_x_dipole_mapping[x_a][i_c] = " << setw(23) << setprecision(15) << -MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[x_a].g_channel[i_c] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].g_channel[i_c] << endline_flush;
	}
      }
    }
    g_x_ij_a = 0.;
    for (int i = 0; i < MC_x_dipole[x_a].g_channel.size(); i++){g_x_ij_a += MC_x_dipole[x_a].alpha[i] * MC_x_dipole[x_a].g_channel[i];}
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      for (int i = 0; i < MC_x_dipole[x_a].g_channel.size(); i++){
	logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].alpha[" << i << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].alpha[i] << "   " << "MC_x_dipole[" << x_a << "].g_channel[" << i << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].g_channel[i] << endline_flush;
      }
      logger << LOG_DEBUG_VERBOSE << "g_x_ij_a = " << g_x_ij_a << endline_flush;
    }
  }

  double z_i = (xbp_all[0][bp_i] * xbp_all[0][bp_a]) / (xbp_all[0][bp_i] * xbp_all[0][bp_a] + xbp_all[0][bp_j] * xbp_all[0][bp_a]);
  csi->dipole[x_a].zuv = z_i;
  double one_minus_z_i = (xbp_all[0][bp_j] * xbp_all[0][bp_a]) / (xbp_all[0][bp_i] * xbp_all[0][bp_a] + xbp_all[0][bp_j] * xbp_all[0][bp_a]);
  double g_alpha = g_x_ij_a * g_propto_pot(one_minus_z_i, 0., 1., exp_ij_a_z, map_technical_x) / (pi * (xbp_all[x_a][bpt_ij] * xbp_all[0][bp_a]));
  if (switch_console_output_phasespace_issue){if (g_alpha < 0.){logger << LOG_WARN << "x_a = " << x_a << "   g_alpha = " << g_alpha << endline_flush;}}
  if (switch_off_new_psp){
    if (xbs_all[x_a][0] == 0.){xbs_all[x_a][0] = x_ij_a * xbs_all[0][0];}
    if (xbsqrts_all[x_a][0] == 0.){xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);}
  }
  ag_psp_dipole(x_a, csi->dipole[x_a - 1].sum_channel());
  for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] = MC_phasespace.g_channel[i] * g_alpha;}
  // invert mappings for VAMP implementation
  if (switch_IS_mode_phasespace == 2 || switch_IS_mode_phasespace == 4){
    vector<double> inv_r(3);
    inv_r[0] = inv_propto_pot(one_minus_x_ij_a, 0., one_minus_x_ij_a_min, exp_ij_a_x, map_technical_x);
    inv_r[1] = inv_propto_pot(one_minus_z_i, 0., 1., exp_ij_a_z, map_technical_x);
    fourvector k_perp = xbp_all[0][bp_j] + one_minus_x_ij_a / x_ij_a * z_i * xbp_all[x_a][bp_a];
    k_perp = -(k_perp - xbp_all[x_a][bpt_ij]) / one_minus_z_i;
    fourvector temp_p_ija = xbp_all[x_a][bpt_ij] + xbp_all[x_a][bp_a];
    fourvector temp_p_a_boost = xbp_all[x_a][bp_a].boost(temp_p_ija);
    k_perp = (k_perp.boost(temp_p_ija)).rotation_inverse(temp_p_a_boost);
    double phi = atan2(k_perp.x2(),k_perp.x1());
    if (phi < 0){phi += 2 * M_PI;}

    // Use fourvector function for phi ???
    
    inv_r[2] = inv_phi(phi);
    if (inv_r[2] <= 0. || inv_r[2] >= 1. || munich_isnan(inv_r[2])){
      if (switch_console_output_phasespace_issue){
	logger << LOG_WARN << "inv_r[2] not in [0; 1] !   r = " << inv_r[2] << "; k_perp.x,y,z = " << k_perp.x1() << ", " << k_perp.x2() << ", " << k_perp.x3() << endline_flush;
      }
      if (munich_isnan(inv_r[2])){inv_r[2] = map_technical_r;}
      else if (inv_r[2] <= 0){inv_r[2] = map_technical_r;}
      else if (inv_r[2] >= 1){inv_r[2] = 1. - map_technical_r;}
    }
    if (switch_IS_mode_phasespace == 2){
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){
        double g_IS_temp;
        double g_IS = 1.;
        for (int j = 0; j < 3; j++){
          random_psp[(i + 1) * 8 - 3 + j]->get_g_IS(inv_r[j], g_IS_temp);
          g_IS *= g_IS_temp;
        }
        MC_phasespace.g_channel[i] *= g_IS;
      }
    }
    else if (switch_IS_mode_phasespace == 4){
      double g_IS_temp;
      double g_IS = 1.;
      for (int j = 0; j < 3; j++){
        random_psp[map_type_to_list_MC_IS[vector<int> {x_a, 6 + j, 0}]]->get_g_IS(inv_r[j], g_IS_temp);
        g_IS *= g_IS_temp;
      }
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] *= g_IS;}
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){logger << LOG_DEBUG_VERBOSE << "MC_phasespace.g_channel[" << setw(5) << i << "] = " << setprecision(15) << setw(23) << MC_phasespace.g_channel[i] << endline_flush;}
  }

  logger.finished();
}


void phasespace_set::ac_psp_RA_group_ai_k(double & sinx_ik_a_min, int x_a){
  static Logger logger("phasespace_set::ac_psp_RA_group_ai_k");
  logger.called();

  int bp_a = csi->dipole[x_a].binary_R_emitter_1();
  int bp_i = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  //  int bpt_ai = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();
  int bp_b = bp_a % 2 + 1;

  double one_minus_x_ik_a_min = 1. - sinx_ik_a_min / xbs_all[0][0];
  double x_ik_a = 1.;
  double one_minus_x_ik_a = 0.;

  if (MC_x_dipole[x_a].switch_optimization == -1){
    one_minus_x_ik_a = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., one_minus_x_ik_a_min, exp_ai_k_x, map_technical_x);
    x_ik_a = 1. - one_minus_x_ik_a;
  }
  else {
    // random_MC_tau[1] -> increase number in calculate_intial_tau_IS_x1x2_IS !!! Why is random_MC_tau[1] used ???
    // random numbers missing for MC_x_dipole mappings !!!
    for (int i_c = 0; i_c < MC_x_dipole[x_a].beta.size(); i_c++){if (random_MC_tau[1] <= MC_x_dipole[x_a].beta[i_c]){MC_x_dipole[x_a].channel = i_c; break;}}
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      for (int i_c = 0; i_c < MC_x_dipole[x_a].beta.size(); i_c++){
	logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].beta[" << i_c << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].beta[i_c] << endline_flush;
      }
      logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].channel = " << MC_x_dipole[x_a].channel << endline_flush;
    }
  
    if (MC_x_dipole[x_a].channel == 0){
      one_minus_x_ik_a = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., one_minus_x_ik_a_min, exp_ai_k_x, map_technical_x);
      x_ik_a = 1. - one_minus_x_ik_a;
    }
    else {
      // multi-channel, where channel 0 reproduces the original version, and channels >0 map potential IS resonances...
      if (MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel] < 0){
	double temp_r = random_manager.get_random(x_a, 6, 0);
	x_ik_a = c_propagator_Breit_Wigner(temp_r, -MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel], sinx_ik_a_min, xbs_all[0][0]) / xbs_all[0][0];
	one_minus_x_ik_a = 1. - x_ik_a;
	if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "s_dipole = " << setw(23) << setprecision(15) << sqrt(x_ik_a * xbs_all[0][0]) << "   " << "x_ik_a = " << setw(23) << setprecision(15) << x_ik_a << "   " << "xbs_all[0][0] = " << setw(23) << setprecision(15) << xbs_all[0][0] << endline_flush; }
      }
    }
  }

  xbp_all[x_a][bp_a] = x_ik_a * xbp_all[0][bp_a];
  xbp_all[x_a][bp_b] = xbp_all[0][bp_b];
  xbp_all[x_a][0] = xbp_all[x_a][bp_a] + xbp_all[x_a][bp_b];
  xbs_all[x_a][0] = x_ik_a * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];
  ac_psp_dipole(x_a, MC_phasespace.channel - csi->dipole[x_a - 1].sum_channel());
  fourvector temp_p_aik = xbp_all[x_a][bp_a] + xbp_all[x_a][bpt_k];
  fourvector temp_p_k_boost = xbp_all[x_a][bp_a].boost(temp_p_aik);
  double sqrt_p2_aik = temp_p_aik.m();
  double u_i = h_propto_pot(random_manager.get_random(x_a, 7, 0), 0., 1., exp_ai_k_u, map_technical_x);
  double one_minus_u_i = 1. - u_i;
  double phi_k_perp = c_phi(random_manager.get_random(x_a, 8, 0));
  double f2 = u_i * one_minus_x_ik_a / (one_minus_u_i * x_ik_a);
  double f = sqrt(f2);
  fourvector khat_perp(-f2 * sqrt_p2_aik, f * sqrt_p2_aik * cos(phi_k_perp), f * sqrt_p2_aik * sin(phi_k_perp), -f2 * sqrt_p2_aik);
  fourvector k_perp = (khat_perp.rotation(temp_p_k_boost)).boost(temp_p_aik.Pinv());

  xbp_all[0][bp_i] = one_minus_x_ik_a / x_ik_a * (1. + u_i) * xbp_all[x_a][bp_a] + u_i * xbp_all[x_a][bpt_k] + one_minus_u_i * k_perp;
  xbp_all[0][bp_k] = one_minus_u_i * (xbp_all[x_a][bpt_k] - k_perp) - one_minus_x_ik_a / x_ik_a * u_i * xbp_all[x_a][bp_a];
  xbp_all[0][bp_a] = xbp_all[x_a][bp_a] / x_ik_a;
  for (int xbi = 4; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else if (xbi == bp_k){}
    else if (xbi < bp_i){xbp_all[0][xbi] = xbp_all[x_a][xbi];}
    else if (xbi > bp_i){xbp_all[0][xbi] = xbp_all[x_a][xbi / 2];}
    else {logger << LOG_FATAL << "xbi = " << xbi << " : Should not happen!" << endline_flush; exit(1);}
  }

  logger.finished();
}


void phasespace_set::ag_psp_RA_group_ai_k(double & sinx_ik_a_min, int x_a){
  static Logger logger("phasespace_set::ag_psp_RA_group_ai_k");
  logger.called();

  int bp_a = csi->dipole[x_a].binary_R_emitter_1();
  int bp_i = csi->dipole[x_a].binary_R_emitter_2();
  int bp_k = csi->dipole[x_a].binary_R_spectator();
  //  int bpt_ai = csi->dipole[x_a].binary_A_emitter();
  int bpt_k = csi->dipole[x_a].binary_A_spectator();
  //  int bp_b = bp_a % 2 + 1;
  if (switch_off_new_psp){
    if (xbs_all[0][bp_i + bp_k] == 0.){xbs_all[0][bp_i + bp_k] = (xbp_all[0][bp_i] + xbp_all[0][bp_k]).m2();}
  }
  double pa_p12 = 2. * (xbp_all[0][bp_a] * (xbp_all[0][bp_i] + xbp_all[0][bp_k]));

  double one_minus_x_ik_a_min = 1. - sinx_ik_a_min / xbs_all[0][0];
  double one_minus_x_ik_a = xbs_all[0][bp_i + bp_k] / pa_p12;
  double x_ik_a = 1. - one_minus_x_ik_a;

  double g_x_ik_a = 0.;
  if (MC_x_dipole[x_a].switch_optimization == -1){
    g_x_ik_a = g_propto_pot(one_minus_x_ik_a, 0., one_minus_x_ik_a_min, exp_ai_k_x, map_technical_x);
  }
  else {
    double x_ik_a_s_part = x_ik_a * xbs_all[0][0];
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "x_ik_a_s_part = " << x_ik_a_s_part << "   sqrt -> " << sqrt(x_ik_a_s_part) << endline_flush; }
    // multi-channel, where channel 0 reproduces the original version, and channels >0 map potential IS resonances...
    for (int i_c = 0; i_c < MC_x_dipole[x_a].g_channel.size(); i_c++){
      if (i_c == 0){
	MC_x_dipole[x_a].g_channel[i_c] = g_propto_pot(one_minus_x_ik_a, 0., one_minus_x_ik_a_min, exp_ai_k_x, map_technical_x);
      }
      else {
	if (MC_x_dipole_mapping[x_a][i_c] < 0){MC_x_dipole[x_a].g_channel[i_c] = xbs_all[0][0] * g_propagator_Breit_Wigner(x_ik_a_s_part, -MC_x_dipole_mapping[x_a][i_c], sinx_ik_a_min, xbs_all[0][0]);}
	///      else if (tau_MC_map[i_c] > 0){MC_x_dipole[x_a].g_channel[i_c] = s_had * g_propagator_vanishing_width(x_ik_a_s_part, M2[tau_MC_map[i_c]], tau_0_s_had, s_had, nuxs);}
	if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	  logger << LOG_DEBUG_VERBOSE << "sinx_ik_a_min = " << setw(23) << setprecision(15) << sinx_ik_a_min << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "x_ik_a_s_part = " << setw(23) << setprecision(15) << x_ik_a_s_part << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "xbs_all[0][0] = " << setw(23) << setprecision(15) << xbs_all[0][0] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "-MC_x_dipole_mapping[x_a][i_c] = " << setw(23) << setprecision(15) << -MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[x_a].g_channel[i_c] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].g_channel[i_c] << endline_flush;
	}
      }
    }
    g_x_ik_a = 0.;
    for (int i = 0; i < MC_x_dipole[x_a].g_channel.size(); i++){g_x_ik_a += MC_x_dipole[x_a].alpha[i] * MC_x_dipole[x_a].g_channel[i];}
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      for (int i = 0; i < MC_x_dipole[x_a].g_channel.size(); i++){
        logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].alpha[" << i << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].alpha[i] << "   " << "MC_x_dipole[" << x_a << "].g_channel[" << i << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].g_channel[i] << endline_flush;
      }
      logger << LOG_DEBUG_VERBOSE << "g_x_ik_a = " << g_x_ik_a << endline_flush;
    }
  }

  double u_i = (xbp_all[0][bp_i] * xbp_all[0][bp_a]) / (xbp_all[0][bp_i] * xbp_all[0][bp_a] + xbp_all[0][bp_k] * xbp_all[0][bp_a]);
  csi->dipole[x_a].zuv = u_i;
  double one_minus_u_i = 1. - u_i;
  double g_alpha = g_x_ik_a * g_propto_pot(u_i, 0., 1., exp_ai_k_u, map_technical_x) / (pi * (xbp_all[x_a][bpt_k] * xbp_all[0][bp_a]));
  if (switch_console_output_phasespace_issue){if (g_alpha < 0.){logger << LOG_WARN << "x_a = " << x_a << "   g_alpha = " << g_alpha << endline_flush;}}
  if (switch_off_new_psp){
    if (xbs_all[x_a][0] == 0.){xbs_all[x_a][0] = x_ik_a * xbs_all[0][0];}
    if (xbsqrts_all[x_a][0] == 0.){xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);}
  }
  ag_psp_dipole(x_a, csi->dipole[x_a - 1].sum_channel());
  for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] = MC_phasespace.g_channel[i] * g_alpha;}
  // invert mappings for VAMP implementation
  if (switch_IS_mode_phasespace == 2 || switch_IS_mode_phasespace == 4){
    vector<double> inv_r(3);
    inv_r[0] = inv_propto_pot(one_minus_x_ik_a, 0., one_minus_x_ik_a_min, exp_ai_k_x, map_technical_x);
    inv_r[1] = inv_propto_pot(u_i, 0., 1., exp_ai_k_u, map_technical_x);
    fourvector k_perp = xbp_all[0][bp_k] + one_minus_x_ik_a / x_ik_a * u_i * xbp_all[x_a][bp_a];
    k_perp = -(k_perp - xbp_all[x_a][bpt_k]) / one_minus_u_i;
    fourvector temp_p_aik = xbp_all[x_a][bp_a] + xbp_all[x_a][bpt_k];
    fourvector temp_p_k_boost = xbp_all[x_a][bp_a].boost(temp_p_aik);
    k_perp = (k_perp.boost(temp_p_aik)).rotation_inverse(temp_p_k_boost);
    double phi = atan2(k_perp.x2(),k_perp.x1());
    if (phi < 0){phi += 2 * M_PI;}

    // Use fourvector function for phi ???
    
    inv_r[2] = inv_phi(phi);
    if (inv_r[2] <= 0. || inv_r[2] >= 1. || munich_isnan(inv_r[2])){
      if (switch_console_output_phasespace_issue){
	logger << LOG_WARN << "inv_r[2] not in [0; 1] !   r = " << inv_r[2] << "; k_perp.x,y,z = " << k_perp.x1() << ", " << k_perp.x2() << ", " << k_perp.x3() << endline_flush;
      }
      if (munich_isnan(inv_r[2])){inv_r[2] = map_technical_r;}
      else if (inv_r[2] <= 0){inv_r[2] = map_technical_r;}
      else if (inv_r[2] >= 1){inv_r[2] = 1. - map_technical_r;}
    }
    if (switch_IS_mode_phasespace == 2){
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){
        double g_IS_temp;
        double g_IS = 1.;
        for (int j = 0; j < 3; j++){
          random_psp[(i + 1) * 8 - 3 + j]->get_g_IS(inv_r[j], g_IS_temp);
          g_IS *= g_IS_temp;
        }
        MC_phasespace.g_channel[i] *= g_IS;
      }
    }
    else if (switch_IS_mode_phasespace == 4){
      double g_IS_temp;
      double g_IS = 1.;
      for (int j = 0; j < 3; j++){
        random_psp[map_type_to_list_MC_IS[vector<int> {x_a, 6 + j, 0}]]->get_g_IS(inv_r[j], g_IS_temp);
        g_IS *= g_IS_temp;
      }
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] *= g_IS;}
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){logger << LOG_DEBUG_VERBOSE << "MC_phasespace.g_channel[" << setw(5) << i << "] = " << setprecision(15) << setw(23) << MC_phasespace.g_channel[i] << endline_flush;}
  }

  logger.finished();
}


void phasespace_set::ac_psp_RA_group_ai_b(double & sinx_i_ab_min, int x_a){
  static Logger logger("phasespace_set::ac_psp_RA_group_ai_b");
  logger.called();

  int bp_a = csi->dipole[x_a].binary_R_emitter_1();
  int bp_i = csi->dipole[x_a].binary_R_emitter_2();
  int bp_b = csi->dipole[x_a].binary_R_spectator();
  //  int bpt_ai = csi->dipole[x_a].binary_A_emitter(); // always bpt_ai == bp_a
  //  int bpt_b = csi->dipole[x_a].binary_A_spectator(); // always bpt_b == bp_b

  double one_minus_x_i_ab_min = 1. - sinx_i_ab_min / xbs_all[0][0];
  double x_i_ab = 1.;
  double one_minus_x_i_ab = 0.;
  
  if (MC_x_dipole[x_a].switch_optimization == -1){
    one_minus_x_i_ab = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., one_minus_x_i_ab_min, exp_ai_b_x, map_technical_x);
    x_i_ab = 1. - one_minus_x_i_ab;
  }
  else {
    // random_MC_tau[1] -> increase number in calculate_intial_tau_IS_x1x2_IS !!! Why is random_MC_tau[1] used ???
    // random numbers missing for MC_x_dipole mappings !!!
    for (int i_c = 0; i_c < MC_x_dipole[x_a].beta.size(); i_c++){if (random_MC_tau[1] <= MC_x_dipole[x_a].beta[i_c]){MC_x_dipole[x_a].channel = i_c; break;}}
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      for (int i_c = 0; i_c < MC_x_dipole[x_a].beta.size(); i_c++){
	logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].beta[" << i_c << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].beta[i_c] << endline_flush;
      }
      logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].channel = " << MC_x_dipole[x_a].channel << endline_flush;
    }
  
    if (MC_x_dipole[x_a].channel == 0){
      one_minus_x_i_ab = h_propto_pot(random_manager.get_random(x_a, 6, 0), 0., one_minus_x_i_ab_min, exp_ai_b_x, map_technical_x);
      x_i_ab = 1. - one_minus_x_i_ab;
    }
    else {
      // multi-channel, where channel 0 reproduces the original version, and channels >0 map potential IS resonances...
      if (MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel] < 0){
	double temp_r = random_manager.get_random(x_a, 6, 0);
	x_i_ab = c_propagator_Breit_Wigner(temp_r, -MC_x_dipole_mapping[x_a][MC_x_dipole[x_a].channel], sinx_i_ab_min, xbs_all[0][0]) / xbs_all[0][0];
	one_minus_x_i_ab = 1. - x_i_ab;
	if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "s_dipole = " << setw(23) << setprecision(15) << sqrt(x_i_ab * xbs_all[0][0]) << "   " << "x_i_ab = " << setw(23) << setprecision(15) << x_i_ab << "   " << "xbs_all[0][0] = " << setw(23) << setprecision(15) << xbs_all[0][0] << endline_flush; }
      }
    }
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      logger << LOG_DEBUG_VERBOSE << "sqrt(sinx_i_ab_min) = " << sqrt(sinx_i_ab_min) << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "xbsqrts_all[0][0] = " << xbsqrts_all[0][0] << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "x_i_ab = " << x_i_ab << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "one_minus_x_i_ab = " << one_minus_x_i_ab << endline_flush;
    }
    if (munich_isnan(x_i_ab)){logger << LOG_WARN << "x_i_ab = " << x_i_ab << endline_flush;}
  }

  xbp_all[x_a][bp_a] = x_i_ab * xbp_all[0][bp_a];
  xbp_all[x_a][bp_b] = xbp_all[0][bp_b];
  xbp_all[x_a][0] = xbp_all[x_a][bp_a] + xbp_all[x_a][bp_b];
  xbs_all[x_a][0] = x_i_ab * xbs_all[0][0];
  xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);
  xbp_all[x_a][xb_out_dipoles] = xbp_all[x_a][0];
  xbs_all[x_a][xb_out_dipoles] = xbs_all[x_a][0];
  xbsqrts_all[x_a][xb_out_dipoles] = xbsqrts_all[x_a][0];
  fourvector temp_p_aib = xbp_all[x_a][bp_a] + xbp_all[x_a][bp_b];
  fourvector temp_p_a_boost = xbp_all[x_a][bp_a].boost(temp_p_aib);
  double sqrt_p2_aib = temp_p_aib.m();
  double v_i = h_propto_pot(random_manager.get_random(x_a, 7, 0), 0., one_minus_x_i_ab, exp_ai_b_v, map_technical_x);
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "v_i = " << v_i << endline_flush; }
  if (munich_isnan(v_i)){logger << LOG_WARN << "v_i = " << v_i << endline_flush;}
  double one_minus_v_i = 1. - v_i;
  double phi_k_perp = c_phi(random_manager.get_random(x_a, 8, 0));
  if (munich_isnan(phi_k_perp)){logger << LOG_WARN << "phi_k_perp = " << phi_k_perp << endline_flush;}
  double f2 = v_i * (one_minus_x_i_ab - v_i) / x_i_ab;
  double f = sqrt(f2);
  fourvector khat_perp(0., f * sqrt_p2_aib * cos(phi_k_perp), f * sqrt_p2_aib * sin(phi_k_perp), 0.);
  fourvector k_perp = (khat_perp.rotation(temp_p_a_boost)).boost(temp_p_aib.Pinv());
  fourvector K = (x_i_ab + v_i) / x_i_ab * xbp_all[x_a][bp_a] + one_minus_v_i * xbp_all[x_a][bp_b] - k_perp;
  if (munich_isnan(K)){logger << LOG_WARN << "K = " << K << endline_flush;}
  fourvector Kt = xbp_all[x_a][bp_a] + xbp_all[x_a][bp_b];
  if (munich_isnan(Kt)){logger << LOG_WARN << "Kt = " << Kt << endline_flush;}
  fourvector KKt = K + Kt;
  xbp_all[0][bp_i] = v_i * xbp_all[x_a][bp_b] + (one_minus_x_i_ab - v_i) / x_i_ab * xbp_all[x_a][bp_a] + k_perp;
  xbp_all[0][bp_a] = xbp_all[x_a][bp_a] / x_i_ab;
  xbp_all[0][bp_b] = xbp_all[x_a][bp_b];
  ac_psp_dipole(x_a, MC_phasespace.channel - csi->dipole[x_a - 1].sum_channel());

  double prod_em_sp = xbp_all[x_a][bp_a] * xbp_all[x_a][bp_b];
  for (int xbi = 4; xbi < xbp_all[0].size(); xbi = xbi * 2){
    if      (xbi == bp_i){}
    else {
      int temp_xbi = 0;
      if      (xbi < bp_i){temp_xbi = xbi;}
      else if (xbi > bp_i){temp_xbi = xbi / 2;}
      double prod_em_xbi = xbp_all[x_a][temp_xbi] * xbp_all[x_a][bp_a];
      double prod_sp_xbi = xbp_all[x_a][temp_xbi] * xbp_all[x_a][bp_b];
      double prod_perp_xbi = xbp_all[x_a][temp_xbi] * k_perp;

      xbp_all[0][xbi] = xbp_all[x_a][temp_xbi]
        - ((2. * x_i_ab + v_i) * prod_em_xbi + (2. - v_i) * x_i_ab * prod_sp_xbi - x_i_ab * prod_perp_xbi) / ((4. * x_i_ab + v_i - x_i_ab * v_i) * prod_em_sp) * KKt
        + (prod_em_xbi + prod_sp_xbi) / prod_em_sp * K;

      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	if (munich_isnan(xbp_all[0][xbi])){
	  logger << LOG_DEBUG_VERBOSE << "i_acc = " << i_acc << "   xbi = " << xbi << "   xbp_all[" << x_a << "][" << temp_xbi << "] = " << xbp_all[x_a][temp_xbi] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "K = " << K << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "Kt = " << Kt << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "x_i_ab = " << x_i_ab << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "v_i = " << v_i << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "phi_k_perp = " << phi_k_perp << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "prod_em_sp = " << prod_em_sp << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "prod_em_xbi = " << prod_em_xbi << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "prod_sp_xbi = " << prod_sp_xbi << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "xbp_all[" << x_a << "][" << bp_a << "] = " << xbp_all[x_a][bp_a] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "xbp_all[" << x_a << "][" << bp_b << "] = " << xbp_all[x_a][bp_b] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "xbp_all[" << x_a << "][" << temp_xbi << "] = " << xbp_all[x_a][temp_xbi] << endline_flush;
	}
      }
    }
  }
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int xbi = 4; xbi < xbp_all[0].size(); xbi = xbi * 2){
      logger << LOG_DEBUG_VERBOSE << "xbp_all[" << 0 << "][" << xbi << "] = " << xbp_all[0][xbi] << endline_flush;
    }
  }

  logger.finished();
}


void phasespace_set::ag_psp_RA_group_ai_b(double & sinx_i_ab_min, int x_a){
  static Logger logger("phasespace_set::ag_psp_RA_group_ai_b");
  logger.called();

  int bp_a = csi->dipole[x_a].binary_R_emitter_1();
  int bp_i = csi->dipole[x_a].binary_R_emitter_2();
  int bp_b = csi->dipole[x_a].binary_R_spectator();
  //  int bpt_ai = csi->dipole[x_a].binary_A_emitter(); // always bpt_ai == bp_a
  //  int bpt_b = csi->dipole[x_a].binary_A_spectator(); // always bpt_b == bp_b

  double one_minus_x_i_ab_min = 1. - sinx_i_ab_min / xbs_all[0][0];
  double one_minus_x_i_ab = 2 * xbp_all[0][bp_i].x0() / xbsqrts_all[0][0];
  double x_i_ab = 1. - one_minus_x_i_ab;
  double g_x_i_ab = 0.;
  if (MC_x_dipole[x_a].switch_optimization == -1){
    g_x_i_ab = g_propto_pot(one_minus_x_i_ab, 0., one_minus_x_i_ab_min, exp_ai_b_x, map_technical_x);
  }
  else {
    double x_i_ab_s_part = x_i_ab * xbs_all[0][0];
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "x_i_ab_s_part = " << x_i_ab_s_part << "   sqrt -> " << sqrt(x_i_ab_s_part) << endline_flush; }
    for (int i_c = 0; i_c < MC_x_dipole[x_a].g_channel.size(); i_c++){
      if (i_c == 0){
	MC_x_dipole[x_a].g_channel[i_c] = g_propto_pot(one_minus_x_i_ab, 0., one_minus_x_i_ab_min, exp_ai_b_x, map_technical_x);
      }
      else {
	if (MC_x_dipole_mapping[x_a][i_c] < 0){MC_x_dipole[x_a].g_channel[i_c] = xbs_all[0][0] * g_propagator_Breit_Wigner(x_i_ab_s_part, -MC_x_dipole_mapping[x_a][i_c], sinx_i_ab_min, xbs_all[0][0]);}
	if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	  logger << LOG_DEBUG_VERBOSE << "sinx_i_ab_min = " << setw(23) << setprecision(15) << sinx_i_ab_min << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "x_i_ab_s_part = " << setw(23) << setprecision(15) << x_i_ab_s_part << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "xbs_all[0][0] = " << setw(23) << setprecision(15) << xbs_all[0][0] << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "MC_x_dipole_mapping.size() = " << MC_x_dipole_mapping.size() << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "MC_x_dipole_mapping[" << x_a << "].size() = " << MC_x_dipole_mapping[x_a].size() << endline_flush;
	  logger << LOG_DEBUG_VERBOSE << "MC_x_dipole.size() = " << MC_x_dipole.size() << endline_flush;
	}
      }
    }
    g_x_i_ab = 0.;
    for (int i = 0; i < MC_x_dipole[x_a].g_channel.size(); i++){g_x_i_ab += MC_x_dipole[x_a].alpha[i] * MC_x_dipole[x_a].g_channel[i];}
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      for (int i = 0; i < MC_x_dipole[x_a].g_channel.size(); i++){
	logger << LOG_DEBUG_VERBOSE << "MC_x_dipole[" << x_a << "].alpha[" << i << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].alpha[i] << "   " << "MC_x_dipole[" << x_a << "].g_channel[" << i << "] = " << setw(23) << setprecision(15) << MC_x_dipole[x_a].g_channel[i] << endline_flush;
      }
      logger << LOG_DEBUG_VERBOSE << "g_x_i_ab = " << g_x_i_ab << endline_flush;
    }
  }

  double v_i = (xbp_all[0][bp_i] * xbp_all[0][bp_a]) / (xbp_all[0][bp_b] * xbp_all[0][bp_a]);
  csi->dipole[x_a].zuv = v_i;
  double g_alpha = g_x_i_ab * g_propto_pot(v_i, 0., one_minus_x_i_ab, exp_ai_b_v, map_technical_x) / (pi * (xbp_all[0][bp_b] * xbp_all[0][bp_a]));
  if (switch_console_output_phasespace_issue){if (g_alpha < 0.){logger << LOG_WARN << "x_a = " << x_a << "   g_alpha = " << g_alpha << endline_flush;}}
  if (switch_off_new_psp){
    if (xbs_all[x_a][0] == 0.){xbs_all[x_a][0] = x_i_ab * xbs_all[0][0];}
    if (xbsqrts_all[x_a][0] == 0.){xbsqrts_all[x_a][0] = sqrt(xbs_all[x_a][0]);}
  }
  ag_psp_dipole(x_a, csi->dipole[x_a - 1].sum_channel());
  for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] = MC_phasespace.g_channel[i] * g_alpha;}
  // invert mappings for VAMP implementation
  if (switch_IS_mode_phasespace == 2 || switch_IS_mode_phasespace == 4){
    vector<double> inv_r(3);
    inv_r[0] = inv_propto_pot(one_minus_x_i_ab, 0., one_minus_x_i_ab_min, exp_ai_b_x, map_technical_x);
    inv_r[1] = inv_propto_pot(v_i, 0., one_minus_x_i_ab, exp_ai_b_v, map_technical_x);
    fourvector k_perp = xbp_all[0][bp_i] - v_i * xbp_all[x_a][bp_b] + (one_minus_x_i_ab - v_i) / x_i_ab * xbp_all[x_a][bp_a];
    fourvector temp_p_aib = xbp_all[x_a][bp_a] + xbp_all[x_a][bp_b];
    fourvector temp_p_a_boost = xbp_all[x_a][bp_a].boost(temp_p_aib);
    k_perp = (k_perp.boost(temp_p_aib)).rotation_inverse(temp_p_a_boost);
    double phi = atan2(k_perp.x2(),k_perp.x1());
    if (phi < 0){phi += 2 * M_PI;}
    inv_r[2] = inv_phi(phi);
    if (inv_r[2] <= 0. || inv_r[2] >= 1. || munich_isnan(inv_r[2])){
      if (switch_console_output_phasespace_issue){
	logger << LOG_WARN << "inv_r[2] not in [0; 1] !   r = " << inv_r[2] << "; k_perp.x,y,z = " << k_perp.x1() << ", " << k_perp.x2() << ", " << k_perp.x3() << endline_flush;
      }
      if (munich_isnan(inv_r[2])){inv_r[2] = map_technical_r;}
      else if (inv_r[2] <= 0){inv_r[2] = map_technical_r;}
      else if (inv_r[2] >= 1){inv_r[2] = 1. - map_technical_r;}
    }
    if (switch_IS_mode_phasespace == 2){
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){
        double g_IS_temp;
        double g_IS = 1.;
        for (int j = 0; j < 3; j++){
          random_psp[(i + 1) * 8 - 3 + j]->get_g_IS(inv_r[j], g_IS_temp);
          g_IS *= g_IS_temp;
        }
        MC_phasespace.g_channel[i] *= g_IS;
      }
    }
    else if (switch_IS_mode_phasespace == 4){
      double g_IS_temp;
      double g_IS = 1.;
      for (int j = 0; j < 3; j++){
        random_psp[map_type_to_list_MC_IS[vector<int> {x_a, 6 + j, 0}]]->get_g_IS(inv_r[j], g_IS_temp);
        g_IS *= g_IS_temp;
      }
      for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){MC_phasespace.g_channel[i] *= g_IS;}
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int i = csi->dipole[x_a - 1].sum_channel(); i < csi->dipole[x_a].sum_channel(); i++){logger << LOG_DEBUG_VERBOSE << "MC_phasespace.g_channel[" << setw(5) << i << "] = " << setprecision(15) << setw(23) << MC_phasespace.g_channel[i] << endline_flush;}
  }
  
  logger.finished();
}

