#include "header.hpp"

void event_set::fill_object_list_basic(){
  static Logger logger("event_set::fill_object_list_basic");
  logger.called();

  object_list.push_back("none");

  object_list.push_back("lep");
  object_list.push_back("lm");
  object_list.push_back("lp");
  object_list.push_back("e");
  object_list.push_back("mu");
  object_list.push_back("tau");
  object_list.push_back("em");
  object_list.push_back("mum");
  object_list.push_back("taum");
  object_list.push_back("ep");
  object_list.push_back("mup");
  object_list.push_back("taup");

  object_list.push_back("nua");
  object_list.push_back("nu");
  object_list.push_back("nux");
  object_list.push_back("nea");
  object_list.push_back("nma");
  object_list.push_back("nta");
  object_list.push_back("ne");
  object_list.push_back("nm");
  object_list.push_back("nt");
  object_list.push_back("nex");
  object_list.push_back("nmx");
  object_list.push_back("ntx");
  
  object_list.push_back("charm");
  object_list.push_back("acharm");
  object_list.push_back("charm_qa");
  object_list.push_back("bottom");
  object_list.push_back("abottom");
  object_list.push_back("bottom_qa");
  object_list.push_back("top_qa");
  object_list.push_back("top");
  object_list.push_back("atop");

  object_list.push_back("v");
  object_list.push_back("w");
  object_list.push_back("wp");
  object_list.push_back("wm");
  object_list.push_back("z");

  object_list.push_back("h");

  object_list.push_back("photon");

  object_list.push_back("jet");
  object_list.push_back("ljet");
  object_list.push_back("hjet");
  object_list.push_back("ajet");

  object_list.push_back("bjet");
  object_list.push_back("bjet_b");
  object_list.push_back("bjet_bx");
  object_list.push_back("cjet");
  object_list.push_back("cjet_c");
  object_list.push_back("cjet_cx");
  object_list.push_back("tjet");
  object_list.push_back("tjet_t");
  object_list.push_back("tjet_tx");

  object_list.push_back("missing");

  for (int i = 0; i < object_list.size(); i++){observed_object[object_list[i]] = i;}
  pda.resize(object_list.size());

  logger.finished();
}


void event_set::fill_object_list_user(){
  static Logger logger("event_set::fill_object_list_user");
  logger.called();

  int old_n = object_list.size();

  // Check if default particles would be overwritten by user input (which is not allowed):
  for (int i_p = 0; i_p < user->particle_name.size(); i_p++){
    for (int i_o = 0; i_o < object_list.size(); i_o++){
      if (object_list[i_o] == user->particle_name[i_p]){
	logger << LOG_FATAL << "The default particle  " << user->particle_name[i_p] << "  cannot be overwritten by a user-defined particle with the same name: Input needs to be changed!" << endline_flush;
	exit(1);
      }
    }
  }
  
  for (int i_p = 0; i_p < user->particle_name.size(); i_p++){
    logger << LOG_DEBUG << "user->particle_name[" << i_p << "] = " << user->particle_name[i_p] << endline_flush;
    object_list.push_back(user->particle_name[i_p]);
    //new
    string temp_init = "";
    pda.push_back(define_particle_set(user->particle_name[i_p], temp_init));
    pda[pda.size() - 1].is_user_defined = true;
    logger << LOG_DEBUG << "pda[" << pda.size() - 1 << "].is_user_defined = " << pda[pda.size() - 1].is_user_defined << endline_flush;
  }

  for (int i = old_n; i < object_list.size(); i++){observed_object[object_list[i]] = i;}
  
  logger << LOG_DEBUG << "AFTER  fill_object_list_user  (event_set::output_object_list_no_acces)" << endline << endline << output_object_list_no_access() << endline_flush;

  logger.finished();
}


void event_set::determination_equivalent_object_lep_lm_lp(string lep, string lm, string lp){
  static Logger logger("event_set::determination_equivalent_object_lep_lm_lp");
  logger.called();

  // check if lep is identical to lm or lp etc.
  if (pda[observed_object[lep]].n_partonlevel == pda[observed_object[lp]].n_partonlevel &&
      pda[observed_object[lp]].n_partonlevel > 0 && pda[observed_object[lm]].n_partonlevel == 0){equivalent_object[lp] = equivalent_object[lep];}
  else if (pda[observed_object[lep]].n_partonlevel == pda[observed_object[lm]].n_partonlevel &&
	   pda[observed_object[lm]].n_partonlevel > 0 && pda[observed_object[lp]].n_partonlevel == 0){equivalent_object[lm] = equivalent_object[lep];}

  logger.finished();
}


void event_set::determination_equivalent_object_lep_e_mu_tau(string lep, string e, string mu, string tau){
  static Logger logger("event_set::determination_equivalent_object_lep_e_mu_tau");
  logger.called();

  // check if lep is identical to e or mu or tau etc.
  if (pda[observed_object[lep]].n_partonlevel == pda[observed_object[e]].n_partonlevel &&
      pda[observed_object[e]].n_partonlevel > 0 &&
      pda[observed_object[mu]].n_partonlevel == 0 &&
      pda[observed_object[tau]].n_partonlevel == 0){equivalent_object[e] = equivalent_object[lep];}
  else if (pda[observed_object[lep]].n_partonlevel == pda[observed_object[mu]].n_partonlevel &&
	   pda[observed_object[mu]].n_partonlevel > 0 &&
	   pda[observed_object[e]].n_partonlevel == 0 &&
	   pda[observed_object[tau]].n_partonlevel == 0){equivalent_object[mu] = equivalent_object[lep];}
  else if (pda[observed_object[lep]].n_partonlevel == pda[observed_object[tau]].n_partonlevel &&
	   pda[observed_object[e]].n_partonlevel > 0 &&
	   pda[observed_object[mu]].n_partonlevel == 0 &&
	   pda[observed_object[tau]].n_partonlevel == 0){equivalent_object[tau] = equivalent_object[lep];}

  logger.finished();
}


void event_set::determination_equivalent_object(){
  static Logger logger("event_set::determination_equivalent_object");
  logger.called();

  map<string, int> protect_observable_object;

  for (int i = 1; i < object_list.size(); i++){
    if (pda[i].is_user_defined){equivalent_object[object_list[i]] = object_list[i];}
    else if (object_list[i] == "ljet" && pda[observed_object["jet"]].n_partonlevel != 0){equivalent_object["ljet"] = "ljet";}
    else if (pda[observed_object[object_list[i]]].n_partonlevel == 0){equivalent_object[object_list[i]] = "none";}
    else {equivalent_object[object_list[i]] = object_list[i];}
  }

  determination_equivalent_object_lep_lm_lp("top_qa", "top", "atop");
  determination_equivalent_object_lep_lm_lp("bottom_qa", "bottom", "abottom");
  determination_equivalent_object_lep_lm_lp("charm_qa", "charm", "acharm");

  determination_equivalent_object_lep_lm_lp("tjet", "tjet_t", "tjet_tx");
  determination_equivalent_object_lep_lm_lp("bjet", "bjet_b", "bjet_bx");
  determination_equivalent_object_lep_lm_lp("cjet", "cjet_c", "cjet_cx");
  
  determination_equivalent_object_lep_lm_lp("lep", "lm", "lp");
  determination_equivalent_object_lep_e_mu_tau("lep", "e", "mu", "tau");
  determination_equivalent_object_lep_lm_lp("e", "em", "ep");
  determination_equivalent_object_lep_lm_lp("mu", "mum", "mup");
  determination_equivalent_object_lep_lm_lp("tau", "taum", "taup");
  determination_equivalent_object_lep_e_mu_tau("lm", "em", "mum", "taum");
  determination_equivalent_object_lep_e_mu_tau("lp", "ep", "mup", "taup");

  determination_equivalent_object_lep_lm_lp("nua", "nu", "nux");
  determination_equivalent_object_lep_e_mu_tau("nua", "nea", "nma", "nta");
  determination_equivalent_object_lep_lm_lp("nea", "ne", "nex");
  determination_equivalent_object_lep_lm_lp("nma", "nm", "nmx");
  determination_equivalent_object_lep_lm_lp("nta", "nt", "ntx");
  determination_equivalent_object_lep_e_mu_tau("nu", "ne", "nm", "nt");
  determination_equivalent_object_lep_e_mu_tau("nux", "nex", "nmx", "ntx");

  determination_equivalent_object_lep_lm_lp("w", "wp", "wm");
  determination_equivalent_object_lep_lm_lp("v", "w", "z");
  determination_equivalent_object_lep_e_mu_tau("v", "z", "wp", "wm");

  logger.finished();
}


void event_set::determination_object_list_selection(){
  static Logger logger("event_set::determination_object_list_selection");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "pds.size() = " << pds.size() << endline_flush;

  object_list_selection.push_back("none");
  pds.push_back(define_particle_set ());

  logger << LOG_DEBUG << setw(5) << "no" << setw(10) << "object" << setw(10) << "n_min" << setw(10) << "n_max" << setw(10) << "pT" << setw(10) << "ET" << setw(10) << "|eta|"  << setw(10) << "|y|"<< endline_flush;

  for (int i_o = 1; i_o < object_list.size(); i_o++){
    int flag_old_object = false;
    if (equivalent_object[object_list[i_o]] != ""){
      for (int i_os = 0; i_os < object_list_selection.size(); i_os++){if (equivalent_object[object_list[i_o]] == object_list_selection[i_os]){flag_old_object = true; break;}}
      if (!flag_old_object){
	int x_o = 0;
	for (int j_o = 0; j_o < object_list.size(); j_o++){if (object_list[j_o] == equivalent_object[object_list[i_o]]){x_o = j_o; break;}}
	object_list_selection.push_back(object_list[x_o]);
	pds.push_back(pda[x_o]);
      }
    }
  }

  for (int i_o = 0; i_o < object_list.size(); i_o++){
    for (int i_os = 0; i_os < object_list_selection.size(); i_os++){
      if (equivalent_object[object_list[i_o]] == object_list_selection[i_os]){equivalent_index[i_o] = i_os;}
    }
  }

  for (int i_os = 0; i_os < object_list_selection.size(); i_os++){
    vector<size_t> temp_index_inverse;
    vector<string> temp_object_inverse;
    for (int i_o = 1; i_o < object_list.size(); i_o++){
      if (object_list_selection[i_os] == equivalent_object[object_list[i_o]]){
	temp_index_inverse.push_back(i_o);
	temp_object_inverse.push_back(object_list[i_o]);
      }
    }
    equivalent_inverse_index[i_os] = temp_index_inverse;
    equivalent_inverse_object[object_list_selection[i_os]] = temp_object_inverse;
  }

  for (int i_os = 0; i_os < object_list_selection.size(); i_os++){
    for (int j_os = 0; j_os < equivalent_inverse_object[object_list_selection[i_os]].size(); j_os++){
      access_object_selection[equivalent_inverse_object[object_list_selection[i_os]][j_os]] = i_os;
    }
  }
  for (int i_o = 0; i_o < object_list.size(); i_o++){
    access_object[object_list[i_o]] = observed_object[equivalent_object[object_list[i_o]]];
  }
  
  logger << LOG_DEBUG << "(event_set::output_equivalent_object)" << endline << endline << output_equivalent_object() << endline_flush;
  logger << LOG_INFO << "(event_set::output_equivalent_inverse_object)" << endline << endline << output_equivalent_inverse_object() << endline_flush;
  logger << LOG_INFO << "(event_set::output_object_list_selection)" << endline << endline << output_object_list_selection() << endline_flush;
  logger << LOG_DEBUG << "(event_set::output_object_list)" << endline << endline << output_object_list() << endline_flush;

  logger.finished();
}


// object_list :               object_list           (index) -> object_list           (name)
// observed_object :           object_list           (name)  -> object_list           (index)     (return only index in object_list, without use of equivalent_object - in contrast to access_object)
// access_object :             object_list           (name)  -> object_list           (index)     (points to index of the equivalent object in object_list)

// equivalent_index :          object_list           (index) -> object_list_selection (index)     (no inversion, could be map<string, vector<size_t> >)
// equivalent_object :         object_list           (name)  -> object_list_selection (name)      (no inversion, could be map<string, vector<string> >)

// object_list_selection:      object_list_selection (index) -> object_list_selection (name)
// (observed_object_selection   object_list_selection (name)  -> object_list_selection (index)     (would - by construction - be identical to access_object_selection) )
// access_object_selection :   object_list           (name)  -> object_list_selection (index)     (points to index of the equivalent object in object_list_selection)

