#include "header.hpp"

void event_set::initialization_complete(){
  static Logger logger("event_set::initialization_complete");
  logger.called();

  determination_index_group_parton();
  // -> index_group_protojet_parton ( -> photon_jet_algorithm : photons added to protojets later)
  // -> index_group_recombination_parton
  // -> index_group_photon_parton
  // -> ps_runtime_subtraction_QCD (phase space optimization with QCD dipoles)
  // -> ps_runtime_subtraction_QEW (phase space optimization with QEW dipoles)
  // -> index_group_missing_parton
  
  determination_object_partonlevel_phasespace();
  // -> ps_n_partonlevel
  
  determination_object_observed_from_partonlevel(); // needs modification !!!  User input should never be overwritten, even if contribution gives zero !!!
  // -> pda[i_p].n_observed_min/max  (reset based on  pda[i_p].n_partonlevel)
  
  determination_equivalent_object(); // Might need adaptation accordingly !!! Could be avoided by checking full object list for n_observed_min - n_partonlevel correspondence !!!
  // -> equivalent_object  (based on  pda[i_p].n_partonlevel)
  determination_object_list_selection();
  // -> object_list_selection
  // -> pds  (equivalent to  pda , but for the objects in object_list_selection, plus "all" [0] and "none" [1])
  // -> observed_object_selection (map from  object_list_selection[i_o]  to  i_o)

  
  determination_index_group(); // Requires  object_list_selection  to be filled beforehand.
  // -> index_group_all_parton
  // -> index_group_original
  // -> index_group_jet

  check_parton_content();
  // -> pda[i_p].n_partonlevel  (reset based on ps_n_partonlevel : maximum among all phase spaces (might change the [0] result with special heavy-jet treatment (like (g->bbx) -> ljet)
  // Removal of phase spaces that do not have the right parton content (not active, only csi->int_end if all phase spaces are removed)
  // - uses  index_group_all_partons  and  ps_runtime_original  to check content ( -> ps_runtime_original not used any more... could presumably be removed !!!)
  // - checks also for vetoes against particles 

  // runtime_order and runtime_order_inverse  are actually not used anywhere.
  /*
  determination_runtime_object();
  // Requires  object_list_selection  to be filled beforehand. 
  // -> runtime_order
  // -> runtime_order_inverse 
  output_runtime_object();
  */

  
  for (int i_f = 0; i_f < name_fiducial_cut.size(); i_f++){fiducial_cut.push_back(fiducialcut(name_fiducial_cut[i_f], *this));}
  if (fiducial_cut.size() > 0){logger << LOG_INFO << "(event_set::output_list_fiducialcut)" << endline << endline << output_list_fiducialcut() << endline_flush;}

  logger << LOG_DEBUG << "csi->int_end = " << csi->int_end << endline_flush;


  cut_ps.resize(csi->n_ps, 0);
  first_non_cut_ps = 0;
  
  p_parton.resize(csi->n_ps, vector<fourvector> (csi->n_particle + 3, fourvector ()));
  if (csi->class_contribution_CS_real){for (int i_a = 1; i_a < csi->n_ps; i_a++){p_parton[i_a].erase(p_parton[i_a].end() - 1, p_parton[i_a].end());}}

  ix_y = 0;
  n_object.resize(object_list.size(), vector<int> (csi->n_ps, 0));

  // 1st index (ix_y) shall be extended to cover rotated events (ix_y = 1)
  event_original.resize(1, vector<vector<particle> > (csi->n_ps));
  for (int i_a = 0; i_a < csi->n_ps; i_a++){event_original[0][i_a].resize(csi->type_parton[i_a].size());}      
  event_photon.resize(1, vector<vector<particle> > (csi->n_ps));
  event_missing.resize(1, vector<particle> (csi->n_ps));
  event_jet.resize(1, vector<vector<particle_history> > (csi->n_ps));

  event_particle_group.resize(1, vector<vector<vector<particle *> > > (csi->n_ps, vector<vector<particle *> > (object_list_selection.size(), vector<particle *> (0))));

  j_g_missing = access_object_selection["missing"];  // j_g -> index in relevant_object_list
  j_g_photon = access_object_selection["photon"];  // j_g -> index in relevant_object_list
  j_g_jet = access_object_selection["jet"];  // j_g -> index in relevant_object_list

  user_particle.resize(user->particle_name.size(), vector<vector<particle> > (csi->n_ps));

  
  // Required only to keep exactly the particle access format of MATRIX_v2.0.0 in user-defined code !!!

  particle_event.resize(1, vector<vector<vector<particle> > > (csi->n_ps));
  for (int i_a = 0; i_a < csi->n_ps; i_a++){particle_event[ix_y][i_a].resize(event_particle_group[ix_y][i_a].size());}


  start_p_parton = p_parton;

  logger.finished();
}


void event_set::initialization_integration(){
  static Logger logger("event_set::initialization_integration");
  logger.called();

  logger << LOG_ERROR << "Not implemented!" << endline_flush;
  exit(1);
  //  Check if a separate initialization step is needed here!!!
  
  logger.finished();
}


void event_set::check_parton_content(){
  static Logger logger("event_set::check_parton_content");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "before removal of non-contributing subprocesses   (event_set::output_object_list)" << endline << endline << output_object_list() << endline_flush;

  // perform these checks before adapting n_observed_min etc. to n_partonlevel (if requirements cannot be fulfilled, runs won't start and write 0 as XS).

  bool flag_remove = false;
  vector<int> remove_phasespace(csi->n_ps, 0);
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_o = 1; i_o < object_list.size(); i_o++){
      int i_r = access_object_selection[object_list[i_o]];
      //      int i_r = observed_object[object_list[i_o]];
      bool flag_original = false;
      for (int i_j = 0; i_j < index_group_original.size(); i_j++){if (i_r == index_group_original[i_j]){flag_original = true;}}
      if (equivalent_object[object_list[i_o]] == "none"){flag_original = true;}

      if ((pda[i_o].n_observed_min > ps_n_partonlevel[i_a][i_o]) && flag_original){
	//      if ((pds[i_r].n_observed_min > ps_n_partonlevel[i_a][i_o]) && flag_original){
	logger << LOG_FATAL << "pda.n_observed_min[" << setw(2) << i_o << "] = " << setw(2) << pda[i_o].n_observed_min << " > " << setw(2) << ps_n_partonlevel[i_a][i_o] << " = n_partonlevel[" << setw(2) << i_o << "]" << " -> dipole " << setw(2) << i_a << " does not have sufficient parton content (" << object_list[i_o] << ")." << endline_flush;
	//	logger << LOG_FATAL << "pds.n_observed_min[" << setw(2) << i_r << "] = " << setw(2) << pds[i_r].n_observed_min << " > " << setw(2) << ps_n_partonlevel[i_a][i_o] << " = n_partonlevel[" << setw(2) << i_o << "]" << " -> dipole " << setw(2) << i_a << " does not have sufficient parton content (" << object_list[i_o] << ")." << endline_flush;
	remove_phasespace[i_a] = 1;
	// Check if things like photons mimicking jets etc. need to be considered, which could let the dipole contribute (mainly photons treated as jets, collinear e+e- pairs as photons, etc. !!!
	flag_remove = true;
	break;
      }
    }
  }

  if (accumulate(remove_phasespace.begin(), remove_phasespace.end(), 0) == remove_phasespace.size()){
    logger.newLine(LOG_FATAL);
    logger << LOG_FATAL << "No contribution can result from this subprocess due to parton content (" << csi->n_ps << " phase spaces)." << endline_flush;
    logger.newLine(LOG_FATAL);
    csi->int_end = 1;
  }
  else if (flag_remove){
    // treatment of removed dipoles has to be added. By now, they are simply taken into account and always cut, which might cause problems !!!
    // Ideally, they should be removed completely (including phase-space generation)
    /*
    logger << LOG_FATAL << "(*RA_dipole).size() = " << (*RA_dipole).size() << endline_flush;
    for (int i_a = csi->n_ps - 1; i_a >= 0; i_a--){
      if (remove_dipole[i_a] == 1){(*RA_dipole).erase((*RA_dipole).begin() + i_a);}
    }
    logger << LOG_FATAL << "(*RA_dipole).size() = " << (*RA_dipole).size() << endline_flush;
    //    initialization_RA((*RA_dipole));

    n_ps = (*RA_dipole).size();
    n_pc = (*RA_dipole).size();
    n_pz = 1;

    csi->type_parton.resize(csi->n_ps);
    for (int i_a = 0; i_a < (*RA_dipole).size(); i_a++){csi->type_parton[i_a] = (*RA_dipole)[i_a].type_parton();}
    */
  }

  for (int i_o = 1; i_o < object_list.size(); i_o++){
    int i_r = access_object_selection[object_list[i_o]];
    //    int i_r = observed_object[object_list[i_o]];
    bool flag_original = false;
    for (int i_j = 0; i_j < index_group_original.size(); i_j++){if (i_r == index_group_original[i_j]){flag_original = true;}}
    if (equivalent_object[object_list[i_o]] == "none"){flag_original = true;}

    if (object_list[i_o] == "lep" || object_list[i_o] == "e" || object_list[i_o] == "lm" || object_list[i_o] == "lp" || object_list[i_o] == "em" || object_list[i_o] == "ep"){
      logger << LOG_DEBUG << "[" << setw(10) << object_list[i_o] << "] : pda[" << setw(2) << i_o << "].is_restricted = " << pda[i_o].is_restricted << "   pda[" << setw(2) << i_o << "].n_observed_max = " << pda[i_o].n_observed_max << "   pda[" << setw(2) << i_o << "].n_partonlevel = " << setw(2) << pda[i_o].n_partonlevel << endline_flush;
      if (i_r > 0){
	logger << LOG_DEBUG << "[" << setw(10) << object_list[i_o] << "] : pds[" << setw(2) << i_r << "].is_restricted = " << pds[i_r].is_restricted << "   pds[" << setw(2) << i_r << "].n_observed_max = " << pds[i_r].n_observed_max << "   pds[" << setw(2) << i_r << "].n_partonlevel = " << setw(2) << pds[i_r].n_partonlevel << endline_flush;
      }
    }

    //    if (!pda[i_o].is_restricted && pda[i_o].n_observed_max < pda[i_o].n_partonlevel && flag_original){
    if (!pds[i_r].is_restricted && pds[i_r].n_observed_max < pds[i_r].n_partonlevel && flag_original){
      logger << LOG_FATAL << "No event will survive event selection because of  n_observed_max  request on  object_list[" << i_o << "] = " << object_list[i_o] << " = " << "object_list_selection[" << i_r << "] = " << object_list_selection[i_r] << " -> Don't start integration." << endline_flush;
      csi->int_end = 1;
    }
  }

  logger.finished();
}

