#include "header.hpp"

void event_set::jet_algorithm_flavour(){
  static Logger logger("jet_algorithm_flavour");
  logger.called();

  if (jet_algorithm == 0){event_jet[ix_y][ix_a].swap(temp_protojet);}
  else if (jet_algorithm == 1){sc_Ellis_Soper_flavour();}
  else if (jet_algorithm == 2){kTrun2_flavour();}
  else if (jet_algorithm == 3){antikT_flavour();}
  else if (jet_algorithm == 4){sc_Ellis_Soper_mod_flavour();}

  logger.finished();
}


void event_set::sc_Ellis_Soper_flavour(){
  static Logger logger("event_set::sc_Ellis_Soper_flavour");
  logger.called();

  int min_ot = 1;
  int min1 = 0;
  int min2 = 0;
  double d_min1 = numeric_limits<double>::infinity();
  vector<vector<double> > d_two(temp_protojet.size(), vector<double> (temp_protojet.size()));
  while (temp_protojet.size() > 0){
    d_min1 = numeric_limits<double>::infinity();
    min_ot = 1;
    for (int i = 0; i < temp_protojet.size(); i++){
      if (temp_protojet[i].pT2XX() < d_min1){d_min1 = temp_protojet[i].pT2XX(); min1 = i;}
    }
    for (int i = 0; i < temp_protojet.size(); i++){
      for (int j = i + 1; j < temp_protojet.size(); j++){
	if (temp_protojet[i].pT2XX() < temp_protojet[j].pT2XX()){d_two[i][j] = temp_protojet[i].pT2XX();}
	else{d_two[i][j] = temp_protojet[j].pT2XX();}
	if (jet_R_definition == 0){d_two[i][j] = d_two[i][j] * R2_eta(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	else if (jet_R_definition == 1){d_two[i][j] = d_two[i][j] * R2_rapidity(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	if (d_two[i][j] < d_min1){d_min1 = d_two[i][j]; min1 = i; min2 = j; min_ot = 2;}
      }
    }
    if (min_ot == 1){
      event_jet[ix_y][ix_a].push_back(temp_protojet[min1]);
      temp_protojet.erase(temp_protojet.begin() + min1);
    }
    else {
      double phi_i = temp_protojet[min1].phiXX();
      double phi_j = temp_protojet[min2].phiXX();
      double ET_i = temp_protojet[min1].pTXX();
      double ET_j = temp_protojet[min2].pTXX();
      double ET = ET_i + ET_j;
      double deltaphi = abs(phi_i - phi_j);
      double phi;
      if (deltaphi <= pi){phi = (ET_i * phi_i + ET_j * phi_j) / ET;}
      else {
	if (phi_i > pi){phi_j = phi_j + f2pi;}
	else           {phi_i = phi_i + f2pi;}
	phi = (ET_i * phi_i + ET_j * phi_j) / ET;
	if (phi > f2pi){phi -= f2pi;}
      }     
      double eta = (ET_i * temp_protojet[min1].etaXX() + ET_j * temp_protojet[min2].etaXX()) / ET;
      double pz = ET / tan(2 * atan(exp(-eta)));
      temp_protojet[min2].momentum = fourvector(sqrt(pow(ET, 2) + pow(pz, 2)), cos(phi) * ET, sin(phi) * ET, pz);
      temp_protojet[min2].calculate_all();
      for (int i_i = 0; i_i < temp_protojet[min2].parton_origin.size(); i_i++){temp_protojet[min1].parton_origin.push_back(temp_protojet[min2].parton_origin[i_i]);}
      for (int i_i = 0; i_i < temp_protojet[min2].flavour.size(); i_i++){temp_protojet[min1].flavour.push_back(temp_protojet[min2].flavour[i_i]);}
      // Ellis-Soper prescription instead of:  temp_protojet[min1] += temp_protojet[min2];
      temp_protojet.erase(temp_protojet.begin() + min2);
    }
  }

  logger.finished();
}


void event_set::sc_Ellis_Soper_mod_flavour(){
  static Logger logger("event_set::sc_Ellis_Soper_mod_flavour");
  logger.called();

  int min_ot = 1;
  int min1 = 0;
  int min2 = 0;
  double d_min1 = numeric_limits<double>::infinity();
  vector<vector<double> > d_two(temp_protojet.size(), vector<double> (temp_protojet.size()));
  while (temp_protojet.size() > 0){
    d_min1 = numeric_limits<double>::infinity();
    min_ot = 1;
    for (int i = 0; i < temp_protojet.size(); i++){
      if (temp_protojet[i].pT2XX() < d_min1){d_min1 = temp_protojet[i].pT2XX(); min1 = i;}
    }
    for (int i = 0; i < temp_protojet.size(); i++){
      for (int j = i + 1; j < temp_protojet.size(); j++){
	if (temp_protojet[i].pT2XX() < temp_protojet[j].pT2XX()){d_two[i][j] = temp_protojet[i].pT2XX();}
	else{d_two[i][j] = temp_protojet[j].pT2XX();}
	if (jet_R_definition == 0){d_two[i][j] = d_two[i][j] * R2_eta(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	else if (jet_R_definition == 1){d_two[i][j] = d_two[i][j] * R2_rapidity(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	if (d_two[i][j] < d_min1){d_min1 = d_two[i][j]; min1 = i; min2 = j; min_ot = 2;}
      }
    }
    if (min_ot == 1){
      event_jet[ix_y][ix_a].push_back(temp_protojet[min1]);
      temp_protojet.erase(temp_protojet.begin() + min1);
    }
    else {
      temp_protojet[min1] += temp_protojet[min2];
      temp_protojet.erase(temp_protojet.begin() + min2);
    }
  }

  logger.finished();
}


void event_set::kTrun2_flavour(){
  static Logger logger("event_set::kTrun2_flavour");
  logger.called();

  int min_ot = 1;
  int min1 = 0;
  int min2 = 0;
  double d_min1 = numeric_limits<double>::infinity();
  vector<vector<double> > d_two(temp_protojet.size(), vector<double> (temp_protojet.size()));
  while (temp_protojet.size() > 0){
    d_min1 = numeric_limits<double>::infinity();
    min_ot = 1;
    for (int i = 0; i < temp_protojet.size(); i++){
      if (temp_protojet[i].pT2XX() < d_min1){d_min1 = temp_protojet[i].pT2XX(); min1 = i;}
    }
    for (int i = 0; i < temp_protojet.size(); i++){
      for (int j = i + 1; j < temp_protojet.size(); j++){
	if (temp_protojet[i].pT2XX() < temp_protojet[j].pT2XX()){d_two[i][j] = temp_protojet[i].pT2XX();}
	else{d_two[i][j] = temp_protojet[j].pT2XX();}
	if (jet_R_definition == 0){d_two[i][j] = d_two[i][j] * R2_eta(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	else if (jet_R_definition == 1){d_two[i][j] = d_two[i][j] * R2_rapidity(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	if (d_two[i][j] < d_min1){d_min1 = d_two[i][j]; min1 = i; min2 = j; min_ot = 2;}
      }
    }
    if (min_ot == 1){
      event_jet[ix_y][ix_a].push_back(temp_protojet[min1]);
      temp_protojet.erase(temp_protojet.begin() + min1);
    }
    else {
      temp_protojet[min1] += temp_protojet[min2];
      temp_protojet.erase(temp_protojet.begin() + min2);
    }
  }

  logger.finished();
}


void event_set::antikT_flavour(){
  static Logger logger("event_set::antikT_flavour");
  logger.called();

  int min_ot = 1;
  int min1 = 0;
  int min2 = 0;
  double d_min1 = numeric_limits<double>::infinity();
  vector<vector<double> > d_two(temp_protojet.size(), vector<double> (temp_protojet.size()));
  while (temp_protojet.size() > 0){
    d_min1 = numeric_limits<double>::infinity();
    min_ot = 1;
    min1 = 0;
    for (int i = 0; i < temp_protojet.size(); i++){
      if (1. / temp_protojet[i].pT2XX() < d_min1){d_min1 = 1. / temp_protojet[i].pT2XX(); min1 = i;}
    }
    for (int i = 0; i < temp_protojet.size(); i++){
      for (int j = i + 1; j < temp_protojet.size(); j++){
	if (temp_protojet[i].pT2XX() > temp_protojet[j].pT2XX()){d_two[i][j] = 1. / temp_protojet[i].pT2XX();}
	else {d_two[i][j] = 1. / temp_protojet[j].pT2XX();}
	if (jet_R_definition == 0){d_two[i][j] = d_two[i][j] * R2_eta(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	else if (jet_R_definition == 1){d_two[i][j] = d_two[i][j] * R2_rapidity(temp_protojet[i], temp_protojet[j]) / jet_R2;}
	if (d_two[i][j] < d_min1){d_min1 = d_two[i][j]; min1 = i; min2 = j; min_ot = 2;}
      }
    }
    if (min_ot == 1){
      event_jet[ix_y][ix_a].push_back(temp_protojet[min1]);
      temp_protojet.erase(temp_protojet.begin() + min1);
    }
    else {
      temp_protojet[min1] += temp_protojet[min2];
      temp_protojet.erase(temp_protojet.begin() + min2);
    }
  }

  logger.finished();
}


