#include "header.hpp"

// Eq. (3.4) without E_gamma prefactor
double event_set::frixione_discr(double delta) {
  if (frixione_n == 1){return (1. - cos(delta)) / (1. - cos(frixione_delta_0));}
  else {return pow((1. - cos(delta)) / (1. - cos(frixione_delta_0)), frixione_n);}
}

double event_set::frixione_discr(double delta, double delta_dynamic) {
  if (frixione_n == 1){return (1. - cos(delta)) / (1. - cos(delta_dynamic));}
  else {return pow((1. - cos(delta)) / (1. - cos(delta_dynamic)), frixione_n);}
}

double event_set::frixione_discriminant_R2(double delta) {
  //replace [(1 - cos R) / (1 - cos R_0)] ^ n  =>  [(R² / R²_0)] ^ n  in Frixione's isolation formula (this matters when R > 1, i.e. at small pT)
  if (frixione_n == 1){return (pow(delta, 2)) / (pow(frixione_delta_0, 2));}
  else {return pow((pow(delta, 2)) / (pow(frixione_delta_0, 2)), frixione_n);}
}

double event_set::frixione_discriminant_R2(double delta, double delta_dynamic) {
  //replace [(1 - cos R) / (1 - cos R_0)] ^ n  =>  [(R² / R²_0)] ^ n  in Frixione's isolation formula (this matters when R > 1, i.e. at small pT)
  if (frixione_n == 1){return (pow(delta, 2)) / (pow(delta_dynamic, 2));}
  else {return pow((pow(delta, 2)) / (pow(delta_dynamic, 2)), frixione_n);}
}

struct frix_parton {
  int index;
  double delta;
  bool operator < (const frix_parton & rhs) const { return delta < rhs.delta; }
};

bool event_set::perform_frixione_isolation(particle & photon){
  static Logger logger("frixione_isolation");
  logger.called();

  static vector<frix_parton> frixione_order_parton;
  static frix_parton tmp_parton;
  static double frixione_delta2_0 = pow(frixione_delta_0, 2);
  static double delta_dynamic = 0.;
  frixione_order_parton.clear();

  if (frixione_isolation == 3 || frixione_isolation == 4 || frixione_isolation == 5){
    delta_dynamic = msi->M_Z / photon.pTXX() / sqrt(frixione_epsilon);
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }
  else if (frixione_isolation == 6){
    delta_dynamic = msi->M_Z / photon.pTXX() / sqrt(frixione_epsilon) * sqrt(2.);
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }
  else if (frixione_isolation == 7){
    delta_dynamic = msi->M_Z / photon.pTXX() / sqrt(frixione_epsilon) * 2.;
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }
  else if (frixione_isolation == 8){
    delta_dynamic = msi->M_Z / photon.pTXX() / sqrt(frixione_epsilon) * 2. * sqrt(2.);
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }
  else if (frixione_isolation == 9){
    delta_dynamic = msi->M_Z / photon.pTXX() / sqrt(frixione_epsilon) * 4.;
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }
  else if (frixione_isolation == 10){
    if (photon.pTXX() < 2 * msi->M_Z){delta_dynamic = frixione_delta_0;}
    else {delta_dynamic = msi->M_Z / photon.pTXX();}
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }
  else if (frixione_isolation == 11){
    delta_dynamic = msi->M_Z / photon.pTXX() / sqrt(frixione_epsilon);
    if (delta_dynamic > frixione_delta_0){delta_dynamic = frixione_delta_0;}
    frixione_delta2_0 = pow(delta_dynamic, 2);
  }

  //    logger << LOG_DEBUG_VERBOSE << "delta_dynamic = " << delta_dynamic << "   msi->M_Z = " << msi->M_Z << "   photon.pTXX() = " << photon.pTXX() << endline_flush;

  // identify partons inside the cone
  //  logger << LOG_DEBUG_VERBOSE << "photon = " << photon.momentum << endline_flush;
  for (int i = 0; i < temp_protojet.size(); i++){
    //    logger << LOG_DEBUG_VERBOSE << "temp_protojet[" << i << "] = " << temp_protojet[i].momentum << endline_flush;
    //    logger << LOG_DEBUG_VERBOSE << "delta_2 = " << delta_2 << endline_flush;
    //    logger << LOG_DEBUG_VERBOSE << "frixione_delta2_0 = " << frixione_delta2_0 << endline_flush;
    double delta_2 = 0.;
    if (frixione_isolation == 5 ||
	frixione_isolation == 6 ||
	frixione_isolation == 7 ||
	frixione_isolation == 8 ||
	frixione_isolation == 9){delta_2 = R2_coshrapidity_cosphi(&photon, &temp_protojet[i]);}
    else {delta_2 = R2_eta(&photon, &temp_protojet[i]);}
    /*
	frixione_isolation == 9){delta_2 = R2_coshrapidity_cosphi(photon.momentum, temp_protojet[i].momentum);}
    else {delta_2 = R2_eta(photon, temp_protojet[i].momentum);}
    */
    
    if (delta_2 < frixione_delta2_0) {
      tmp_parton.index = i;
      tmp_parton.delta = sqrt(delta_2);
      frixione_order_parton.push_back(tmp_parton);
    }
  }
  sort(frixione_order_parton.begin(), frixione_order_parton.end());
  //  logger << LOG_DEBUG_VERBOSE << "frixione_order_parton.size() = " << frixione_order_parton.size() << endline_flush;
  if (switch_output_cutinfo){
    info_cut << "photon = " << photon.momentum << endline;
    for (int i = 0; i < frixione_order_parton.size(); i++){
      info_cut << "frixione_order_parton[" << i << " -> " << frixione_order_parton[i].index << "] = " << temp_protojet[frixione_order_parton[i].index] << "   delta = " << frixione_order_parton[i].delta << endline;
    }
  }

  bool isolation = true;
  double E_i_sum = 0.;
  for (int i = 0; i < frixione_order_parton.size(); i++){
    double H_i = 0.;
    if      (frixione_isolation == 1){H_i = frixione_epsilon * photon.pTXX() * frixione_discr(frixione_order_parton[i].delta);} // standard definition
    else if (frixione_isolation == 2){H_i = frixione_fixed_ET_max * frixione_discr(frixione_order_parton[i].delta);}
    else if (frixione_isolation == 3){H_i = frixione_epsilon * photon.pTXX() * frixione_discr(frixione_order_parton[i].delta, delta_dynamic);} // dynamic cone (photon.pTXX() dependent)
    else if (frixione_isolation == 4){H_i = frixione_epsilon * photon.pTXX() * frixione_discriminant_R2(frixione_order_parton[i].delta, delta_dynamic);} // delta definition via (R²/R_0²)^n !!!
    else if (frixione_isolation == 5){H_i = frixione_epsilon * photon.pTXX() * frixione_discriminant_R2(frixione_order_parton[i].delta, delta_dynamic);} // delta definition via (R²/R_0²)^n !!!
    else if (frixione_isolation == 6){H_i = frixione_epsilon * photon.pTXX() * frixione_discriminant_R2(frixione_order_parton[i].delta, delta_dynamic);} // delta definition via (R²/R_0²)^n !!!
    else if (frixione_isolation == 7){H_i = frixione_epsilon * photon.pTXX() * frixione_discriminant_R2(frixione_order_parton[i].delta, delta_dynamic);} // delta definition via (R²/R_0²)^n !!!
    else if (frixione_isolation == 8){H_i = frixione_epsilon * photon.pTXX() * frixione_discriminant_R2(frixione_order_parton[i].delta, delta_dynamic);} // delta definition via (R²/R_0²)^n !!!
    else if (frixione_isolation == 9){H_i = frixione_epsilon * photon.pTXX() * frixione_discriminant_R2(frixione_order_parton[i].delta, delta_dynamic);} // delta definition via (R²/R_0²)^n !!!
    else if (frixione_isolation == 10){H_i = frixione_epsilon * photon.pTXX() * frixione_discr(frixione_order_parton[i].delta, delta_dynamic);} // dynamic cone (photon.pTXX() dependent)
    else if (frixione_isolation == 11){H_i = frixione_epsilon * photon.pTXX() * frixione_discr(frixione_order_parton[i].delta, delta_dynamic);} // dynamic cone (photon.pTXX() dependent)
    else if (frixione_isolation == 12){H_i = frixione_fixed_ET_max * frixione_discriminant_R2(frixione_order_parton[i].delta);}
    else {assert(false);}

    E_i_sum += temp_protojet[frixione_order_parton[i].index].pTXX();

    if (switch_output_cutinfo){
      for (int i = 0; i < frixione_order_parton.size(); i++){
	info_cut << "E (inside cone with delta < delta_0 = " << setw(23) << setprecision(15) << frixione_order_parton[i].delta << ") = " << E_i_sum << endline;
      }
    }

    if (E_i_sum > H_i){
      if (switch_output_cutinfo){info_cut << "E (inside cone with delta < delta_0 = " << setw(23) << setprecision(15) << frixione_order_parton[i].delta << ") = " << E_i_sum << " > " << H_i << " = H_i -> Photon is not isolated." << endline;}
      isolation = false;
      break;
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << isolation << endline_flush; }
  logger.finished();
  return isolation;
}
