#include "header.hpp"

// boosts system to the specified transverse momentum. the azimuth is generated uniformly
void event_set::perform_QT_boost_cms(){
  static Logger logger("event_set::perform_QT_boost_cms");
  logger.called();

  //  i_a = 0: standard point (CMS), not boosted to 2-dimensional qT
  //  i_a = 1: point with finite 2-dim qT
  for (int i_p = 1; i_p < p_parton[0].size(); i_p++){p_parton[1][i_p] = p_parton[0][i_p].boost(psi->recoil_system.Pinv());}

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    stringstream temp_ss;
    for (int i_a = 0; i_a < csi->n_ps; i_a++){
      for (int i_p = 1; i_p < p_parton[i_a].size(); i_p++){temp_ss << "p_parton[" << i_a << "][" << i_p << "] = " << p_parton[i_a][i_p] << "   m2 = " << p_parton[i_a][i_p].m2() << endline;}
      temp_ss << endline;
    }
    logger << LOG_DEBUG_POINT << endline << temp_ss.str() << endline_flush;
    logger << LOG_DEBUG_POINT << setw(25) << "psi->recoil_system" << " = " << psi->recoil_system << endline_flush;
    logger << LOG_DEBUG_POINT << setw(25) << "sum CMS FS momenta" << " = " << p_parton[1][3] + p_parton[1][4] << endline_flush;
  }
  
  logger.finished();
}


void event_set::perform_QT_boost_recoil(){
  static Logger logger("event_set::perform_QT_boost_recoil");
  logger.called();

  double sqrts = 2 * psi->E; 
  double Q2 = psi->xbs_all[0][0];
  //  double Q = sqrt(psi->xbs_all[0][0]);
  double qt2 = psi->recoil_qT2;
  //  double qt = sqrt(qt2);
  double qt = psi->recoil_qT;
  double y = log(sqrt(psi->x_pdf[1] / psi->x_pdf[2]));
  double mt = sqrt(Q2+qt2);
  double qz = 0.5 * mt * (exp(y) - exp(-y));
  double q0 = 0.5 * mt * (exp(y) + exp(-y));
  double phi = psi->recoil_phi; 
    
  fourvector q_CS = fourvector(q0, qt*cos(phi), qt*sin(phi), qz);

  logger << LOG_DEBUG_POINT << "phi = " << phi << endline_flush;
  logger << LOG_DEBUG_POINT << "q_CS (after qT_phi rotation) = " << q_CS << "   y = " << q_CS.rapidity() << "   eta = " << q_CS.eta() << "   m = " << q_CS.m() << endline_flush;
  
  double kt1 = q_CS.x1()/2. ;  
  double kt2 = q_CS.x2()/2. ;   //  CS FRAME PRESCRIPTION
  

  double zeta1 = 1./Q2/2. * ( Q2+ 2.*( q_CS.x1()*kt1+q_CS.x2()*kt2 )
			      + sqrt( pow((Q2 + 2.*( q_CS.x1()*kt1+q_CS.x2()*kt2)),2)
				      - 4.*pow(mt,2)*(pow(kt1,2)+pow(kt2,2)))) ; 

  double qP1 = (q_CS.x0()-q_CS.x3())*sqrts/2.;
  //  double qP2 = (q_CS.x0()+q_CS.x3())*sqrts/2.;

  logger << LOG_DEBUG_POINT << "kt1 = " << kt1 << endline_flush;
  logger << LOG_DEBUG_POINT << "kt2 = " << kt2 << endline_flush;
  
  
  p_parton[1][1] = fourvector( sqrts/2.*( zeta1*Q2/2./qP1 + (pow(kt1,2)+pow(kt2,2))/zeta1*qP1/Q2/pow(sqrts,2)*2.), kt1, kt2,
			       sqrts/2.* (zeta1*Q2/2./qP1 - (pow(kt1,2)+pow(kt2,2))/zeta1*qP1/Q2/pow(sqrts,2)*2.));
  p_parton[1][2] = q_CS - p_parton[1][1];
  /*
  p_parton[1][2] = fourvector( q_CS.x0()-p_parton[1][1].x0(),
			       q_CS.x1()-p_parton[1][1].x1(),
			       q_CS.x2()-p_parton[1][1].x2(),
			       q_CS.x3()-p_parton[1][1].x3() );
  */
  for (int i_p = 3; i_p < p_parton[0].size(); i_p++){p_parton[1][i_p] = p_parton[0][i_p].boostinv(q_CS, Q2);}
  
  logger.finished();
}
