#include "header.hpp"

void event_set::output_parton_momenta_LAB(){
  static Logger logger("event_set::output_photon_recombination");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   Parton-level momenta (hadronic CMS = LAB frame):" << endline;
  for (int i_p = 1; i_p < p_parton[ix_a].size(); i_p++){
    info_cut << "[" << right << setw(2) << ix_a << "]   [" << setw(2) << i_p << " -> " << setw(3) << csi->type_parton[0][i_p] << "] = " << event_original[ix_y][ix_a][i_p].momentum << " m = " << setw(15) << setprecision(8) << sqrt(abs(event_original[ix_y][ix_a][i_p].m2XX())) << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_photon_recombination(){
  static Logger logger("event_set::output_photon_recombination");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   photon_recombination (" << photon_recombination << ")   ---   index_group_photon_parton[" << ix_a << "].size() = " << index_group_photon_parton[ix_a].size() << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   " << "Summary after photon recombination:" << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   index_unrecombined_photon.size() = " << index_unrecombined_photon.size() << endline;
  if (index_unrecombined_photon.size() < index_group_photon_parton[ix_a].size()){
    info_cut << "[" << right << setw(2) << ix_a << "]   Final-state parton-level momenta (hadronic CMS = LAB frame) after photon recombination:" << endline;
    for (int i_p = 3; i_p < p_parton[ix_a].size(); i_p++){
      info_cut << "[" << right << setw(2) << ix_a << "]   type_parton[" << i_p << "] = " << setw(3) << csi->type_parton[ix_a][i_p] << " -> " << event_original[ix_y][ix_a][i_p].momentum << endline;
    }
  }
  else {
    info_cut << "[" << right << setw(2) << ix_a << "]   No photons have been recombined." << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_runtime_original(size_t i_g, size_t j_g){
  static Logger logger("event_set::output_runtime_original");
  logger.called();

  if (i_g == 0){
    info_cut << "[" << right << setw(2) << ix_a << "]   runtime_original" << "   ---   index_group_original.size() = " << index_group_original.size() << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]   " << left << setw(10) << object_list_selection[j_g] << " -> index_group_original[" << setw(2) << i_g << "]  -> j_g = " << setw(2) << j_g << "   index_group_all_parton[ix_a = " << right << setw(2) << ix_a << "][j_g = " << setw(2) << j_g << "].size() = " << index_group_all_parton[ix_a][j_g].size() << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   j_g = " << j_g << "   " << pds[j_g] << endline;
  for (size_t j_p = 0; j_p < index_group_all_parton[ix_a][j_g].size(); j_p++){
    size_t i_p = index_group_all_parton[ix_a][j_g][j_p];  // i_p -> position in  event_original[ix_y][ix_a]  of partons that belong to the respective object class (ix_a dependent !!!)
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << "all " << setw(2) << i_p << " -> " << "momentum = " << event_original[ix_y][ix_a][i_p].momentum << endline;
    bool check_particle_rapidity = pds[j_g].check_rapidity_eta(&event_original[ix_y][ix_a][i_p]);
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << setw(10) << "";
    info_cut << "|y|   = " << right << setw(15) << setprecision(8) << abs(event_original[ix_y][ix_a][i_p].rapidityXX()) << " < " << left << setw(5) << pds[j_g].define_y << "   ";
    info_cut << "|eta| = " << right << setw(15) << setprecision(8) << abs(event_original[ix_y][ix_a][i_p].etaXX()) << " < " << left << setw(5) << pds[j_g].define_eta << "   ";
    info_cut << "-->   " << check_particle_rapidity << endline;
    bool check_particle_pT = pds[j_g].check_pT_ET(&event_original[ix_y][ix_a][i_p]);
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << setw(10) << "";
    info_cut << "pT    = " << right << setw(15) << setprecision(8) << abs(event_original[ix_y][ix_a][i_p].pTXX()) << " > " << left << setw(5) << pds[j_g].define_pT << "   ";
    info_cut << "ET    = " << right << setw(15) << setprecision(8) << abs(event_original[ix_y][ix_a][i_p].ETXX()) << " > " << left << setw(5) << pds[j_g].define_ET << "   ";
    info_cut << "-->   " << check_particle_pT << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << pds[j_g].n_observed_min << " <= n_object = " << n_object[j_g][ix_a] << " <= " << pds[j_g].n_observed_max << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_runtime_original_summary(){
  static Logger logger("event_set::output_runtime_original_summary");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   " << "Summary after runtime_original:" << endline;
  for (size_t i_g = 0; i_g < index_group_original.size(); i_g++){ // i_g -> counter for objects contained in index_group_original
    size_t j_g = index_group_original[i_g];  // j_g -> index in relevant_object_list
    //    size_t k_g = equivalent_index[j_g];  // k_g -> index in object_list
    for (size_t i_i = 0; i_i < event_particle_group[ix_y][ix_a][j_g].size(); i_i++){
      info_cut << "[" << right << setw(2) << ix_a << "]" << "   ";
      if (i_i == 0){info_cut << left << setw(10) << object_list_selection[j_g] << "    ";}
      else {info_cut << left << setw(10) << "" << "    ";}
      info_cut << setw(2) << i_i << " -> " << "momentum = " << event_particle_group[ix_y][ix_a][j_g][i_i]->momentum << "   " << "pT = " << setprecision(8) << setw(15) << event_particle_group[ix_y][ix_a][j_g][i_i]->pTXX() << endline;
    }
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_runtime_missing(){
  static Logger logger("event_set::output_runtime_missing");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   runtime_missing " << "   ---   index_group_missing_parton[" << ix_a << "].size() = " << index_group_missing_parton[ix_a].size() << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   " << left << setw(10) << object_list_selection[j_g_missing] << " -> j_g = " << setw(2) << j_g_missing << "   " << "   index_group_missing_parton[ix_a = " << right << setw(2) << ix_a << "].size() = " << index_group_missing_parton[ix_a].size() << endline;
  for (size_t i_i = 0; i_i < index_group_missing_parton[ix_a].size(); i_i++){
    int i_p = index_group_missing_parton[ix_a][i_i];
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << "all " << setw(2) << i_p << " -> " << "momentum = " << event_original[ix_y][ix_a][i_p].momentum << endline;
  }
  bool check_missing = pds[j_g_missing].check_pT_ET(&event_missing[ix_y][ix_a]);
  info_cut << "[" << right << setw(2) << ix_a << "]   " << left << setw(10) << "" << "    " << setw(10) << "";
  info_cut << "pT    = " << setw(10) << event_particle_group[ix_y][ix_a][j_g_missing][0]->pTXX() << " > " << left << setw(8) << pds[j_g_missing].define_pT << "   ";
  info_cut << "ET    = " << setw(10) << event_particle_group[ix_y][ix_a][j_g_missing][0]->ETXX() << " > " << left << setw(8) << pds[j_g_missing].define_ET << "   ";
  info_cut << "-->   " << check_missing << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_runtime_missing_summary(){
  static Logger logger("event_set::output_runtime_missing_summary(");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   Summary after runtime_missing:" << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   " << left << setw(10) << object_list_selection[j_g_missing] << "    " << "   -> " << "momentum = " << setw(2) << event_particle_group[ix_y][ix_a][j_g_missing][0]->momentum << "   " << "pT = " << setprecision(8) << setw(15) << event_particle_group[ix_y][ix_a][j_g_missing][0]->pTXX() << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_protojet_determination(){
  static Logger logger("event_set::output_protojet_determination");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   Selection of protojets   ---   jet_algorithm (" << jet_algorithm << ")  -  index_group_protojet_parton[" << ix_a << "].size() = " << index_group_protojet_parton[ix_a].size() << endline;
  for (size_t i_l = 0; i_l < index_group_protojet_parton[ix_a].size(); i_l++){
    size_t i_p = index_group_protojet_parton[ix_a][i_l];
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "protojet" << "    " << "all " << setw(2) << i_p << " -> " << "momentum = " << temp_protojet[i_l].momentum << "   flavour = ( " << temp_protojet[i_l].flavour[0] << " )   origin = ( " << temp_protojet[i_l].parton_origin[0] << " )" << endline;
   }
  if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << endline;}

  logger.finished();
}


void event_set::output_frixione_isolation(){
  static Logger logger("event_set::output_frixione_isolation");
  logger.called();
  
  info_cut << "[" << right << setw(2) << ix_a << "]   frixione_isolation (" << frixione_isolation << ")   ---   index_unrecombined_photon.size() = " << index_unrecombined_photon.size() << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   " << left << setw(10) << object_list_selection[j_g_photon] << " -> j_g = " << setw(2) << j_g_photon << "   " << "   index_unrecombined_photon.size() = " << index_unrecombined_photon.size() << "   ---   photon_photon_recombination (" << photon_photon_recombination << ")" << endline;
  output_isolated_photon();
  
  logger.finished();
}


void event_set::output_frixione_isolation_multiphoton(bool flag_recombination_photon_photon){
  static Logger logger("event_set::output_recombination_photon_photon");
  logger.called();
  
  if (flag_recombination_photon_photon){
    info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   photon_photon_recombination (" <<  photon_photon_recombination << ")   ---   index_unrecombined_photon.size() = " << index_unrecombined_photon.size() << "  -  photon-photon-recombination is applied." << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   PHOTON-PHOTON-RECOMBINATION" << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
    output_isolated_photon();
  }
  else {
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << "No isolated photons are recombined." << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << pds[j_g_photon].n_observed_min << " <= n_object = " << n_object[j_g_photon][ix_a] << " <= " << pds[j_g_photon].n_observed_max << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_isolated_photon(){
  static Logger logger("event_set::output_isolated_photon");
  logger.called();

  for (size_t i_i = 0; i_i < event_photon[ix_y][ix_a].size(); i_i++){
    bool check_particle_rapidity = pds[j_g_photon].check_rapidity_eta(&event_photon[ix_y][ix_a][i_i]);
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << "photon " << setw(2) << i_i << " -> " << "momentum = " << event_photon[ix_y][ix_a][i_i].momentum << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << setw(13) << "";
    info_cut << "|y|   = " << right << setw(15) << setprecision(8) << abs(event_photon[ix_y][ix_a][i_i].rapidityXX()) << " < " << left << setw(5) << pds[j_g_photon].define_y << "   ";
    info_cut << "|eta| = " << right << setw(15) << setprecision(8) << abs(event_photon[ix_y][ix_a][i_i].etaXX()) << " < " << left << setw(5) << pds[j_g_photon].define_eta << "   ";
    info_cut << "-->   " << check_particle_rapidity << endline;
    bool check_particle_pT = pds[j_g_photon].check_pT_ET(&event_photon[ix_y][ix_a][i_i]);
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << setw(13) << "";
    info_cut << "pT    = " << right << setw(15) << setprecision(8) << abs(event_photon[ix_y][ix_a][i_i].pTXX()) << " > " << left << setw(5) << pds[j_g_photon].define_pT << "   ";
    info_cut << "ET    = " << right << setw(15) << setprecision(8) << abs(event_photon[ix_y][ix_a][i_i].ETXX()) << " > " << left << setw(5) << pds[j_g_photon].define_ET << "   ";
    info_cut << "-->   " << check_particle_pT << endline;
  }

  logger.finished();
}


void event_set::output_frixione_isolation_summary(){
  static Logger logger("event_set::output_frixione_isolation_summary");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   Summary after Frixione isolation:" << endline;
  for (size_t i_i = 0; i_i < event_particle_group[ix_y][ix_a][j_g_photon].size(); i_i++){
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   ";
    if (i_i == 0){info_cut << left << setw(10) << object_list_selection[j_g_photon] << "    ";}
    else {info_cut << left << setw(10) << "" << "    ";}
    info_cut << setw(2) << i_i << " -> " << "momentum = " << setw(2) << event_particle_group[ix_y][ix_a][j_g_photon][i_i]->momentum << "   " << "pT = " << setprecision(8) << setw(15) << event_particle_group[ix_y][ix_a][j_g_photon][i_i]->pTXX() << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  
  logger.finished();
}


void event_set::output_photon_jet_algorithm(){
  static Logger logger("event_set::output_photon_jet_algorithm");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]" << "   photon_jet_algorithm (" <<  photon_jet_algorithm << ")   ---   index_unisolated_photon.size() = " << index_unisolated_photon.size() << endline;
  if (index_unisolated_photon.size()){
    if (photon_jet_algorithm){
      for (size_t i_l = 0; i_l < temp_protojet.size(); i_l++){
	info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "protojet" << "    " << "protojet " << setw(2) << i_l << " -> " << "momentum = " << temp_protojet[i_l].momentum << "   flavour = ( " << temp_protojet[i_l].flavour[0] << " )   origin = ( " << temp_protojet[i_l].parton_origin[0] << " )" << endline;
      }
      if (switch_output_cutinfo){info_cut << "[" << right << setw(2) << ix_a << "]" << endline;}
    }
    else {
      size_t j_g = access_object_selection["ajet"];
      //      size_t k_g = observed_object["ajet"];
      for (size_t i_i = 0; i_i < event_particle_group[ix_y][ix_a][j_g].size(); i_i++){
	info_cut << "[" << right << setw(2) << ix_a << "]" << "   ";
	if (i_i == 0){info_cut << left << setw(10) << object_list_selection[j_g] << "    ";}
	else {info_cut << left << setw(10) << "" << "    ";}
	info_cut << setw(2) << i_i << " -> " << "momentum = " << setw(2) << event_particle_group[ix_y][ix_a][j_g][i_i]->momentum << "   " << "pT = " << setprecision(8) << setw(15) << event_particle_group[ix_y][ix_a][j_g][i_i]->pTXX() << endline;
      }
    }
  }
  else {
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   no non-isolated photons" << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
    
  logger.finished();
}


void event_set::output_protojet_removal_beam(bool check_removal){
  static Logger logger("event_set::output_protojet_removal_beam");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   removal of protojets too close to beam   ---   |y| > parton_y_max = " << parton_y_max << "  or  |eta| > parton_eta_max = " << parton_eta_max << endline;
  if (check_removal){
    info_cut << "[" << right << setw(2) << ix_a << "]   protojets after removal of those too close to beam:" << endline;
    for (size_t i_i = 0; i_i < temp_protojet.size(); i_i++){
      info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "protojet" << "    " << " -> " << "momentum = " << temp_protojet[i_i].momentum << "   flavour = ( " << temp_protojet[i_i].flavour[0] << " )   origin = ( " << temp_protojet[i_i].parton_origin[0] << " )" << endline;
    }
  }
  else {info_cut << "[" << right << setw(2) << ix_a << "]   no protojets removed." << endline;}
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  
  logger.finished();
}


void event_set::output_jet_algorithm_protojet(){
  static Logger logger("event_set::photon_jet_algorithm_protojet");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]" << "   jet_algorithm (" <<  jet_algorithm << ")   ---   temp_protojet.size() = " << temp_protojet.size() << endline;
  if (temp_protojet.size() > 0){info_cut << "[" << right << setw(2) << ix_a << "]" << endline;}
  for (size_t i_i = 0; i_i < temp_protojet.size(); i_i++){
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   temp_protojet[" << setw(2) << i_i << "] -> " << temp_protojet[i_i].momentum << "   flavour = ( " << temp_protojet[i_i].flavour[0] << " )   origin = ( " << temp_protojet[i_i].parton_origin[0] << " )" << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_jet_algorithm_jet(){
  static Logger logger("event_set::photon_jet_algorithm_jet");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]" << "   result of jet_algorithm (" <<  jet_algorithm << ")   ---   event_jet.size() = " << event_jet[0][ix_a].size() << endline;
  if (event_jet[0][ix_a].size() > 0){info_cut << "[" << right << setw(2) << ix_a << "]" << endline;}
  for (size_t i_p = 0; i_p < event_jet[0][ix_a].size(); i_p++){
    stringstream jet_flavour_ss;
    jet_flavour_ss << "( ";
    for (size_t i_q = 0; i_q < event_jet[0][ix_a][i_p].flavour.size(); i_q++){
      jet_flavour_ss << event_jet[0][ix_a][i_p].flavour[i_q];
      if (i_q < event_jet[0][ix_a][i_p].flavour.size() - 1){jet_flavour_ss << " ";}
    }
    jet_flavour_ss << " )";

    stringstream jet_parton_origin_ss;
    jet_parton_origin_ss << "( ";
    for (size_t i_q = 0; i_q < event_jet[0][ix_a][i_p].parton_origin.size(); i_q++){
      jet_parton_origin_ss << event_jet[0][ix_a][i_p].parton_origin[i_q];
      if (i_q < event_jet[0][ix_a][i_p].parton_origin.size() - 1){jet_parton_origin_ss << " ";}
    }
    jet_parton_origin_ss << " )";

    info_cut << "[" << right << setw(2) << ix_a << "]" << "   event_jet[" << setw(2) << i_p << "] -> " << event_jet[0][ix_a][i_p].momentum << "   flavour = " << jet_flavour_ss.str() << "   origin = " << jet_parton_origin_ss.str() << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
    
  logger.finished();
}


void event_set::output_frixione_jet_removal(){
  static Logger logger("event_set::output_frixione_jet_removal");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]" << "   frixione_jet_removal (" << frixione_jet_removal << ")   ---   frixione_isolation = " << frixione_isolation << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]   event_jet.size() = " << event_jet[0][ix_a].size() << endline;
    
  logger.finished();
}


void event_set::output_summary_frixione_jet_removal(bool check_removal){
  static Logger logger("event_set::output_summary_frixione_jet_removal");
  logger.called();

  if (check_removal){
    info_cut << "[" << right << setw(2) << ix_a << "]   Jets after removal of those too close to isolated photons:" << endline;
    for (size_t i_p = 0; i_p < event_jet[0][ix_a].size(); i_p++){
      info_cut << "[" << right << setw(2) << ix_a << "]" << "   event_jet[" << i_p << "]   pT = " << event_jet[0][ix_a][i_p].pTXX() << "   ET = " << event_jet[0][ix_a][i_p].ETXX() << "   flavour = " << event_jet[0][ix_a][i_p].flavour[0] << "   origin = " << event_jet[0][ix_a][i_p].parton_origin[0] << endline;
    }
  }
  else {info_cut << "[" << right << setw(2) << ix_a << "]   No jets removed." << endline;}
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
    
  logger.finished();
}


void event_set::output_jet_interpretation(){
  static Logger logger("event_set::output_jet_interpretation");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]" << "   Interpretation of individual jets   ---   event_jet.size() = " << event_jet[0][ix_a].size() << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
     
  logger.finished();
}


void event_set::output_jet_algorithm_result(size_t i_g, size_t j_g){
  static Logger logger("event_set::output_jet_algorithm_result");
  logger.called();

  if (i_g == 0){
    info_cut << "[" << right << setw(2) << ix_a << "]   Filling of jet containers   ---   index_group_jet.size() = " << index_group_jet.size() << endline;
    info_cut << "[" << right << setw(2) << ix_a << "]" << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << object_list_selection[j_g] << " -> index_group_jet[" << setw(2) << i_g << "] -> " << "  -> j_g = " << setw(2) << j_g << endline;

  for (size_t i_j = 0; i_j < event_jet[0][ix_a].size(); i_j++){
    if      (j_g == access_object_selection["jet"] && event_jet[0][ix_a][i_j].identity == 22){continue;}
    else if (j_g == access_object_selection["ajet"] && event_jet[0][ix_a][i_j].identity != 22){continue;}
    else if (j_g == access_object_selection["ljet"] && event_jet[0][ix_a][i_j].identity != 0){continue;}
    else if (j_g == access_object_selection["bjet"] && abs(event_jet[0][ix_a][i_j].identity) != 5){continue;}
    else if (j_g == access_object_selection["bjet_b"] && event_jet[0][ix_a][i_j].identity != +5){continue;}
    else if (j_g == access_object_selection["bjet_bx"] && event_jet[0][ix_a][i_j].identity != -5){continue;}
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << "jet " << setw(2) << i_j << " -> " << "momentum = " << event_jet[0][ix_a][i_j].momentum << endline;
    bool check_particle_rapidity = pds[j_g].check_rapidity_eta(&event_jet[0][ix_a][i_j]);
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << setw(10) << "";
    info_cut << "|y|   = " << right << setw(15) << setprecision(8) << abs(event_jet[0][ix_a][i_j].rapidityXX()) << " < " << left << setw(5) << pds[j_g].define_y << "   ";
    info_cut << "|eta| = " << right << setw(15) << setprecision(8) << abs(event_jet[0][ix_a][i_j].etaXX()) << " < " << left << setw(5) << pds[j_g].define_eta << "   ";
    info_cut << "-->   " << check_particle_rapidity << endline;
    bool check_particle_pT = pds[j_g].check_pT_ET(&event_jet[0][ix_a][i_j]);
    info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << left << setw(10) << "" << "    " << setw(10) << "";
    info_cut << "pT    = " << right << setw(15) << setprecision(8) << abs(event_jet[0][ix_a][i_j].pTXX()) << " > " << left << setw(5) << pds[j_g].define_pT << "   ";
    info_cut << "ET    = " << right << setw(15) << setprecision(8) << abs(event_jet[0][ix_a][i_j].ETXX()) << " > " << left << setw(5) << pds[j_g].define_ET << "   ";
    info_cut << "-->   " << check_particle_pT << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << "   " << setw(10) << "" << "    " << pds[j_g].n_observed_min << " <= n_object[" << setw(2) << j_g << "][" << setw(2) <<ix_a << "] = " << n_object[j_g][ix_a] << " <= " << pds[j_g].n_observed_max << endline;
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}


void event_set::output_jet_algorithm_result_summary(){
  static Logger logger("event_set::output_jet_algorithm_result_summary");
  logger.called();

  info_cut << "[" << right << setw(2) << ix_a << "]   Summary after runtime_jet_recombination:" << endline;
  for (size_t i_g = 0; i_g < index_group_jet.size(); i_g++){
    size_t j_g = index_group_jet[i_g];  // j_g -> index in relevant_object_list
    //    size_t k_g = equivalent_index[j_g];  // k_g -> index in object_list
    info_cut << "[" << right << setw(2) << ix_a << "]   number of objects in group [j_g = " << setw(2) << j_g << " -> " << "j_g = " << setw(2) << j_g << " -> " << setw(10) << object_list_selection[j_g] << "][" << right << setw(2) << ix_a << "].size() = " << event_particle_group[ix_y][ix_a][j_g].size() << endline;
  }
  info_cut << "[" << right << setw(2) << ix_a << "]" << endline;

  logger.finished();
}

