#include "header.hpp"

void qTsubtraction_basic::event_selection_qTcut(int x_a){
  static Logger logger("qTsubtraction_basic::event_selection_qTcut");
  logger.called();

  if (esi->switch_output_cutinfo){
    for (int i_p = 1; i_p < esi->p_parton[x_a].size(); i_p++){
      esi->info_cut << "[" << setw(2) << x_a << "]   type_parton[" << i_p << "] = " << setw(3) << csi->type_parton[x_a][i_p] << " -> " << esi->p_parton[x_a][i_p] << " m2 = " << esi->p_parton[x_a][i_p].m2() << endline_flush;
    }
    for (int i_p = 1; i_p < esi->p_parton[x_a].size(); i_p++){
      esi->info_cut << "[" << setw(2) << x_a << "]   type_parton[" << i_p << "] = " << setw(3) << csi->type_parton[x_a][i_p] << " -> " << esi->event_original[esi->ix_y][esi->ix_a][i_p].momentum << " m2 = " << esi->event_original[esi->ix_y][esi->ix_a][i_p].m2XX() << endline_flush;
    }
  }
  /*
  // old version (would not work in EW and MIX corrections with Born-level photons):
  QT_Q = fourvector();
  if (esi->switch_output_cutinfo){esi->info_cut << "QT_Q = " << QT_Q << endline;}
  for (int i_p = 3; i_p < 3 + csi->n_particle_born; i_p++){
    QT_Q = QT_Q + esi->p_parton[x_a][i_p];
    if (esi->switch_output_cutinfo){esi->info_cut << "QT_Q = " << QT_Q << endline;}
  }
  if (esi->switch_output_cutinfo){esi->info_cut << "old: QT_Q = " << QT_Q << "   QT_Q.pT = " << QT_Q.pT() << endline;}
  */

  // For more sophisticated determination of QT_Q (in particular if photons are involved in NNLO MIX or QEW) !!!
  QT_Q = fourvector();
  if (esi->switch_output_cutinfo){
    esi->info_cut << "csi->n_particle      = " << csi->n_photon_born << endline;
    esi->info_cut << "csi->n_particle_born = " << csi->n_photon_born << endline;
    esi->info_cut << "csi->n_photon_born  = " << csi->n_photon_born << endline;
    esi->info_cut << "QT_Q (reset)                                = " << QT_Q << endline;
  }
  for (int i_p = 3; i_p < csi->type_parton[x_a].size(); i_p++){
    if (esi->switch_output_cutinfo){ esi->info_cut << "csi->type_parton[" << x_a << "][" << i_p << "] = " << csi->type_parton[x_a][i_p] << endline; }
    if (csi->type_parton[x_a][i_p] >= -csi->msi->N_f && csi->type_parton[x_a][i_p] <= csi->msi->N_f){ continue; }
    if (csi->type_parton[x_a][i_p] == 22){ continue; }
    QT_Q += esi->event_original[esi->ix_y][esi->ix_a][i_p].momentum;
    if (esi->switch_output_cutinfo){ esi->info_cut << "QT_Q [i_p = " << setw(2) << i_p << "]                             = " << QT_Q << endline; }
  }
  if (esi->switch_output_cutinfo){ esi->info_cut << "QT_Q (after adding all partons but photons) = " << QT_Q << endline; }
  // Boost does not affect the (relevant) pT components !!!

  for (int i_i = 0; i_i < csi->n_photon_born; i_i++){
    if (esi->switch_output_cutinfo){ esi->info_cut << "esi->event_particle_group[" << esi->ix_y << "][" << x_a << "][" << esi->j_g_photon << "][" << i_i << "]->momentum = " << esi->event_particle_group[esi->ix_y][x_a][esi->j_g_photon][i_i]->momentum << endline; }
    if (i_i > esi->event_particle_group[esi->ix_y][x_a][esi->j_g_photon].size()){ logger << LOG_WARN << "Should not happen!" << endline_flush; continue; }
    QT_Q += esi->event_particle_group[esi->ix_y][x_a][esi->j_g_photon][i_i]->momentum;
    if (esi->switch_output_cutinfo){ esi->info_cut << "QT_Q [isolated photon[i_i = " << setw(2) << i_i << "]]            = " << QT_Q << endline; }
  }
  if (esi->switch_output_cutinfo){ esi->info_cut << "QT_Q (after adding photons)                 = " << QT_Q << endline; }

  QT_QT = QT_Q.pT();
  if (esi->switch_output_cutinfo){ esi->info_cut << "QT_QT     = " << QT_QT << endline; }
  if (switch_qTcut == 1){
    QT_sqrtQ2 = QT_Q.m();
    if (esi->switch_output_cutinfo){ esi->info_cut << "QT_sqrtQ2 = " << QT_sqrtQ2 << endline; }
    if (QT_QT / QT_sqrtQ2 < min_qTcut / 100.){ (esi->cut_ps)[x_a] = -1; }
    // very exceptional case: Q² is numerically negative since too close to 0 !!!
    else if (QT_Q.m2() < 0.){ (esi->cut_ps)[x_a] = n_qTcut - 1; }
    else {
      if (binning_qTcut == "linear"){
	if (esi->switch_output_cutinfo){ esi->info_cut << "(QT_QT / QT_sqrtQ2 - min_qTcut / 100.) * 100) / step_qTcut = " << (QT_QT / QT_sqrtQ2 - min_qTcut / 100.) * 100 / step_qTcut << endline; }
	double temp_rcut = ((QT_QT / QT_sqrtQ2 - min_qTcut / 100.) * 100) / step_qTcut;
	esi->cut_ps[x_a] = GSL_MIN_INT(int(temp_rcut), n_qTcut - 1);
	if (esi->cut_ps[x_a] < 0 && temp_rcut > double(n_qTcut - 1)){ esi->cut_ps[x_a] = n_qTcut - 1; }
      }
      //      else if (binning_qTcut == "logarithmic"){} // add fast version for logarithmic distribution of qTcut
      else {
	double temp_value_cut = QT_QT / QT_sqrtQ2 * 100;
	for (int i_q = n_qTcut - 1; i_q >= 0; i_q--){
	  if (temp_value_cut > value_qTcut[i_q]){ esi->cut_ps[x_a] = i_q; break; }
	}
      }
    }
    //    logger << LOG_DEBUG << "esi->cut_ps[" << x_a << "] = " << (esi->cut_ps)[x_a] << endline_flush;
    //    logger << LOG_DEBUG << endline << esi->info_cut.str();
  }
  else if (switch_qTcut == 2){
    if (QT_QT < min_qTcut){ esi->cut_ps[x_a] = -1; }// return;}
    else {
      if (binning_qTcut == "linear"){
	esi->cut_ps[x_a] = GSL_MIN_INT(int((QT_QT - min_qTcut) / step_qTcut), n_qTcut - 1);
      }
      //      else if (binning_qTcut == "logarithmic"){} // add fast version for logarithmic distribution of qTcut
      else {
	double temp_value_cut = QT_QT;
	for (int i_q = n_qTcut - 1; i_q >= 0; i_q--){
	  if (temp_value_cut > value_qTcut[i_q]){ esi->cut_ps[x_a] = i_q; break; }
	}
      }
    }
  }

  logger.finished();
}


void qTsubtraction_basic::event_selection_recoil_qTcut(){
  static Logger logger("qTsubtraction_basic::event_selection_qTcut");
  logger.called();

  if (psi->user->switch_value[psi->user->switch_map["CTfromPTmode"]]){
    if (psi->recoil_r < min_qTcut / 100.){psi->recoil_rcut = -1;}
    else {
      if (binning_qTcut == "linear"){
	if (esi->switch_output_cutinfo){esi->info_cut << "(psi->recoil_r - min_qTcut / 100.) * 100) / step_qTcut = " << (psi->recoil_r - min_qTcut / 100.) * 100 / step_qTcut << endline;}
	double temp_rcut = ((psi->recoil_r - min_qTcut / 100.) * 100) / step_qTcut;
	psi->recoil_rcut = GSL_MIN_INT(int(temp_rcut), n_qTcut - 1);
      }
      //      else if (binning_qTcut == "logarithmic"){} // add fast version for logarithmic distribution of qTcut
      else {
	double temp_value_cut = psi->recoil_r * 100;
	for (int i_q = n_qTcut - 1; i_q >= 0; i_q--){
	  if (temp_value_cut > value_qTcut[i_q]){psi->recoil_rcut = i_q; break;}
	}
      }
    }
  }
  else {
    if (psi->recoil_r < min_qTcut / 100.){psi->recoil_rcut = 0;}
    else {
      if (binning_qTcut == "linear"){
	if (esi->switch_output_cutinfo){esi->info_cut << "(psi->recoil_r - min_qTcut / 100.) * 100) / step_qTcut = " << (psi->recoil_r - min_qTcut / 100.) * 100 / step_qTcut << endline;}
	double temp_rcut = ((psi->recoil_r - min_qTcut / 100.) * 100) / step_qTcut;
	psi->recoil_rcut = int(temp_rcut) + 1;
	if (psi->recoil_rcut >= n_qTcut){psi->recoil_rcut = -1;}
      }
      //      else if (binning_qTcut == "logarithmic"){} // add fast version for logarithmic distribution of qTcut
      else {
	psi->recoil_rcut = -1;
	double temp_value_cut = psi->recoil_r * 100;
	for (int i_q = n_qTcut - 1; i_q >= 0; i_q--){
	  if (temp_value_cut > value_qTcut[i_q]){psi->recoil_rcut = i_q + 1; break;}
	}
      }
    }
  }

  logger.finished();
}

