#include "header.hpp"

particle_history::particle_history(){}

particle_history::~particle_history(){}

particle_history::particle_history(fourvector _momentum, int i_a, int i_p, event_set * _esi) : particle(_momentum){
  static Logger logger("particle_history::particle_history (momentum, i_a, i_p, esi)");
  logger.called();

  esi = _esi;
  int temp_type = esi->csi->type_parton[i_a][i_p];
  if (abs(esi->csi->type_parton[i_a][i_p]) <= esi->msi->N_f){temp_type = 0;}
  parton_origin = vector<int> (1, i_p);
  flavour = vector<int> (1, temp_type);

  logger.finished();
}

particle_history::particle_history(fourvector _momentum, double _M, int i_a, int i_p, event_set * _esi) : particle(_momentum, _M){
  static Logger logger("particle_history::particle_history (momentum, M, i_a, i_p, esi)");
  logger.called();

  esi = _esi;
  int temp_type = esi->csi->type_parton[i_a][i_p];
  if (abs(esi->csi->type_parton[i_a][i_p]) <= esi->msi->N_f){temp_type = 0;}
  parton_origin = vector<int> (1, i_p);
  flavour = vector<int> (1, temp_type);

  logger.finished();
}

particle_history::particle_history(fourvector _momentum, double _M, double _M2, int i_a, int i_p, event_set * _esi) : particle(momentum, _M, _M2){
  static Logger logger("particle_history::particle_history (momentum, M, M2, i_a, i_p, esi)");
  logger.called();

  esi = _esi;
  int temp_type = esi->csi->type_parton[i_a][i_p];
  if (abs(esi->csi->type_parton[i_a][i_p]) <= esi->msi->N_f){temp_type = 0;}
  parton_origin = vector<int> (1, i_p);
  flavour = vector<int> (1, temp_type);

  logger.finished();
}


// Requires an improved implementation !!!
void particle_history::determination_identity(){
  static Logger logger("particle_history::determination_identity");
  logger.called();

  // Jets are categorized only according to the content of the "heaviest" quark/antiquark flavour and photons.
  // qjet    -> for any number of quarks/antiquarks of that flavour if massive, and for n_q != n_qx if massless
  // qjet_q  -> for excess of flavoured quarks (in number), i.e. n_q > n_qx, independent of quark mass
  // qjet_qx -> for excess of flavoured antiquarks (in number), i.e. n_qx > n_q, independent of quark mass
  // ljet    -> no flavours beyond light-jet threshold, or no net flavour in massless case

  // input variable required to decide the light-jet threshold: natural choice: m_q,
  // but IR-finite distinction also possible for bjets from massless bottom (anti)quarks
  // same for cjet from charm (anti)quarks, but flavour symmetry in generation would need to be switched off !!!

  
  int content_highest_flavour = 0;
  int content_photon = 0;
  for (int i_q = 0; i_q < flavour.size(); i_q++){
    if (flavour[i_q] != 22 && abs(flavour[i_q]) > content_highest_flavour){content_highest_flavour = abs(flavour[i_q]);}
    else if (flavour[i_q] == 22){content_photon++;}
  }
  
  identity = 0;
  if (content_highest_flavour != 0){
    for (int i_q = 0; i_q < flavour.size(); i_q++){
      if (abs(flavour[i_q]) == content_highest_flavour){identity += flavour[i_q];}
    }
  }
  if (identity == 0 && esi->msi->M[content_highest_flavour] != 0.){identity = 11 * content_highest_flavour;}
  else if (identity != 0 && abs(identity) != content_highest_flavour){
    if (identity > 0){identity = content_highest_flavour;}
    else {identity = -content_highest_flavour;}
  }
  
  /*
  identity = accumulate(flavour.begin(), flavour.end(), 0);
  
  if (esi->msi->M_b > 0. && identity == 0){
    // no IR-safe defintion in case of massless b quarks - b and bx are combined to bottom jet: switch needed !!!
    for (int i_q = 0; i_q < flavour.size(); i_q++){if (abs(flavour[i_q]) == 5){identity = 55; break;}}
  }

  if (identity == 0){} // light jet  or  b-bx jet
  else if (identity == 5 || identity == -5 || identity == 55){} // bjet_b  or  bjet_bx
  else {
  */

  if (content_photon > 0){
    double E_phot = 0.;
    double E_had = 0.;
    int count_phot = 0;
    for (int i_l = 0; i_l < flavour.size(); i_l++){
      if (flavour[i_l] == 22){
	E_phot += (esi->event_particle_group[esi->ix_y][esi->ix_a][0][parton_origin[i_l]]->momentum).x0();
	count_phot++;
      }
      else {
	E_had += (esi->event_particle_group[esi->ix_y][esi->ix_a][0][parton_origin[i_l]]->momentum).x0();
      }
    }

    if (esi->switch_output_cutinfo){esi->info_cut << "E_phot = " << E_phot << "   E_had = " << E_had << endline;}

    if (E_phot / (E_phot + E_had) >= esi->photon_E_threshold_ratio){identity = 22;}
    //    else {identity = identity - count_phot * 22;}
    logger << LOG_DEBUG_VERBOSE << "E_phot = " << E_phot << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "E_had = " << E_had << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "identity = " << identity << endline_flush;
  }

  logger.finished();
}


particle_history particle_history::operator + (const particle_history & p2) const{
  particle_history psum;
  psum.momentum = momentum + p2.momentum;
  psum.reset_all();
  //  psum.calculate_all();
  for (int i_i = 0; i_i < p2.parton_origin.size(); i_i++){psum.parton_origin.push_back(p2.parton_origin[i_i]);}
  for (int i_i = 0; i_i < p2.flavour.size(); i_i++){psum.flavour.push_back(p2.flavour[i_i]);}
  return psum;
}


particle_history & particle_history::operator += (const particle_history & p2) {
  momentum = momentum + p2.momentum;
  reset_all();
  //  calculate_all();
  for (int i_i = 0; i_i < p2.parton_origin.size(); i_i++){parton_origin.push_back(p2.parton_origin[i_i]);}
  for (int i_i = 0; i_i < p2.flavour.size(); i_i++){flavour.push_back(p2.flavour[i_i]);}
  return *this;
}

