#include "header.hpp"

munich::munich(){}

munich::munich(int argc, char *argv[], string basic_process_class, string identifier){
  static Logger logger("munich::munich");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  Log::setLogThreshold(LOG_INFO);
  walltime_start();

  logger << LOG_DEBUG << "argc = " << argc << endline_flush;
  if (argc == 0){logger << LOG_ERROR << "No input specified." << endline_flush; exit(1);}
  else if (argc < 3){
    if (argc == 1){
      cin >> subprocess;
      logger << LOG_DEBUG << "console input: subprocess = " << subprocess << endline_flush;
    }
    else if (argc == 2){
      logger << LOG_DEBUG << "argv[1] = " << argv[1] << endline_flush;
      subprocess = argv[1];
    }

    isi = new inputparameter_set(basic_process_class, subprocess, identifier);
  }
  else if (argc > 2){
    logger << LOG_DEBUG << "argv[1] = " << argv[1] << endline_flush;
    order = argv[1];

    logger << LOG_DEBUG << "argv[2] = " << argv[2] << endline_flush;
    if (order == "result"){cout << "calculate the results specified in " << argv[2] << endline_flush;}
    else if (order == "distribution"){cout << "calculate the results specified in " << argv[2] << endline_flush;}
    else if (order == "scaleband"){cout << "calculate the results specified in " << argv[2] << endline_flush;}
    infilename = argv[2];

   infilename_scaleband = "";
    if (argc == 4){
      logger << LOG_DEBUG << "argv[3] = " << argv[3] << endline_flush;
      infilename_scaleband = argv[3];
    }

    isi = new inputparameter_set(basic_process_class, "");
  }

  get_path_rundirectory();
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::initialization(){
  static Logger logger("munich::initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  user = new user_defined();
  user->initialization(isi);

  msi = new model_set();
  msi->initialization(isi, user);
  //  csi->msi = msi;

  csi->initialization(isi, msi);

  esi->initialization(isi, csi, msi, user);

  psi->initialization(isi, csi, msi, esi, user);
  esi->psi = psi;
  esi->qT_basic = new qTsubtraction_basic(isi, csi, esi, psi);

  osi->initialization(isi, csi, msi, esi, psi, user);

  // For compatibility with "Bornloop" parameter... to be removed at some point !!!
  if (user->string_value[user->string_map["model"]] == "Bornloop"){csi->class_contribution_loopinduced = 1;}

  // finally not needed, I guess !!!
  osi->xmunich = this;

  // Isolate some qTsubtraction related information from observable_set (to be rearranged completely) !!!
  /////  qT_basic = new qTsubtraction_basic(isi, csi, esi, psi);
  /////  esi->qT_basic->output();
  /////  esi->qT_basic = qT_basic;

  logger.newLine(LOG_DEBUG);
  logger << LOG_DEBUG << "Settings from input files:" << endl << endl << isi << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::run_initialization(){
  static Logger logger("munich::run_initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (asi == NULL){
    if (osi->switch_OL){
#ifdef OPENLOOPS
      asi = new amplitude_OpenLoops_set();
#endif
    }
    else if (osi->switch_RCL){
#ifdef RECOLA
      asi = new amplitude_Recola_set();
#endif
    }
  }
  /*
  // Maybe shift here and split initialization and initialization_complete ???
  asi->initialization(osi, csi, msi, esi, isi);
  */
  osi->asi = asi;
  // Should not be done here !!!
  //  psi->initialization_masses();

  // No duplicated information stored in csi should be required in psi in the future !!!
  //  csi->initialization_contribution_order(); // moved to void contribution_set::initialization_complete() !!!
  //  psi->initialization_contribution_order(); // moved to void phasespace_set::initialization_complete() !!!


  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

void munich::run_integration(){
  static Logger logger("munich::run_integration");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "csi->type_contribution = " << csi->type_contribution << endline_flush;
  logger << LOG_DEBUG << "csi->type_correction = " << csi->type_correction << endline_flush;

  if (csi->type_contribution == "born"){integration_born();}
  if (csi->type_contribution == "loop"){integration_born();}
  if (csi->type_contribution == "L2I"){integration_born();}

  if (csi->type_contribution == "VA" && csi->type_correction == "QCD"){integration_VA_QCD();}
  if (csi->type_contribution == "CA" && csi->type_correction == "QCD"){integration_CA_QCD();}
  if (csi->type_contribution == "RA" && csi->type_correction == "QCD"){integration_RA_QCD();}
  if (csi->type_contribution == "RA" && csi->type_correction == "---"){integration_born();}

  if (csi->type_contribution == "L2VA" && csi->type_correction == "QCD"){integration_VA_QCD();}
  if (csi->type_contribution == "L2CA" && csi->type_correction == "QCD"){integration_CA_QCD();}
  if (csi->type_contribution == "L2RA" && csi->type_correction == "QCD"){integration_RA_QCD();}
  if (csi->type_contribution == "L2RA" && csi->type_correction == "---"){integration_born();}

  if (csi->type_contribution == "VA" && csi->type_correction == "QEW"){integration_VA_QEW();}
  if (csi->type_contribution == "CA" && csi->type_correction == "QEW"){integration_CA_QEW();}
  if (csi->type_contribution == "RA" && csi->type_correction == "QEW"){integration_RA_QEW();}
  if (csi->type_contribution == "VA" && csi->type_correction == "MIX"){integration_VA_MIX();}
  if (csi->type_contribution == "RA" && csi->type_correction == "MIX"){integration_RA_MIX();}

  if (csi->type_contribution == "VT" && csi->type_correction == "QCD"){integration_VT_QCD();}
  if (csi->type_contribution == "CT" && csi->type_correction == "QCD"){integration_CT_QCD();}
  if (csi->type_contribution == "RT" && csi->type_correction == "QCD"){integration_born();}
  if (csi->type_contribution == "RT" && csi->type_correction == "---"){integration_born();}
  if (csi->type_contribution == "PT" && csi->type_correction == "QCD"){integration_PT_QCD();}

  if (csi->type_contribution == "VT2" && csi->type_correction == "QCD"){integration_VT2_QCD();}
  if (csi->type_contribution == "CT2" && csi->type_correction == "QCD"){integration_CT2_QCD();}
  if (csi->type_contribution == "RVA" && csi->type_correction == "QCD"){integration_VA_QCD();}
  if (csi->type_contribution == "RCA" && csi->type_correction == "QCD"){integration_CA_QCD();}
  if (csi->type_contribution == "RRA" && csi->type_correction == "QCD"){integration_RA_QCD();}
  if (csi->type_contribution == "RRA" && csi->type_correction == "---"){integration_born();}
  if (csi->type_contribution == "PT2" && csi->type_correction == "QCD"){integration_PT2_QCD();}

  if (csi->type_contribution == "L2VT" && csi->type_correction == "QCD"){integration_VT_QCD();}
  if (csi->type_contribution == "L2CT" && csi->type_correction == "QCD"){integration_CT_QCD();}
  if (csi->type_contribution == "L2RT" && csi->type_correction == "QCD"){integration_born();}
  if (csi->type_contribution == "L2RT" && csi->type_correction == "---"){integration_born();}
  if (csi->type_contribution == "L2PT" && csi->type_correction == "QCD"){integration_PT_QCD();}

  //  suggested new names:
  //  B.NLL
  //  VT.NLL
  //  B.NNLL
  //  VT.NNLL
  //  VT2.NNLL
  if (csi->type_contribution == "NLL_LO" && csi->type_correction == "QCD"){integration_VT_QCD();}
  if (csi->type_contribution == "NLL_NLO" && csi->type_correction == "QCD"){integration_VT_QCD();}
  if (csi->type_contribution == "NNLL_LO" && csi->type_correction == "QCD"){integration_VT_QCD();}
  if (csi->type_contribution == "NNLL_NLO" && csi->type_correction == "QCD"){integration_VT_QCD();}
  if (csi->type_contribution == "NNLL_NNLO" && csi->type_correction == "QCD"){integration_VT2_QCD();}
  //  CT -> CT.res
  //  CT2 -> CT2.res

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::initialization_integration(){
  static Logger logger("munich::initialization_integration");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  csi->initialization_complete();
  esi->initialization_complete();
  osi->initialization_complete();
  // amplitude initialization here separately ???
  //  asi->initialization(osi, csi, msi, esi, isi);
  //  esi->initialization_complete();
  psi->initialization_complete();
  /*
    if (csi->class_contribution_CS_real){psi->initialization_complete_RA();}
    else {psi->initialization_complete();}
  */
  osi->initialization_integration();
  // New position (only initialize if int_end != 0 ???) !!!
  asi->initialization(osi, csi, msi, esi, isi);

  save_osi = *osi;
  save_psi = *psi;
  rsi = runresumption_set(*osi, *psi, save_osi, save_psi);
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::initialization_integration_check(){
  static Logger logger("munich::initialization_integration_check");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "csi->mass_parton[0][1] = " << csi->mass_parton[0][1] << endline_flush;
  logger << LOG_DEBUG << "csi->mass_parton[0][2] = " << csi->mass_parton[0][2] << endline_flush;
  logger << LOG_DEBUG << "psi->n_events_max = " <<  psi->n_events_max<< endline_flush;
  logger << LOG_DEBUG << "csi->int_end = " << csi->int_end << endline_flush;
  logger << LOG_DEBUG << "osi->int_end = " << osi->int_end << endline_flush;

  
  /*
  if (csi->mass_parton[0][1] > 0. || csi->mass_parton[0][2] > 0.){
    logger << LOG_WARN << "Massive incoming partons are not allowed." << endline_flush;
    osi->int_end = 1;
  }
  */
  if (psi->n_events_max == 0){
    logger << LOG_WARN << "Zero events are requested." << endline_flush;
    osi->int_end = 1;
  }

  // check if csi->int_end is set for RA contributions !!!
  //  if (!csi->class_contribution_CS_real){if (csi->int_end != 0){osi->int_end = csi->int_end;}}
  if (csi->int_end != 0){osi->int_end = csi->int_end;}
  logger << LOG_DEBUG << "csi->int_end = " << csi->int_end << endline_flush;
  logger << LOG_DEBUG << "osi->int_end = " << osi->int_end << endline_flush;

  if (asi->process_id == -1){
    if (asi->switch_amplitude_not_required){
      logger << LOG_WARN << "Required amplitudes do not exist and are thus understood to vanish." << endline_flush;
      osi->int_end = 1;
    }
    else {
      logger << LOG_FATAL << "Required amplitudes do not exist." << endline_flush;
      if (!osi->int_end){exit(1);}
    }
  }
  
  if (osi->int_end){osi->output_zero_contribution_complete();}
  if (psi->error_readin){logger << LOG_FATAL << "psi->error_readin" << endline_flush; osi->int_end = 1; exit(1);}

  if (user->int_value[user->int_map["rescaling_exponent"]] != 0){psi->hcf = psi->hcf * pow(user->double_value[user->double_map["rescaling_factor"]], user->int_value[user->int_map["rescaling_exponent"]]);}

  osi->initialization_runtime();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::walltime_start(){
  static Logger logger("munich::walltime_start");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  struct tm y2k_tm = {0};
  y2k_tm.tm_hour = 0;
  y2k_tm.tm_min = 0;
  y2k_tm.tm_sec = 0;
  y2k_tm.tm_year = 100;
  y2k_tm.tm_mon = 0;
  y2k_tm.tm_mday = 1;
  y2k_time = mktime (&y2k_tm);

  start_time_point = chrono::system_clock::now();
  start_time = chrono::system_clock::to_time_t(start_time_point);

  //  string temp_start_time = ctime(&start_time);
  //  temp_start_time = temp_start_time.substr(0, temp_start_time.size() - 1);
  //  logger << LOG_INFO << temp_start_time << "   ---   " << setprecision(16) << difftime(start_time, y2k_time) << " seconds since y2k" << endline_flush;

  start_CPU_time = clock();

  logger << LOG_INFO << ctime(&start_time);
  logger << LOG_DEBUG << setprecision(16) << noshowpoint << difftime(start_time, y2k_time) << " seconds since y2k" << endline_flush;
  logger.newLine(LOG_INFO);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

void munich::walltime_end(){
  static Logger logger("munich::walltime_end");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  end_time_point = chrono::system_clock::now();
  end_time = chrono::system_clock::to_time_t(end_time_point);

  std::chrono::duration<double> elapsed_seconds = end_time_point - start_time_point;
  logger.newLine(LOG_INFO);
  logger << LOG_INFO << "Wall time:  elapsed in seconds:  " << left << setprecision(15) << setw(23) << elapsed_seconds.count() << endline;
  //  double wall_sec_d = elapsed_seconds.count();
  int wall_sec = int(elapsed_seconds.count());
  int wall_min = 0;
  int wall_h = 0;
  if (wall_sec >= 3600) {
    wall_h = int(wall_sec / 3600);
    wall_sec -= wall_h * 3600;
  }
  if (wall_sec >= 60) {
    wall_min = int(wall_sec / 60);
    wall_sec -= wall_min * 60;
  }
  end_CPU_time = clock();
  double CPU_seconds = 1. * (end_CPU_time - start_CPU_time) / CLOCKS_PER_SEC;
  logger << LOG_INFO << "CPU time:   elapsed in seconds:  " << setprecision(15) << setw(23) << CPU_seconds << endline;
  logger << LOG_INFO << "CPU time / wall time ratio =     " << left << setprecision(15) << setw(23) << CPU_seconds / elapsed_seconds.count() << endline;

  int CPU_sec = int(CPU_seconds);
  int CPU_min = 0;
  int CPU_h = 0;
  if (CPU_sec >= 3600) {
    CPU_h = int(CPU_sec / 3600);
    CPU_sec -= CPU_h * 3600;
  }
  if (CPU_sec >= 60) {
    CPU_min = int(CPU_sec / 60);
    CPU_sec -= CPU_min * 60;
  }


  logger << LOG_INFO << "Wall time:  evaluation time = " << right << setw(3) << wall_h << "h " << right << setw(3) << wall_min << "min " << right << setw(3) << wall_sec << "sec     n_gen = " << right << setw(10) << psi->i_gen << "   n_acc = " << right << setw(10) << psi->i_acc << endline;
  logger << LOG_INFO << "CPU time:   evaluation time = " << right << setw(3) << CPU_h << "h " << right << setw(3) << CPU_min << "min " << right << setw(3) << CPU_sec << "sec     n_gen = " << right << setw(10) << psi->i_gen << "   n_acc = " << right << setw(10) << psi->i_acc << endline;
  logger << LOG_INFO << "Wall time:  evaluation time in seconds = " << right << setw(7) << wall_h * 3600 + wall_min * 60 + wall_sec << "     n_gen = " << right << setw(10) << psi->i_gen << "   n_acc = " << right << setw(10) << psi->i_acc << endline;
  logger << LOG_INFO << "CPU time:   evaluation time in seconds = " << right << setw(7) << CPU_h * 3600 + CPU_min * 60 + CPU_sec << "     n_gen = " << right << setw(10) << psi->i_gen << "   n_acc = " << right << setw(10) << psi->i_acc << endline;

  logger << LOG_INFO << ctime(&end_time);
  logger << LOG_DEBUG << setprecision(16) << noshowpoint << difftime(end_time, y2k_time) << " seconds since y2k" << endline_flush;
  logger.newLine(LOG_INFO);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

void munich::walltime_now(){
  static Logger logger("munich::walltime_end");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  now_time_point = chrono::system_clock::now();
  now_time = chrono::system_clock::to_time_t(now_time_point);
  //  double passedtime_now_y2k = difftime(now_time, y2k_time);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::get_summary(){
  static Logger logger("munich::get_summary");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ysi->csi = csi;
  ysi->isi = isi;
  ysi->osi = osi;

  ysi->order = order;
  ysi->infilename = infilename;
  ysi->infilename_scaleband = infilename_scaleband;

  ysi->name_mainrundirectory = name_mainrundirectory;

  logger << LOG_DEBUG << "name_mainrundirectory = " << name_mainrundirectory << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "ysi->order = " << ysi->order << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "munich: order = " << order << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "ysi->infilename = " << ysi->infilename << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "munich: infilename = " << infilename << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "ysi->infilename_scaleband = " << ysi->infilename_scaleband << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "munich: infilename_scaleband = " << infilename_scaleband << endline_flush;

  ysi->get_summary();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void munich::get_path_rundirectory(){
  static Logger logger("munich::get_path_rundirectory");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  char buff[FILENAME_MAX];
  char * test = getcwd(buff, FILENAME_MAX);
  logger << LOG_DEBUG << test << endline_flush;
  logger << LOG_DEBUG << buff << endline_flush;
  path_rundirectory = string(buff);
  int i_s4 = -1;
  int i_s5 = -1;
  size_t counter = 0;
  size_t final_counter = 4;
  for (int i_s = 0; i_s < path_rundirectory.size(); i_s++){
    int j_s = path_rundirectory.size() - i_s;
    if (path_rundirectory[j_s] == '/'){
      counter++;
      logger << LOG_DEBUG << "path_rundirectory.substr(" << j_s + 1 << ") = " << path_rundirectory.substr(j_s + 1) << endline_flush;
      if (counter == 1 && path_rundirectory.substr(j_s + 1) == "result"){final_counter = 1;}
      if (counter == final_counter){i_s4 = j_s;}
      if (counter == final_counter + 1){i_s5 = j_s; break;}
    }
  }
  name_rundirectory = path_rundirectory.substr(i_s4 + 1);
  name_mainrundirectory = path_rundirectory.substr(i_s5 + 1, i_s4 - i_s5 - 1);
  path_mainrundirectory = path_rundirectory.substr(0, i_s4);

  logger << LOG_DEBUG << "path_rundirectory = " << path_rundirectory << endline_flush;
  logger << LOG_DEBUG << "path_mainrundirectory = " << path_mainrundirectory << endline_flush;
  logger << LOG_DEBUG << "name_rundirectory = " << name_rundirectory << endline_flush;
  logger << LOG_DEBUG << "name_mainrundirectory = " << name_mainrundirectory << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

  
 
