#include "header.hpp"

void munich::integration_RA_QCD(){
  static Logger logger("munich::integration_RA_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  initialization_integration();
  asi->calculate_ME2check_RA_QCD(psi);
  initialization_integration_check();

  // Maybe simplify nan-checks throughout ???
  while (osi->int_end == 0){
    rsi.perform_iteration_step();
    if (osi->int_end == 1){break;}
    psi->calculate_IS();
    psi->calculate_IS_RA();
    if (psi->RA_x_a == 0){psi->ac_psp(0, psi->MC_phasespace.channel);}
    else {psi->ac_psp_RA_dipoles();}
    if (munich_isnan(psi->xbp_all)){errorhandling_c_xbpsp(); continue;}
    psi->correct_phasespacepoint_real();
    psi->determine_dipole_phasespace_RA();
    esi->determine_p_parton();
    if (munich_isnan(esi->p_parton)){errorhandling_c_psp(); continue;}
    psi->correct_phasespacepoint_dipole(); // Which phase-space is actually corrected here (psi -- osi) ???
    esi->perform_event_selection();
    if (osi->switch_RS == 1){for (int i_a = 1; i_a < csi->n_ps; i_a++){esi->cut_ps[i_a] = -1;}}
    else if (osi->switch_RS == 2){esi->cut_ps[0] = -1;}
    esi->first_non_cut_ps = -1;
    for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){if (esi->cut_ps[i_a] > -1){esi->first_non_cut_ps = i_a; break;}}
    if (esi->first_non_cut_ps == -1){psi->handling_cut_psp(); continue;}
    // from qT subtraction contribution RRA:
    for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){if (esi->cut_ps[i_a] > -1){if (osi->change_cut[esi->cut_ps[i_a] + 1] == 0){osi->change_cut[esi->cut_ps[i_a] + 1] = 1;}}}
    if (osi->switch_console_output_tau_0){psi->output_check_tau_0();}
    if (osi->switch_output_testpoint){osi->output_testpoint_input();}
    psi->ag_psp(0, 0);
    psi->ag_psp_RA_dipoles();
    psi->calculate_g_tot();
    if (munich_isnan(psi->g_tot)){errorhandling_RA_gtot(); continue;}
    if (csi->type_contribution == "L2RA" || user->string_value[user->string_map["model"]] == "Bornloop"){
      osi->determine_techcut_RA();
      if (psi->RA_techcut){psi->RA_techcut = 0; psi->handling_techcut_psp(); continue;}
    }
    asi->calculate_ME2_RA_QCD();
    if (!(csi->type_contribution == "L2RA" || user->string_value[user->string_map["model"]] == "Bornloop")){
      osi->determine_techcut_RA();
    }
    if (csi->class_contribution_qTcut_explicit){if (psi->RA_techcut){for (int i_a = 0; i_a < csi->n_ps; i_a++){if (esi->cut_ps[i_a] != -1){esi->qT_basic->counter_killed_qTcut[esi->cut_ps[i_a]]++;}}}}
    if (psi->RA_techcut){psi->RA_techcut = 0; psi->handling_techcut_psp(); continue;}
    // Check for matrix elements that are considered unstable and thus set to zero:
    if (csi->type_contribution == "L2RA" || user->string_value[user->string_map["model"]] == "Bornloop"){
      for (int i_a = 0; i_a < csi->n_ps; i_a++){
	//	logger << LOG_INFO << "esi->cut_ps[" << i_a << "] = " << esi->cut_ps[i_a] << "   osi->RA_ME2[" << i_a << "] = " << osi->RA_ME2[i_a] << endline_flush;
	if (esi->cut_ps[i_a] != -1 && osi->RA_ME2[i_a] == 0.){
	  psi->RA_techcut = 1;
	  logger << LOG_DEBUG << right << setw(12) << psi->i_gen << " [" << setw(3) << psi->i_nan << "] " << " " << setw(10) << psi->i_acc << " (" << setw(7) << psi->i_tec << ") " << " " << setw(10) << psi->i_rej << " " << left << "   Matrix element set to zero in dipole no. " << setw(2) << i_a << "   -> psi->RA_techcut = " << psi->RA_techcut << endline_flush;
	}
      }
      if (psi->RA_techcut){
	psi->i_tec++;
	psi->RA_techcut = 0;
	psi->handling_techcut_psp();
	continue;
      }
    }

    double temp_sum_RA_ME2 = accumulate(osi->RA_ME2.begin(), osi->RA_ME2.end(), 0.);
    if (munich_isinf(temp_sum_RA_ME2)){errorhandling_RA_me2(3); continue;}
    if (munich_isnan(temp_sum_RA_ME2)){errorhandling_RA_me2(0); continue;}
    if (!(osi->check_vanishing_ME2_end)){handling_vanishing_me2(); if (osi->flag_vanishing_ME2){continue;}}
    if (psi->g_tot == 0.){errorhandling_RA_me2(1); continue;} // does that really happen?
    if (munich_isinf(psi->g_tot)){psi->handling_cut_psp(); continue;}
    osi->determine_scale_complete();
    if (csi->class_contribution_qTcut_explicit){for (int i_a = 0; i_a < csi->n_ps; i_a++){if (esi->cut_ps[i_a] != -1){esi->qT_basic->counter_acc_qTcut[esi->cut_ps[i_a]]++;}}}
    // add proper errorhandling, also for mu_fact !!!
    if (osi->switch_CV){if (munich_isnan(osi->RA_value_factor_alpha_S)){errorhandling_alpha_S(); continue;}}
    if (osi->switch_TSV){if (munich_isnan(osi->value_relative_factor_alpha_S)){errorhandling_alpha_S(); continue;}}
    if (osi->switch_moment){osi->moments();}
    if (osi->switch_CV){osi->calculate_pdf_LHAPDF_RA_CV();}
    if (osi->switch_TSV){osi->calculate_pdf_LHAPDF_list_TSV();}
    if (osi->switch_CV){osi->determine_integrand_RA();}
    if (munich_isnan(osi->var_RA_ME2)){errorhandling_RA_me2(1); continue;}
    if (osi->switch_CV){if (munich_isnan(osi->var_RA_ME2_CV)){errorhandling_RA_me2(1); continue;}}
    if (osi->switch_CV){if (munich_isnan(osi->var_RA_ME2_CV)){errorhandling_RA_me2(2); continue;}}
    if (psi->i_gen <= esi->switch_output_check_selection){osi->output_check_selection_psp();}
    if (osi->switch_output_cancellation_check){osi->output_cancellation_RA();}
    if (osi->switch_TSV){osi->determine_psp_weight_TSV();}
    if (osi->switch_CV){osi->determine_psp_weight_RA();}
    if (osi->switch_output_maxevent){osi->output_integrand_maximum_RA();}
    if (osi->switch_distribution){osi->determine_distribution_complete();}
    if (!psi->end_optimization){
      osi->determine_techcut_optimization_RA();
      if (osi->switch_optimization_modifier){osi->calculate_optimization_modifier();}
      psi->psp_optimization_complete(osi->integrand, osi->this_psp_weight, osi->this_psp_weight2, osi->optimization_modifier);
    }
    psi->i_acc++; psi->i_acc_step++;
    osi->determine_runtime();
  }
  osi->output_finalization_integration();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
