#include "header.hpp"

void munich::integration_VT_QCD(){
  static Logger logger("munich::integration_VT_QCD");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  initialization_integration();
  asi->calculate_ME2check_VT_QCD();
  initialization_integration_check();
  
  while (osi->int_end == 0){
    rsi.perform_iteration_step();
    if (osi->int_end == 1){break;}
    psi->calculate_IS();
    if (!csi->switch_VTX){
      // calculate z1z2 only for accepted points (but use the same random number) !!!
      psi->calculate_initial_collinear_z1z2_IS();
      if (psi->switch_resummation){psi->calculate_IS_QT();}
    }
    else {psi->rng.access_random_number(2 * psi->n_random_z);} // to get identical phase space points as for (csi->switch_VTX)
    // z1/z2 could be calculated only after event selection !!!
    if (!psi->identical_psp){
      psi->ac_psp(0, psi->MC_phasespace.channel);
      esi->determine_p_parton();
      if (munich_isnan(esi->p_parton)){errorhandling_c_psp(); continue;}
    }

#ifdef MORE
    if (osi->switch_resummation){
      double Q = sqrt(psi->xbs_all[0][0]);
      double y = log(sqrt(psi->x_pdf[1] / psi->x_pdf[2]));
      performQTBoost_cms(psi->QT_qt2, Q, y, esi->p_parton);
    }
#endif

    esi->perform_event_selection();
    if (esi->cut_ps[0] == -1){psi->handling_cut_psp(); continue;}
    if (osi->dynamical_Qres){osi->QT_Qres = osi->QT_Qres_prefactor*sqrt(psi->xbs_all[0][0]);}
    if (osi->switch_console_output_tau_0){psi->output_check_tau_0();}
    if (osi->switch_output_testpoint){osi->output_testpoint_input();}
    if (!psi->identical_psp){psi->ag_psp(0, 0);}
    psi->calculate_g_tot();
    if (munich_isinfnan(psi->g_tot)){errorhandling_gtot(); continue;}
    osi->determine_scale_complete();
    if (!psi->identical_psp){
      asi->calculate_ME2_VT_QCD(); // change order -> to before scale determination ???
      osi->ME2 = osi->VA_V_ME2 + osi->VA_X_ME2;
      if (munich_isnan(osi->ME2)){errorhandling_me2(); continue;}
      // asi->failed_evaluation  needs to be set by amplitude class !!!
      if (asi->failed_evaluation){ errorhandling_amplitude(); continue; }
    }
    if (!(osi->check_vanishing_ME2_end)){handling_vanishing_me2(); if (osi->flag_vanishing_ME2){continue;}}

    if (osi->switch_moment){osi->moments();}
    if (osi->switch_CV){osi->calculate_pdf_LHAPDF_list_CV();}
    if (osi->switch_TSV){osi->calculate_pdf_LHAPDF_list_TSV();}
    osi->determine_integrand_CX_ncollinear_VT_QCD();
    if (osi->switch_TSV){osi->determine_psp_weight_TSV();}
    osi->determine_psp_weight_QT();
    if (osi->switch_output_maxevent){osi->output_integrand_maximum_VA();}
    if (osi->switch_distribution){osi->determine_distribution_complete();}
    if (osi->switch_optimization_modifier){osi->calculate_optimization_modifier();}
    psi->psp_optimization_complete(osi->integrand, osi->this_psp_weight, osi->this_psp_weight2, osi->optimization_modifier);
    psi->i_acc++; psi->i_acc_step++;
    if (csi->n_particle == 1){psi->identical_psp = true;}
    osi->determine_runtime();
  }
  osi->output_finalization_integration();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
