#include "header.hpp"

dddistribution::dddistribution(){
  static Logger logger("dddistribution::dddistribution");
  logger.called();

  name = "";
  d_1 = 0;
  d_2 = 0;
  n_bins = 0;
  step = 0.;
  xdistribution temp;
  distribution_1 = temp;
  distribution_2 = temp;
  mirror_type = -1;

  logger.finished();
}

dddistribution::dddistribution(string _name, int _d_1, int _d_2, observable_set *_osi){
  static Logger logger("dddistribution::dddistribution");
  logger.called();

  osi = _osi;
  name = _name;
  d_1 = _d_1;
  d_2 = _d_2;
  n_bins = osi->dat[d_1].n_bins * osi->dat[d_2].n_bins;
  step = osi->dat[d_1].step * osi->dat[d_2].step;
  distribution_1 = osi->dat[d_1];
  distribution_2 = osi->dat[d_2];

  logger << LOG_DEBUG_VERBOSE << "before mirror_type" << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "osi->csi->type_parton.size() = " << osi->csi->type_parton.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "osi->csi->type_parton[0].size() = " << osi->csi->type_parton[0].size() << endline_flush;
  if (osi->csi->type_parton[0].size() > 2){
    //      if (osi->csi->type_parton[0][1] != osi->csi->type_parton[0][2]){mirror_type = -1;} // probably wrong for symmetric dddistributions !!!
    if (osi->csi->type_parton[0][1] == osi->csi->type_parton[0][2]){mirror_type = -1;}
    else if (distribution_1.symm == 0 && distribution_2.symm == 0){mirror_type = 0;}
    else if (distribution_1.symm != 0 || distribution_2.symm != 0){mirror_type = 1;}
    else {logger << LOG_ERROR << "dddistribution  " << name << " : Invalid symmtetry structure!" << endline_flush; exit(1);}
  }
  logger << LOG_DEBUG_VERBOSE << "mirror_type = " << mirror_type << endline_flush;

  bin.resize(osi->n_ps);
  bin_max.resize(osi->n_ps);

  logger.finished();
}


void dddistribution::determineBin() {
  static Logger logger("dddistribution::determineBin");
  logger.called();

  // for every phase space point ???
  bin.resize(osi->n_ps);
  bin_max.resize(osi->n_ps);

  mirror_bin.resize(osi->n_ps);
  mirror_bin_max.resize(osi->n_ps);

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    logger << LOG_DEBUG_VERBOSE << setw(20) << name << "   mirror_type = " << mirror_type << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "distribution_1.name = " << distribution_1.xdistribution_name << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "distribution_2.name = " << distribution_2.xdistribution_name << endline_flush;
  }
  for (int i_a = 0; i_a < osi->n_ps; i_a++){
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
      logger << LOG_DEBUG_VERBOSE << "distribution_1.bin.size() = " << distribution_1.bin.size() << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "distribution_2.bin.size() = " << distribution_2.bin.size() << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "osi->dat[" << d_1 << "].bin.size() = " << osi->dat[d_1].bin.size() << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "osi->dat[" << d_2 << "].bin.size() = " << osi->dat[d_2].bin.size() << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "distribution_1.n_bins = " << distribution_1.n_bins << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "distribution_2.n_bins = " << distribution_2.n_bins << endline_flush;
    }
    if (osi->dat[d_1].bin[i_a] == -1 || osi->dat[d_2].bin[i_a] == -1){
      bin[i_a] = -1;
    }
    else {
      bin[i_a] = osi->dat[d_1].bin[i_a] * distribution_2.n_bins + osi->dat[d_2].bin[i_a];
    }

    if (mirror_type == -1){ //  actually no mirror_bin needed !!!
      mirror_bin[i_a] = bin[i_a];
    }
    else if (mirror_type == 0){
      //    if (osi->csi->type_parton[i_a][1] != osi->csi->type_parton[i_a][2]){
      //      if (distribution_1.symm == 0 && distribution_2.symm == 0){ //  asymmetric distribution 1, asymmetric distribution 2
      mirror_bin[i_a] = bin[i_a];
    }
    else if (mirror_type == 1){
      //      else if (distribution_1.symm == 0 && distribution_2.symm == 1){ //  asymmetric distribution 1, symmetric distribution 2
      if (distribution_1.symm == 0 && distribution_2.symm == 1){ //  asymmetric distribution 1, symmetric distribution 2
	mirror_bin[i_a] = osi->dat[d_1].bin[i_a] * distribution_2.n_bins + (distribution_2.n_bins - 1 - osi->dat[d_2].bin[i_a]);

	//	logger << LOG_DEBUG_VERBOSE << setw(20) << name << "          bin[" << i_a << "] = " << bin[i_a] << "   d1-bin = " << setw(3) << osi->dat[d_1].bin[i_a] << "   d2-bin = " << osi->dat[d_2].bin[i_a] << endline_flush;
	//	logger << LOG_DEBUG_VERBOSE << setw(20) << "" << "   mirror_bin[" << i_a << "] = " << mirror_bin[i_a] << endline_flush;
      }
      else if (distribution_1.symm == 1 && distribution_2.symm == 0){ //  symmetric distribution 1, asymmetric distribution 2
	mirror_bin[i_a] = (distribution_1.n_bins - 1 - osi->dat[d_1].bin[i_a]) * distribution_2.n_bins + osi->dat[d_2].bin[i_a];
      }
      else if (distribution_1.symm == 1 && distribution_2.symm == 1){ //  symmetric distribution 1, symmetric distribution 2
	mirror_bin[i_a] = n_bins - 1 - bin[i_a];
      }
      else {logger << LOG_FATAL << "No allowed case chosen." << endline_flush; exit(1);}
    }
    else {logger << LOG_ERROR << "dddistribution  " << name << " : Invalid symmtetry structure!" << endline_flush; exit(1);}

    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << setw(20) << name << "   bin[" << i_a << "] = " << bin[i_a] << "   mirror_bin[" << i_a << "] = " << mirror_bin[i_a] << endline_flush; }
  }

  logger.finished();
}


void dddistribution::calculate_distribution_bin(){
  static Logger logger("dddistribution::calculate_distribution_bin");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
    for (int i_q = 0; i_q < osi->distribution_no_qTcut.size(); i_q++){
      stringstream out_dnq;
      for (int i_qa = 0; i_qa < osi->distribution_no_qTcut_phasespace[i_q].size(); i_qa++){
	out_dnq << osi->distribution_no_qTcut_phasespace[i_q][i_qa] << " ";
      }
      logger << LOG_DEBUG_VERBOSE << "osi->distribution_no_qTcut_phasespace [" << i_q << "] @ " << setw(3) << osi->distribution_no_qTcut[i_q] << " = " << out_dnq.str() << endline_flush;
    }
  }

  distribution_no_bin.clear();
  distribution_no_bin_phasespace.clear();
  distribution_no_bin_no_qTcut_phasespace.clear();

  if (osi->n_ps == 1){
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << name << "   osi->n_ps = 1" << endline_flush; }
    /////    int i_d = dat.size() + i_ddd;
    if (distribution_1.typeCumulative != CUMULATIVE_NONE && distribution_2.typeCumulative != CUMULATIVE_NONE){
      // to be filled !!! dddistributions don't seem to be implemented for cumulative distribution
    }
    else if (distribution_1.typeCumulative == CUMULATIVE_NONE && distribution_2.typeCumulative != CUMULATIVE_NONE){
      // to be filled !!! dddistributions don't seem to be implemented for cumulative distribution
    }
    else if (distribution_1.typeCumulative != CUMULATIVE_NONE && distribution_2.typeCumulative == CUMULATIVE_NONE){
      // to be filled !!! dddistributions don't seem to be implemented for cumulative distribution
    }
    else if (distribution_1.typeCumulative == CUMULATIVE_NONE && distribution_2.typeCumulative == CUMULATIVE_NONE){
      //      logger << LOG_DEBUG_VERBOSE << name << "   bin[0] = " << bin[0] << endline_flush;
      if (bin[0] != -1){
	distribution_no_bin.resize(1, bin[0]);
	//	  distribution_no_bin = bin; // could still be -1 --- does the same ???
	distribution_no_bin_phasespace.resize(1, vector<int> (1, 0)); // -> (1, 0) only 1 phasespace, namely 0; resize(1, ...) -> only 1 bin
	distribution_no_bin_no_qTcut_phasespace.resize(1, vector<vector<int> > (1, vector<int> (1, 0)));

	/////	for (int i_q = 0; i_q <= cut_ps[0]; i_q++){bin_count_TSV[i_q][i_d][dddat[i_ddd].bin[0]]++;}// ??? < or <= ???  // mirrored contributions ???
      }
    }
    else {
      logger << LOG_FATAL << "dddat(" << name << "):   osi->n_ps = 1   typeCumulative combination not implemented." << endline_flush; exit(1);
    }
  }
  else if (osi->n_ps > 1){
    if (distribution_1.typeCumulative != CUMULATIVE_NONE || distribution_2.typeCumulative != CUMULATIVE_NONE){
      // to be filled !!! dddistributions don't seem to be implemented for cumulative distribution
    }
    else if (distribution_1.typeCumulative == CUMULATIVE_NONE && distribution_2.typeCumulative == CUMULATIVE_NONE){
      ////////////////////////////////////////////////////////////////////////////////////////////
      //  osi->n_ps > 1  //  (dddat[i_ddd].distribution_1&2).typeCumulative == CUMULATIVE_NONE  //
      ////////////////////////////////////////////////////////////////////////////////////////////

      // better avoid erase and fill a temporary vector, sort it, and then fill distribution_no_bin ??? Check if result is identical !!!
      /*
      vector<int> temp_bin = bin;
      sort(temp_bin.begin(), temp_bin.end());
      for (int i_b = 0; i_b < osi->n_ps; i_b++){
	if (temp_bin[i_b] == 1;){ continue; }
	if (distribution_no_bin.size() == 0){ distribution_no_bin.push_back(temp_bin[i_b]); }
	else { if (temp_bin[i_b] > distribution_no_bin[distribution_no_bin.size() - 1]){ distribution_no_bin.push_back(temp_bin[i_b]); } }
      }
      */      
      distribution_no_bin = bin;
      sort(distribution_no_bin.begin(), distribution_no_bin.end());
      for (int i_b = osi->n_ps - 1; i_b > 0; i_b--){
	if (distribution_no_bin[i_b] == distribution_no_bin[i_b - 1]){distribution_no_bin.erase(distribution_no_bin.begin() + i_b);}
      }
      if (distribution_no_bin[0] == -1){distribution_no_bin.erase(distribution_no_bin.begin());}

      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	logger << LOG_DEBUG_VERBOSE << "distribution_no_bin(" << name << ").size() = " << distribution_no_bin.size() << endline_flush; 
	stringstream out_dnb;
	for (int i_b = 0; i_b < distribution_no_bin.size(); i_b++){out_dnb << distribution_no_bin[i_b] << "   ";}
	logger << LOG_DEBUG_VERBOSE << "distribution_no_bin (" << name << ") = " << out_dnb.str() << endline_flush;
      }

      distribution_no_bin_phasespace.resize(distribution_no_bin.size());
      for (int i_a = 0; i_a < osi->n_ps; i_a++){
	for (int i_b = 0; i_b < distribution_no_bin.size(); i_b++){
	  if (distribution_no_bin[i_b] == bin[i_a]){distribution_no_bin_phasespace[i_b].push_back(i_a); break;}
	}
      }

      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	for (int x_b = 0; x_b < distribution_no_bin_phasespace.size(); x_b++){
	  stringstream out_dnbp;
	  for (int x_a = 0; x_a < distribution_no_bin_phasespace[x_b].size(); x_a++){
	    out_dnbp << distribution_no_bin_phasespace[x_b][x_a] << " ";
	  }
	  logger << LOG_DEBUG_VERBOSE << "distribution_no_bin_phasespace (" << name << ")[" << x_b << "] @ " << setw(3) << distribution_no_bin[x_b] << " = " << out_dnbp.str() << endline_flush;
	}
      }

      distribution_no_bin_no_qTcut_phasespace.resize(distribution_no_bin.size(), vector<vector<int> > (osi->distribution_no_qTcut.size()));
      for (int i_b = 0; i_b < distribution_no_bin.size(); i_b++){
	for (int i_q = 0; i_q < osi->distribution_no_qTcut.size(); i_q++){
	  for (int i_ba = 0; i_ba < distribution_no_bin_phasespace[i_b].size(); i_ba++){
	    for (int i_qa = 0; i_qa < osi->distribution_no_qTcut_phasespace[i_q].size(); i_qa++){
	      //	      logger << LOG_DEBUG_VERBOSE << "i_b = " << i_b << "   i_q = " << i_q << "   i_ba = " << i_ba << "   i_qa = " << i_qa << endline_flush;
	      if (distribution_no_bin_phasespace[i_b][i_ba] == osi->distribution_no_qTcut_phasespace[i_q][i_qa]){
		distribution_no_bin_no_qTcut_phasespace[i_b][i_q].push_back(distribution_no_bin_phasespace[i_b][i_ba]);
	      }
	    }
	  }
	}
      }

      if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){
	for (int i_b = 0; i_b < distribution_no_bin.size(); i_b++){
	  for (int i_q = 0; i_q < osi->distribution_no_qTcut.size(); i_q++){
	    stringstream out_dnbnqp;
	    for (int i_x = 0; i_x < distribution_no_bin_no_qTcut_phasespace[i_b][i_q].size(); i_x++){
	      out_dnbnqp << distribution_no_bin_no_qTcut_phasespace[i_b][i_q][i_x] << " ";
	    }
	    logger << LOG_DEBUG_VERBOSE << "distribution_no_bin_no_qTcut_phasespace [" << i_b << "][" << i_q << "] @ bin -> " << setw(3) << distribution_no_bin[i_b] << " @ qTcut -> " << setw(3) << osi->distribution_no_qTcut[i_q] << " = " << out_dnbnqp.str() << endline_flush;
	  }
	}
      }
    }
    else {
      logger << LOG_FATAL << "dddat(" << name << "):   typeCumulative combination not implemented." << endline_flush; exit(1);
    }
  }

  logger.finished();
}


ostream & operator << (ostream &s, const dddistribution &sp){
  static Logger logger("dddistribution");
  logger.called();

  stringstream temp_ss;
  temp_ss << setw(20) << sp.name;
  temp_ss << "   n_bins = " << setw(3) << sp.n_bins;
  temp_ss << "   mirror_type = " << setw(3) << sp.mirror_type;
  temp_ss << "   step = " << setw(3) << sp.step << " (unused)";
  temp_ss << "   d1 = " << setw(20) << (sp.distribution_1).xdistribution_name << " (" << setw(2) << sp.d_1 << ")";
  temp_ss << sp.distribution_1;
  temp_ss << "   d2 = " << setw(20) << (sp.distribution_2).xdistribution_name << " (" << setw(2) << sp.d_2 << ")";
  temp_ss << sp.distribution_2;
  temp_ss << endl;

  s << temp_ss.str() << endl;
  logger << LOG_DEBUG_VERBOSE << temp_ss.str() << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished - return" << endline_flush;
  return s;
}
