#include "header.hpp"

void manager_TSV::initialization_input_predefinition_TSV(inputparameter_set * isi){
  static Logger logger("manager_TSV::initialization_predefinition_TSV");
  logger.called();

  // scale variation related pre-definitions

  switch_TSV = 0;
  for (int i_i = 0; i_i < isi->input_observable.size(); i_i++){
    if (isi->input_observable[i_i].variable == "switch_TSV"){switch_TSV = atoi(isi->input_observable[i_i].value.c_str());}
  }

  int counter_scaleset = -1;
  name_set_TSV.resize(0);
  for (int i_i = 0; i_i < isi->input_observable.size(); i_i++){
    if (isi->input_observable[i_i].variable == "name_set_TSV" || isi->input_observable[i_i].variable == "scaleset"){
      int flag = -1;
      for (int i_s = 0; i_s < counter_scaleset + 1; i_s++){
	if (isi->input_observable[i_i].value == name_set_TSV[i_s]){flag = i_s; break;}
      }
      logger << LOG_DEBUG_VERBOSE << "flag = " << flag << endline_flush;
      if (flag == -1){
	counter_scaleset++;
	name_set_TSV.push_back(isi->input_observable[i_i].value);
      }
    }
  }

  
  // to deal with differences between two selected scale choices, acting as a local subtraction for each other; or to reconstruct scale variation results

  int counter_scalediffset = -1;
  name_diff_set_TSV.resize(0);
  for (int i_i = 0; i_i < isi->input_observable.size(); i_i++){
    if (isi->input_observable[i_i].variable == "name_diff_set_TSV"){
      int flag = -1;
      for (int i_s = 0; i_s < counter_scalediffset + 1; i_s++){
	if (isi->input_observable[i_i].value == name_diff_set_TSV[i_s]){flag = i_s; break;}
      }
      logger << LOG_DEBUG_VERBOSE << "flag = " << flag << endline_flush;
      if (flag == -1){
	counter_scalediffset++;
	name_diff_set_TSV.push_back(isi->input_observable[i_i].value);
      }
    }
  }

  logger.finished();
}


void manager_TSV::initialization_set_default(){
  static Logger logger("manager_TSV::initialization");
  logger.called();

  logger << LOG_DEBUG << "number of scalesets: " << name_set_TSV.size() << endline_flush;
  for (int i_s = 0; i_s < name_set_TSV.size(); i_s++){
    logger << LOG_DEBUG << "name_set_TSV[" << setw(2) << i_s << "] = " << name_set_TSV[i_s] << endline_flush;
  }
  logger.newLine(LOG_DEBUG);

  n_set_TSV = name_set_TSV.size();
  logger << LOG_DEBUG << "n_set_TSV = " << n_set_TSV << endline_flush;

  central_scale_TSV.resize(n_set_TSV);
  central_scale_ren_TSV.resize(n_set_TSV);
  central_scale_fact_TSV.resize(n_set_TSV);
  relative_central_scale_TSV.resize(n_set_TSV);
  relative_central_scale_ren_TSV.resize(n_set_TSV);
  relative_central_scale_fact_TSV.resize(n_set_TSV);

  factor_scale_TSV.resize(n_set_TSV);
  factor_scale_ren_TSV.resize(n_set_TSV);
  factor_scale_fact_TSV.resize(n_set_TSV);

  dynamic_scale_TSV.resize(n_set_TSV);
  dynamic_scale_ren_TSV.resize(n_set_TSV);
  dynamic_scale_fact_TSV.resize(n_set_TSV);
  min_qTcut_TSV.resize(n_set_TSV);
  max_qTcut_TSV.resize(n_set_TSV);
  min_qTcut_distribution_TSV.resize(n_set_TSV);
  max_qTcut_distribution_TSV.resize(n_set_TSV);

  no_central_scale_ren_TSV.resize(n_set_TSV);
  // !!! no input parameter !!!

  no_central_scale_fact_TSV.resize(n_set_TSV);
  // !!! no input parameter !!!

  for (int i_s = 0; i_s < n_set_TSV; i_s++){
    central_scale_TSV[i_s] = empty_double;
    central_scale_ren_TSV[i_s] = empty_double;
    central_scale_fact_TSV[i_s] = empty_double;
    relative_central_scale_TSV[i_s] = empty_double;
    relative_central_scale_ren_TSV[i_s] = empty_double;
    relative_central_scale_fact_TSV[i_s] = empty_double;
    dynamic_scale_TSV[i_s] = empty_int;
    dynamic_scale_ren_TSV[i_s] = empty_int;
    dynamic_scale_fact_TSV[i_s] = empty_int;
    factor_scale_TSV[i_s] = empty_int;
    factor_scale_ren_TSV[i_s] = empty_int;
    factor_scale_fact_TSV[i_s] = empty_int;
    min_qTcut_TSV[i_s] = empty_double;
    max_qTcut_TSV[i_s] = empty_double;
    min_qTcut_distribution_TSV[i_s] = empty_double;
    max_qTcut_distribution_TSV[i_s] = empty_double;
  }

  
  // to deal with differences between two selected scale choices, acting as a local subtraction for each other; or to reconstruct scale variation results

  n_diff_set_TSV = name_diff_set_TSV.size();

  name_diff_set_plus_TSV.resize(n_diff_set_TSV, "");
  name_diff_set_minus_TSV.resize(n_diff_set_TSV, "");
  no_diff_set_plus_TSV.resize(n_diff_set_TSV, -1);
  no_diff_set_minus_TSV.resize(n_diff_set_TSV, -1);

  n_extended_set_TSV = n_set_TSV + n_diff_set_TSV;

  n_scale_TSV.resize(n_extended_set_TSV);
  n_scale_ren_TSV.resize(n_extended_set_TSV);
  n_scale_fact_TSV.resize(n_extended_set_TSV);

  switch_distribution_TSV.resize(n_extended_set_TSV);
  switch_moment_TSV.resize(n_extended_set_TSV);

  max_n_integrand_TSV.resize(n_extended_set_TSV);
  // !!! no input parameter !!!

  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    n_scale_TSV[i_s] = empty_int;
    n_scale_ren_TSV[i_s] = empty_int;
    n_scale_fact_TSV[i_s] = empty_int;

    switch_distribution_TSV[i_s] = empty_int;
    switch_moment_TSV[i_s] = empty_int;
  }

  name_extended_set_TSV.resize(n_extended_set_TSV, "");
  for (int i_s = 0; i_s < n_set_TSV; i_s++){
    name_extended_set_TSV[i_s] = name_set_TSV[i_s];
  }
  for (int i_s = 0; i_s < n_diff_set_TSV; i_s++){
    name_extended_set_TSV[n_set_TSV + i_s] = name_diff_set_TSV[i_s];
  }

  
  /////  switch_reference = "";
  name_reference_TSV = "" ;
  no_reference_TSV = -1;
  no_scale_ren_reference_TSV = -1;
  no_scale_fact_reference_TSV = -1;
  no_qTcut_reference_TSV = 0;

  logger.finished();
}


void manager_TSV::initialization_input(inputparameter_set * isi){
  static Logger logger("manager_TSV::initialization");
  logger.called();

  // Error message if  present_scaleset / present_scalediffset  is not set !!!
  int present_scaleset = -1;
  int present_scalediffset = -1;

  for (int i_i = 0; i_i < isi->input_observable.size(); i_i++){
    logger << LOG_DEBUG_VERBOSE << "isi->input_observable[" << i_i << "] = " << isi->input_observable[i_i] << endline_flush;

    if (isi->input_observable[i_i].variable == ""){}

    ////////////////////////////////////////
    //  scale variation parameters - TSV  //
    ////////////////////////////////////////

    else if (isi->input_observable[i_i].variable == "switch_TSV"){switch_TSV = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "name_set_TSV" || isi->input_observable[i_i].variable == "scaleset"){ // to be removed...
      for (int i_s = 0; i_s < n_set_TSV; i_s++){
	if (isi->input_observable[i_i].value == name_set_TSV[i_s]){present_scaleset = i_s; break;}
      }
    }
    else if (isi->input_observable[i_i].variable == "central_scale_TSV"){central_scale_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "central_scale_ren_TSV"){central_scale_ren_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "central_scale_fact_TSV"){central_scale_fact_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "relative_central_scale_TSV"){relative_central_scale_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "relative_central_scale_ren_TSV"){relative_central_scale_ren_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "relative_central_scale_fact_TSV"){relative_central_scale_fact_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "n_scale_TSV"){n_scale_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "n_scale_ren_TSV"){n_scale_ren_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "n_scale_fact_TSV"){n_scale_fact_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "factor_scale_TSV"){factor_scale_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "factor_scale_ren_TSV"){factor_scale_ren_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "factor_scale_fact_TSV"){factor_scale_fact_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "dynamic_scale_TSV"){dynamic_scale_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "dynamic_scale_ren_TSV"){dynamic_scale_ren_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "dynamic_scale_fact_TSV"){dynamic_scale_fact_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "min_qTcut_TSV"){min_qTcut_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "max_qTcut_TSV"){max_qTcut_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "switch_distribution_TSV"){switch_distribution_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "min_qTcut_distribution_TSV"){min_qTcut_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "max_qTcut_distribution_TSV"){max_qTcut_TSV[present_scaleset] = atof(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "switch_moment_TSV"){switch_moment_TSV[present_scaleset] = atoi(isi->input_observable[i_i].value.c_str());}

    else if (isi->input_observable[i_i].variable == "name_diff_set_TSV"){ // to be removed...
      for (int i_s = 0; i_s < n_diff_set_TSV; i_s++){
	if (isi->input_observable[i_i].value == name_diff_set_TSV[i_s]){present_scalediffset = i_s; break;}
      }
    }
    else if (isi->input_observable[i_i].variable == "name_diff_set_plus_TSV"){name_diff_set_plus_TSV[present_scalediffset] = isi->input_observable[i_i].value;}
    else if (isi->input_observable[i_i].variable == "name_diff_set_minus_TSV"){name_diff_set_minus_TSV[present_scalediffset] = isi->input_observable[i_i].value;}

    //    else if (isi->input_observable[i_i].variable == "switch_reference"){switch_reference = isi->input_observable[i_i].value;}
    else if (isi->input_observable[i_i].variable == "name_reference_TSV"){name_reference_TSV = isi->input_observable[i_i].value;}
    else if (isi->input_observable[i_i].variable == "no_reference_TSV"){no_reference_TSV = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "no_scale_ren_reference_TSV"){no_scale_ren_reference_TSV = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "no_scale_fact_reference_TSV"){no_scale_fact_reference_TSV = atoi(isi->input_observable[i_i].value.c_str());}
    else if (isi->input_observable[i_i].variable == "no_qTcut_reference_TSV"){no_qTcut_reference_TSV = atoi(isi->input_observable[i_i].value.c_str());}

  }
    
  logger.finished();
}


void manager_TSV::initialization_after_input(){
  static Logger logger("manager_TSV::initialization");
  logger.called();

  max_dyn_ren = 0;
  max_dyn_fact = 0;
  
  if (switch_TSV){initialization_basic_TSV();}
  else {n_set_TSV = 0;} // should not be required !!!

  // Could be partially or completely shifted to TSV initialization !!!

  for (int i_s = 0; i_s < n_diff_set_TSV; i_s++){
    for (int j_s = 0; j_s < n_set_TSV; j_s++){
      if (name_diff_set_plus_TSV[i_s] == name_set_TSV[j_s]){no_diff_set_plus_TSV[i_s] = j_s;}
      if (name_diff_set_minus_TSV[i_s] == name_set_TSV[j_s]){no_diff_set_minus_TSV[i_s] = j_s;}
    }
    if (no_diff_set_plus_TSV[i_s] == -1){
      logger << LOG_FATAL << "name_diff_set_TSV[" << i_s << "] = " << name_diff_set_TSV[i_s] << "   has an invalid plus scale: name_diff_set_plus_TSV[" << i_s << "] = " << name_diff_set_plus_TSV[i_s] << endline_flush;
      exit(1);
    }
    if (no_diff_set_minus_TSV[i_s] == -1){
      logger << LOG_FATAL << "name_diff_set_TSV[" << i_s << "] = " << name_diff_set_TSV[i_s] << "   has an invalid minus scale: name_diff_set_minus_TSV[" << i_s << "] = " << name_diff_set_minus_TSV[i_s] << endline_flush;
      exit(1);
    }
  }

  for (int i_s = 0; i_s < n_set_TSV; i_s++){
    logger << LOG_DEBUG << "name_set_TSV[" << i_s << "] = " << name_set_TSV[i_s] << endline_flush;
  }
  logger.newLine(LOG_DEBUG);
  for (int i_s = 0; i_s < n_diff_set_TSV; i_s++){
    logger << LOG_DEBUG << "name_diff_set_TSV[" << i_s << "] = " << name_diff_set_TSV[i_s] << endline_flush;
  }
  logger.newLine(LOG_DEBUG);
  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    logger << LOG_DEBUG << "name_extended_set_TSV[" << i_s << "] = " << name_extended_set_TSV[i_s] << endline_flush;
  }
  logger.newLine(LOG_DEBUG);

  // A number of variables need to be filled !!!
  if (switch_TSV){
    //    logger << LOG_DEBUG << "name_reference_TSV          = " << name_reference_TSV << endline_flush;
    for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
      //      logger << LOG_DEBUG << "name_extended_set_TSV[" << i_s << "] = " << name_extended_set_TSV[i_s] << endline_flush;
      if (name_reference_TSV == name_extended_set_TSV[i_s]){
	no_reference_TSV = i_s;
	//	logger << LOG_DEBUG << "no_reference_TSV          = " << no_reference_TSV << endline_flush;
	break;
      }
    }

    if (no_reference_TSV){
      logger << LOG_DEBUG << "default output from TSV:" << endline_flush;
      logger << LOG_DEBUG << "name_reference_TSV          = " << name_reference_TSV << endline_flush;
      logger << LOG_DEBUG << "no_reference_TSV            = " << no_reference_TSV << endline_flush;
      logger << LOG_DEBUG << "no_scale_ren_reference_TSV  = " << no_scale_ren_reference_TSV << endline_flush;
      logger << LOG_DEBUG << "no_scale_fact_reference_TSV = " << no_scale_fact_reference_TSV << endline_flush;
      logger << LOG_DEBUG << "no_qTcut_reference_TSV      = " << no_qTcut_reference_TSV << endline_flush;
    }
    //    else {logger << LOG_DEBUG << "No reference scale defined. Standard used for default output." << endline_flush;}

    logger.newLine(LOG_DEBUG);
  }
  
  logger.finished();
}


 
