#include "header.hpp"

void xdistribution::initialization_particle(){
  static Logger logger("xdistribution::initialization_particle");
  logger.called();

  for (int i_p=0; i_p<all_particle.size(); i_p++) {
    vector<ParticleID> new_group;
    for (int i_s=0; i_s<all_particle[i_p].size(); i_s++) {
      logger << LOG_DEBUG << "new particle " << i_p << ", " << i_s << endline_flush;
      ParticleID new_particle;
      new_particle.type=all_particle_group[i_p][i_s];
      // type to be replaced by x_os eventually !!!
      new_particle.x_os = osi->esi->access_object_selection[osi->esi->object_list[all_particle_group[i_p][i_s]]];
      // new 20211024 !!!
      new_particle.multiplicity = 0;
      
      new_particle.number=all_particle[i_p][i_s];
      logger << LOG_DEBUG << new_particle.type << ", " << new_particle.number << endline_flush;
      if (required_subparticle[i_p][i_s] == 2){
        new_particle.required_definition = true;
        new_particle.required_exists = true;
	//  new_particle.required_exists = false; // (should be true, but is not used so far in checkRequirements
      }
      else if (required_subparticle[i_p][i_s] == 1){
	new_particle.required_definition = false;
        new_particle.required_exists = true;
      }
      else if (required_subparticle[i_p][i_s] == 0){
	new_particle.required_definition = false;
        new_particle.required_exists = false;
      }
      else if (required_subparticle[i_p][i_s] == -1){
	new_particle.required_definition = true;
        new_particle.required_exists = false;
      }
      
      new_group.push_back(new_particle);
    }
    particles.push_back(new_group);
  }

  logger.finished();
}


void xdistribution::initialization_type_cumulative(){
  static Logger logger("xdistribution::initialization_type_cumulative");
  logger.called();

  if (xdistribution_type == "multiplicity+" ||
      xdistribution_type == "pTcut" ||
      xdistribution_type == "mcut" ||
      xdistribution_type == "Mcut" ||
      xdistribution_type == "absdmcut" ||
      xdistribution_type == "absdMcut" ||
      xdistribution_type == "taucut" ||
      xdistribution_type == "pTmincut" ||
      xdistribution_type == "mcutpTlep20" ||
      xdistribution_type == "McutpTlep20" ||
      xdistribution_type == "mcutpTlep100" ||
      xdistribution_type == "McutpTlep100") {
    typeCumulative = CUMULATIVE_UPPER;
    // observable >= 'left bin edge x' -> bins from 0 to 'observable_max' are filled (check if +1 or not !!!).
  }
  else if (xdistribution_type == "pTveto" ||
	   xdistribution_type == "absetamaxveto" ||
	   xdistribution_type == "absetamaxvetopTlep20" ||
	   xdistribution_type == "absetamaxvetopTlep100") {
    typeCumulative = CUMULATIVE_LOWER;
  }
  else if (xdistribution_type == "pTvetoexcl") {
    typeCumulative = CUMULATIVE_BOTH;
  }
  else {
    typeCumulative = CUMULATIVE_NONE;
    // 'left bin edge x' <= observable < 'left bin edge x+1' -> bin 'observable = x' is filled.
  }

  logger.finished();
}


void xdistribution::initialization_range(string & _type_binning, int _n_bins, double _start, double _end, double _step, string _edges){
  static Logger logger("xdistribution::initialization_range");
  logger.called();

  logger << LOG_DEBUG << "_type_binning = " << _type_binning << endline_flush;
  logger << LOG_DEBUG << "_start = " << _start << endline_flush;
  logger << LOG_DEBUG << "_end = " << _end << endline_flush;
  logger << LOG_DEBUG << "_step = " << _step << endline_flush;
  logger << LOG_DEBUG << "_n_bins = " << _n_bins << endline_flush;

  type_binning = _type_binning;

  // Actually _start and _end could reasonably be 0 - check !!!
  int type_input = 0;
  if (_n_bins == 0){type_input = 1;}
  else if (_step == 0.){type_input = 4;}
  else if (_end == 0.){type_input = 3;}
  else if (_start == 0.){type_input = 2;}

  if (type_binning == "linear"){
    logger << LOG_DEBUG << "linear: type_input = " << type_input << endline_flush;
    if (type_input == 0){
      n_bins = _n_bins;
      start = _start;
      end = _end;
      step = _step;
      if ((end - start) != (n_bins * step)){logger << LOG_FATAL << "ill-defined distribution input " << xdistribution_name << " (" << xdistribution_type << ") : start = " << start << "   end = " << end << "   step = " << step << "   n_bins = " << n_bins << endline_flush; exit(1);}
    }
    else if (type_input == 1){
      // no n_bins
      start = _start;
      end = _end;
      step = _step;
      if (step <= 0. || n_bins < 1){logger << LOG_FATAL << "ill-defined distribution input " << xdistribution_name << " (" << xdistribution_type << ") : start = " << start << "   end = " << end << "   step = " << step << endline_flush; exit(1);}
      n_bins = int((end - start) / step);
      double dn_bins = (end - start) / step;
      if (int(dn_bins) != n_bins){
 	double eps = 1.e-12;
	if (abs(dn_bins - n_bins - 1) < eps){ n_bins = int(dn_bins + eps); logger << LOG_DEBUG << "n_bins  had to be corrected to a proper integer value:" << dn_bins << " -> " << n_bins << endline_flush; }
	else { logger << LOG_FATAL << "Input parameters do not correspond to a reasonable integer value of n_bins." << endline_flush; exit (1); }
      }
    }
    else if (type_input == 2){
      // no start
      n_bins = _n_bins;
      step = _step;
      end = _end;
      start = end - n_bins * step;
    }
    else if (type_input == 3){
      // no end
      n_bins = _n_bins;
      start = _start;
      step = _step;
      end = start + n_bins * step;
    }
    else if (type_input == 4){
      // no step
      n_bins = _n_bins;
      if (n_bins <= 0){logger << LOG_FATAL << "ill-defined distribution input " << xdistribution_name << " (" << xdistribution_type << ") : start = " << start << "   end = " << end << "   n_bins = " << n_bins << endline_flush; exit(1);}
      start = _start;
      end = _end;
      step = (end - start) / n_bins;
    }
    if (step <= 0. || n_bins < 1){logger << LOG_FATAL << "ill-defined distribution input " << xdistribution_name << " (" << xdistribution_type << ") : start = " << start << "   end = " << end << "   step = " << step << endline_flush; exit(1);}

    bin_edge.resize(n_bins + 1);
    bin_width.resize(n_bins);
    for (int i_b = 0; i_b < n_bins; i_b++){
      bin_edge[i_b] = start + i_b * step;
      bin_width[i_b] = step;
    }
    bin_edge[n_bins] = end;

    logger << LOG_DEBUG << "start = " << start << endline_flush;
    logger << LOG_DEBUG << "end = " << end << endline_flush;
    logger << LOG_DEBUG << "step = " << step << endline_flush;
    logger << LOG_DEBUG << "n_bins = " << n_bins << endline_flush;

  }
  else if (_type_binning == "logarithmic"){
    if (type_input != 4){logger << LOG_FATAL << "ill-defined distribution; equally-sized bins are not compatible with logarithmic binning " << xdistribution_name << " (" << xdistribution_type << "!" << endline_flush; exit(1);}
    n_bins = _n_bins;
    if (_start <= 0 || _end <= 0){logger << LOG_FATAL << "ill-defined distribution; start/end are not compatible with logarithmic binning " << xdistribution_name << " (" << xdistribution_type << "!" << endline_flush; exit(1);}
    start = log10(_start);
    end = log10(_end);
    step = (end - start) / n_bins;
    bin_edge.resize(n_bins + 1);
    bin_width.resize(n_bins);
    for (int i_b = 0; i_b < n_bins; i_b++){
      bin_edge[i_b] = pow(10., start + i_b * step);
      bin_width[i_b] = bin_edge[i_b] * (pow(10., step) - 1);
      //      bin_edge[i_b] = start + i_b * step;
      //      bin_width[i_b] = pow(10., bin_edge[i_b]) * (pow(10., step) - 1);
      logger << LOG_DEBUG << "bin_edge[" << i_b << "] = " << setprecision(8) << setw(15) << bin_edge[i_b] << "   " << setprecision(8) << setw(15) << bin_width[i_b] << endline_flush;
    }
    bin_edge[n_bins] = pow(10., end);
  }
  else if (_type_binning == "irregular"){
    edges = _edges;
    logger << LOG_DEBUG << "irregular: edges = " << edges << endline_flush;
    vector<string> vs_edge(1);
    for (int i_b = 0; i_b < edges.size(); i_b++){
      logger << LOG_DEBUG << "i_b = " << i_b << "   vs_edge.size() = " << vs_edge.size() << endline_flush;
      if (edges[i_b] == ':'){vs_edge.push_back("");}
      else if (edges[i_b] != ':'){vs_edge[vs_edge.size() - 1].push_back(edges[i_b]);}
    }
    n_bins = vs_edge.size() - 1;
    for (int i_b = 0; i_b < n_bins; i_b++){logger << LOG_DEBUG << "vs_edge[" << i_b << "] = " << setprecision(8) << setw(15) << vs_edge[i_b] << endline_flush;}
    bin_edge.resize(n_bins + 1);
    bin_width.resize(n_bins);
    for (int i_b = 0; i_b < n_bins + 1; i_b++){bin_edge[i_b] = atof(vs_edge[i_b].c_str());}
    for (int i_b = 0; i_b < n_bins; i_b++){bin_width[i_b] = bin_edge[i_b + 1] - bin_edge[i_b];}
    for (int i_b = 0; i_b < n_bins; i_b++){if (bin_width[i_b] < 0.){logger << LOG_FATAL << "ill-defined distribution; irregular bins are not correctly ordered: " << xdistribution_name << " (" << xdistribution_type << "!" << endline_flush; exit(1);}}

    step = 1.;
    logger << LOG_DEBUG << xdistribution_name << " (" << xdistribution_type << ")" << endline_flush;
    for (int i_b = 0; i_b < n_bins; i_b++){logger << LOG_DEBUG << "bin_edge[" << i_b << "] = " << setprecision(8) << setw(15) << bin_edge[i_b] << "   " << setprecision(8) << setw(15) << bin_width[i_b] << endline_flush;}

    start = bin_edge[0];
    end = bin_edge[n_bins];
  }

  logger.finished();
}


void xdistribution::initialization_symmetry(){
  static Logger logger("xdistribution::initialization_symmetry");
  logger.called();

  // default if distribution is not listed in the following:
  symm = 0;
  
  if      ((xdistribution_type == "pT") ||
	   (xdistribution_type == "pTratio") ||
	   (xdistribution_type == "pTveto") ||
	   (xdistribution_type == "pTvetoexcl") ||
	   (xdistribution_type == "pTcut") ||
	   (xdistribution_type == "absy") ||
	   (xdistribution_type == "abseta") ||
	   (xdistribution_type == "dabsy") ||
	   (xdistribution_type == "dabseta") ||
	   (xdistribution_type == "absdy") ||
	   (xdistribution_type == "absdeta") ||
	   (xdistribution_type == "absdabsy") ||
	   (xdistribution_type == "absdabseta") ||
	   (xdistribution_type == "absymax") ||
	   (xdistribution_type == "absetamax") ||
	   (xdistribution_type == "absymin") ||
	   (xdistribution_type == "absetamin") ||
	   (xdistribution_type == "m") ||
	   (xdistribution_type == "mmin") ||
	   (xdistribution_type == "mmax") ||
	   (xdistribution_type == "mcut") ||
	   (xdistribution_type == "dm") ||
	   (xdistribution_type == "absdm") ||
	   (xdistribution_type == "absdmcut") ||
	   (xdistribution_type == "dR") ||
	   (xdistribution_type == "ET") ||
	   (xdistribution_type == "ETproj") ||
	   (xdistribution_type == "mT") ||
	   (xdistribution_type == "mTATLAS") ||
	   (xdistribution_type == "mTCMS") ||
	   (xdistribution_type == "m34_CMS") ||
	   (xdistribution_type == "mTcluster") ||
	   (xdistribution_type == "XS") ||
	   (xdistribution_type == "taucut") ||
	   (xdistribution_type == "muR") ||
	   (xdistribution_type == "muF") ||
	   (xdistribution_type == "multiplicity") ||
	   (xdistribution_type == "multiplicity+") ||
	   (xdistribution_type == "dphi") ||
	   (xdistribution_type == "dphi_pi") ||
	   xdistribution_type == "pTmincut" ||
	   xdistribution_type == "mcutpTlep20" ||
	   xdistribution_type == "McutpTlep20" ||
	   xdistribution_type == "mcutpTlep100" ||
	   xdistribution_type == "McutpTlep100" ||
	   xdistribution_type == "absetamaxveto" ||
	   xdistribution_type == "absetamaxvetopTlep20" ||
	   xdistribution_type == "absetamaxvetopTlep100" ||
	   xdistribution_type == "abscosthetastar" ||
	   xdistribution_type == "phistar" ||
	   xdistribution_type == "abscosthetaS" ||
           xdistribution_type == "absysum"){symm = 0;}
  else if ((xdistribution_type == "y") ||
	   (xdistribution_type == "eta") ||
	   (xdistribution_type == "dy") ||
	   (xdistribution_type == "deta") ||
	   (xdistribution_type == "etadiff") ||
    	   (xdistribution_type == "ysum")){
    // make sure distribution is defined symmetrically
    if (start != -end){
      logger << LOG_FATAL << "For distribution type " << xdistribution_type << " only symmetric ranges are supported (with even number of bins)." << endline_flush;
      exit(1);
    }
    else if (n_bins % 2){
      logger << LOG_FATAL << "For distribution type " << xdistribution_type << " only an even number of bins is supported (with a symmetric range)." << endline_flush;
      exit(1);
    }
    symm = 1;
  }
  else if ((xdistribution_type == "phi") ||
	   (xdistribution_type == "phi_pair_CMS") ||
	   (xdistribution_type == "costheta12_CMS") ||
	   //	   (xdistribution_type == "dpT_djet_pT") ||
	   (xdistribution_type == "costheta")){
    if (all_particle.size() == 1){
      // make sure distribution is defined symmetrically
      if (start != -end){
	logger << LOG_FATAL << "For distribution type " << xdistribution_type << " and a single particle, only symmetric ranges are supported (with even number of bins)." << endline_flush;
	exit(1);
      }
      else if (n_bins % 2){
	logger << LOG_FATAL << "For distribution type " << xdistribution_type << " and a single particle, only an even number of bins is supported (with a symmetric range)." << endline_flush;
	exit(1);
      }
      symm = 1;
    }
    else {symm = 0;}
  }

  logger.finished();
}
