#include "header.hpp"
#include "manager.psp.hpp"

manager_psp::manager_psp(){
  static Logger logger("manager_psp::manager_psp");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

manager_psp::manager_psp(int _x_psp, int _x_a, phasespace_set * _psi){
  static Logger logger("manager_psp::manager_psp (x_psp, x_a, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_psp, _x_a, _psi);

  logger.finished();
}

manager_psp::~manager_psp(){
  static Logger logger("manager_psp::~manager_psp");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void manager_psp::initialization(int _x_psp, int _x_a, phasespace_set * _psi){
  static Logger logger("manager_psp::initialization");
  logger.called();

  x_psp = _x_psp;
  x_a = _x_a;

  psi = _psi;

  if (x_psp == 0){ i_xb = vectorbinary_from_binary(psi->xb_all_max[x_a]); }
  else { i_xb = vectorbinary_from_binary(x_psp); }

  if (i_xb.size() == 0){ logger << LOG_WARN << " Particle " << x_psp << " without content created (x_a = " << x_a << ")." << endline_flush; }
  else {
    if (i_xb.size() == 1){ switch_one_particle = true; }
    else { switch_one_particle = false; }

    switch_IS_FS = false;
    for (int i_p = 0; i_p < i_xb.size(); i_p++){
      if (i_xb[i_p] < 3){ i_xb_IS.push_back(i_xb[i_p]); switch_IS_FS = true; }
      else { i_xb_FS.push_back(i_xb[i_p]); }
    }
  }
  
  logger.finished();
}


void manager_psp::initialization_integration(){
  static Logger logger("manager_psp::initialization_integration");
  logger.called();

  p = & psi->xbp_all[x_a][x_psp];
  s = & psi->xbs_all[x_a][x_psp];
  sqrts = & psi->xbsqrts_all[x_a][x_psp];

  available_p = false;
  if (!switch_one_particle){
    available_s = true;
    available_sqrts = true;
  }
  else {
    available_s = false;
    available_sqrts = false;
  }

  available_level = 0;
 
  logger.finished();
}


void manager_psp::calculate_p(){
  static Logger logger("manager_psp::calculate_p");
  logger.called();

  if (switch_IS_FS){
    *p = psi->xbp_all[x_a][i_xb[0]];
    available_p = true;
    for (int i_p = 0; i_p < i_xb_FS.size(); i_p++){ *p -= psi->xbp_all[x_a][i_xb_FS[i_p]]; }
  }
  else {
    *p = psi->xbp_all[x_a][i_xb[0]];
    for (int i_p = 1; i_p < i_xb_FS.size(); i_p++){ *p += psi->xbp_all[x_a][i_xb_FS[i_p]]; }
  }
  available_p = true;
 
  logger.finished();
}


void manager_psp::calculate_s(){
  static Logger logger("manager_psp::calculate_s");
  logger.called();

  if (!available_p){ calculate_p(); }
  *s = p->m2();
  if (switch_IS_FS){ if (*s > 0.){ *s = 0.;} }
  else { if (*s < 0.){ *s = 0.;} }
  available_s = true;
 
  logger.finished();
}


void manager_psp::calculate_sqrts(){
  static Logger logger("manager_psp::calculate_sqrts");
  logger.called();

  if (!available_s){ calculate_s(); }
  *sqrts = sqrt(*s);
  available_sqrts = true;
 
  logger.finished();
}


string manager_psp::info(){
  static Logger logger("manager_psp::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "psp[" << x_a << "][" << setw(3) << x_psp << "]";
  temp_ss << " - i_xb = ";
  for (int i_p = 1; i_p < i_xb.size(); i_p++){temp_ss << setw(4) << i_xb[i_p];}
  //  temp_ss << " - required = " << required;
  temp_ss << endline;
  temp_ss << "available_p = " << available_p;
  if (available_p){ temp_ss << "   p = " << *p;}
  temp_ss << endline;
  temp_ss << "available_s = " << available_s;
  if (available_s){ temp_ss << "   p = " << *p;}
  temp_ss << endline;
  temp_ss << "available_sqrts = " << available_sqrts;
  if (available_sqrts){ temp_ss << "   p = " << *p;}
  temp_ss << endline;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}


