#include "header.hpp"
#include "mapping.decay.hpp"

mapping_decay::mapping_decay(){
  static Logger logger("mapping_decay::mapping_decay");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

mapping_decay::mapping_decay(int _x_d, int _x_a, int _out, int _out_1, int _out_2, phasespace_set * _psi){
  static Logger logger("mapping_decay::mapping_decay (x_a, m, out, in_1, in_2, out_1, out_2, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_d, _x_a, _out, _out_1, _out_2, _psi);

  logger.finished();
}

mapping_decay::~mapping_decay(){
  static Logger logger("mapping_decay::~mapping_decay");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void mapping_decay::initialization(int _x_d, int _x_a, string & content, phasespace_set * _psi){
  static Logger logger("mapping_decay::initialization(x_d, x_a, content, psi)");
  logger.called();

  vector<string> entries(1);
  for (size_t i_i = 0; i_i < content.size(); i_i++){
    if (content[i_i] == ' ' && i_i != content.size() - 1){ entries.push_back(""); }
    else if (content[i_i] != ' '){ entries[entries.size() - 1].push_back(content[i_i]); }
  }
  if (entries.size() != 3){ logger << LOG_FATAL << "decay[" << _x_d << "][" << _x_a << "]  does not have the right content!" << endline_flush; }
  
  int _out = atoi(entries[0].c_str());
  int _out_1 = atoi(entries[1].c_str());
  int _out_2 = atoi(entries[2].c_str());

  initialization(_x_d, _x_a, _out, _out_1, _out_2, _psi);

  logger.finished();
}


void mapping_decay::initialization(int _x_d, int _x_a, int _out, int _out_1, int _out_2, phasespace_set * _psi){
  static Logger logger("mapping_decay::initialization");
  logger.called();

  x_d = _x_d;
  x_a = _x_a;

  out = _out;
  out_1 = _out_1;
  out_2 = _out_2;

  psi = _psi;

  required_xbp = {out, out_1};
  required_xbs = {out, out_1, out_2};
  required_xbsqrts = {};

  /*
  p_out = & psi->xbp_all[x_a][out];
  p_out_1 = & psi->xbp_all[x_a][out_1];
  p_out_2 = & psi->xbp_all[x_a][out_2];
  
  s_out = & psi->xbs_all[x_a][out];
  s_out_1 = & psi->xbs_all[x_a][out_1] ;
  s_out_2 = & psi->xbs_all[x_a][out_2];

  sqrts_out = & psi->xbsqrts_all[x_a][out];

  g_d = & psi->g_d[x_a][x_d];
  logger << LOG_DEBUG << "g_d = " << g_d << "   has been set." << endline_flush;
    
  // switch_pTthreshold only makes sense for first s-channel decay !!!
  if (psi->csi->class_contribution_QT_powercorrection && psi->switch_decay_pTthreshold && out == psi->xb_max - 4 ){
    switch_pTthreshold = 1;
    exp_pTthreshold_cth = psi->exp_decay_pTthreshold_cth;
    pT2_threshold = pow(psi->exp_decay_pTthreshold_cth, 2);
  }
  else { switch_pTthreshold = 0; }

  if (psi->switch_decay_cth && psi->exp_decay_cth != 1.){
    switch_mapping_cth = 1;
    exp_cth = psi->exp_decay_cth;
  }
  else { switch_mapping_cth = 0; }

  // rotation is obsolete in partonic CMS - reconsider dipole case !!!
  if (psi->switch_decay_system && !((x_a == 0 && out == psi->xb_max - 4) || (x_a > 0 && out == psi->xb_max_dipoles - 4))){ switch_rotate_system = 1; }
  else { switch_rotate_system = 0; }
  
  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 4, x_d}]];
  random_psp_r2 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 5, x_d}]];

  stringstream temp_ss;
  temp_ss << "decay";
  if (switch_rotate_system == 0){ temp_ss << ""; }
  else if (switch_rotate_system == 1){ temp_ss << "_system"; }
  if (switch_mapping_cth == 0){ temp_ss << ""; }
  else if (switch_mapping_cth == 1){ temp_ss << "_exp" << exp_cth; }
  temp_ss << "_" << out << "__" << out_1 << "_" << out_2;
  name = temp_ss.str();
  */
  stringstream temp_ss_d_costheta;
  temp_ss_d_costheta << "costheta_of_decay";
  temp_ss_d_costheta << "_" << out << "__" << out_1 << "_" << out_2;
  MC_IS_name_d_costheta = temp_ss_d_costheta.str();
  
  stringstream temp_ss_d_phi;
  temp_ss_d_phi << "phi_of_decay";
  temp_ss_d_phi << "_" << out << "__" << out_1 << "_" << out_2;
  MC_IS_name_d_phi = temp_ss_d_phi.str();
  
  logger << LOG_DEBUG << "decay[" << setw(5) << x_d << "] =  " << name << "  has been initialized." << endline_flush;

  logger.finished();
}


void mapping_decay::initialization_integration(){
  static Logger logger("mapping_decay::initialization_integration");
  logger.called();

  p_out = & psi->xbp_all[x_a][out];
  p_out_1 = & psi->xbp_all[x_a][out_1];
  p_out_2 = & psi->xbp_all[x_a][out_2];
  
  s_out = & psi->xbs_all[x_a][out];
  s_out_1 = & psi->xbs_all[x_a][out_1];
  s_out_2 = & psi->xbs_all[x_a][out_2];

  sqrts_out = & psi->xbsqrts_all[x_a][out];

  g_d = & psi->g_d[x_a][x_d];
  logger << LOG_DEBUG_VERBOSE << "g_d = " << g_d << "   has been set." << endline_flush;

  // switch_pTthreshold only makes sense for first s-channel decay !!!
  if (psi->csi->class_contribution_QT_powercorrection && psi->switch_decay_pTthreshold && out == psi->xb_max - 4 ){
    switch_pTthreshold = 1;
    exp_pTthreshold_cth = psi->exp_decay_pTthreshold_cth;
    pT2_threshold = pow(psi->decay_pTthreshold_cth, 2);
  }
  else { switch_pTthreshold = 0; }

  if (psi->switch_decay_cth && psi->exp_decay_cth != 1.){
    switch_mapping_cth = 1;
    exp_cth = psi->exp_decay_cth;
  }
  else { switch_mapping_cth = 0; }

  // rotation is obsolete in partonic CMS - reconsider dipole case !!!
  if (psi->switch_decay_system && !((x_a == 0 && out == psi->xb_max - 4) || (x_a > 0 && out == psi->xb_max_dipoles - 4))){ switch_rotate_system = 1; }
  else { switch_rotate_system = 0; }
  
  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 4, x_d}]];
  random_psp_r2 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 5, x_d}]];

  stringstream temp_ss;
  temp_ss << "decay";
  if (switch_rotate_system == 0){ temp_ss << ""; }
  else if (switch_rotate_system == 1){ temp_ss << "_system"; }
  if (switch_mapping_cth == 0){ temp_ss << ""; }
  else if (switch_mapping_cth == 1){ temp_ss << "_exp" << exp_cth; }
  temp_ss << "_" << out << "__" << out_1 << "_" << out_2;
  name = temp_ss.str();
  /*
  stringstream temp_ss_d_costheta;
  temp_ss_d_costheta << "costheta_of_decay";
  temp_ss_d_costheta << "_" << out << "__" << out_1 << "_" << out_2;
  MC_IS_name_d_costheta = temp_ss_d_costheta.str();
  
  stringstream temp_ss_d_phi;
  temp_ss_d_phi << "phi_of_decay";
  temp_ss_d_phi << "_" << out << "__" << out_1 << "_" << out_2;
  MC_IS_name_d_phi = temp_ss_d_phi.str();
  */
  available = false;

  logger.finished();
}


void mapping_decay::calculate_c(){
  static Logger logger("mapping_decay::calculate_c");
  logger.called();

  // Only for: psi->switch_IS_mode_phasespace == 4
  r2 = random_psp_r2->get_random();
  r1 = random_psp_r1->get_random();

  lambda_out_012 = lambda(*s_out, *s_out_1, *s_out_2);
  sqrt_lambda_out_012 = sqrt(lambda_out_012);
  k10 = fourvector(((*s_out + *s_out_1 - *s_out_2) / (2 * *sqrts_out)), 0., 0., sqrt_lambda_out_012 / (2 * *sqrts_out));

  if (switch_pTthreshold){
    pT2_max = lambda_out_012 / (4 * *s_out);
    if (pT2_threshold <= pT2_max){
      costheta_pTcut = sqrt(1. - pT2_threshold / pT2_max);
      costheta = psi->c_costheta_pTcut(r1, costheta_pTcut, exp_pTthreshold_cth);
    }
    else { costheta = psi->c_costheta(r1); }
  }
  else if (switch_mapping_cth){ costheta = psi->h_costheta_collinear(r1, exp_cth); }
  else { costheta = psi->c_costheta(r1); }

  phi = psi->c_phi(r2);

  if (switch_rotate_system){ *p_out_1 = ((k10.rotation(phi, costheta)).rotation(*p_out)).boost(p_out->Pinv(), *s_out); }
  else { *p_out_1 = k10.rotation(phi, costheta).boost(p_out->Pinv(), *s_out); }
  psi->psp[x_a][out_1].register_p();
  *p_out_2 = *p_out - *p_out_1;
  psi->psp[x_a][out_2].register_p();

  //  logger << LOG_INFO << "i_gen = " << setw(8) << psi->i_gen << "*p_out_1 = " << *p_out_1 << "   phi = " << phi << "   costheta = " << costheta << endline_flush;
  //  logger << LOG_INFO << "i_gen = " << setw(8) << psi->i_gen << "   r1 = " << setprecision(15) << setw(23) << r1 << "   r2 = " << setprecision(15) << setw(23) << r2 << endline_flush;

  available = true;

  if (munich_isnan(*p_out_1)){ errorhandling_c(); }

  logger.finished();
}


void mapping_decay::calculate_g(){
  static Logger logger("mapping_decay::calculate_g");
  logger.called();

  if (!available){
    lambda_out_012 = lambda(*s_out, *s_out_1, *s_out_2);
    sqrt_lambda_out_012 = sqrt(lambda_out_012);
    // meaning of auxiliary vector k10 differs for these cases !!!
    //    k10 = fourvector(((*s_out + *s_out_1 - *s_out_2) / (2 * *sqrts_out)), 0., 0., sqrt_lambda_out_012 / (2 * *sqrts_out));
    if (switch_pTthreshold){
      k10 = p_out_1->boost(*p_out, *s_out);
      costheta = k10.costheta();
      pT2_max = lambda_out_012 / (4 * *s_out);
      costheta_pTcut = sqrt(1. - pT2_threshold / pT2_max);
    }
    else if (switch_mapping_cth){
      if (switch_rotate_system){ k10 = (p_out_1->boost(*p_out, *s_out)).rotation_inverse(*p_out); }
      else { k10 = p_out_1->boost(*p_out, *s_out); }
      costheta = k10.costheta();
    }
  }

  if (switch_pTthreshold){
    if (pT2_threshold <= pT2_max){ g_d_costheta = psi->g_costheta_pTcut(costheta, costheta_pTcut, exp_pTthreshold_cth); }
    else { g_d_costheta = 0.5; }
  }
  else if (switch_mapping_cth){ g_d_costheta = psi->g_costheta_collinear(costheta, exp_cth); }
  else { g_d_costheta = 0.5; }

  *g_d = g_d_costheta * (4 * *s_out) / (pi * sqrt_lambda_out_012);

  if (!(*g_d > 0.)){ errorhandling_g(); }

  logger.finished();
}


// maybe directly access memebers r1 and r2 ???
void mapping_decay::calculate_inv_r(){
  static Logger logger("mapping_decay::calculate_inv_r");
  logger.called();

  //  if (switch_rotate_system){ k10 = (*p_out_1.boost(*p_out, *s_out)).rotation_inverse(*p_out); }
  //  else { k10 = *p_out_1.boost(*p_out, *s_out); }
  //  costheta = k10.costheta();
  //  logger << LOG_DEBUG << setw(50) << name << "available = " << available << endline_flush;
  if (!available){
    //    phi = k10.phi();
    //    r2 = psi->inv_phi(phi);
    if (switch_pTthreshold){
      if (pT2_threshold <= pT2_max){ r1 = psi->inv_costheta_pTcut(costheta, costheta_pTcut, exp_pTthreshold_cth); }
      else { r1 = psi->inv_costheta(costheta); }
    }
    else if (switch_mapping_cth){ r1 = psi->inv_costheta_collinear(costheta, exp_cth); }
    else {
      if (switch_rotate_system){ k10 = (p_out_1->boost(*p_out, *s_out)).rotation_inverse(*p_out); }
      else { k10 = p_out_1->boost(*p_out, *s_out); }
      costheta = k10.costheta();
      r1 = psi->inv_costheta(costheta);
    }
    phi = k10.phi();
    r2 = psi->inv_phi(phi);
    if (!(r1 > 0. && r1 < 1.)){ errorhandling_inv_r(); }
  }
  random_psp_r1->get_g_IS(r1, g_IS_r1);
  random_psp_r2->get_g_IS(r2, g_IS_r2);
  *g_d *= g_IS_r1 * g_IS_r2;

  //  logger << LOG_DEBUG << setw(50) << name << "inv_r[4][" << setw(3) << x_d << "] = " << setprecision(15) << setw(23) << r1 << "   g_IS_temp = " << g_IS_r1 << endline_flush;
  //  logger << LOG_DEBUG << setw(50) << name << "inv_r[5][" << setw(3) << x_d << "] = " << setprecision(15) << setw(23) << r2 << "   g_IS_temp = " << g_IS_r2 << endline_flush;
 
  logger.finished();
}


void mapping_decay::errorhandling_c(){
  static Logger logger("mapping_decay::errorhandling_c");
  logger.called();

  n_fail_c++;

  logger << LOG_DEBUG << "Debug output since point creation failed!" << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << out << "] = " << *s_out << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << out_1 << "] = " << *s_out_1 << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << out_2 << "] = " << *s_out_2 << endline_flush;

  logger.finished();
}

  
void mapping_decay::errorhandling_g(){
  static Logger logger("mapping_decay::errorhandling_g");
  logger.called();

  n_fail_g++;
  
  if (*g_d < 0.){
    logger << LOG_DEBUG << "g_d < 0.   out = " << out << "   out_1 = " << out_1 << "   out_2 = " << out_2 << "   *s_out = " << *s_out << endline_flush;
  }

  logger.finished();
}

 
void mapping_decay::errorhandling_inv_r(){
  static Logger logger("mapping_decay::errorhandling_inv_r");
  logger.called();

  n_fail_inv_r++;
  
    if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "r1 is NaN in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; k10.x, k10.y = " << k10.x1() << ", " << k10.x2() << endline_flush;
    r1 = 0.5; // why ???
  }
  else if (r1 <= 0.){
    logger << LOG_DEBUG << "r1 < 0 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; k10.x, k10.y = " << k10.x1() << ", " << k10.x2() << endline_flush;
    r1 = psi->map_technical_r;
  }
  else if (r1 >= 1.){
    logger << LOG_DEBUG << "r1 > 1 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 << "; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    r1 = 1. - psi->map_technical_r;
  }

  if (munich_isnan(r2)){
    logger << LOG_DEBUG << "r2 is NaN in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r2 = " << r2 <<"; k10.z, r = " << k10.x3() << ", " << k10.r() << endline_flush;
    logger << LOG_DEBUG << *p_out << endline_flush;
    r2 = 0.5;
  }
  else if (r2 <= 0.){
    logger << LOG_DEBUG << "r2 < 0 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r2 = " << r2 <<"; k10.z, r = " << k10.x3() << ", " << k10.r() << endline_flush;
    logger << LOG_DEBUG << *p_out << endline_flush;
    r2 = psi->map_technical_r;
  }
  else if (r2 >= 1.){
    logger << LOG_DEBUG << "r2 > 1 in inv_decay!" << endline_flush;
    logger << LOG_DEBUG << "r2 = " << r2 <<"; k10.z, r=" << k10.x3() << ", " << k10.r() << endline_flush;
    logger << LOG_DEBUG << *p_out << endline_flush;
    r2 = 1. - psi->map_technical_r;
  }
  
  logger.finished();
}


string mapping_decay::info() const {
  static Logger logger("mapping_decay::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "decay[" << setw(2) << x_a << "][" << setw(3) << x_d << "] = " << name;
  
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}


