#include "header.hpp"
#include "mapping.propagator.hpp"

mapping_propagator::mapping_propagator(){
  static Logger logger("mapping_propagator::mapping_propagator");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

mapping_propagator::mapping_propagator(int _x_p, int _x_a, int _m, int _out, int _x_smin, int _x_smax, phasespace_set * _psi){
  static Logger logger("mapping_propagator::mapping_propagator (x_p, x_a, m, out, x_smin, s_smax, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_p, _x_a, _m, _out, _x_smin, _x_smax, _psi);

  logger.finished();
}

mapping_propagator::~mapping_propagator(){
  static Logger logger("mapping_propagator::~mapping_propagator");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void mapping_propagator::initialization(int _x_p, int _x_a, string & content, phasespace_set * _psi){
  static Logger logger("mapping_propagator::initialization (x_p, x_a, content, psi)");
  logger.called();

  vector<string> entries(1);
  for (size_t i_i = 0; i_i < content.size(); i_i++){
    if (content[i_i] == ' ' && i_i != content.size() - 1){ entries.push_back(""); }
    else if (content[i_i] != ' '){ entries[entries.size() - 1].push_back(content[i_i]); }
  }
  if (entries.size() != 4){ logger << LOG_FATAL << "propagator[" << _x_p << "][" << _x_a << "]  does not have the right content!" << endline_flush; }
  
  int _m = atoi(entries[0].c_str());
  int _out = atoi(entries[1].c_str());
  int _x_smin = atoi(entries[2].c_str());
  int _x_smax = atoi(entries[3].c_str());

  initialization(_x_p, _x_a, _m, _out, _x_smin, _x_smax, _psi);

  logger.finished();
}


void mapping_propagator::initialization(int _x_p, int _x_a, int _m, int _out, int _x_smin, int _x_smax, phasespace_set * _psi){
  static Logger logger("mapping_propagator::initialization");
  logger.called();

  x_p = _x_p;
  x_a = _x_a;

  m = _m;
  
  out = _out;
  x_smin = _x_smin;
  x_smax = _x_smax;
  
  psi = _psi;

  smin = & psi->smin[x_a][x_smin];
  smax = & psi->smax[x_a][x_smax];

  required_xbp = {};
  required_xbs = {out};
  //  required_xbsqrts = {};
  required_xbsqrts = smin->get_required_xbsqrts();
  for (int i_r = 0; i_r < smax->get_required_xbsqrts().size(); i_r++){ required_xbsqrts.push_back(smax->get_required_xbsqrts()[i_r]); }

  /*
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   psi = " << psi << endline_flush;
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   psi->xbs_all.size() = " << psi->xbs_all.size()  << endline_flush;
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   psi->xbs_all[" << x_a << "].size() = " << psi->xbs_all[x_a].size()  << endline_flush;
  s_out = & psi->xbs_all[x_a][out];
  sqrts_out = & psi->xbsqrts_all[x_a][out];
  smin_out = & psi->v_smin[x_a][x_smin];
  smax_out = & psi->v_smax[x_a][x_smax];
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   s_out, etc. have been set." << endline_flush;
  smin = & psi->smin[x_a][x_smin];
  smax = & psi->smax[x_a][x_smax];

  g_p = & psi->g_p[x_a][x_p];
  logger << LOG_DEBUG << "g_p = " << g_p << "   has been set." << endline_flush;

  mass = psi->M[m];
  mass2 = psi->M2[m];
  Gamma = psi->map_Gamma[m];
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   mass, etc. have been set." << endline_flush;

  // narrow-width propagator - make sure that *s_out = (mass² of resonance) has already been set at this stage !!!
  if (*s_out != 0.){ switch_propagator_type = 0; }
  else if (psi->map_Gamma[m] == 0.){ switch_propagator_type = 2; }
  else { switch_propagator_type = 1; }
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   switch_propagator_type = " << switch_propagator_type << endline_flush;

  if (switch_propagator_type == 0){ *g_p = 1. / (pi * mass * Gamma); }
  
  if (switch_propagator_type == 2){
    nuexp = psi->nuxs;
    if (nuexp == 1.){ switch_nuexp = 1;}
    else { switch_nuexp = 0; }
    //    if (psi->M2[m] == 0.){ mass2 = psi->mass0; }
    //    else { mass2 = psi->M2[m] + psi->mass0; }
    mass2_mapping = psi->M2[m] + psi->mass0; 
  }
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   before psi->random_psp" << endline_flush;
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   random index: " << psi->map_type_to_list_MC_IS[vector<int> {x_a, 0, x_p}] << endline_flush;
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   psi->random_psp.size() = " << psi->random_psp.size() << endline_flush;
  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 0, x_p}]];
  logger << LOG_DEBUG << "x_p = " << setw(3) << x_p << "   after  psi->random_psp" << endline_flush;

  stringstream temp_ss;
  temp_ss << "propagator_" << "";
  if (switch_propagator_type == 0){ temp_ss << "OS"; }
  else if (switch_propagator_type == 1){ temp_ss << "BW"; }
  else if (switch_propagator_type == 2){ temp_ss << "VW"; }
  
//  if (switch_propagator_type == 0){ temp_ss << "onshell_NWA"; }
//  else if (switch_propagator_type == 1){ temp_ss << "Breit_Wigner"; }
//  else if (switch_propagator_type == 2){ temp_ss << "vanishing_width"; }
  
  temp_ss << "_" << m << "_" << out << "_smin" << x_smin << "_smax" << x_smax;
  name = temp_ss.str();
  */
  stringstream temp_ss_p;
  temp_ss_p << "propagator_" << m << "_smin" << x_smin << "_smax" << x_smax;
  MC_IS_name_p = temp_ss_p.str();
  
  logger << LOG_DEBUG << "propagator[" << setw(5) << x_p << "] =  " << name << "  has been initialized." << endline_flush;

  logger.finished();
}


void mapping_propagator::initialization_integration(){
  static Logger logger("mapping_propagator::initialization_integration");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   psi = " << psi << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   psi->xbs_all.size() = " << psi->xbs_all.size()  << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   psi->xbs_all[" << x_a << "].size() = " << psi->xbs_all[x_a].size()  << endline_flush;
  s_out = & psi->xbs_all[x_a][out];
  sqrts_out = & psi->xbsqrts_all[x_a][out];
  smin_out = & psi->v_smin[x_a][x_smin];
  smax_out = & psi->v_smax[x_a][x_smax];
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   s_out, etc. have been set." << endline_flush;
  //  smin = & psi->smin[x_a][x_smin];
  //  smax = & psi->smax[x_a][x_smax];

  g_p = & psi->g_p[x_a][x_p];
  logger << LOG_DEBUG_VERBOSE << "g_p = " << g_p << "   has been set." << endline_flush;

  mass = psi->M[m];
  mass2 = psi->M2[m];
  Gamma = psi->map_Gamma[m];
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   mass, etc. have been set." << endline_flush;

  // narrow-width propagator - make sure that *s_out = (mass² of resonance) has already been set at this stage !!!
  if (*s_out != 0.){ switch_propagator_type = 0; }
  else if (psi->map_Gamma[m] == 0.){ switch_propagator_type = 2; }
  else { switch_propagator_type = 1; }
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   switch_propagator_type = " << switch_propagator_type << endline_flush;

  if (switch_propagator_type == 0){ *g_p = 1. / (pi * mass * Gamma); }
  
  if (switch_propagator_type == 2){
    nuexp = psi->nuxs;
    if (nuexp == 1.){ switch_nuexp = 1;}
    else { switch_nuexp = 0; }
    //    if (psi->M2[m] == 0.){ mass2 = psi->mass0; }
    //    else { mass2 = psi->M2[m] + psi->mass0; }
    mass2_mapping = psi->M2[m] + psi->mass0; 
  }
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   before psi->random_psp" << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   random index: " << psi->map_type_to_list_MC_IS[vector<int> {x_a, 0, x_p}] << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   psi->random_psp.size() = " << psi->random_psp.size() << endline_flush;
  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 0, x_p}]];
  logger << LOG_DEBUG_VERBOSE << "x_p = " << setw(3) << x_p << "   after  psi->random_psp" << endline_flush;

  stringstream temp_ss;
  temp_ss << "propagator_" << "";
  if (switch_propagator_type == 0){ temp_ss << "OS"; }
  else if (switch_propagator_type == 1){ temp_ss << "BW"; }
  else if (switch_propagator_type == 2){ temp_ss << "VW"; }
  //  if (switch_propagator_type == 0){ temp_ss << "onshell_NWA"; }
  //  else if (switch_propagator_type == 1){ temp_ss << "Breit_Wigner"; }
  //  else if (switch_propagator_type == 2){ temp_ss << "vanishing_width"; }
  
  temp_ss << "_" << m << "_" << out << "_smin" << x_smin << "_smax" << x_smax;
  name = temp_ss.str();
  /*
  stringstream temp_ss_p;
  temp_ss_p << "propagator_" << m << "_smin" << x_smin << "_smax" << x_smax;
  MC_IS_name_p = temp_ss_p.str();
  */
  available = false;

  logger.finished();
}


void mapping_propagator::calculate_c(){
  static Logger logger("mapping_propagator::calculate_c");
  logger.called();

  smin->get();
  smax->get();
  // Only for: psi->switch_IS_mode_phasespace == 4
  r1 = random_psp_r1->get_random();
  if (switch_propagator_type == 1){ calculate_Breit_Wigner_c(); }
  else if (switch_propagator_type == 2){ calculate_vanishing_width_c(); }
  //  else { calculate_narrow_width_c(); }
  available = true;
  if (!(*s_out > 0.)){ errorhandling_c(); }

  logger.finished();
}


void mapping_propagator::calculate_g(){
  static Logger logger("mapping_propagator::calculate_g");
  logger.called();

  //  smin->get();
  //  smax->get();
  if (switch_propagator_type == 1){ calculate_Breit_Wigner_g(); }
  else if (switch_propagator_type == 2) { calculate_vanishing_width_g(); }
  //  else { calculate_narrow_width_g(); }
  if (!(*g_p > 0.)){ errorhandling_g(); }

  logger.finished();
}


void mapping_propagator::calculate_inv_r(){
  static Logger logger("mapping_propagator::calculate_inv_r");
  logger.called();

  //  logger << LOG_DEBUG << setw(45) << name << "available = " << available << endline_flush;
  if (!available){
    if (switch_propagator_type == 1){ calculate_Breit_Wigner_inv_r(); }
    else if (switch_propagator_type == 2) { calculate_vanishing_width_inv_r(); }
    //    else { calculate_narrow_width_inv_r(); }   
    if (!(r1 > 0. && r1 < 1.)){ errorhandling_inv_r(); }
  }
  random_psp_r1->get_g_IS(r1, g_IS_r1);
  *g_p *= g_IS_r1;

  //  logger << LOG_DEBUG << setw(45) << name << "psi->MC_phasespace.channel = " << psi->MC_phasespace.channel << endline_flush;
  //  logger << LOG_DEBUG << setw(45) << name << "inv_r[0][" << setw(3) << x_p << "] = " << setprecision(15) << setw(23) << r1 << "   g_IS_temp = " << g_IS_r1 << endline_flush;
 
  logger.finished();
}


void mapping_propagator::errorhandling_c(){
  static Logger logger("mapping_propagator::errorhandling_c");
  logger.called();

  n_fail_c++;
  logger << LOG_DEBUG << "Debug output since point creation failed!" << endline_flush;
  logger << LOG_WARN << "xbsqrts_all[" << x_a << "][" << out << "] is NaN (r1 =" << r1 << ", m = " << m << ", smin = " << *smin_out << ", smax = " << *smax_out << ", Gamma = " << Gamma << endline_flush;

  if (switch_propagator_type == 1){ errorhandling_Breit_Wigner_c(); }
  else if (switch_propagator_type == 2){ errorhandling_vanishing_width_c(); }
  //  else { errorhandling_narrow_width_c(); }

  logger.finished();
}

  
void mapping_propagator::errorhandling_g(){
  static Logger logger("mapping_propagator::errorhandling_g");
  logger.called();

  n_fail_g++;
  if (psi->switch_console_output_phasespace_issue){
    if (*smax_out <= *smin_out){
      logger << LOG_WARN << "smin = " << *smin_out << "   smax = " << *smax_out << ", m = " << m << ", out = " << out << endline_flush;
      for (int i_p = 0; i_p < psi->xbs_all[x_a].size(); i_p++){
	if (psi->xbs_all[x_a][i_p] != 0.){
	  logger << LOG_WARN << "xbs_all[" << x_a << "][" << i_p << "] = " << psi->xbs_all[x_a][i_p] << endline_flush;
	}
      }
    }
  }
  
  if (munich_isnan(*g_p)){
    logger << LOG_WARN << "smin = " << *smin_out << "   s[" << out << "] = " << *s_out << "   smax = " << *smax_out << "   g_p = " << *g_p << "   gtot = nan from s-channel mapping" << endline_flush;
  }
  if (*g_p < 0.){
    logger << LOG_WARN << "g_p < 0.   smin = " << *smin_out << "   s[" << out << "] = " << *s_out << "   smax = " << *smax_out << "   g_p = " << *g_p << "   mass = " << mass << "   Gamma = " << Gamma << endline_flush;
  }
  
  if (switch_propagator_type == 1){ errorhandling_Breit_Wigner_g(); }
  else if (switch_propagator_type == 2){ errorhandling_vanishing_width_g(); }
  //  else { errorhandling_narrow_width_g(); }

  logger.finished();
}

 
void mapping_propagator::errorhandling_inv_r(){
  static Logger logger("mapping_propagator::errorhandling_inv_r");
  logger.called();

  n_fail_inv_r++;
  if (switch_propagator_type == 1){ errorhandling_Breit_Wigner_inv_r(); }
  else if (switch_propagator_type == 2){ errorhandling_vanishing_width_inv_r(); }
  //  else { errorhandling_narrow_width_inv_r(); }

  logger.finished();
}


void mapping_propagator::calculate_Breit_Wigner_limits(){
  static Logger logger("mapping_propagator::calculate_Breit_Wigner_limits");
  logger.called();

  // see Bronstein, Sec. 2.8.7, Eq. (2.7)
  arg1 = (*smin_out - mass2) / (mass * Gamma);
  arg2 = (*smax_out - mass2) / (mass * Gamma);
  if (arg1 * arg2 > -1.){delta_y = atan((arg2 - arg1) / (1 + arg1 * arg2));}
  else if (arg2 > 0.){delta_y = M_PI + atan((arg2 - arg1) / (1. + arg1 * arg2));}
  else {delta_y = -M_PI + atan((arg2 - arg1) / (1. + arg1 * arg2));}
  // see Bronstein, Sec. 2.7.2.3, Eq. (2.87)

  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    logger << LOG_DEBUG_POINT << "*s_out      = " << *s_out << endline_flush;
    logger << LOG_DEBUG_POINT << "*smin_out   = xbsmin_all[" << x_a << "][" << out << "] = " << *smin_out << endline_flush;
    logger << LOG_DEBUG_POINT << "*smax_out   = xbsmax_all[" << x_a << "][" << out << "] = " << *smax_out << endline_flush;
    logger << LOG_DEBUG_POINT << "mass        = " << mass << endline_flush;
    logger << LOG_DEBUG_POINT << "mass2       = " << mass2 << endline_flush;
    logger << LOG_DEBUG_POINT << "Gamma       = " << Gamma << endline_flush;
    logger << LOG_DEBUG_POINT << "arg1        = " << arg1 << endline_flush;
    logger << LOG_DEBUG_POINT << "arg2        = " << arg2 << endline_flush;
    logger << LOG_DEBUG_POINT << "delta_y     = " << delta_y << endline_flush;
  }

  logger.finished();
}


void mapping_propagator::calculate_Breit_Wigner_c(){
  static Logger logger("mapping_propagator::calculate_Breit_Wigner_c");
  logger.called();

  calculate_Breit_Wigner_limits();
  *s_out = mass * Gamma * (arg1 + tan(r1 * delta_y)) / (1 - arg1 * tan(r1 * delta_y)) + mass2;
  psi->psp[x_a][out].register_s();
  // Check why required !!!
  *sqrts_out = sqrt(*s_out);
  psi->psp[x_a][out].register_sqrts();

  logger.finished();
}


void mapping_propagator::calculate_Breit_Wigner_g(){
  static Logger logger("mapping_propagator::calculate_Breit_Wigner_g");
  logger.called();

  if (!available){ calculate_Breit_Wigner_limits(); }
  *g_p = (mass * Gamma) / (delta_y * (pow(*s_out - mass2, 2) + mass2 * pow(Gamma, 2)));

  logger.finished();
}


void mapping_propagator::calculate_Breit_Wigner_inv_r(){
  static Logger logger("mapping_propagator::calculate_Breit_Wigner_inv_r");
  logger.called();

  double atan_arg1 = atan(arg1);
  r1 = (atan((*s_out - mass2) / (mass * Gamma)) - atan_arg1) / (atan(arg2) - atan_arg1);

  logger.finished();
}


void mapping_propagator::errorhandling_Breit_Wigner_c(){
  static Logger logger("mapping_propagator::errorhandling_Breit_Wigner_c");
  logger.called();


  logger.finished();
}

  
void mapping_propagator::errorhandling_Breit_Wigner_g(){
  static Logger logger("mapping_propagator::errorhandling_Breit_Wigner_g");
  logger.called();


  logger.finished();
}

 
void mapping_propagator::errorhandling_Breit_Wigner_inv_r(){
  static Logger logger("mapping_propagator::errorhandling_Breit_Wigner_inv_r");
  logger.called();

  if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "r1 is NaN in inv_propagator_Breit_Wigner!" << endline_flush;
    logger << LOG_DEBUG << "    r = " << setw(23) << setprecision(15) << r1 << ";            smin = " << setw(23) << setprecision(15) << *smin_out << ";      s = " << setw(23) << setprecision(15) << *s_out << ";   smax = " << setw(23) << setprecision(15) << *smax_out << endline_flush;
    logger << LOG_DEBUG << "M[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << mass << ";   Gamma[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << Gamma << ";   arg1 = " << setw(23) << setprecision(15) << arg1 << ";   arg2 = " << setw(23) << setprecision(15) << arg2 << endline_flush;
    if (*s_out <= *smin_out){r1 = psi->map_technical_s;}
    else if (*s_out >= *smax_out){r1 = 1. - psi->map_technical_s;}
    else {
      if (psi->switch_console_output_phasespace_issue){
        logger << LOG_WARN << "   smin = " << *smin_out << "   s = " << *s_out << "   smax = " << *smax_out << endline;
	logger << LOG_WARN << "   arg1 = " << arg1 << "   arg2 = " << arg2 << endline;
	logger << LOG_WARN << "   M[" << m << "] = " << mass << "   M2[" << m << "] = " << mass2 << "   Gamma[" << m << "] = " << Gamma << endline_flush;
        logger << LOG_ERROR << "r1 remains NaN in inv_propagator_Breit_Wigner!" << endline_flush;
      }
      r1 = psi->map_technical_s;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r1 <= 0.) {
    logger << LOG_DEBUG << "r1 <= 0 in inv_propagator_Breit_Wigner!" << endline_flush;
    logger << LOG_DEBUG << "    r1 = " << setw(23) << setprecision(15) << r1 << ";            smin = " << setw(23) << setprecision(15) << *smin_out << ";      s = " << setw(23) << setprecision(15) << *s_out << ";   smax = " << setw(23) << setprecision(15) << *smax_out << endline_flush;
    logger << LOG_DEBUG << "M[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << mass << ";   Gamma[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << Gamma << ";   arg1 = " << setw(23) << setprecision(15) << arg1 << ";   arg2 = " << setw(23) << setprecision(15) << arg2 << endline_flush;
    r1 = psi->map_technical_s;
  }
  else if (r1 >= 1.) {
    logger << LOG_DEBUG << "r1 >= 1 in inv_propagator_Breit_Wigner!" << endline_flush;
    logger << LOG_DEBUG << "    r1 = " << setw(23) << setprecision(15) << r1 << ";            smin = " << setw(23) << setprecision(15) << *smin_out << ";      s = " << setw(23) << setprecision(15) << *s_out << ";   smax = " << setw(23) << setprecision(15) << *smax_out << endline_flush;
    logger << LOG_DEBUG << "M[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << mass << ";   Gamma[" << setw(2) << m << "] = " << setw(23) << setprecision(15) << Gamma << ";   arg1 = " << setw(23) << setprecision(15) << arg1 << ";   arg2 = " << setw(23) << setprecision(15) << arg2 << endline_flush;
    r1 = 1. - psi->map_technical_s;
  }

  logger.finished();
}


void mapping_propagator::calculate_vanishing_width_limits(){
  static Logger logger("mapping_propagator::calculate_vanishing_width_limits");
  logger.called();

  if (switch_nuexp){
    term_smin_mass2 = log(*smin_out - mass2_mapping);
    term_smax_mass2 = log(*smax_out - mass2_mapping);
  }
  else {
    term_smin_mass2 = pow(*smin_out - mass2_mapping, 1. - nuexp);
    term_smax_mass2 = pow(*smax_out - mass2_mapping, 1. - nuexp);
  }

  logger.finished();
}


void mapping_propagator::calculate_vanishing_width_c(){
  static Logger logger("mapping_propagator:calculate_vanishing_width_c");
  logger.called();

  calculate_vanishing_width_limits();
  if (switch_nuexp){ *s_out = exp(r1 * term_smax_mass2 + (1. - r1) * term_smin_mass2) + mass2_mapping; }
  else { *s_out = pow(r1 * term_smax_mass2 + (1. - r1) * term_smin_mass2, (1. / (1. - nuexp))) + mass2_mapping; }
  *sqrts_out = sqrt(*s_out);

  logger.finished();
}


void mapping_propagator::calculate_vanishing_width_g(){
  static Logger logger("mapping_propagator:calculate_vanishing_width_g");
  logger.called();

  if (!available){ calculate_vanishing_width_limits(); }
  if (switch_nuexp){ *g_p = 1. / ((term_smax_mass2 - term_smin_mass2) * (*s_out - mass2_mapping)); }
  else { *g_p = (1. - nuexp) / ((term_smax_mass2 - term_smin_mass2) * pow(*s_out - mass2_mapping, nuexp)); }

  logger.finished();
}


void mapping_propagator::calculate_vanishing_width_inv_r(){
  static Logger logger("mapping_propagator::calculate_vanishing_width_inv_r");
  logger.called();

  if (switch_nuexp){ r1 = (log(*s_out - mass2_mapping) - term_smin_mass2) / (term_smax_mass2 - term_smin_mass2); }
  else { r1 = (pow(*s_out - mass2_mapping, 1. - nuexp) - term_smin_mass2) / (term_smax_mass2 - term_smin_mass2); }

  logger.finished();
}


void mapping_propagator::errorhandling_vanishing_width_c(){
  static Logger logger("mapping_propagator::errorhandling_vanishing_width_c");
  logger.called();

  if (munich_isnan(*s_out)) {
    logger << LOG_DEBUG << "s               = " << *s_out << endline_flush;
    logger << LOG_DEBUG << "smin            = " << *smin_out << endline_flush;
    logger << LOG_DEBUG << "smax            = " << *smax_out << endline_flush;
    logger << LOG_DEBUG << "term_smax_mass2 = " << term_smax_mass2 << endline_flush;
    logger << LOG_DEBUG << "term_smin_mass2 = " << term_smin_mass2 << endline_flush;
    logger << LOG_DEBUG << "mass2           = " << mass2 << endline_flush;
    logger << LOG_DEBUG << "mass2_mapping   = " << mass2_mapping << endline_flush;
    logger << LOG_DEBUG << "nuexp           = " << nuexp << endline_flush;
  }

  logger.finished();
}

  
void mapping_propagator::errorhandling_vanishing_width_g(){
  static Logger logger("mapping_propagator::errorhandling_vanishing_width_g");
  logger.called();
 
  logger << LOG_DEBUG_VERBOSE << "smin = " << *smin_out << "  s = " << *s_out << "   smax = " << *smax_out << "   mass2 = " << mass2 << "   smin - mass0 = " << *smin_out - mass2_mapping << " ^(1-nuexp = " << nuexp << ") = " << pow(*smin_out - mass2_mapping, 1. - nuexp) << endline_flush;

  logger << LOG_DEBUG << "g_p < 0.   smin = " << *smin_out << "   s = " << *s_out << "   smax = " << *smax_out << "   g_p = " << *g_p << "   mass2 = " << mass2_mapping << endline_flush;

  if (*smax_out <= *smin_out) {
    if (psi->switch_console_output_phasespace_issue){
      logger << LOG_WARN << "smin = " << *smin_out << "  s = " << *s_out << "   smax = " << *smax_out << "   mass2_mapping = " << mass2_mapping << "   smin - mass2_mapping = " << *smin_out - mass2_mapping << " ^(1-nuexp = " << nuexp  << ") = " << pow(*smin_out - mass2_mapping, 1. - nuexp) << endline_flush;
      logger << LOG_ERROR << "g_p is set to infinity." << endline_flush;
    }
    *g_p = std::numeric_limits<double>::infinity();
  }

  logger.finished();
}

 
void mapping_propagator::errorhandling_vanishing_width_inv_r(){
  static Logger logger("mapping_propagator::errorhandling_vanishing_width_inv_r");
  logger.called();

  if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "s               = " << *s_out << endline_flush;
    logger << LOG_DEBUG << "smin            = " << *smin_out << endline_flush;
    logger << LOG_DEBUG << "smax            = " << *smax_out << endline_flush;
    logger << LOG_DEBUG << "term_smin_mass2 = " << term_smin_mass2 << endline_flush;
    logger << LOG_DEBUG << "term_smax_mass2 = " << term_smax_mass2 << endline_flush;
    logger << LOG_DEBUG << "mass2_mapping   = " << mass2_mapping << endline_flush;
    logger << LOG_DEBUG << "nuexp           = " << nuexp << endline_flush;
    logger << LOG_DEBUG << "r1 is NaN in inv_propagator_vanishing_width!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 << "; smin, s, smax=" << *smin_out << ", " << *s_out << ", " << *smax_out << endline_flush;
    if (*s_out <= *smin_out){r1 = psi->map_technical_r;}
    else if (*s_out >= *smax_out){r1 = 1. - psi->map_technical_r;}
    else {
      if (psi->switch_console_output_phasespace_issue){
        logger << LOG_ERROR << "r1 remains NaN in inv_propagator_vanishing_width!" << endline_flush;
	logger << LOG_WARN << "r1 = " << r1 << "; smin = " << *smin_out << ", s = " << *s_out << ", smax = " << *smax_out << "   mass2_mapping = " << mass2_mapping << "   nuexp = " << nuexp << endline_flush;
      }
      r1 = psi->map_technical_r;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r1 <= 0.){
    logger << LOG_DEBUG << "r1 <= 0 in inv_propagator_vanishing_width!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 << "; smin = " << *smin_out << ", s = " << *s_out << ", smax = " << *smax_out << "   mass2_mapping = " << mass2_mapping << "   nuexp = " << nuexp << endline_flush;
    r1 = psi->map_technical_r;
  }
  if (r1 >= 1.){
    logger << LOG_DEBUG << "r1 >= 1 in inv_propagator_vanishing_width!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 << "; smin = " << *smin_out << ", s = " << *s_out << ", smax = " << *smax_out << "   mass2_mapping = " << mass2_mapping << "   nuexp = " << nuexp << endline_flush;
    r1 = 1. - psi->map_technical_r;
  }

  logger.finished();
}

/*
void mapping_propagator::c_propagator_narrow_width(){}

void mapping_propagator::g_propagator_narrow_width(int m){
  static Logger logger("mapping_propagator::g_propagator_narrow_width");
  logger.called();

  *g_p = 1. / (pi * mass * Gamma);

  logger.finished();
}

void mapping_propagator::inv_propagator_narrow_width(){}
*/


string mapping_propagator::info() const {
  static Logger logger("mapping_propagator::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "propagator[" << setw(2) << x_a << "][" << setw(3) << x_p << "] = " << name;
  
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}


