#include "header.hpp"
#include "mapping.smax.hpp"

mapping_smax::mapping_smax(){
  static Logger logger("mapping_smax::mapping_smax");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

mapping_smax::mapping_smax(int _x_smax, int _x_a, int _in, vector<int> _out, phasespace_set * _psi){
  static Logger logger("mapping_smax::mapping_smax (x_smax, x_a, out, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_smax, _x_a, _in, _out, _psi);

  logger.finished();
}

mapping_smax::~mapping_smax(){
  static Logger logger("mapping_smax::~mapping_smax");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void mapping_smax::initialization(int _x_smax, int _x_a, string & content, phasespace_set * _psi){
  static Logger logger("mapping_smax::initialization (x_smax, x_a, content, psi)");
  logger.called();

  vector<string> entries(1);
  for (size_t i_i = 0; i_i < content.size(); i_i++){
    if (content[i_i] == ' ' && i_i != content.size() - 1){ entries.push_back(""); }
    else if (content[i_i] != ' '){ entries[entries.size() - 1].push_back(content[i_i]); }
  }
  if (entries.size() < 2){ logger << LOG_FATAL << "smax[" << _x_smax << "][" << _x_a << "]  does not have the right content!" << endline_flush; }
  
  int _p_smax_process = atoi(entries[0].c_str());
  vector<int> _p_smax_xb(entries.size() - 1);
  for (size_t i_i = 1; i_i < entries.size(); i_i++){ _p_smax_xb[i_i - 1] = atoi(entries[i_i].c_str()); }
  
  initialization(_x_smax, _x_a, _p_smax_process, _p_smax_xb, _psi);

  logger.finished();
}


void mapping_smax::initialization(int _x_smax, int _x_a, int _p_smax_process, vector<int> _p_smax_xb, phasespace_set * _psi){
  static Logger logger("mapping_smax::initialization");
  logger.called();

  x_smax = _x_smax;
  x_a = _x_a;

  p_smax_xb = _p_smax_xb;
  p_smax_process = _p_smax_process;
  // p_smax_process could be determined at runtime to deal with NWA calculations !!!!

  psi = _psi;

  if (p_smax_xb.size() == 1){ switch_trivial = true; }
  else { switch_trivial = false; }
  
  xb_combination = accumulate(p_smax_xb.begin(), p_smax_xb.end(), 0);

  required_xbsqrts = {p_smax_xb.begin() + 1, p_smax_xb.end()}; 

  available = false;

  logger << LOG_DEBUG << "xb_combination = " << xb_combination << endline_flush;
  logger << LOG_DEBUG << "p_smax_xb[0] = " << p_smax_xb[0] << endline_flush;
  for (int i_p = 1; i_p < p_smax_xb.size(); i_p++){ logger << LOG_DEBUG << "p_smax_xb[" << i_p << "] = " << p_smax_xb[i_p] << endline_flush; }

  logger.finished();
}


void mapping_smax::initialization_integration(){
  static Logger logger("mapping_smax::initialization_integration");
  logger.called();

  smax = & psi->v_smax[x_a][x_smax];

  logger.finished();
}

//v_smax[x_a][  1] = pow(xbsqrts_all[x_a][0] - sqrtsmin_opt[x_a][76], 2);

void mapping_smax::calculate(){
  static Logger logger("mapping_smax::calculate");
  logger.called();

  if (switch_trivial){ *smax = pow(psi-> xbsqrts_all[x_a][p_smax_process] - psi->sqrtsmin_opt[x_a][xb_combination], 2); }
  else {
    double sqrtsmax_sum = psi->sqrtsmin_opt[x_a][p_smax_xb[0]];
    for (int i_p = 1; i_p < p_smax_xb.size(); i_p++){ sqrtsmax_sum += psi->xbsqrts_all[x_a][p_smax_xb[i_p]]; }
    *smax = pow(psi-> xbsqrts_all[x_a][p_smax_process] - GSL_MAX_DBL(sqrtsmax_sum, psi->sqrtsmin_opt[x_a][xb_combination]), 2);
  }
  available = true;
 
  logger.finished();
}


string mapping_smax::info() const {
  static Logger logger("mapping_smax::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "v_smax[x_a][" << setw(3) << x_smax << "] = pow(sqrtsmin_opt[x_a][" << p_smax_process << "] - ";
  //  if (level_complexity == 0){
  if (switch_trivial){ temp_ss << "sqrtsmin_opt[x_a][" << xb_combination << "], 2);"; }
  else if (p_smax_xb[0] == 0 && p_smax_xb.size() == 2){ temp_ss << "xbsqrts_all[x_a][" << p_smax_xb[1] << "], 2);"; }
  else {
    temp_ss << "GSL_MAX_DBL(";
    if (p_smax_xb[0] != 0){ temp_ss << "sqrtsmin_opt[x_a][" << p_smax_xb[0] << "]"; }
    for (int i_p = 1; i_p < p_smax_xb.size(); i_p++){
      if (!(p_smax_xb[0] == 0 && i_p == 1)){temp_ss << " + "; }
      temp_ss << "xbsqrts_all[x_a][" << p_smax_xb[i_p] << "]";
    }
    temp_ss << ", sqrtsmin_opt[x_a][" << xb_combination << "]), 2);";
  }
  temp_ss << "  //  xb_smax = {" << setw(3) << p_smax_xb[0];
  for (int i_p = 1; i_p < p_smax_xb.size(); i_p++){ temp_ss << setw(3) << p_smax_xb[i_p]; }
  temp_ss << "}";
      
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}


