#include "header.hpp"
#include "mapping.smin.hpp"

mapping_smin::mapping_smin(){
  static Logger logger("mapping_smin::mapping_smin");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

mapping_smin::mapping_smin(int _x_smin, int _x_a, vector<int> _p_smin_xb, phasespace_set * _psi){
  static Logger logger("mapping_smin::mapping_smin (x_smin, x_a, p_smin_xb, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_smin, _x_a, _p_smin_xb, _psi);

  logger.finished();
}

mapping_smin::~mapping_smin(){
  static Logger logger("mapping_smin::~mapping_smin");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void mapping_smin::initialization(int _x_smin, int _x_a, string & content, phasespace_set * _psi){
  static Logger logger("mapping_smin::initialization (x_smin, x_a, content, psi)");
  logger.called();

  vector<string> entries(1);
  for (size_t i_i = 0; i_i < content.size(); i_i++){
    if (content[i_i] == ' ' && i_i != content.size() - 1){ entries.push_back(""); }
    else if (content[i_i] != ' '){ entries[entries.size() - 1].push_back(content[i_i]); }
  }
  if (entries.size() < 1){ logger << LOG_FATAL << "smin[" << _x_smin << "][" << _x_a << "]  does not have the right content!" << endline_flush; }

  vector<int> _p_smin_xb(entries.size());
  for (size_t i_i = 0; i_i < entries.size(); i_i++){ _p_smin_xb[i_i] = atoi(entries[i_i].c_str()); }
  
  initialization(_x_smin, _x_a, _p_smin_xb, _psi);

  logger.finished();
}


void mapping_smin::initialization(int _x_smin, int _x_a, vector<int> _p_smin_xb, phasespace_set * _psi){
  static Logger logger("mapping_smin::initialization");
  logger.called();

  x_smin = _x_smin;
  x_a = _x_a;

  p_smin_xb = _p_smin_xb;

  psi = _psi;
  
  if (p_smin_xb.size() == 1){ switch_trivial = true; }
  else { switch_trivial = false; }

  xb_combination = accumulate(p_smin_xb.begin(), p_smin_xb.end(), 0);

  required_xbsqrts = {p_smin_xb.begin() + 1, p_smin_xb.end()}; 
    
  logger.finished();
}


void mapping_smin::initialization_integration(){
  static Logger logger("mapping_smin::initialization_integration");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "before smin = & psi->v_smin[x_a][x_smin];" << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_smin = " << x_smin << endline_flush;
  
  logger << LOG_DEBUG_VERBOSE << "psi = " << psi << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "psi->v_smin.size() = " << psi->v_smin.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "psi->v_smin[x_a].size() = " << psi->v_smin[x_a].size() << endline_flush;

  smin = & psi->v_smin[x_a][x_smin];
  logger << LOG_DEBUG_VERBOSE << "after smin = & psi->v_smin[x_a][x_smin];" << endline_flush;

  available = false;

  logger << LOG_DEBUG << "xb_combination = " << xb_combination << endline_flush;
  logger << LOG_DEBUG << "p_smin_xb[0] = " << p_smin_xb[0] << endline_flush;
  for (int i_p = 1; i_p < p_smin_xb.size(); i_p++){ logger << LOG_DEBUG << "p_smin_xb[" << i_p << "] = " << p_smin_xb[i_p] << endline_flush; }

  logger.finished();
}


void mapping_smin::calculate(){
  static Logger logger("mapping_smin::calculate");
  logger.called();

  if (switch_trivial){ *smin = psi->smin_opt[x_a][xb_combination]; }
  else {
    double sqrtsmin_sum = psi->sqrtsmin_opt[x_a][p_smin_xb[0]];
    for (int i_p = 1; i_p < p_smin_xb.size(); i_p++){ sqrtsmin_sum += psi->xbsqrts_all[x_a][p_smin_xb[i_p]]; }
    *smin = GSL_MAX_DBL(pow(sqrtsmin_sum, 2), psi->smin_opt[x_a][xb_combination]);
  }
  available = true;
 
  logger.finished();
}


string mapping_smin::info() const {
  static Logger logger("mapping_smin::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "v_smin[x_a][" << setw(3) << x_smin << "] = ";
  if (switch_trivial){ temp_ss << "smin_opt[x_a][" << xb_combination << "];";}
  else {
    temp_ss << "GSL_MAX_DBL(pow(";
    if (p_smin_xb[0] != 0){
      temp_ss << "sqrtsmin_opt[x_a][" << p_smin_xb[0] << "]";
    }
    for (int i_p = 1; i_p < p_smin_xb.size(); i_p++){
      if (!(p_smin_xb[0] == 0 && i_p == 1)){temp_ss << " + "; }
      temp_ss << " xbsqrts_all[x_a][" << p_smin_xb[i_p] << "]";
    }
    temp_ss << ", 2), smin_opt[x_a][" << xb_combination << "]);";
  }
  temp_ss << "  //  xb_smin = {" << setw(3) << p_smin_xb[0];
  for (int i_p = 1; i_p < p_smin_xb.size(); i_p++){ temp_ss << setw(3) << p_smin_xb[i_p]; }
  temp_ss << "}";

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}


