#include "header.hpp"
#include "mapping.tchannel.hpp"

mapping_tchannel::mapping_tchannel(){
  static Logger logger("mapping_tchannel::mapping_tchannel");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

mapping_tchannel::mapping_tchannel(int _x_t, int _x_a, int _m, int _out, int _in_1, int _in_2, int _out_1, int _out_2, phasespace_set * _psi){
  static Logger logger("mapping_tchannel::mapping_tchannel (x_a, m, out, in_1, in_2, out_1, out_2, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_t, _x_a, _m, _out, _in_1, _in_2, _out_1, _out_2, _psi);

  logger.finished();
}

mapping_tchannel::~mapping_tchannel(){
  static Logger logger("mapping_tchannel::~mapping_tchannel");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void mapping_tchannel::initialization(int _x_t, int _x_a, string & content, phasespace_set * _psi){
  static Logger logger("mapping_tchannel::initialization(x_t, x_a, content, psi)");
  logger.called();

  vector<string> entries(1);
  for (size_t i_i = 0; i_i < content.size(); i_i++){
    if (content[i_i] == ' ' && i_i != content.size() - 1){ entries.push_back(""); }
    else if (content[i_i] != ' '){ entries[entries.size() - 1].push_back(content[i_i]); }
  }
  if (entries.size() != 6){ logger << LOG_FATAL << "tchannel[" << _x_t << "][" << _x_a << "]  does not have the right content!" << endline_flush; }
  
  int _m = atoi(entries[0].c_str());
  int _out = atoi(entries[1].c_str());
  int _in_1 = atoi(entries[2].c_str());
  int _in_2 = atoi(entries[3].c_str());
  int _out_1 = atoi(entries[4].c_str());
  int _out_2 = atoi(entries[5].c_str());

  initialization(_x_t, _x_a, _m, _out, _in_1, _in_2, _out_1, _out_2, _psi);

  logger.finished();
}


void mapping_tchannel::initialization(int _x_t, int _x_a, int _m, int _out, int _in_1, int _in_2, int _out_1, int _out_2, phasespace_set * _psi){
  static Logger logger("mapping_tchannel::initialization");
  logger.called();

  x_t = _x_t;
  x_a = _x_a;

  m = _m;

  out = _out;
  in_1 = _in_1;
  in_2 = _in_2;
  out_1 = _out_1;
  out_2 = _out_2;
  in_1_out_1 = in_1 + out_1;
  
  psi = _psi;

  required_xbp = {out, in_1, out_1};
  required_xbs = {in_1 + out_1, out, in_1, in_2, out_1, out_2};
  required_xbsqrts = {};
  
  /*
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi = " << psi << endline_flush;
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi->xbs_all.size() = " << psi->xbs_all.size()  << endline_flush;
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi->xbs_all[" << x_a << "].size() = " << psi->xbs_all[x_a].size() << endline_flush;
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi->xbp_all.size() = " << psi->xbp_all.size()  << endline_flush;
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi->xbp_all[" << x_a << "].size() = " << psi->xbp_all[x_a].size() << endline_flush;
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi->xbsqrts_all.size() = " << psi->xbsqrts_all.size()  << endline_flush;
  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   psi->xbsqrts_all[" << x_a << "].size() = " << psi->xbsqrts_all[x_a].size() << endline_flush;
  
  p_out = & psi->xbp_all[x_a][out];
  p_in_1 = & psi->xbp_all[x_a][in_1];
  p_in_2 = & psi->xbp_all[x_a][in_2];
  p_out_1 = & psi->xbp_all[x_a][out_1];
  p_out_2 = & psi->xbp_all[x_a][out_2];
  p_in_1_out_1 = & psi->xbp_all[x_a][in_1 + out_1];
  
  s_out = & psi->xbs_all[x_a][out];
  s_out_1 = & psi->xbs_all[x_a][out_1];
  s_out_2 = & psi->xbs_all[x_a][out_2];
  s_in_1 = & psi->xbs_all[x_a][in_1] ;
  s_in_2 = & psi->xbs_all[x_a][in_2];
  s_in_1_out_1 = & psi->xbs_all[x_a][in_1 + out_1];

  sqrts_out = & psi->xbsqrts_all[x_a][out];

  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   s_out, etc. have been set." << endline_flush;

  g_t = & psi->g_t[x_a][x_t];
  logger << LOG_DEBUG << "g_t = " << g_t << "   has been set." << endline_flush;
  
  nuexp = psi->nuxt;
  if (psi->M2[m] != 0.){ mass2_mapping = -psi->M2[m]; }
  else { mass2_mapping = psi->mass0; }

  logger << LOG_DEBUG << "x_t = " << setw(3) << x_t << "   nuexp, mass2_mapping, etc. have been set." << endline_flush;

  switch_opt_one_particle = 0;
  mass2_out_1 = 0.;
  // possibly directly via contribution csi->... ???
  if (psi->MC_optswitch){
    logger << LOG_DEBUG << "psi->xb_content_n.size() = " << psi->xb_content_n.size() << endline_flush;
    if (psi->xb_content_n[out_1] == 1){
      smin_out_1 = psi->xbs_all[x_a][psi->xb_content_xb[out_1][0]];
      pTcut_out_1 = psi->mapping_cut_pT[abs(psi->csi->type_parton[0][psi->xb_content[out_1][0]])];
      if (*s_out_1 == 0.){
	switch_opt_one_particle = 1;
      }
      else {
	switch_opt_one_particle = 2;
	mass2_out_1 = psi->csi->mass2_parton[0][psi->xb_content[out_1][0]];
      }
    }
  }
  logger << LOG_DEBUG << "switch_opt_one_particle = " << switch_opt_one_particle << endline_flush;

  // switch_pTthreshold only makes sense for first s-channel decay !!!
  if (psi->csi->class_contribution_QT_powercorrection && psi->switch_decay_pTthreshold && out == psi->xb_max - 4 ){ switch_pTthreshold = 1; }
  else { switch_pTthreshold = 0; }

  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 2, x_t}]];
  random_psp_r2 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 3, x_t}]];

  stringstream temp_ss;
  temp_ss << "tchannel" << "";
  if (switch_opt_one_particle == 0){ temp_ss << ""; }
  else if (switch_opt_one_particle == 1){ temp_ss << "_opt"; }
  temp_ss << "_" << m << "_" << out << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  name = temp_ss.str();
  */
    // to be set later !!!


  // Use something like  switch_new_psp_with_old_psp_grid !!!
  bool switch_temp_name_opt = false;
  if (in_1 < 3 || in_2 < 3){ switch_temp_name_opt = true; }
  // no impact on optimization, just to match the oldPS name (in order to use oldPS grids)
  
  //  if (psi->MC_optswitch){
  {
    logger << LOG_DEBUG_VERBOSE << "psi->xb_content_n.size() = " << psi->xb_content_n.size() << endline_flush;
    // opt should actually be used only if incoming parton is of pure IS type:
    //    if (in_1 < 3 || in_2 < 3){ switch_temp_name_opt = true; }
    //    if (psi->xb_content_n[out_1] == 1 && in_1 < 3){ switch_temp_name_opt = true; }
    // if correct, "in_1 < 3" is not required below !!!
    //    if (psi->xb_content_n[out_1] == 1){ switch_temp_name_opt = 1; }
    // distrinction between 1 and 2 comes later !!!
  }

  // Check if  "  && in_1 < 3 "  is correct !!!
   stringstream temp_ss_t_t;
  temp_ss_t_t << "t_of_tchannel";
  if (psi->switch_new_psp_with_old_psp_grid){ if (switch_temp_name_opt){ temp_ss_t_t << "_opt"; } }
  temp_ss_t_t << "_" << m <<  "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  //  temp_ss_t_t << "_" << m << "_" << out << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  MC_IS_name_t_t = temp_ss_t_t.str();
  logger << LOG_DEBUG << "tchannel[" << setw(5) << x_t << "] : MC_IS_name_t_t   = " << MC_IS_name_t_t << endline_flush;

  stringstream temp_ss_t_phi;
  temp_ss_t_phi << "phi_of_tchannel";
  if (psi->switch_new_psp_with_old_psp_grid){ if (switch_temp_name_opt){ temp_ss_t_phi << "_opt"; } }
  temp_ss_t_phi << "_" << m << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  //  temp_ss_t_phi << "_" << m << "_" << out << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  MC_IS_name_t_phi = temp_ss_t_phi.str();
  logger << LOG_DEBUG << "tchannel[" << setw(5) << x_t << "] : MC_IS_name_t_phi = " << MC_IS_name_t_phi << endline_flush;

  logger << LOG_DEBUG << "tchannel[" << setw(5) << x_t << "] =  " << name << "  has been initialized." << endline_flush;

  logger.finished();
}


void mapping_tchannel::initialization_integration(){
  static Logger logger("mapping_tchannel::initialization_integration");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi = " << psi << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi->xbs_all.size() = " << psi->xbs_all.size()  << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi->xbs_all[" << x_a << "].size() = " << psi->xbs_all[x_a].size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi->xbp_all.size() = " << psi->xbp_all.size()  << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi->xbp_all[" << x_a << "].size() = " << psi->xbp_all[x_a].size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi->xbsqrts_all.size() = " << psi->xbsqrts_all.size()  << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   psi->xbsqrts_all[" << x_a << "].size() = " << psi->xbsqrts_all[x_a].size() << endline_flush;
  
  p_out = & psi->xbp_all[x_a][out];
  p_in_1 = & psi->xbp_all[x_a][in_1];
  p_in_2 = & psi->xbp_all[x_a][in_2];
  p_out_1 = & psi->xbp_all[x_a][out_1];
  p_out_2 = & psi->xbp_all[x_a][out_2];
  p_in_1_out_1 = & psi->xbp_all[x_a][in_1_out_1];
  
  s_out = & psi->xbs_all[x_a][out];
  s_out_1 = & psi->xbs_all[x_a][out_1];
  s_out_2 = & psi->xbs_all[x_a][out_2];
  s_in_1 = & psi->xbs_all[x_a][in_1] ;
  s_in_2 = & psi->xbs_all[x_a][in_2];
  s_in_1_out_1 = & psi->xbs_all[x_a][in_1_out_1];

  sqrts_out = & psi->xbsqrts_all[x_a][out];

  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   s_out, etc. have been set." << endline_flush;

  g_t = & psi->g_t[x_a][x_t];
  logger << LOG_DEBUG_VERBOSE << "g_t = " << g_t << "   has been set." << endline_flush;
  
  nuexp = psi->nuxt;
  if (psi->M2[m] != 0.){ mass2_mapping = -psi->M2[m]; }
  else { mass2_mapping = psi->mass0; }

  logger << LOG_DEBUG_VERBOSE << "x_t = " << setw(3) << x_t << "   nuexp, mass2_mapping, etc. have been set." << endline_flush;

  switch_opt_one_particle = 0;
  mass2_out_1 = 0.;
  // Check if extra condition is required, like " && in_1 < 3 "
  
  // possibly directly via contribution csi->... ???
  if (psi->MC_optswitch){
    logger << LOG_DEBUG_VERBOSE << "psi->xb_content_n.size() = " << psi->xb_content_n.size() << endline_flush;
    if (psi->xb_content_n[out_1] == 1){
      //      smin_out_out_1 = psi->xbs_all[x_a][psi->xb_content_xb[out_1][0]];
      smin_out_out_1 = 0.;
      for (int i_p = 0; i_p < psi->xb_content_xb[out - out_1].size(); i_p++){ smin_out_out_1 += psi->xbsqrts_all[x_a][psi->xb_content_xb[out - out_1][i_p]]; }
      smin_out_out_1 = pow(smin_out_out_1, 2);

      pTcut_out_1 = psi->mapping_cut_pT[abs(psi->csi->type_parton[0][psi->xb_content[out_1][0]])];
      if (*s_out_1 == 0.){
	switch_opt_one_particle = 1;
      }
      else {
	switch_opt_one_particle = 2;
	mass2_out_1 = psi->csi->mass2_parton[0][psi->xb_content[out_1][0]];
      }
    }
  }
  logger << LOG_DEBUG_VERBOSE << "switch_opt_one_particle = " << switch_opt_one_particle << endline_flush;

  // switch_pTthreshold only makes sense for first s-channel decay !!!
  /*
  if (psi->csi->class_contribution_QT_powercorrection && psi->switch_decay_pTthreshold && out == psi->xb_max - 4 ){ switch_pTthreshold = 1; }
  else { switch_pTthreshold = 0; }
  */
  
  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 2, x_t}]];
  random_psp_r2 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 3, x_t}]];

  stringstream temp_ss;
  temp_ss << "tchannel" << "";
  if (switch_opt_one_particle == 0){ temp_ss << ""; }
  else if (switch_opt_one_particle == 1){ temp_ss << "_opt"; }
  temp_ss << "_" << m << "_" << out << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  name = temp_ss.str();
  /*
  stringstream temp_ss_t_t;
  temp_ss_t_t << "t_of_tchannel";
  if (switch_opt_one_particle == 1){ temp_ss_t_t << "_opt"; }
  temp_ss_t_t << "_" << m << "_" << out << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  MC_IS_name_t_t = temp_ss_t_t.str();
  
  stringstream temp_ss_t_phi;
  temp_ss_t_phi << "phi_of_tchannel";
  if (switch_opt_one_particle == 1){ temp_ss_t_t << "_opt"; }
  temp_ss_t_phi << "_" << m << "_" << out << "__" << in_1 << "_" << in_2 << "__" << out_1 << "_" << out_2;
  MC_IS_name_t_phi = temp_ss_t_phi.str();
  */
  available = false;
  
  logger << LOG_DEBUG_VERBOSE << name << "   psi->MC_optswitch = " << psi->MC_optswitch << endline_flush;

  logger.finished();
}


void mapping_tchannel::calculate_propagator_limits(){
  static Logger logger("mapping_tchannel::calculate_propagator_limits");
  logger.called();

  term1 = (*s_out + *s_out_1 - *s_out_2) * (*s_out + *s_in_1 - *s_in_2);
  term2in = sqrt(lambda(*s_out, *s_in_1, *s_in_2));
  term2out = sqrt(lambda(*s_out, *s_out_1, *s_out_2));
  term2 = term2in * term2out;
  tmin = *s_out_1 + *s_in_1 - (term1 + term2) / (2 * *s_out);
  tmax = *s_out_1 + *s_in_1 - (term1 - term2) / (2 * *s_out);
  //  if (tmin > 0.){ logger << LOG_DEBUG << "tmin = " << tmin << endline_flush; }
  //  if (tmax > 0.){ logger << LOG_DEBUG << "tmax = " << tmax << endline_flush; tmax = 0.; }
  if (tmax > 0.){ tmax = 0.; }
  
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    logger << LOG_DEBUG_POINT << "*s_out   = xbp_all[" << x_a << "][" << out << "].m2() = " << psi->xbp_all[x_a][out].m2() << "   psi->xbp_all][" << x_a << "][" << out << "] = " << psi->xbp_all[x_a][out] << endline_flush;
    logger << LOG_DEBUG_POINT << "*s_out   = xbs_all[" << x_a << "][" << out << "] = " << *s_out << endline_flush;
    logger << LOG_DEBUG_POINT << "*s_out_1 = xbs_all[" << x_a << "][" << out_1 << "] = " << *s_out_1 << endline_flush;
    logger << LOG_DEBUG_POINT << "*s_out_2 = xbs_all[" << x_a << "][" << out_2 << "] = " << *s_out_2 << endline_flush;
    logger << LOG_DEBUG_POINT << "*s_in_1  = xbs_all[" << x_a << "][" << in_1 << "] = " << *s_in_1 << endline_flush;
    logger << LOG_DEBUG_POINT << "*s_in_2  = xbs_all[" << x_a << "][" << in_2 << "] = " << *s_in_2 << endline_flush;
    logger << LOG_DEBUG_POINT << "t        = xbs_all[" << x_a << "][" << in_1_out_1 << "] = " << *s_in_1_out_1 << endline_flush;
    logger << LOG_DEBUG_POINT << "term1 = " << term1 << endline_flush;
    logger << LOG_DEBUG_POINT << "lambda(*s_out, *s_in_1, *s_in_2) = " << lambda(*s_out, *s_in_1, *s_in_2) << endline_flush;
    logger << LOG_DEBUG_POINT << "term2in = " << term2in << endline_flush;
    logger << LOG_DEBUG_POINT << "lambda(*s_out, *s_out_1, *s_out_2) = " << lambda(*s_out, *s_out_1, *s_out_2) << endline_flush;
    logger << LOG_DEBUG_POINT << "term2out = " << term2out << endline_flush;
    logger << LOG_DEBUG_POINT << "tmin = " << tmin << endline_flush;
    logger << LOG_DEBUG_POINT << "tmax = " << tmax << endline_flush;
  }
    
  // This should be the only difference between previous tchannel and tchannel_opt ???
  if (switch_opt_one_particle){
    tmax_cut = 0.;
    if (switch_opt_one_particle == 1){
      tmax_cut = .5 * (*s_out - smin_out_out_1) * (sqrt(1. - (4. * *s_out * pTcut_out_1) / pow((*s_out - smin_out_out_1), 2.)) - 1.);
    }
    else if (switch_opt_one_particle == 2){
      tmax_cut = mass2_out_1 + .5 * (*s_out + mass2_out_1 - smin_out_out_1) * (sqrt(1. - (4. * *s_out * (mass2_out_1 + pTcut_out_1) / pow((*s_out + mass2_out_1 - smin_out_out_1), 2.))) - 1.);
    }
    if ((tmax_cut < tmax) && (tmin < tmax_cut)){ tmax = tmax_cut; }
  }
  
  logger.finished();
}


void mapping_tchannel::calculate_c(){
  static Logger logger("mapping_tchannel::calculate_c");
  logger.called();

  // Only for: psi->switch_IS_mode_phasespace == 4
  r2 = random_psp_r2->get_random();
  r1 = random_psp_r1->get_random();

  // presumably not required at all !!!
  //  if (switch_pTthreshold){ psi->c_decay(x_a, out, out_1, out_2, r1, r2); return; }
  
  calculate_propagator_limits();
  calculate_propagator_c();
  term3 = 2 * *sqrts_out;
  p10 = fourvector(((*s_out + *s_in_1 - *s_in_2) / term3), 0., 0., (term2in / term3));
  p20 = fourvector(((*s_out - *s_in_1 + *s_in_2) / term3), 0., 0., -(term2in / term3));
  k10 = fourvector(((*s_out + *s_out_1 - *s_out_2) / term3), 0., 0., (term2out / term3));
  p1b = p_in_1->boost(*p_out, *s_out);
  phi = psi->c_phi(r2);
  costheta = ((2 * *s_out * (*s_in_1_out_1 - *s_out_1 - *s_in_1)) + term1) / term2;
  k10 = k10.rotation_beam(phi, costheta);
  k20 = p10 + p20 - k10;
  *p_out_1 = (k10.rotation(p1b)).boost(p_out->Pinv(), *s_out);
  psi->psp[x_a][out_1].register_p();
  *p_out_2 = (k20.rotation(p1b)).boost(p_out->Pinv(), *s_out);
  psi->psp[x_a][out_2].register_p();
  *p_in_1_out_1 = *p_in_1 - *p_out_1;
  psi->psp[x_a][in_1_out_1].register_p();
  available = true;
  if (munich_isnan(*p_out_1)){ errorhandling_c(); }

  logger.finished();
}


void mapping_tchannel::calculate_g(){
  static Logger logger("mapping_tchannel::calculate_g");
  logger.called();

  //  if (switch_pTthreshold){ *g_t = psi->g_decay(x_a, out, out_1, out_2); return; }

  if (!available){ calculate_propagator_limits(); }
  calculate_propagator_g();
  *g_t = g_t_propagator * (2 * term2in) / pi;
  if (!(*g_t > 0.)){ errorhandling_g(); }

  logger.finished();
}


// maybe directly access memebers r1 and r2 ???
void mapping_tchannel::calculate_inv_r(){
  static Logger logger("mapping_tchannel::calculate_inv_r");
  logger.called();

  //  if (switch_pTthreshold){ psi->inv_decay(x_a, out, out_1, out_2, r1, r2); return; }
  
  // tmin, tmax, etc. either available from calculate_g, or already from calculate_c !!!
  //  logger << LOG_DEBUG << setw(47) << name << "available = " << available << endline_flush;
  if (!available){
    calculate_propagator_inv_r();
    //    r1 = inv_t_propagator(-*s_in_1_out_1, m, -tmin, -tmax);
    p1b = p_in_1->boost(*p_out, *s_out);
    k10 = p_out_1->boost(*p_out, *s_out).rotation_inverse(p1b);
    phi = k10.phi();
    r2 = phi / (2 * M_PI);
    // Would this work as well ???
    //  r2 = inv_phi(k10.phi());
    if (!(r1 > 0. && r1 < 1.)){ errorhandling_inv_r(); }
  }
  random_psp_r1->get_g_IS(r1, g_IS_r1);
  random_psp_r2->get_g_IS(r2, g_IS_r2);
  *g_t *= g_IS_r1 * g_IS_r2;

  //  logger << LOG_DEBUG << setw(47) << name << "inv_r[2][" << setw(3) << x_t << "] = " << setprecision(15) << setw(23) << r1 << "   g_IS_temp = " << g_IS_r1 << endline_flush;
  //  logger << LOG_DEBUG << setw(47) << name << "inv_r[3][" << setw(3) << x_t << "] = " << setprecision(15) << setw(23) << r2 << "   g_IS_temp = " << g_IS_r2 << endline_flush;

  logger.finished();
}


void mapping_tchannel::errorhandling_c(){
  static Logger logger("mapping_tchannel::errorhandling_c");
  logger.called();

  n_fail_c++;
  
  logger << LOG_DEBUG << "Debug output since point creation failed!" << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << out << "] = " << *s_out << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << out_1 << "] = " << *s_out_1 << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << out_2 << "] = " << *s_out_2 << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << in_1 << "] = " << *s_in_1 << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << in_2 << "] = " << *s_in_2 << endline_flush;
  logger << LOG_DEBUG << "xbs_all[" << x_a << "][" << in_1_out_1 << "] = " << *s_in_1_out_1 << endline_flush;
  logger << LOG_DEBUG << "tmin = " << tmin << endline_flush;
  logger << LOG_DEBUG << "tmax = " << tmax << endline_flush;
  logger << LOG_DEBUG << "tmax_cut = " << tmax_cut << endline_flush;

  logger.finished();
}

  
void mapping_tchannel::errorhandling_g(){
  static Logger logger("mapping_tchannel::errorhandling_g");
  logger.called();

  n_fail_g++;
  
  if (*g_t < 0.){
    *g_t = 0.;
    logger << LOG_DEBUG << "g_t < 0." << "   out = " << out << "   in_1 = " << in_1 << "   in_2 = " << in_2 << "   out_1 = " << out_1 << "   out_2 = " << out_2 << endline_flush;
  }
  if (munich_isnan(*g_t)){
    logger << LOG_DEBUG << "error output:  g_t  is  NaN !" << endline_flush;
    for (int i_p = 1; i_p < psi->xbp_all[x_a].size(); i_p *= 2){
      logger << LOG_DEBUG << "psi->xbp_all[" << x_a << "][" << setw(3) << i_p << "] = " << psi->xbp_all[x_a][i_p] << endline_flush;
    }
    logger << LOG_DEBUG << "m = " << m << endline_flush;
    logger << LOG_DEBUG << "out = " << out << endline_flush;
    logger << LOG_DEBUG << "in_1 = " << in_1 << endline_flush;
    logger << LOG_DEBUG << "in_2 = " << in_2 << endline_flush;
    logger << LOG_DEBUG << "out_1 = " << out_1 << endline_flush;
    logger << LOG_DEBUG << "out_2 = " << out_2 << endline_flush;
    logger << LOG_DEBUG << "term1 = " << term1 << endline_flush;
    logger << LOG_DEBUG << "term2 = " << term2 << endline_flush;
    logger << LOG_DEBUG << "tmin = " << tmin << endline_flush;
    logger << LOG_DEBUG << "tmax = " << tmax << endline_flush;
    logger << LOG_DEBUG << "tmax_cut = " << tmax_cut << endline_flush;
    logger << LOG_DEBUG << "t = *s_in_1_out_1 = " << *s_in_1_out_1 << endline_flush;
    //    logger << LOG_DEBUG << "g_t_propagator(-*s_in_1_out_1, m, -tmin, -tmax) = " << g_t_propagator(-*s_in_1_out_1, m, -tmin, -tmax) << endline_flush;
    logger << LOG_DEBUG << "g_t_propagator = " << g_t_propagator << endline_flush;
    logger << LOG_DEBUG << "lambda(*s_out, *s_in_1, *s_in_2) = " << lambda(*s_out, *s_in_1, *s_in_2) << endline_flush;
    logger << LOG_DEBUG << "term2in = " << term2in << endline_flush;
    logger << LOG_DEBUG << "lambda(*s_out, *s_out_1, *s_out_2) = " << lambda(*s_out, *s_out_1, *s_out_2) << endline_flush;
    logger << LOG_DEBUG << "term2out = " << term2out << endline_flush;
    logger << LOG_DEBUG << "xbs_all[x_a][out = " << out << "] = " << *s_out << endline_flush;
    logger << LOG_DEBUG << "xbs_all[x_a][out_1 = " << out_1 << "] = " << *s_out_1 << endline_flush;
    logger << LOG_DEBUG << "xbs_all[x_a][out_2 = " << out_2 << "] = " << *s_out_2 << endline_flush;
  }
  
  logger.finished();
}

 
void mapping_tchannel::errorhandling_inv_r(){
  static Logger logger("mapping_tchannel::errorhandling_inv_r");
  logger.called();

  n_fail_inv_r++;
  
  // Set r2 to  map_technical_r  and  1. - map_technical_r  instead ???
  //  if (r2 <= 0){
  if (!(r2 >= 0)) {
    logger << LOG_DEBUG << "r2<0 in inv_tchannel_opt!" << endline_flush;
    logger << LOG_DEBUG << "r2=" << r2 <<"; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    r2 = 0.;
  }
  if ((r2 > 1)) {
    logger << LOG_DEBUG << "r2>1 in inv_tchannel_opt!" << endline_flush;
    logger << LOG_DEBUG << "r2=" << r2 <<"; k10.x, k10.y=" << k10.x1() << ", " << k10.x2() << endline_flush;
    r2 = 1.;
  }
  
  logger.finished();
}


void mapping_tchannel::calculate_propagator_c(){
  static Logger logger("mapping_tchannel::calculate_propagator_c");
  logger.called();

  if (nuexp != 1.){ *s_in_1_out_1 = -pow(r1 * pow((-tmax) - mass2_mapping, 1 - nuexp) + (1 - r1) * pow((-tmin) - mass2_mapping, 1. - nuexp), (1. / (1. - nuexp))) - mass2_mapping; }
  else { *s_in_1_out_1 = -exp(r1 * log((-tmax) - mass2_mapping) + (1. - r1) * log((-tmin) - mass2_mapping)) - mass2_mapping; }

  if (!(*s_in_1_out_1 < 0.)){ errorhandling_propagator_c(); }

  logger.finished();
}


void mapping_tchannel::calculate_propagator_g(){
  static Logger logger("mapping_tchannel::calculate_propagator_g");
  logger.called();

  if (nuexp != 1.){ g_t_propagator = -(1. - nuexp) / ((pow((-tmax) - mass2_mapping, 1. - nuexp) - pow((-tmin) - mass2_mapping, 1. - nuexp)) * pow(-*s_in_1_out_1 - mass2_mapping, nuexp)); }
  else { g_t_propagator = -1. / ((log((-tmax) - mass2_mapping) - log((-tmin) - mass2_mapping)) * (-*s_in_1_out_1 - mass2_mapping)); }
  
  if (!(g_t_propagator > 0.)){ errorhandling_propagator_g(); }

  logger.finished();
}


void mapping_tchannel::calculate_propagator_inv_r(){
  static Logger logger("mapping_tchannel::calculate_propagator_inv_r");
  logger.called();
  
  if (nuexp != 1.) {
    double tmin_m0 = pow((-tmin) - mass2_mapping, 1. - nuexp);
    r1 = (pow(-*s_in_1_out_1 - mass2_mapping, 1. - nuexp) - tmin_m0) / (pow((-tmax) - mass2_mapping, 1. - nuexp) - tmin_m0);
  }
  else { r1 = log((-*s_in_1_out_1 - mass2_mapping) / ((-tmin) - mass2_mapping)) / log(((-tmax) - mass2_mapping) / ((-tmin) - mass2_mapping)); }
  
  if (!(r1 > 0. && r1 < 1.)){ errorhandling_propagator_inv_r(); }
  
  logger.finished();
}


void mapping_tchannel::errorhandling_propagator_c(){
  static Logger logger("mapping_tchannel::errorhandling_propagator_c");
  logger.called();

  logger << LOG_DEBUG << "*s_in_1_out_1 > 0.   tmin = " << tmin << "   *s_in_1_out_1 = " << *s_in_1_out_1 << "   tmax = " << tmax << endline_flush;
 
  logger.finished();
}

  
void mapping_tchannel::errorhandling_propagator_g(){
  static Logger logger("mapping_tchannel::errorhandling_propagator_g");
  logger.called();

  if (munich_isnan(g_t_propagator)) {
    logger << LOG_DEBUG << "g_t_propagator == nan in g_t_propagator!   tmin = " << setprecision(15) << setw(23) << tmin << "   *s_in_1_out_1 = " << setprecision(15) << setw(23) << *s_in_1_out_1 << "   tmax = " << setprecision(15) << setw(23) << tmax << "   gtot = nan from t-channel propagator mapping" << endline_flush;
  }

  if (g_t_propagator < 0.){ logger << LOG_DEBUG << "g_t_propagator = " << g_t_propagator << " > 0.   tmin = " << tmin << "   *s_in_1_out_1 = " << *s_in_1_out_1 << "   tmax = " << tmax << endline_flush; }

  logger.finished();
}

 
void mapping_tchannel::errorhandling_propagator_inv_r(){
  static Logger logger("mapping_tchannel::errorhandling_propagator_inv_r");
  logger.called();

  //  static double temp_map_technical_r = 1.e-8; -> psi->map_technical_t
  if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "r1 is NaN in inv_t_propagator!  r1 = " << setprecision(15) << setw(23) << r1 << "   tmin = " << setprecision(15) << setw(23) << tmin << "   *s_in_1_out_1 = " << setprecision(15) << setw(23) << *s_in_1_out_1 << "   tmax = " << setprecision(15) << setw(23) << tmax << endline_flush;
    if (*s_in_1_out_1 >= tmin){ r1 = psi->map_technical_t; }
    else if (*s_in_1_out_1 <= tmax){ r1 = 1. - psi->map_technical_t; }
    else {
      if (psi->switch_console_output_phasespace_issue){ logger << LOG_ERROR << "r1 remains NaN in inv_t_propagator!" << endline_flush; }
      r1 = psi->map_technical_t;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r1 <= 0.){
    logger << LOG_DEBUG << "r1 <= 0 in inv_t_propagator!  r1 = " << setprecision(15) << setw(23) << r1 << "   tmin = " << setprecision(15) << setw(23) << tmin << "   *s_in_1_out_1 = " << setprecision(15) << setw(23) << *s_in_1_out_1 << "   tmax = " << setprecision(15) << setw(23) << tmax << endline_flush;
    r1 = psi->map_technical_t;
  }
  else if (r1 >= 1.){
    logger << LOG_DEBUG << "r1 >= 1 in inv_t_propagator!  r1 = " << setprecision(15) << setw(23) << r1 << "   tmin = " << setprecision(15) << setw(23) << tmin << "   *s_in_1_out_1 = " << setprecision(15) << setw(23) << *s_in_1_out_1 << "   tmax = " << setprecision(15) << setw(23) << tmax << endline_flush;
    r1 = 1. - psi->map_technical_t;
  }

  logger.finished();
}


string mapping_tchannel::info() const {
  static Logger logger("mapping_tchannel::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "tchannel[" << setw(2) << x_a << "][" << setw(3) << x_t << "] = " << name;
  
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}


string mapping_tchannel::output_extended_info(){
  static Logger logger("output_extended_info");
  logger.called();

  stringstream temp_ss;

  // to be filled if required !!!

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}

