#include "header.hpp"
#include "mapping.timelikeinvariant.hpp"

mapping_timelikeinvariant::mapping_timelikeinvariant(){
  static Logger logger("mapping_timelikeinvariant::mapping_timelikeinvariant");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }
 
  logger.finished();
}

mapping_timelikeinvariant::mapping_timelikeinvariant(int _x_f, int _x_a, int _out, int _x_smin, int _x_smax, phasespace_set * _psi){
  static Logger logger("mapping_timelikeinvariant::mapping_timelikeinvariant (x_a, m, out, in_1, in_2, out_1, out_2, psi)");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "constructor called..." << endline_flush; }

  initialization(_x_f, _x_a, _out, _x_smin, _x_smax, _psi);

  logger.finished();
}

mapping_timelikeinvariant::~mapping_timelikeinvariant(){
  static Logger logger("mapping_timelikeinvariant::~mapping_timelikeinvariant");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "destructor called..." << endline_flush; }

  logger.finished();
}


void mapping_timelikeinvariant::initialization(int _x_f, int _x_a, string & content, phasespace_set * _psi){
  static Logger logger("mapping_timelikeinvariant::initialization(x_f, x_a, content, psi)");
  logger.called();

  vector<string> entries(1);
  for (size_t i_i = 0; i_i < content.size(); i_i++){
    if (content[i_i] == ' ' && i_i != content.size() - 1){ entries.push_back(""); }
    else if (content[i_i] != ' '){ entries[entries.size() - 1].push_back(content[i_i]); }
  }
  if (entries.size() != 3){ logger << LOG_FATAL << "timelikeinvariant[" << _x_f << "][" << _x_a << "]  does not have the right content!" << endline_flush; }
  
  int _out = atoi(entries[0].c_str());
  int _x_smin = atoi(entries[1].c_str());
  int _x_smax = atoi(entries[2].c_str());

  initialization(_x_f, _x_a, _out, _x_smin, _x_smax, _psi);

  logger.finished();
}


void mapping_timelikeinvariant::initialization(int _x_f, int _x_a, int _out, int _x_smin, int _x_smax, phasespace_set * _psi){
  static Logger logger("mapping_timelikeinvariant::initialization");
  logger.called();

  x_f = _x_f;
  x_a = _x_a;

  out = _out;
  x_smin = _x_smin;
  x_smax = _x_smax;

  psi = _psi;

  smin = & psi->smin[x_a][x_smin];
  smax = & psi->smax[x_a][x_smax];

  required_xbp = {};
  required_xbs = {out};
  //  required_xbsqrts = {};
  required_xbsqrts = smin->get_required_xbsqrts();
  for (int i_r = 0; i_r < smax->get_required_xbsqrts().size(); i_r++){ required_xbsqrts.push_back(smax->get_required_xbsqrts()[i_r]); }

  /*
  s_out = & psi->xbs_all[x_a][out];
  sqrts_out = & psi->xbsqrts_all[x_a][out];
  smin_out = & psi->v_smin[x_a][x_smin];
  smax_out = & psi->v_smax[x_a][x_smax];

  g_f = & psi->g_f[x_a][x_f];
  logger << LOG_DEBUG << "g_f = " << g_f << "   has been set." << endline_flush;

  smin = & psi->smin[x_a][x_smin];
  smax = & psi->smax[x_a][x_smax];

  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 1, x_f}]];

  stringstream temp_ss;
  temp_ss << "timelikeinvariant_" << "";
  temp_ss << "_" << "_" << out << "_smin" << x_smin << "_smax" << x_smax;
  name = temp_ss.str();
  */
  stringstream temp_ss_f;
  temp_ss_f << "timelikeinvariant_smin" << x_smin << "_smax" << x_smax;
  MC_IS_name_f = temp_ss_f.str();
  
  logger << LOG_DEBUG << "timelikeinvariant[" << setw(5) << x_f << "] =  " << name << "  has been initialized." << endline_flush;

  logger.finished();
}


void mapping_timelikeinvariant::initialization_integration(){
  static Logger logger("mapping_timelikeinvariant::initialization_integration");
  logger.called();

  s_out = & psi->xbs_all[x_a][out];
  sqrts_out = & psi->xbsqrts_all[x_a][out];
  smin_out = & psi->v_smin[x_a][x_smin];
  smax_out = & psi->v_smax[x_a][x_smax];

  g_f = & psi->g_f[x_a][x_f];
  logger << LOG_DEBUG_VERBOSE << "g_f = " << g_f << "   has been set." << endline_flush;

  //  smin = & psi->smin[x_a][x_smin];
  //  smax = & psi->smax[x_a][x_smax];

  random_psp_r1 = psi->random_psp[psi->map_type_to_list_MC_IS[vector<int> {x_a, 1, x_f}]];

  stringstream temp_ss;
  temp_ss << "timelikeinvariant_" << "";
  temp_ss << "_" << "_" << out << "_smin" << x_smin << "_smax" << x_smax;
  name = temp_ss.str();
  /*
  stringstream temp_ss_f;
  temp_ss_f << "timelikeinvariant_smin" << x_smin << "_smax" << x_smax;
  MC_IS_name_f = temp_ss_f.str();
  */
  available = false;

  logger.finished();
}


void mapping_timelikeinvariant::calculate_c(){
  static Logger logger("mapping_timelikeinvariant::calculate_c");
  logger.called();

  smin->get();
  smax->get();
  // Only for: psi->switch_IS_mode_phasespace == 4
  r1 = random_psp_r1->get_random();
  *s_out = r1 * *smax_out + (1. - r1) * *smin_out;
  psi->psp[x_a][out].register_s();
  // Check why required !!!
  *sqrts_out = sqrt(*s_out);
  psi->psp[x_a][out].register_sqrts();
  available = true;
  if (!(*s_out > 0.)){ errorhandling_c(); }

  logger.finished();
}


void mapping_timelikeinvariant::calculate_g(){
  static Logger logger("mapping_timelikeinvariant::calculate_g");
  logger.called();

  //  smin->get();
  //  smax->get();
  *g_f = 1. / (*smax_out - *smin_out);
  if (!(*g_f > 0.)){ errorhandling_g(); }

  logger.finished();
}


void mapping_timelikeinvariant::calculate_inv_r(){
  static Logger logger("mapping_timelikeinvariant::calculate_inv_r");
  logger.called();

  if (!available){ r1 = (*s_out - *smin_out) / (*smax_out - *smin_out); }
  random_psp_r1->get_g_IS(r1, g_IS_r1);
  *g_f *= g_IS_r1;

  //  logger << LOG_DEBUG << setw(45) << name << "inv_r[0][" << setw(3) << x_f << " = " << setprecision(15) << setw(23) << r1 << "   g_IS_temp = " << g_IS_r1 << endline_flush;

  logger.finished();
}


void mapping_timelikeinvariant::errorhandling_c(){
  static Logger logger("mapping_timelikeinvariant::errorhandling_c");
  logger.called();

  n_fail_c++;
  logger << LOG_DEBUG << "Debug output since point creation failed!" << endline_flush;

  logger.finished();
}

  
void mapping_timelikeinvariant::errorhandling_g(){
  static Logger logger("mapping_timelikeinvariant::errorhandling_g");
  logger.called();

  n_fail_g++;
  if (*g_f < 0.){  }

  logger.finished();
}

 
void mapping_timelikeinvariant::errorhandling_inv_r(){
  static Logger logger("mapping_timelikeinvariant::errorhandling_inv_r");
  logger.called();

  n_fail_inv_r++;
  //  static double temp_map_technical_r = 1.e-8; -> psi->map_technical_r
  if (munich_isnan(r1)) {
    logger << LOG_DEBUG << "r1 is NaN in inv_timelikeinvariant!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; smin, s, smax = " << *smin_out << ", " << *s_out << ", " << *smax_out << endline_flush;
    if (*s_out <= *smin_out){r1 = psi->map_technical_r;}
    else if (*s_out >= *smax_out){r1 = 1. - psi->map_technical_r;}
    else {
      if (psi->switch_console_output_phasespace_issue){
        logger << LOG_ERROR << "r1 remains NaN in inv_timelikeinvariant!" << endline_flush;
      }
      r1 = psi->map_technical_r;
      // Otherwise program might stop !!! Reason for "nan" should be clarified !!!
    }
  }
  else if (r1 <= 0.){
    logger << LOG_DEBUG << "r1 < 0 in inv_timelikeinvariant!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; smin, s, smax = " << *smin_out << ", " << *s_out << ", " << *smax_out << endline_flush;
    r1 = psi->map_technical_r;
  }
  if (r1 >= 1.){
    logger << LOG_DEBUG << "r1 > 1 in inv_timelikeinvariant!" << endline_flush;
    logger << LOG_DEBUG << "r1 = " << r1 <<"; smin, s, smax = " << *smin_out << ", " << *s_out << ", " << *smax_out << endline_flush;
    r1 = 1. - psi->map_technical_r;
  }
  
  logger.finished();
}


string mapping_timelikeinvariant::info() const {
  static Logger logger("mapping_timelikeinvariant::info");
  logger.called();

  stringstream temp_ss;
  temp_ss << "timelikeinvariant[" << setw(2) << x_a << "][" << setw(3) << x_f << "] = " << name;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << temp_ss.str() << endline_flush; }
  logger.finished();
  return temp_ss.str();
}




