#include "header.hpp"

void phasespace_set::initialization_psp_dat(int x_a){
  static Logger logger("phasespace_set::initialization_psp_dat");
  logger.called();

  stringstream identifier_ss;
  identifier_ss << csi->subprocess_name_new_format[x_a] << " " << csi->phasespace_order_alpha_s[x_a] << " " << csi->phasespace_order_alpha_e[x_a] << " " << csi->phasespace_order_interference[x_a]; 
  string identifier_subprocess = identifier_ss.str();
  
  string order_phasespace;
  if (csi->class_contribution_phasespace == 0 || (x_a > 0 && csi->class_contribution_phasespace == 1)){ order_phasespace = "born"; }
  else if (csi->class_contribution_phasespace == 1 || (x_a > 0 && csi->class_contribution_phasespace == 2)){ order_phasespace = "real"; }
  else if (csi->class_contribution_phasespace == 2){ order_phasespace = "doublereal"; }

  string filename_phasespace_dat;
  filename_phasespace_dat = csi->path_MUNICH + "psg/" + csi->identifier + "/" + csi->identifier + ".phasespace." + order_phasespace + ".dat";

  logger << LOG_DEBUG << "filename_phasespace_dat = " << filename_phasespace_dat << endline_flush;

  ifstream input(filename_phasespace_dat);
  if (!input.is_open()) {
    logger << LOG_FATAL << "filename_phasespace_dat = " << filename_phasespace_dat << " could not be opened." << endline_flush; //exit(1);
  }
  string line;
  int line_counter = 0;
  int data_active = 0;
  logger << LOG_DEBUG << "identifier_subprocess = " << identifier_subprocess << endline_flush;
  while (getline(input, line)){
    line_counter++;
    logger << LOG_DEBUG << "line_counter = " << setw(5) << line_counter << " :   " << line << endline_flush;
    if (line == identifier_subprocess){
      logger << LOG_DEBUG << "Required phase space information is found in line "<< line_counter << " : " << identifier_subprocess << endline_flush;
      data_active = 1;
    }
    else if (line == "" && data_active){ logger << LOG_DEBUG << "Required phase space information ended in line "<< line_counter << " : " << identifier_subprocess << endline_flush; break; }
    else if (data_active){
      logger << LOG_DEBUG_VERBOSE << "data_active = " << data_active << endline_flush;
      vector<string> input_line;
      int start = -1;
      for (size_t i_i = 0; i_i < line.size(); i_i++){
	if (line[i_i] == '|'){
	  input_line.push_back(line.substr(start + 1, i_i - start - 1));
	  start = i_i + 1;
	}
      }
      if (start == -1){
	logger << LOG_DEBUG << "Ignore line: " << line << endline_flush;
	continue;
      }
      //      for (size_t i_i = 0; i_i < input_line.size(); i_i++){ logger << LOG_DEBUG << "input_line[" << i_i << "] = " << input_line[i_i] << "" << endline_flush; }
      // determine number of entries:
      vector<string> entries(1);
      for (size_t i_i = 0; i_i < input_line[0].size(); i_i++){
	if (input_line[0][i_i] == ' '){ entries.push_back(""); }
	else { entries[entries.size() - 1].push_back(input_line[0][i_i]); }
      }
      string name = entries[0];
      int n_entries = atoi(entries[1].c_str());
      logger << LOG_DEBUG << name << "  has  " << n_entries << "  entries." << endline_flush;
      if (data_active == 1){ // smin translation
	smin[x_a].resize(n_entries);
	for (int i_smin = 0; i_smin < n_entries; i_smin++){ smin[x_a][i_smin].initialization(i_smin, x_a, input_line[i_smin + 1], this); }
	for (int i_smin = 0; i_smin < smin[x_a].size(); i_smin++){ logger << LOG_DEBUG << "smin[" << x_a << "][" << i_smin << "] = " <<  smin[x_a][i_smin].info() << endline_flush; }
      }
      else if (data_active == 2){ // smax translation
	smax[x_a].resize(n_entries);
	for (int i_smax = 0; i_smax < n_entries; i_smax++){ smax[x_a][i_smax].initialization(i_smax, x_a, input_line[i_smax + 1], this); }
	for (int i_smax = 0; i_smax < smax[x_a].size(); i_smax++){ logger << LOG_DEBUG << "smax[" << x_a << "][" << i_smax << "] = " <<  smax[x_a][i_smax].info() << endline_flush; }
      }
      else if (data_active == 3){ // propagator translation
	propagator[x_a].resize(n_entries);
	for (int i_p = 0; i_p < n_entries; i_p++){ propagator[x_a][i_p].initialization(i_p, x_a, input_line[i_p + 1], this); }
	for (int i_p = 0; i_p < propagator[x_a].size(); i_p++){ logger << LOG_DEBUG << "propagator[" << x_a << "][" << i_p << "] = " <<  propagator[x_a][i_p].info() << endline_flush; }
      }
      else if (data_active == 4){ // timelikeinvariant translation
	timelikeinvariant[x_a].resize(n_entries);
	for (int i_f = 0; i_f < n_entries; i_f++){ timelikeinvariant[x_a][i_f].initialization(i_f, x_a, input_line[i_f + 1], this); }
	for (int i_f = 0; i_f < timelikeinvariant[x_a].size(); i_f++){ logger << LOG_DEBUG << "timelikeinvariant[" << x_a << "][" << i_f << "] = " <<  timelikeinvariant[x_a][i_f].info() << endline_flush; }
      }
      else if (data_active == 5){ // tchannel translation
	tchannel[x_a].resize(n_entries);
	for (int i_t = 0; i_t < n_entries; i_t++){ tchannel[x_a][i_t].initialization(i_t, x_a, input_line[i_t + 1], this); }
	for (int i_t = 0; i_t < tchannel[x_a].size(); i_t++){ logger << LOG_DEBUG << "tchannel[" << x_a << "][" << i_t << "] = " <<  tchannel[x_a][i_t].info() << endline_flush; }
      }
      else if (data_active == 6){ // decay translation
	decay[x_a].resize(n_entries);
	for (int i_d = 0; i_d < n_entries; i_d++){ decay[x_a][i_d].initialization(i_d, x_a, input_line[i_d + 1], this); }
	for (int i_d = 0; i_d < decay[x_a].size(); i_d++){ logger << LOG_DEBUG << "decay[" << x_a << "][" << i_d << "] = " <<  decay[x_a][i_d].info() << endline_flush; }
      }
      else if (data_active == 7){ // channel translation
      logger << LOG_DEBUG_VERBOSE << "fullchannel.size() = " << fullchannel.size() << endline_flush;
      fullchannel[x_a].resize(n_entries);
      logger << LOG_DEBUG_VERBOSE << "fullchannel[x_a].size() = " << fullchannel[x_a].size() << endline_flush;
	for (int i_ch = 0; i_ch < n_entries; i_ch++){ fullchannel[x_a][i_ch].initialization(i_ch, x_a, input_line[i_ch + 1], this); }
	for (int i_ch = 0; i_ch < fullchannel[x_a].size(); i_ch++){ logger << LOG_DEBUG << "fullchannel[" << x_a << "][" << i_ch << "] = " <<  fullchannel[x_a][i_ch].info() << endline_flush; }
      }
      data_active++;
    }
  }

  if (!data_active){
    logger << LOG_FATAL << "Required phase space information for  " << identifier_subprocess << "  has not been found." << endline_flush;
    exit(1);	
  }

  logger.finished();
}


void phasespace_set::initialization_required_xb(){
  static Logger logger("phasespace_set::initialization_required_xb");
  logger.called();
    
  logger << LOG_DEBUG << "get_required_xbp/s/sqrts determination:" << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "xb_content_n.size() = " << xb_content_n.size() << endline_flush;

  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    for (int i_p = 0; i_p < propagator[i_a].size(); i_p++){
      for (int i_i = 0; i_i < propagator[i_a][i_p].get_required_xbp().size(); i_i++){
	if (xb_content_n[propagator[i_a][i_p].get_required_xbp()[i_i]] == 1){ continue; }
	if (propagator[i_a][i_p].get_required_xbp()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false;
	for (int j_i = 0; j_i < required_xbp[i_a].size(); j_i++){
	  if (required_xbp[i_a][j_i] == propagator[i_a][i_p].get_required_xbp()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbp[i_a].push_back(propagator[i_a][i_p].get_required_xbp()[i_i]); }
      }
    }
    for (int i_p = 0; i_p < propagator[i_a].size(); i_p++){
      for (int i_i = 0; i_i < propagator[i_a][i_p].get_required_xbs().size(); i_i++){
	if (xb_content_n[propagator[i_a][i_p].get_required_xbs()[i_i]] == 1){ continue; }
	if (propagator[i_a][i_p].get_required_xbs()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbs[i_a].size(); j_i++){
	  if (required_xbs[i_a][j_i] == propagator[i_a][i_p].get_required_xbs()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbs[i_a].push_back(propagator[i_a][i_p].get_required_xbs()[i_i]); }
      }
    }
    for (int i_p = 0; i_p < propagator[i_a].size(); i_p++){
      for (int i_i = 0; i_i < propagator[i_a][i_p].get_required_xbsqrts().size(); i_i++){
	if (xb_content_n[propagator[i_a][i_p].get_required_xbsqrts()[i_i]] == 1){ continue; }
	if (propagator[i_a][i_p].get_required_xbsqrts()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbsqrts[i_a].size(); j_i++){
	  if (required_xbsqrts[i_a][j_i] == propagator[i_a][i_p].get_required_xbsqrts()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbsqrts[i_a].push_back(propagator[i_a][i_p].get_required_xbsqrts()[i_i]); }
      }
    }

    for (int i_f = 0; i_f < timelikeinvariant[i_a].size(); i_f++){
      for (int i_i = 0; i_i < timelikeinvariant[i_a][i_f].get_required_xbp().size(); i_i++){
	if (xb_content_n[timelikeinvariant[i_a][i_f].get_required_xbp()[i_i]] == 1){ continue; }
	if (timelikeinvariant[i_a][i_f].get_required_xbp()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbp[i_a].size(); j_i++){
	  if (required_xbp[i_a][j_i] == timelikeinvariant[i_a][i_f].get_required_xbp()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbp[i_a].push_back(timelikeinvariant[i_a][i_f].get_required_xbp()[i_i]); }
      }
    }
    for (int i_f = 0; i_f < timelikeinvariant[i_a].size(); i_f++){
      for (int i_i = 0; i_i < timelikeinvariant[i_a][i_f].get_required_xbs().size(); i_i++){
	if (xb_content_n[timelikeinvariant[i_a][i_f].get_required_xbs()[i_i]] == 1){ continue; }
	if (timelikeinvariant[i_a][i_f].get_required_xbs()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbs[i_a].size(); j_i++){
	  if (required_xbs[i_a][j_i] == timelikeinvariant[i_a][i_f].get_required_xbs()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbs[i_a].push_back(timelikeinvariant[i_a][i_f].get_required_xbs()[i_i]); }
      }
    }
    for (int i_f = 0; i_f < timelikeinvariant[i_a].size(); i_f++){
      for (int i_i = 0; i_i < timelikeinvariant[i_a][i_f].get_required_xbsqrts().size(); i_i++){
	if (xb_content_n[timelikeinvariant[i_a][i_f].get_required_xbsqrts()[i_i]] == 1){ continue; }
	if (timelikeinvariant[i_a][i_f].get_required_xbsqrts()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbsqrts[i_a].size(); j_i++){
	  if (required_xbsqrts[i_a][j_i] == timelikeinvariant[i_a][i_f].get_required_xbsqrts()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbsqrts[i_a].push_back(timelikeinvariant[i_a][i_f].get_required_xbsqrts()[i_i]); }
      }
    }

    for (int i_t = 0; i_t < tchannel[i_a].size(); i_t++){
      for (int i_i = 0; i_i < tchannel[i_a][i_t].get_required_xbp().size(); i_i++){
	if (xb_content_n[tchannel[i_a][i_t].get_required_xbp()[i_i]] == 1){ continue; }
	if (tchannel[i_a][i_t].get_required_xbp()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbp[i_a].size(); j_i++){
	  if (required_xbp[i_a][j_i] == tchannel[i_a][i_t].get_required_xbp()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbp[i_a].push_back(tchannel[i_a][i_t].get_required_xbp()[i_i]); }
      }
    }
    for (int i_t = 0; i_t < tchannel[i_a].size(); i_t++){
      for (int i_i = 0; i_i < tchannel[i_a][i_t].get_required_xbs().size(); i_i++){
	if (xb_content_n[tchannel[i_a][i_t].get_required_xbs()[i_i]] == 1){ continue; }
	if (tchannel[i_a][i_t].get_required_xbs()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbs[i_a].size(); j_i++){
	  if (required_xbs[i_a][j_i] == tchannel[i_a][i_t].get_required_xbs()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbs[i_a].push_back(tchannel[i_a][i_t].get_required_xbs()[i_i]); }
      }
    }
    for (int i_t = 0; i_t < tchannel[i_a].size(); i_t++){
      for (int i_i = 0; i_i < tchannel[i_a][i_t].get_required_xbsqrts().size(); i_i++){
	if (xb_content_n[tchannel[i_a][i_t].get_required_xbsqrts()[i_i]] == 1){ continue; }
	if (tchannel[i_a][i_t].get_required_xbsqrts()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbsqrts[i_a].size(); j_i++){
	  if (required_xbsqrts[i_a][j_i] == tchannel[i_a][i_t].get_required_xbsqrts()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbsqrts[i_a].push_back(tchannel[i_a][i_t].get_required_xbsqrts()[i_i]); }
      }
    }

    for (int i_d = 0; i_d < decay[i_a].size(); i_d++){
      for (int i_i = 0; i_i < decay[i_a][i_d].get_required_xbp().size(); i_i++){
	if (xb_content_n[decay[i_a][i_d].get_required_xbp()[i_i]] == 1){ continue; }
	if (decay[i_a][i_d].get_required_xbp()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbp[i_a].size(); j_i++){
	  if (required_xbp[i_a][j_i] == decay[i_a][i_d].get_required_xbp()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbp[i_a].push_back(decay[i_a][i_d].get_required_xbp()[i_i]); }
      }
    }
    for (int i_d = 0; i_d < decay[i_a].size(); i_d++){
      for (int i_i = 0; i_i < decay[i_a][i_d].get_required_xbs().size(); i_i++){
	if (xb_content_n[decay[i_a][i_d].get_required_xbs()[i_i]] == 1){ continue; }
	if (decay[i_a][i_d].get_required_xbs()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbs[i_a].size(); j_i++){
	  if (required_xbs[i_a][j_i] == decay[i_a][i_d].get_required_xbs()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbs[i_a].push_back(decay[i_a][i_d].get_required_xbs()[i_i]); }
      }
    }
    for (int i_d = 0; i_d < decay[i_a].size(); i_d++){
      for (int i_i = 0; i_i < decay[i_a][i_d].get_required_xbsqrts().size(); i_i++){
	if (xb_content_n[decay[i_a][i_d].get_required_xbsqrts()[i_i]] == 1){ continue; }
	if (decay[i_a][i_d].get_required_xbsqrts()[i_i] == xb_all_out[i_a]){ continue; }
	bool flag = false; 
	for (int j_i = 0; j_i < required_xbsqrts[i_a].size(); j_i++){
	  if (required_xbsqrts[i_a][j_i] == decay[i_a][i_d].get_required_xbsqrts()[i_i]){ flag = true; break; }
	}
	if (!flag){ required_xbsqrts[i_a].push_back(decay[i_a][i_d].get_required_xbsqrts()[i_i]); }
      }
    }
  }

  logger << LOG_DEBUG << "Requirements on p/s/sqrts, without 1-parton momenta, before sorting" << endline << output_required_xb() << endline_flush;

  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    sort(required_xbp[i_a].begin(), required_xbp[i_a].end());
    sort(required_xbs[i_a].begin(), required_xbs[i_a].end());
    sort(required_xbsqrts[i_a].begin(), required_xbsqrts[i_a].end());
  }
  logger << LOG_DEBUG << "Requirements on p/s/sqrts, without 1-parton momenta, after sorting" << endline << output_required_xb() << endline_flush;

  // Add additional requirements for dipole channels:
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < csi->n_ps; i_a++){
      if (csi->dipole[i_a].massive()){ continue; }
      if (csi->dipole[i_a].type_dipole() == 1 || csi->dipole[i_a].type_dipole() == 2){
	bool flag = false;
	int new_required_xbs = csi->dipole[i_a].binary_R_emitter_1() + csi->dipole[i_a].binary_R_emitter_2();
	for (int j_i = 0; j_i < required_xbs[0].size(); j_i++){
	  if (required_xbs[0][j_i] == new_required_xbs){ flag = true; break; }
	}
	if (!flag){ required_xbs[0].push_back(new_required_xbs); }
      }
      else if (csi->dipole[i_a].type_dipole() == 3){
	bool flag = false;
	int new_required_xbs = csi->dipole[i_a].binary_R_emitter_2() + csi->dipole[i_a].binary_R_spectator();
	for (int j_i = 0; j_i < required_xbs[0].size(); j_i++){
	  if (required_xbs[0][j_i] == new_required_xbs){ flag = true; break; }
	}
	if (!flag){ required_xbs[0].push_back(new_required_xbs); }
      }
    }
  }

  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    sort(required_xbp[i_a].begin(), required_xbp[i_a].end());
    sort(required_xbs[i_a].begin(), required_xbs[i_a].end());
    sort(required_xbsqrts[i_a].begin(), required_xbsqrts[i_a].end());
  }
  
  logger << LOG_DEBUG << "Requirements on p/s/sqrts with dipole additions" << endline << output_required_xb() << endline_flush;
  
  logger.finished();
}


string phasespace_set::output_required_xb(){
  static Logger logger("phasespace_set::output_required_xb");
  logger.called();
    
  stringstream temp_ss;
  for (int i_a = 0; i_a < csi->n_ps; i_a++){
    temp_ss << "Required p/s/sqrts for phasespace i_a = " << i_a << " :" << endline;
    //    temp_ss.str(""); temp_ss.clear();
    temp_ss << "required_xbp[" << setw(2) << i_a << "] = {";
    for (int i_r = 0; i_r < required_xbp[i_a].size(); i_r++){
      if (i_r > 0){ temp_ss << ", "; }
      temp_ss << setw(3) << required_xbp[i_a][i_r];
    }
    temp_ss << "}" << endline;
    //    logger << LOG_DEBUG << temp_ss.str() << endline_flush;
      
    //    temp_ss.str(""); temp_ss.clear();
    temp_ss << "required_xbs[" << setw(2) << i_a << "] = {";
    for (int i_r = 0; i_r < required_xbs[i_a].size(); i_r++){
      if (i_r > 0){ temp_ss << ", "; }
      temp_ss << setw(3) << required_xbs[i_a][i_r];
    }
    temp_ss << "}" << endline;
    //    logger << LOG_DEBUG << temp_ss.str() << endline_flush;
      
    //    temp_ss.str(""); temp_ss.clear();
    temp_ss << "required_xbsqrts[" << setw(2) << i_a << "] = {";
    for (int i_r = 0; i_r < required_xbsqrts[i_a].size(); i_r++){
      if (i_r > 0){ temp_ss << ", "; }
      temp_ss << setw(3) << required_xbsqrts[i_a][i_r];
    }
    temp_ss << "}" << endline;
    //    logger << LOG_DEBUG << temp_ss.str() << endline_flush;
  }
   
  logger.finished();
  return temp_ss.str();
}


void phasespace_set::initialization_reconstruction_former_input(int x_a){
  static Logger logger("phasespace_set::initialization_reconstruction_former_input");
  logger.called();

  // only for identity checks:
  vector<vector<vector<int> > > old_needed_v_smin = needed_v_smin;
  vector<vector<vector<int> > > old_needed_v_smax = needed_v_smax;
  vector<vector<vector<int> > > old_needed_g_p = needed_g_p;
  vector<vector<vector<int> > > old_needed_g_f = needed_g_f;
  vector<vector<vector<int> > > old_needed_g_t = needed_g_t;
  vector<vector<vector<int> > > old_needed_g_d = needed_g_d;
  vector<vector<double> > old_g_p = g_p;
  vector<vector<double> > old_g_f = g_f;
  vector<vector<double> > old_g_t = g_t;
  vector<vector<double> > old_g_d = g_d;

  vector<vector<vector<int> > > old_c_p = c_p;
  vector<vector<vector<int> > > old_c_f = c_f;
  vector<vector<vector<int> > > old_c_t = c_t;
  vector<vector<vector<int> > > old_c_d = c_d;

  vector<vector<vector<string> > > old_MC_IS_name_type = MC_IS_name_type;

  needed_v_smin.clear();
  needed_v_smax.clear();
  needed_g_p.clear();
  needed_g_f.clear();
  needed_g_t.clear();
  needed_g_d.clear();
  g_p.clear();
  g_f.clear();
  g_t.clear();
  g_d.clear();

  c_p.clear();
  c_f.clear();
  c_t.clear();
  c_d.clear();

  MC_IS_name_type.clear();

  needed_v_smin.resize(csi->n_ps);
  needed_v_smax.resize(csi->n_ps);
  needed_g_p.resize(csi->n_ps);
  needed_g_f.resize(csi->n_ps);
  needed_g_t.resize(csi->n_ps);
  needed_g_d.resize(csi->n_ps);

  g_p.resize(csi->n_ps);
  g_f.resize(csi->n_ps);
  g_t.resize(csi->n_ps);
  g_d.resize(csi->n_ps);

  c_p.resize(csi->n_ps);
  c_f.resize(csi->n_ps);
  c_t.resize(csi->n_ps);
  c_d.resize(csi->n_ps);

  MC_IS_name_type.resize(csi->n_ps);

  initialization_phasespace_structure(x_a);

  if (Log::getLogThreshold() <= LOG_DEBUG){
    stringstream temp_ss;
    logger << LOG_DEBUG << "check identity of  needed_v_smin" << endline_flush;
    for (int i_smin = 0; i_smin < smin[x_a].size(); i_smin++){
      if (old_needed_v_smin.size() > x_a){
	if (old_needed_v_smin[x_a].size() > i_smin){
	  if (old_needed_v_smin[x_a][i_smin] == needed_v_smin[x_a][i_smin]){ continue; }
	  temp_ss.str(""); temp_ss.clear();
	  temp_ss << "old_needed_v_smin[" << setw(2) << x_a << "][" << setw(3) << i_smin << "] = {";
	  for (int i_n = 0; i_n < old_needed_v_smin[x_a][i_smin].size(); i_n++){
	    if (i_n > 0){ temp_ss << ", "; }
	    temp_ss << setw(3) << old_needed_v_smin[x_a][i_smin][i_n];
	  }
	  temp_ss << "}";
	  logger << LOG_DEBUG << temp_ss.str() << endline_flush;
	}
      }

      temp_ss.str(""); temp_ss.clear();
      temp_ss << "    needed_v_smin[" << setw(2) << x_a << "][" << setw(3) << i_smin << "] = {";
      for (int i_n = 0; i_n < needed_v_smin[x_a][i_smin].size(); i_n++){
	if (i_n > 0){ temp_ss << ", "; }
	temp_ss << setw(3) << needed_v_smin[x_a][i_smin][i_n];
      }
      temp_ss << "}";
      logger << LOG_DEBUG << temp_ss.str() << endline_flush;
    }

    logger << LOG_DEBUG << "check identity of  needed_v_smax" << endline_flush;
    for (int i_smax = 0; i_smax < smax[x_a].size(); i_smax++){
      if (old_needed_v_smax.size() > x_a){
	if (old_needed_v_smax[x_a].size() > i_smax){
	  if (old_needed_v_smax[x_a][i_smax] == needed_v_smax[x_a][i_smax]){ continue; }
	  temp_ss.str(""); temp_ss.clear();
	  temp_ss << "old_needed_v_smax[" << setw(2) << x_a << "][" << setw(3) << i_smax << "] = {";
	  for (int i_n = 0; i_n < old_needed_v_smax[x_a][i_smax].size(); i_n++){
	    if (i_n > 0){ temp_ss << ", "; }
	    temp_ss << setw(3) << old_needed_v_smax[x_a][i_smax][i_n];
	  }
	  temp_ss << "}";
	  logger << LOG_DEBUG << temp_ss.str() << endline_flush;
	}
      }
 
      temp_ss.str(""); temp_ss.clear();
      temp_ss << "    needed_v_smax[" << setw(2) << x_a << "][" << setw(3) << i_smax << "] = {";
      for (int i_n = 0; i_n < needed_v_smax[x_a][i_smax].size(); i_n++){
	if (i_n > 0){ temp_ss << ", "; }
	temp_ss << setw(3) << needed_v_smax[x_a][i_smax][i_n];
      }
      temp_ss << "}";
      logger << LOG_DEBUG << temp_ss.str() << endline_flush;
    }

    logger << LOG_DEBUG << "check identity of  needed_g_p" << endline_flush;
    for (int i_p = 0; i_p < needed_g_p[x_a].size(); i_p++){
      if (old_needed_g_p.size() > x_a){
	if (old_needed_g_p[x_a].size() > i_p){
	  if (old_needed_g_p[x_a][i_p] == needed_g_p[x_a][i_p]){ continue; }
	  temp_ss.str(""); temp_ss.clear();
	  temp_ss << "old_needed_g_p[" << setw(2) << x_a << "][" << setw(3) << i_p << "] = {";
	  for (int i_n = 0; i_n < old_needed_g_p[x_a][i_p].size(); i_n++){
	    if (i_n > 0){ temp_ss << ", "; }
	    temp_ss << setw(3) << old_needed_g_p[x_a][i_p][i_n];
	  }
	  temp_ss << "}";
	  logger << LOG_DEBUG << temp_ss.str() << endline_flush;
	}
      }
      temp_ss.str(""); temp_ss.clear();
      temp_ss << "    needed_g_p[" << setw(2) << x_a << "][" << setw(3) << i_p << "] = {";
      for (int i_n = 0; i_n < needed_g_p[x_a][i_p].size(); i_n++){
	if (i_n > 0){ temp_ss << ", "; }
	temp_ss << setw(3) << needed_g_p[x_a][i_p][i_n];
      }
      temp_ss << "}";
      logger << LOG_DEBUG << temp_ss.str() << endline_flush;
    }

    logger << LOG_DEBUG << "check identity of  needed_g_f" << endline_flush;
    for (int i_f = 0; i_f < needed_g_f[x_a].size(); i_f++){
      if (old_needed_g_f.size() > x_a){
	if (old_needed_g_f[x_a].size() > i_f){
	  if (old_needed_g_f[x_a][i_f] == needed_g_f[x_a][i_f]){ continue; }
	  temp_ss.str(""); temp_ss.clear();
	  temp_ss << "old_needed_g_f[" << setw(2) << x_a << "][" << setw(3) << i_f << "] = {";
	  for (int i_n = 0; i_n < old_needed_g_f[x_a][i_f].size(); i_n++){
	    if (i_n > 0){ temp_ss << ", "; }
	    temp_ss << setw(3) << old_needed_g_f[x_a][i_f][i_n];
	  }
	  temp_ss << "}";
	  logger << LOG_DEBUG << temp_ss.str() << endline_flush;
	}
      }
      temp_ss.str(""); temp_ss.clear();
      temp_ss << "    needed_g_f[" << setw(2) << x_a << "][" << setw(3) << i_f << "] = {";
      for (int i_n = 0; i_n < needed_g_f[x_a][i_f].size(); i_n++){
	if (i_n > 0){ temp_ss << ", "; }
	temp_ss << setw(3) << needed_g_f[x_a][i_f][i_n];
      }
      temp_ss << "}";
      logger << LOG_DEBUG << temp_ss.str() << endline_flush;
    }

    logger << LOG_DEBUG << "check identity of  needed_g_t" << endline_flush;
    for (int i_t = 0; i_t < needed_g_t[x_a].size(); i_t++){
      if (old_needed_g_t.size() > x_a){
	if (old_needed_g_t[x_a].size() > i_t){
	  if (old_needed_g_t[x_a][i_t] == needed_g_t[x_a][i_t]){ continue; }
	  temp_ss.str(""); temp_ss.clear();
	  temp_ss << "old_needed_g_t[" << setw(2) << x_a << "][" << setw(3) << i_t << "] = {";
	  for (int i_n = 0; i_n < old_needed_g_t[x_a][i_t].size(); i_n++){
	    if (i_n > 0){ temp_ss << ", "; }
	    temp_ss << setw(3) << old_needed_g_t[x_a][i_t][i_n];
	  }
	  temp_ss << "}";
	  logger << LOG_DEBUG << temp_ss.str() << endline_flush;
	}
      }
      temp_ss.str(""); temp_ss.clear();
      temp_ss << "    needed_g_t[" << setw(2) << x_a << "][" << setw(3) << i_t << "] = {";
      for (int i_n = 0; i_n < needed_g_t[x_a][i_t].size(); i_n++){
	if (i_n > 0){ temp_ss << ", "; }
	temp_ss << setw(3) << needed_g_t[x_a][i_t][i_n];
      }
      temp_ss << "}";
      logger << LOG_DEBUG << temp_ss.str() << endline_flush;
    }

    logger << LOG_DEBUG << "check identity of  needed_g_d" << endline_flush;
    for (int i_d = 0; i_d < needed_g_d[x_a].size(); i_d++){
      if (old_needed_g_d.size() > x_a){
	if (old_needed_g_d[x_a].size() > i_d){
	  if (old_needed_g_d[x_a][i_d] == needed_g_d[x_a][i_d]){ continue; }
	  temp_ss.str(""); temp_ss.clear();
	  temp_ss << "old_needed_g_d[" << setw(2) << x_a << "][" << setw(3) << i_d << "] = {";
	  for (int i_n = 0; i_n < old_needed_g_d[x_a][i_d].size(); i_n++){
	    if (i_n > 0){ temp_ss << ", "; }
	    temp_ss << setw(3) << old_needed_g_d[x_a][i_d][i_n];
	  }
	  temp_ss << "}";
	  logger << LOG_DEBUG << temp_ss.str() << endline_flush;
	}
      }
      temp_ss.str(""); temp_ss.clear();
      temp_ss << "    needed_g_d[" << setw(2) << x_a << "][" << setw(3) << i_d << "] = {";
      for (int i_n = 0; i_n < needed_g_d[x_a][i_d].size(); i_n++){
	if (i_n > 0){ temp_ss << ", "; }
	temp_ss << setw(3) << needed_g_d[x_a][i_d][i_n];
      }
      temp_ss << "}";
      logger << LOG_DEBUG << temp_ss.str() << endline_flush;
    }

    logger << LOG_DEBUG << "check identity of  MC_IS_name_type" << endline_flush;
    for (int i_t = 0; i_t < MC_IS_name_type[x_a].size(); i_t++){
      for (int i_r = 0; i_r < MC_IS_name_type[x_a][i_t].size(); i_r++){
	if (old_MC_IS_name_type.size() > x_a){
	  if (old_MC_IS_name_type[x_a].size() > i_t){
	    if (old_MC_IS_name_type[x_a][i_t].size() > i_r){
	      if (old_MC_IS_name_type[x_a][i_t][i_r] == MC_IS_name_type[x_a][i_t][i_r]){ continue; }
	      logger << LOG_DEBUG << "old_MC_IS_name_type[" << setw(2) << x_a << "][" << setw(3) << i_t << "][" << setw(3) << i_r << "] = " << old_MC_IS_name_type[x_a][i_t][i_r] << endline_flush;
	    }
	  }
	}
	logger << LOG_DEBUG << "    MC_IS_name_type[" << setw(2) << x_a << "][" << setw(3) << i_t << "][" << setw(3) << i_r << "] = " << MC_IS_name_type[x_a][i_t][i_r] << endline_flush;
      }
    }
    logger << LOG_DEBUG << "identity checks done" << endline_flush;
  }

  logger.finished();
}


void phasespace_set::initialization_phasespace_structure(int x_a){
  static Logger logger("phasespace_set::initialization_phasespace_structure");
  logger.called();

  v_smin[x_a].resize(smin[x_a].size());
  v_smax[x_a].resize(smax[x_a].size());

  g_p[x_a].resize(propagator[x_a].size());
  g_f[x_a].resize(timelikeinvariant[x_a].size());
  g_t[x_a].resize(tchannel[x_a].size());
  g_d[x_a].resize(decay[x_a].size());

  c_p[x_a].resize(fullchannel[x_a].size());
  c_f[x_a].resize(fullchannel[x_a].size());
  c_t[x_a].resize(fullchannel[x_a].size());
  c_d[x_a].resize(fullchannel[x_a].size());
  for (int i_c = 0; i_c < fullchannel[x_a].size(); i_c++){
    c_p[x_a][i_c] = fullchannel[x_a][i_c].get_x_p();
    c_f[x_a][i_c] = fullchannel[x_a][i_c].get_x_f();
    c_t[x_a][i_c] = fullchannel[x_a][i_c].get_x_t();
    c_d[x_a][i_c] = fullchannel[x_a][i_c].get_x_d();
  }
  
  needed_v_smin[x_a].resize(smin[x_a].size());
  needed_v_smax[x_a].resize(smax[x_a].size());

  needed_g_p[x_a].resize(propagator[x_a].size());
  needed_g_f[x_a].resize(timelikeinvariant[x_a].size());
  needed_g_t[x_a].resize(tchannel[x_a].size());
  needed_g_d[x_a].resize(decay[x_a].size());

  for (int i_c = 0; i_c < fullchannel[x_a].size(); i_c++){
    for (int i_p = 0; i_p < fullchannel[x_a][i_c].propagator.size(); i_p++){
      needed_v_smin[x_a][fullchannel[x_a][i_c].propagator[i_p]->get_x_smin()].push_back(i_c);
      needed_v_smax[x_a][fullchannel[x_a][i_c].propagator[i_p]->get_x_smax()].push_back(i_c);
      needed_g_p[x_a][fullchannel[x_a][i_c].propagator[i_p]->get_x_p()].push_back(i_c);
    }
    for (int i_f = 0; i_f < fullchannel[x_a][i_c].timelikeinvariant.size(); i_f++){
      needed_v_smin[x_a][fullchannel[x_a][i_c].timelikeinvariant[i_f]->get_x_smin()].push_back(i_c);
      needed_v_smax[x_a][fullchannel[x_a][i_c].timelikeinvariant[i_f]->get_x_smax()].push_back(i_c);
      needed_g_f[x_a][fullchannel[x_a][i_c].timelikeinvariant[i_f]->get_x_f()].push_back(i_c);
    }
    for (int i_t = 0; i_t < fullchannel[x_a][i_c].tchannel.size(); i_t++){
      needed_g_t[x_a][fullchannel[x_a][i_c].tchannel[i_t]->get_x_t()].push_back(i_c);
    }
    for (int i_d = 0; i_d < fullchannel[x_a][i_c].decay.size(); i_d++){
      needed_g_d[x_a][fullchannel[x_a][i_c].decay[i_d]->get_x_d()].push_back(i_c);
    }
  }

  // Check if n_dipoles dependent part needs to be reset here !!!
  if (x_a == 0){ MC_IS_name_type[0].resize(6); }
  else { MC_IS_name_type[x_a].resize(9); }

  MC_IS_name_type[x_a][0].resize(propagator[x_a].size());
  for (int i_p = 0; i_p < propagator[x_a].size(); i_p++){ MC_IS_name_type[x_a][0][i_p] = propagator[x_a][i_p].MC_IS_name_p; }
  MC_IS_name_type[x_a][1].resize(timelikeinvariant[x_a].size());
  for (int i_f = 0; i_f < timelikeinvariant[x_a].size(); i_f++){ MC_IS_name_type[x_a][1][i_f] = timelikeinvariant[x_a][i_f].MC_IS_name_f; }
  MC_IS_name_type[x_a][2].resize(tchannel[x_a].size());
  MC_IS_name_type[x_a][3].resize(tchannel[x_a].size());
  for (int i_t = 0; i_t < tchannel[x_a].size(); i_t++){
    MC_IS_name_type[x_a][2][i_t] = tchannel[x_a][i_t].MC_IS_name_t_t;
    MC_IS_name_type[x_a][3][i_t] = tchannel[x_a][i_t].MC_IS_name_t_phi;
  }
  MC_IS_name_type[x_a][4].resize(decay[x_a].size());
  MC_IS_name_type[x_a][5].resize(decay[x_a].size());
  for (int i_d = 0; i_d < decay[x_a].size(); i_d++){
    MC_IS_name_type[x_a][4][i_d] = decay[x_a][i_d].MC_IS_name_d_costheta;
    MC_IS_name_type[x_a][5][i_d] = decay[x_a][i_d].MC_IS_name_d_phi;
  }
  
  logger.finished();
}


