#include "header.hpp"

void phasespace_set::initialization_complete(){
  static Logger logger("phasespace_set::initialization_complete");
  logger.called();

  // Shifted here from  void munich::run_initialization() - Sets physical (msi) and technical mass and width parameters !!!
  initialization_masses();

  xb_all_max.resize(csi->n_ps, xb_max);
  xb_all_out.resize(csi->n_ps, xb_max - 4);
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < csi->n_ps; i_a++){
      xb_all_max[i_a] = xb_max_dipoles;
      xb_all_out[i_a] = xb_out_dipoles;
    }
  }

  if (!switch_off_new_psp){
    psp.resize(csi->n_ps);
    smin.resize(csi->n_ps);
    smax.resize(csi->n_ps);
    propagator.resize(csi->n_ps);
    timelikeinvariant.resize(csi->n_ps);
    tchannel.resize(csi->n_ps);
    decay.resize(csi->n_ps);
    fullchannel.resize(csi->n_ps);
  }
 
  initialization_particle_content();

  // Shifted here from  void phasespace_set::initialization_contribution_order(), called from void munich::run_initialization() !!!
  MC_n_channel_phasespace.resize(1);
  MC_sum_channel_phasespace.resize(1);

  if (csi->class_contribution_CS_real){
    logger << LOG_DEBUG << "Initialization needs to be completely imported from  initialization_complete_RA" << endline_flush;

    initialization_minimum_tau();
    initialization_minimum_phasespacecut();

    n_dipoles = csi->dipole.size();
    // shifted to  initialization_phasespace_RA()  to avoid  determination_MCchannels  calls !!!
    /*
    MC_n_channel_phasespace[0] = determination_MCchannels(0);
    MC_sum_channel_phasespace[0] = MC_n_channel_phasespace[0];
    csi->dipole[0].set_phasespace_variable(MC_n_channel_phasespace[0], MC_sum_channel_phasespace[0], 0);
    for (int i_a = 1; i_a < csi->dipole.size(); i_a++){
      if (csi->dipole[i_a].type_dipole() == 1 && !(switch_off_RS_mapping_ij_k)){MC_n_channel_phasespace.push_back(determination_MCchannels_dipole(i_a));}
      else if (csi->dipole[i_a].type_dipole() == 2 && !(switch_off_RS_mapping_ij_a)){MC_n_channel_phasespace.push_back(determination_MCchannels_dipole(i_a));}
      else if (csi->dipole[i_a].type_dipole() == 3 && !(switch_off_RS_mapping_ai_k)){MC_n_channel_phasespace.push_back(determination_MCchannels_dipole(i_a));}
      else if (csi->dipole[i_a].type_dipole() == 5 && !(switch_off_RS_mapping_ai_b)){MC_n_channel_phasespace.push_back(determination_MCchannels_dipole(i_a));}
      else {MC_n_channel_phasespace.push_back(0);}
      MC_sum_channel_phasespace.push_back(MC_sum_channel_phasespace[i_a - 1] + MC_n_channel_phasespace[i_a]);
      logger << LOG_DEBUG_VERBOSE << "MC_n_channel_phasespace[i_a = " << setw(2) << i_a << "] = " << setw(5) << MC_n_channel_phasespace[i_a] << "   " << "MC_sum_channel_phasespace[i_a = " << setw(2) << i_a << "] = " << setw(5) << MC_sum_channel_phasespace[i_a] << endline_flush;
    }
    MC_n_channel = accumulate(MC_n_channel_phasespace.begin(), MC_n_channel_phasespace.end(), 0);
    for (int i_a = 1; i_a < csi->dipole.size(); i_a++){
      int temp_massive = 0;
      if (M[abs(csi->type_parton[0][csi->dipole[i_a].no_R_emitter_1()])] != 0. ||
	  M[abs(csi->type_parton[0][csi->dipole[i_a].no_R_emitter_2()])] != 0. ||
	  M[abs(csi->type_parton[0][csi->dipole[i_a].no_R_spectator()])] != 0.){temp_massive = 1;}
      csi->dipole[i_a].set_phasespace_variable(MC_n_channel_phasespace[i_a], MC_sum_channel_phasespace[i_a], temp_massive);
    }
    */
    // imported from  initialization_complete_RA  until here
  }
  else {
    initialization_phasespace_born();
    //    ax_psp(0);
    if (!switch_off_new_psp){
      initialization_psp_dat(0);
      initialization_reconstruction_former_input(0);
    }
    
    MC_n_channel = fullchannel[0].size();
    //    MC_n_channel = determination_MCchannels(0);
    // Check if this identification causes problems !!!
    MC_n_channel_phasespace[0] = MC_n_channel;
    // Check if these two calls need to be at different positions for RA and rest !!!
    initialization_minimum_tau();
    initialization_minimum_phasespacecut();
  }
    
  if (csi->class_contribution_CS_virtual){
    if (csi->type_correction == "QCD"){ csi->QCD_selection_phasespace_singularity(); }
    // QEW/MIX ???
  }

  logger << LOG_DEBUG_VERBOSE << "csi->type_contribution = " << csi->type_contribution << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "csi->phasespace_order_alpha_s[0] = " << csi->phasespace_order_alpha_s[0] << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "csi->phasespace_order_alpha_e[0] = " << csi->phasespace_order_alpha_e[0] << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "csi->phasespace_order_interference[0] = " << csi->phasespace_order_interference[0] << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "csi->no_process_parton[0] = " << csi->no_process_parton[0] << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "MC_n_channel = " << MC_n_channel << endline_flush;

  if (csi->class_contribution_CS_real){
    logger << LOG_DEBUG << "Initialization needs to be completely imported from  initialization_complete_RA" << endline_flush;

    initialization_optimization();

    initialization_phasespace_subprocess_dipole_RA();
    initialization_phasespace_optimization_fiducialcut();
    optimize_minv();
    initialization_phasespace_subprocess_optimization_dipole_RA();
    initialization_phasespace_RA();
  }
  else {
    initialization_phasespace_subprocess();
    initialization_phasespace_optimization_fiducialcut();
    optimize_minv();
    initialization_phasespace_subprocess_optimization();

    ac_tau_psp(0, tau_MC_map);

    // Check if different order is required for RA and rest !!!
    initialization_optimization();
  }

  initialization_general_IS_name_list();
  initialization_MC_IS_name_list();

  if (csi->type_contribution == "RT"){initialization_fake_dipole_mapping_RT();}
  if (csi->type_contribution == "L2RT"){initialization_fake_dipole_mapping_RT();}
  if (csi->type_contribution == "RCA"){initialization_fake_dipole_mapping_RT();}
  if (csi->type_contribution == "RVA"){initialization_fake_dipole_mapping_RT();}
  if (csi->type_contribution == "RJ"){initialization_fake_dipole_mapping_RT();}

  initialization_phasespace_MC_tau();

  initialization_optimization_grid();

  initialization_MC_rng();
  initialization_MC();

  if (csi->class_contribution_QT_virtual){
    /*
    if (csi->type_contribution == "VT" ||
      csi->type_contribution == "L2VT" ||
      csi->type_contribution == "VT2" ||
      csi->type_contribution == "VJ" ||
      csi->type_contribution == "L2VJ" ||
      csi->type_contribution == "CV2" ){
    */
    if (switch_resummation){initialization_phasespace_IS_QT();}
  }
  if (csi->class_contribution_QT_collinear){
    /*
    if (csi->type_contribution == "CT" ||
      csi->type_contribution == "L2CT" ||
      csi->type_contribution == "CT2" ||
      csi->type_contribution == "CJ" ||
      csi->type_contribution == "L2CJ" ||
      csi->type_contribution == "CJ2" ){
    */
    // why in general (not only for resummation) ??? not active anyway, but maybe issues if non-initialized variables ???
    initialization_phasespace_IS_QT();
  }
  if (csi->class_contribution_QT_powercorrection){
    /*
    if (csi->type_contribution == "PT" ||
      csi->type_contribution == "L2PT" ||
      csi->type_contribution == "PT2"){
    */
    // should be required in general here !!!
    initialization_phasespace_IS_QT();
  }

  if (csi->class_contribution_CS_real){
    for (int i = 0; i < csi->singular_region_list.size(); i++){
      logger << LOG_DEBUG << csi->singular_region_list[i][0] << " " << csi->singular_region_list[i][1] << "   " << csi->singular_region_name[csi->singular_region_list[i][0]][csi->singular_region_list[i][1]] << endline_flush;
    }
  }

  if (!switch_off_new_psp){
    required_xbp.resize(csi->n_ps);
    required_xbs.resize(csi->n_ps);
    required_xbsqrts.resize(csi->n_ps);
     
    for (int i_a = 0; i_a < csi->n_ps; i_a++){
      for (int i_smin = 0; i_smin < smin[i_a].size(); i_smin++){ smin[i_a][i_smin].initialization_integration(); }
      for (int i_smax = 0; i_smax < smax[i_a].size(); i_smax++){ smax[i_a][i_smax].initialization_integration(); }
      for (int i_p = 0; i_p < propagator[i_a].size(); i_p++){ propagator[i_a][i_p].initialization_integration(); }
      for (int i_f = 0; i_f < timelikeinvariant[i_a].size(); i_f++){ timelikeinvariant[i_a][i_f].initialization_integration(); }
      for (int i_t = 0; i_t < tchannel[i_a].size(); i_t++){ tchannel[i_a][i_t].initialization_integration(); }
      for (int i_d = 0; i_d < decay[i_a].size(); i_d++){ decay[i_a][i_d].initialization_integration(); }
    }
    initialization_required_xb();
    for (int i_a = 0; i_a < csi->n_ps; i_a++){
      if (csi->class_contribution_QT_powercorrection){ psp[i_a].resize(start_xbp_all[0].size()); }
      else { psp[i_a].resize(start_xbp_all[i_a].size()); }
      for (int bi_p = 0; bi_p < psp[i_a].size(); bi_p++){ psp[i_a][bi_p].initialization(bi_p, i_a, this); }
      for (int bi_p = 0; bi_p < psp[i_a].size(); bi_p++){ psp[i_a][bi_p].initialization_integration(); }
    }
  }
  
  logger.finished();
}

//#include "old.phasespace.initialization.run.cpp"


void phasespace_set::initialization_particle_content(){
  static Logger logger("phasespace_set::initialization_particle_content");
  logger.called();

  xb_content_n.resize(xb_max);
  xb_content.resize(xb_max);
  xb_content_xb.resize(xb_max);
  
  for (size_t i_x = 0; i_x < xb_max; i_x++){
    xb_content[i_x] = vectorint_from_binary(i_x);
    xb_content_xb[i_x] = vectorbinary_from_binary(i_x);
    xb_content_n[i_x] = xb_content[i_x].size();
  }
    
  for (size_t i_x = 0; i_x < xb_max; i_x++){
    stringstream content_index;
    for (size_t i_p = 0; i_p < xb_content[i_x].size(); i_p++){ content_index << setw(4) << right << xb_content[i_x][i_p]; }
    stringstream content_xb;
    for (size_t i_p = 0; i_p < xb_content_xb[i_x].size(); i_p++){ content_xb << setw(4) << right << xb_content_xb[i_x][i_p]; }
    logger << LOG_DEBUG << "content[" << setw(4) << i_x << "]  ->  " << "n = " << setw(2) << xb_content_n[i_x] << "  --  " << "i_p -> " << setw(30) << content_index.str() << "  --  " << "xb -> " <<setw(30) << content_xb.str() << endline_flush;
  }

  logger.finished();
}


void phasespace_set::initialization_phasespace_optimization_fiducialcut(){
  static Logger logger("phasespace_set::initialization_phasespace_optimization_fiducialcut");
  logger.called();

  for (int i_f = 0; i_f < esi->fiducial_cut.size(); i_f++){
    logger << LOG_DEBUG << "esi->fiducial_cut[" << i_f << "] = " << esi->fiducial_cut[i_f].name << endline << esi->fiducial_cut[i_f] << endline_flush;
    if (esi->fiducial_cut[i_f].observable == "M" && (esi->fiducial_cut[i_f].type == LOWER || esi->fiducial_cut[i_f].type == WINDOW)){

      // Don't derive generation cuts if M is different for dipole kinematics !!!
     
      for (int i_a = 0; i_a < n_dipoles; i_a++){
	logger << LOG_DEBUG << "[i_a = " << i_a << "]   Check if cut may be applied on generation level for phase space:" << endline_flush;
	bool apply_generation_cut = true;
	if (csi->class_contribution_CS_real){
	  for (int i_p = 0; i_p < esi->fiducial_cut[i_f].type_particle.size(); i_p++){
	    for (size_t j_p = 0; j_p < esi->index_group_all_parton[i_a][esi->access_object_selection[esi->fiducial_cut[i_f].type_particle[i_p]]].size(); j_p++){
	      size_t x_p = esi->index_group_all_parton[i_a][esi->access_object_selection[esi->fiducial_cut[i_f].type_particle[i_p]]][j_p];
	      logger << LOG_DEBUG << "[i_a = " << i_a << "]   x_p = esi->index_group_all_parton[i_a][[esi->fiducial_cut[" << i_f << "].type_particle[" << i_p << "]]][" << j_p << "] = " << x_p << endline_flush;
	      if (csi->type_correction == "QCD" || csi->type_correction == "MIX"){
		for (int i_j = 0; i_j < esi->ps_runtime_subtraction_QCD[i_a].size(); i_j++){
		  //		  logger << LOG_DEBUG << "[i_a = " << i_a << "]   esi->object_list_selection[esi->ps_runtime_subtraction_QCD[" << i_a << "][" << i_j << "] = " << esi->ps_runtime_subtraction_QCD[i_a][i_j] << "] = " << esi->object_list_selection[esi->ps_runtime_subtraction_QCD[i_a][i_j]] << endline_flush;
		  logger << LOG_DEBUG << "[i_a = " << i_a << "]   Check if  esi->ps_runtime_subtraction_QCD[" << i_a << "][" << i_j << "] = " << esi->ps_runtime_subtraction_QCD[i_a][i_j] << " =?= " << x_p << " = x_p" << endline_flush;
		  if (x_p == esi->ps_runtime_subtraction_QCD[i_a][i_j]){apply_generation_cut = false; break;}
		}
	      }
	      if (csi->type_correction == "QEW" || csi->type_correction == "MIX"){
		for (int i_j = 0; i_j < esi->ps_runtime_subtraction_QEW[i_a].size(); i_j++){
		  //		  logger << LOG_DEBUG << "[i_a = " << i_a << "]   esi->object_list_selection[esi->ps_runtime_subtraction_QEW[" << i_a << "][" << i_j << "] = " << esi->ps_runtime_subtraction_QEW[i_a][i_j] << "] = " << esi->object_list_selection[esi->ps_runtime_subtraction_QEW[i_a][i_j]] << endline_flush;
		  logger << LOG_DEBUG << "[i_a = " << i_a << "]   Check if  esi->ps_runtime_subtraction_QEW[" << i_a << "][" << i_j << "] = " << esi->ps_runtime_subtraction_QEW[i_a][i_j] << " =?= " << x_p << " = x_p" << endline_flush;
		  if (x_p == esi->ps_runtime_subtraction_QEW[i_a][i_j]){apply_generation_cut = false; break;}
		}
	      }
	      if (!apply_generation_cut){break;}
	    }
	    if (!apply_generation_cut){break;}
	  }
	}
      	if (!apply_generation_cut){
	  logger << LOG_DEBUG << "[i_a = " << i_a << "]   Cut may not be applied on generation level!" << endline_flush;
	  continue;
	}

	vector<bool> temp_binary(csi->type_parton[i_a].size(), false);
	logger << LOG_DEBUG << "[i_a = " << i_a << "]   esi->fiducial_cut[" << i_f << "] = " << esi->fiducial_cut[i_f].name << " can be used to determine invariant-mass limits in phase space generation." << endline_flush;
	for (int i_p = 0; i_p < esi->fiducial_cut[i_f].type_particle.size(); i_p++){
	  logger << LOG_DEBUG << "[i_a = " << i_a << "]   esi->fiducial_cut[" << i_f << "].type_particle[" << i_p << "] = " << esi->fiducial_cut[i_f].type_particle[i_p] << endline_flush;
	  for (size_t j_p = 0; j_p < esi->index_group_all_parton[i_a][esi->access_object_selection[esi->fiducial_cut[i_f].type_particle[i_p]]].size(); j_p++){
	    size_t x_p = esi->index_group_all_parton[i_a][esi->access_object_selection[esi->fiducial_cut[i_f].type_particle[i_p]]][j_p];
	    temp_binary[x_p] = true;
	    logger << LOG_DEBUG << "[i_a = " << i_a << "]   temp_binary[" << x_p << "] = true" << endline_flush;
	  }
	}
	int temp_sum_binary = 0;
	for (int i_p = 3; i_p < temp_binary.size(); i_p++){if (temp_binary[i_p]){temp_sum_binary += pow(2, i_p - 1);}}
	if (temp_sum_binary > 0 && esi->fiducial_cut[i_f].cut_value_lower > sqrtsmin_opt[i_a][temp_sum_binary]){
	  sqrtsmin_opt[i_a][temp_sum_binary] = esi->fiducial_cut[i_f].cut_value_lower;
	}
	logger << LOG_DEBUG << "sqrtsmin_opt[i_a = " << i_a << "][temp_sum_binary = " << temp_sum_binary << "] = " << esi->fiducial_cut[i_f].cut_value_lower << " = esi->fiducial_cut[" << i_f << "].cut_value_lower" << endline_flush;
      }
    }
  }

  logger.finished();
}


void phasespace_set::initialization_optimization(){
  static Logger logger("phasespace_set::initialization_optimization ()");
  logger.called();

  logger << LOG_DEBUG << "run_mode = " << run_mode << endline_flush;
  logger << LOG_DEBUG << "switch_output_weights = " << switch_output_weights << endline_flush;
  logger << LOG_DEBUG << "switch_n_events_opt = " << switch_n_events_opt << endline_flush;
  
  //////////////////////////////////////////
  //  new weight-optimization parameters  //
  //////////////////////////////////////////

  if (run_mode == "grid" || run_mode == "gridacc"){
    for (int i_m = 0; i_m < input_MC.size(); i_m++){if (input_MC[i_m]->switch_optimization == 2){input_MC[i_m]->switch_optimization = 1;}}
    for (int i_s = 0; i_s < input_IS.size(); i_s++){if (input_IS[i_s]->switch_optimization == 2){input_IS[i_s]->switch_optimization = 1;}}
  }
  if (run_mode == "gridrefine"){
    for (int i_m = 0; i_m < input_MC.size(); i_m++){if (input_MC[i_m]->switch_optimization == 2){input_MC[i_m]->switch_optimization = 3;}}
    for (int i_s = 0; i_s < input_IS.size(); i_s++){if (input_IS[i_s]->switch_optimization == 2){input_IS[i_s]->switch_optimization = 3;}}
  }
  
  if (switch_step_mode_grid){i_step_mode = & i_gen_step;}
  else {i_step_mode = & i_acc_step;}

  if (!(run_mode == "grid" || run_mode == "gridacc" || run_mode == "gridrefine")){
    switch_output_weights = 0;
    end_optimization = 1;
    switch_output_optimization = 0;
  }

  if (!csi->class_contribution_collinear){input_IS_z1z2.switch_optimization = -1;}
  if (!csi->class_contribution_CS_real){input_MC_x_dipole.switch_optimization = -1;}

  initialization_filename();

  logger.finished();
}


void phasespace_set::initialization_optimization_grid(){
  static Logger logger("phasespace_set::initialization_optimization_grid");
  logger.called();

  //////////////////////////////////////////
  //  new weight-optimization parameters  //
  //////////////////////////////////////////

  opt_n_events_min = 0;
  n_events_MC_opt = 0;

  input_MC_phasespace.n_channel = MC_n_channel;
  
  if (MC_n_channel == 1){input_MC_phasespace.switch_optimization = -1;}
  if (tau_MC_map.size() == 1){input_MC_tau.switch_optimization = -1;}
  
  if (MC_IS_name_list.size() == 0){
   logger << LOG_DEBUG << "MC_IS_name_list.size() = " << MC_IS_name_list.size() << endline_flush;
   logger << LOG_DEBUG << "switch_IS_mode_phasespace = " << switch_IS_mode_phasespace << endline_flush;
   for (int i_s = 0; i_s < input_IS.size(); i_s++){if (input_IS[i_s]->name.substr(0, 5) == "IS_PS"){input_IS[i_s]->switch_optimization = -1;}}
    switch_IS_mode_phasespace = 0;
  }

  // Introduce parameter to allow factor wrt.  n_step  to set each  n_event_per_step !!! E.g: factor_event_per_step, n_event_factor_per_step ??? 
  // Use to overwrite n_event_per_step, or use the larger value ???

  for (int i_m = 0; i_m < input_MC.size(); i_m++){
    if (input_MC[i_m]->n_event_per_step != 0 && input_MC[i_m]->n_event_per_step < n_step){input_MC[i_m]->n_event_per_step = n_step;}
  }
  for (int i_s = 0; i_s < input_IS.size(); i_s++){
    if (input_IS[i_s]->n_event_per_step != 0 && input_IS[i_s]->n_event_per_step < n_step){input_IS[i_s]->n_event_per_step = n_step;}
  }

  for (int i_m = 0; i_m < input_MC.size(); i_m++){
    if (input_MC[i_m]->switch_optimization == 1 || input_MC[i_m]->switch_optimization == 3){
      if (input_MC[i_m]->n_optimization_step * input_MC[i_m]->n_event_per_step > n_events_MC_opt){n_events_MC_opt = input_MC[i_m]->n_optimization_step * input_MC[i_m]->n_event_per_step;}
    }
  }
  for (int i_s = 0; i_s < input_IS.size(); i_s++){
    if (input_IS[i_s]->switch_optimization == 1 || input_IS[i_s]->switch_optimization == 3){
      if (input_IS[i_s]->n_optimization_step * input_IS[i_s]->n_event_per_step > n_events_MC_opt){n_events_MC_opt = input_IS[i_s]->n_optimization_step * input_IS[i_s]->n_event_per_step;}
    }
  }

  // extra steps for tau, x1x2, z1z2, qTres !!!
  for (int i_s = 0; i_s < input_IS.size(); i_s++){
    if (input_IS[i_s]->name == "IS_tau" || input_IS[i_s]->name == "IS_x1x2" || input_IS[i_s]->name == "IS_z1z2" || input_IS[i_s]->name == "IS_qTres"){
      logger << LOG_DEBUG << "input_IS[" << i_s << "]->n_optimization_step = " << input_IS[i_s]->n_optimization_step << endline_flush;
      logger << LOG_DEBUG << "input_IS[" << i_s << "]->n_optimization_step_extra = " << input_IS[i_s]->n_optimization_step_extra << endline_flush;
      logger << LOG_DEBUG << "n_events_MC_opt = " << n_events_MC_opt << endline_flush;
      logger << LOG_DEBUG << "input_IS[" << i_s << "]->n_event_per_step = " << input_IS[i_s]->n_event_per_step << endline_flush;

      if (input_IS[i_s]->n_optimization_step_extra > 0){
	input_IS[i_s]->n_optimization_step = n_events_MC_opt / input_IS[i_s]->n_event_per_step + input_IS[i_s]->n_optimization_step_extra;
      }
      if (input_IS[i_s]->n_optimization_step == 0){input_IS[i_s]->n_optimization_step = 5;}
    }
  }

  // final determination of opt_n_events_min
  for (int i_m = 0; i_m < input_MC.size(); i_m++){
    if ((input_MC[i_m]->switch_optimization == 1 || input_MC[i_m]->switch_optimization == 3) && (n_events_MC_opt > opt_n_events_min)){opt_n_events_min = n_events_MC_opt;}
    logger << LOG_DEBUG << "i_m = " << i_m << "   opt_n_events_min = " << opt_n_events_min << endline_flush;
  }
  for (int i_s = 0; i_s < input_IS.size(); i_s++){
    if ((input_IS[i_s]->switch_optimization == 1 || input_IS[i_s]->switch_optimization == 3) && (input_IS[i_s]->n_optimization_step * input_IS[i_s]->n_event_per_step > opt_n_events_min)){opt_n_events_min = input_IS[i_s]->n_optimization_step * input_IS[i_s]->n_event_per_step;}
    logger << LOG_DEBUG << "i_s = " << i_s << "   opt_n_events_min = " << opt_n_events_min << endline_flush;
  }

  // final determination of n_optimization_repetition
  
  if (switch_n_events_opt){
    n_events_min = opt_n_events_min;
    n_events_max = opt_n_events_min;
  }
  if (n_step > 0){
    n_repetition = n_events_max / n_step;
    for (int i_m = 0; i_m < input_MC.size(); i_m++){input_MC[i_m]->n_optimization_repetition = input_MC[i_m]->n_event_per_step / n_step;}
    for (int i_s = 0; i_s < input_IS.size(); i_s++){input_IS[i_s]->n_optimization_repetition = input_IS[i_s]->n_event_per_step / n_step;}
  }

  logger << LOG_DEBUG << "n_events_max" << " = " << n_events_max << endline_flush;
  logger << LOG_DEBUG << "n_step" << " = " << n_step << endline_flush;
  logger << LOG_DEBUG << "n_step_max" << " = " << n_repetition << endline_flush;



  for (int i_m = 0; i_m < input_MC.size(); i_m++){logger << LOG_DEBUG << "input_MC[" << i_m << "] = " << *input_MC[i_m] << endline_flush;}
  for (int i_s = 0; i_s < input_IS.size(); i_s++){logger << LOG_DEBUG << "input_IS[" << i_s << "] = " << *input_IS[i_s] << endline_flush;}


  
  MC_phasespace = multichannel_set(input_MC_phasespace, filename_weight_writeout, filename_weight_readin);

  // Why is MC_phasespace initialization separated from the rest ???
  
  logger << LOG_DEBUG << "NEW: opt_n_events_min:   " << opt_n_events_min << endline_flush;

  //////////////////////////////////////////

  logger.finished();
}


void phasespace_set::initialization_filename(){
  static Logger logger("phasespace_set::initialization_filename");
  logger.called();

  filename_weight_writeout = "weights/weights_" + csi->subprocess + ".dat";
  filename_weight_readin = "../../../../" + weight_in_directory + "/weights/weights_" + csi->subprocess + ".dat";
  if (switch_output_weights){system_execute(logger, "mkdir weights");}

  logger.finished();
}


void phasespace_set::initialization_masses(){
  static Logger logger("phasespace_set::initialization_masses");
  logger.called();

  // Extension to improve mappings in loop-induced processes (should not be in msi, but only here - as well as possible values of map_Gamma (->psi) - in contrast to reg_Gamma (->asi)) !!!

  M.resize(50, 0.);
  M2.resize(50, 0.);
  Gamma.resize(50, 0.);
  cM2.resize(50, 0.);
  map_Gamma.resize(50, 0.);
  reg_Gamma.resize(50, 0.);

  for (int i = 0; i < 26; i++){
    M[i] = msi->M[i];
    M2[i] = msi->M2[i];
    Gamma[i] = msi->Gamma[i];
    cM2[i] = msi->cM2[i];
    map_Gamma[i] = msi->map_Gamma[i];
    reg_Gamma[i] = msi->reg_Gamma[i];
  }

  for (int i = 1; i < 7; i++){
    if (M[i] != 0.){
      M[30 + i] = 2 * M[i];
      M2[30 + i] = pow(M[30 + i], 2);
      if (Gamma[i] != 0.){Gamma[30 + i] = 10 * Gamma[i];}
      else {Gamma[30 + i] = 0.1 * M[30 + i];}
      map_Gamma[30 + i] = Gamma[30 + i];
      cM2[30 + i] = pow(M[30 + i], 2) - ri * M[30 + i] * Gamma[30 + i];

      M[40 + i] = 2 * M[i];
      M2[40 + i] = pow(M[40 + i], 2);
      if (Gamma[i] != 0.){Gamma[40 + i] = 10 * Gamma[i];}
      else {Gamma[40 + i] = 0.1 * M[40 + i];}
      map_Gamma[40 + i] = Gamma[40 + i];
      cM2[40 + i] = pow(M[40 + i], 2) - ri * M[40 + i] * Gamma[40 + i];
    }
  }

  for (int i_pdg = 0; i_pdg < M.size(); i_pdg++){
    if ((i_pdg > 6 && i_pdg < 11) || (i_pdg > 16 && i_pdg < 22) || (i_pdg > 25 && i_pdg < 31) || (i_pdg > 36 && i_pdg < 41) || (i_pdg > 46)){continue;}
    logger << LOG_DEBUG << "M[" << setw(2) << i_pdg << "]  = " << right << setprecision(8) << setw(15) << M[i_pdg] << "   map_Gamma = " << right << setprecision(8) << setw(15) << map_Gamma[i_pdg] << "   Gamma[" << setw(2) << i_pdg << "]  = " << right << setprecision(8) << setw(15) << Gamma[i_pdg] << endline_flush;
    if (i_pdg == 25){logger.newLine(LOG_DEBUG);}
  }

  logger.finished();
}


void phasespace_set::initialization_general_IS_name_list(){
  static Logger logger("phasespace_set::initialization_general_MC_IS_name_list");
  logger.called();

  general_IS_name_list.push_back("IS_tau");
  general_IS_name_list.push_back("IS_x1x2");
  if (csi->class_contribution_collinear){
    index_IS_z1z2.resize(3);
    IS_z1z2.resize(3);
    for (int i_c = 1; i_c < 3; i_c++){
      stringstream temp_name_ss;
      temp_name_ss << "IS_z1z2[" << i_c << "]";
      string temp_name = temp_name_ss.str();
     general_IS_name_list.push_back(temp_name);
    }
  }
  if (switch_resummation){
    general_IS_name_list.push_back("IS_qTres");
  }
  if (csi->class_contribution_QT_powercorrection){
    general_IS_name_list.push_back("IS_qTres");
  }
  n_general_IS_name_list = general_IS_name_list.size();
 
  logger.finished();
}


void phasespace_set::initialization_MC_IS_name_type(){
  static Logger logger("phasespace_set::initialization_MC_IS_name_type");
  logger.called();

  MC_IS_name_type.resize(n_dipoles);
  MC_IS_name_type[0].resize(6);
  if (n_dipoles > 1){for (int i_a = 1; i_a < n_dipoles; i_a++){MC_IS_name_type[i_a].resize(9);}}

  logger.finished();
}


void phasespace_set::initialization_MC_IS_name_list(){
  static Logger logger("phasespace_set::initialization_MC_IS_name_list");
  logger.called();

  // start counting variables after the general ones:
  int counter = n_general_IS_name_list;
  //  int counter = 0;
  MC_IS_name_list.resize(0);
  for (int i_a = 0; i_a < MC_IS_name_type.size(); i_a++){
    for (int i_t = 0; i_t < MC_IS_name_type[i_a].size(); i_t++){
      for (int i_i = 0; i_i < MC_IS_name_type[i_a][i_t].size(); i_i++){
	MC_IS_name_list.push_back(MC_IS_name_type[i_a][i_t][i_i]);
	vector<int> temp = {i_a, i_t, i_i};
	map_list_to_type_MC_IS[counter] = temp;
	map_type_to_list_MC_IS[temp] = counter;
	counter++;
      }
    }
  }
  n_MC_IS_name_list = MC_IS_name_list.size();

  logger.finished();
}


void phasespace_set::initialization_phasespace_born(){
  static Logger logger("phasespace_set::initialization_phasespace_born");
  logger.called();

  n_dipoles = 1;

  initialization_MC_IS_name_type();

  c_p.resize(1);
  c_f.resize(1);
  c_t.resize(1);
  c_d.resize(1);

  v_smin.resize(1);
  v_smax.resize(1);

  g_p.resize(1);
  g_f.resize(1);
  g_t.resize(1);
  g_d.resize(1);

  needed_v_smin.resize(1);
  needed_v_smax.resize(1);
  needed_g_p.resize(1);

  needed_g_f.resize(1);
  needed_g_t.resize(1);
  needed_g_d.resize(1);

  start_xbp_all.resize(1);
  start_xbs_all.resize(1);
  start_xbsqrts_all.resize(1);

  MC_optswitch = 1;

  logger.finished();
}


void phasespace_set::initialization_phasespace_RA(){
  static Logger logger("phasespace_set::initialization_phasespace_RA");
  logger.called();

  // Check if this debug output is needed !!!
  if (Log::getLogThreshold() <= LOG_DEBUG_POINT){
    stringstream temp_prc;
    for (int i_a = 0; i_a < csi->no_process_parton.size(); i_a++){
      temp_prc << "csi->no_process_parton[" << setw(2) << i_a << "] = " << setw(3) << csi->no_process_parton[i_a] << "     ";
      temp_prc << "csi->swap_parton[" << setw(2) << i_a << "] = ";
      for (int i_p = 1; i_p < csi->swap_parton[i_a].size(); i_p++){temp_prc << csi->swap_parton[i_a][i_p] << "  ";}
      if (i_a < csi->no_process_parton.size() - 1){temp_prc << endline;}
    }
    logger << LOG_DEBUG << "Process numbers and permutations:" << endline << endline << temp_prc.str() << endline_flush;
    logger.newLine(LOG_DEBUG_POINT);
  }

  initialization_MC_IS_name_type();

  c_p.resize(n_dipoles);
  c_f.resize(n_dipoles);
  c_t.resize(n_dipoles);
  c_d.resize(n_dipoles);

  v_smin.resize(n_dipoles);
  v_smax.resize(n_dipoles);

  g_p.resize(n_dipoles);
  g_f.resize(n_dipoles);
  g_t.resize(n_dipoles);
  g_d.resize(n_dipoles);

  needed_v_smin.resize(n_dipoles);
  needed_v_smax.resize(n_dipoles);
  needed_g_p.resize(n_dipoles);
  needed_g_f.resize(n_dipoles);
  needed_g_t.resize(n_dipoles);
  needed_g_d.resize(n_dipoles);

  MC_x_dipole_mapping.resize(n_dipoles);
  dipole_x.resize(n_dipoles, 0.);

  MC_n_channel_phasespace.resize(n_dipoles);
  MC_sum_channel_phasespace.resize(n_dipoles);
  
  MC_optswitch = 0;

  // later !!!
  /*
  MC_n_channel = MC_sum_channel_phasespace[n_dipoles - 1];
  for (int i_a = 0; i_a < n_dipoles; i_a++){
    logger << LOG_DEBUG << "MC_n_channel_phasespace[" << setw(2) << i_a << "] = " << setw(4) << MC_n_channel_phasespace[i_a] << "   MC_sum_channel_phasespace[" << setw(2) << i_a << "] = " << setw(4) << MC_sum_channel_phasespace[i_a] << endline_flush;
  }
  logger << LOG_DEBUG << "MC_n_channel = " << MC_n_channel << endline_flush;
  */
  if (n_dipoles == 1){ switch_off_RS_mapping = 1; }
  // Switch off dipole mapping with 2->1 kinematics
  // Otherwise, delta function need to be implemented for x-type dipole variables !!!
  for (int i_a = 1; i_a < n_dipoles; i_a++){ if (start_xbs_all[i_a][xb_max / 2 - 4]){ switch_off_RS_mapping = 1; } }
  logger << LOG_DEBUG << "switch_off_RS_mapping = " << switch_off_RS_mapping << endline_flush;
  
  //  if (switch_off_RS_mapping){ MC_n_channel = MC_n_channel_phasespace[0]; }

  //  ax_psp(0);
  initialization_psp_dat(0);
  initialization_phasespace_structure(0);
  MC_n_channel_phasespace[0] = fullchannel[0].size();
  MC_sum_channel_phasespace[0] = MC_n_channel_phasespace[0];
  csi->dipole[0].set_phasespace_variable(MC_n_channel_phasespace[0], MC_sum_channel_phasespace[0], 0);
  if (switch_off_RS_mapping){
    //    MC_n_channel = MC_n_channel_phasespace[0];
    //    MC_n_channel_phasespace.resize(n_dipoles);
    //    logger << LOG_DEBUG << "MC_n_channel = " << MC_n_channel << endline_flush;
  }
  else {
    for (int i_a = 1; i_a < n_dipoles; i_a++){
      if ((csi->dipole[i_a].type_dipole() == 1 && !switch_off_RS_mapping_ij_k) ||
	  (csi->dipole[i_a].type_dipole() == 2 && !switch_off_RS_mapping_ij_a) ||
	  (csi->dipole[i_a].type_dipole() == 3 && !switch_off_RS_mapping_ai_k) ||
	  (csi->dipole[i_a].type_dipole() == 5 && !switch_off_RS_mapping_ai_b)){
	// needs to be modified once  tau_MC_map  is also read in from psp_dat files !!!
	initialization_psp_dat(i_a);
	initialization_phasespace_structure(i_a);
	MC_n_channel_phasespace[i_a] = fullchannel[i_a].size();
	//	MC_n_channel_phasespace.push_back(fullchannel[i_a].size());
      }
      //      else {MC_n_channel_phasespace.push_back(0);}
      MC_sum_channel_phasespace[i_a] = MC_sum_channel_phasespace[i_a - 1] + MC_n_channel_phasespace[i_a];
      //      MC_sum_channel_phasespace.push_back(MC_sum_channel_phasespace[i_a - 1] + MC_n_channel_phasespace[i_a]);
      logger << LOG_DEBUG_VERBOSE << "MC_n_channel_phasespace[i_a = " << setw(2) << i_a << "] = " << setw(5) << MC_n_channel_phasespace[i_a] << "   " << "MC_sum_channel_phasespace[i_a = " << setw(2) << i_a << "] = " << setw(5) << MC_sum_channel_phasespace[i_a] << endline_flush;

      logger << LOG_DEBUG_VERBOSE << "MC_IS_name_type[" << i_a << "].size() = " << MC_IS_name_type[i_a].size() << endline_flush;
      //    }
    
      //    for (int i_a = 1; i_a < csi->dipole.size(); i_a++){
      int temp_massive = 0;
      if (M[abs(csi->type_parton[0][csi->dipole[i_a].no_R_emitter_1()])] != 0. ||
	  M[abs(csi->type_parton[0][csi->dipole[i_a].no_R_emitter_2()])] != 0. ||
	  M[abs(csi->type_parton[0][csi->dipole[i_a].no_R_spectator()])] != 0.){temp_massive = 1;}
      csi->dipole[i_a].set_phasespace_variable(MC_n_channel_phasespace[i_a], MC_sum_channel_phasespace[i_a], temp_massive);
    }
    initialization_phasespace_subprocess_dipole_IS_RA();
  }

  MC_n_channel = accumulate(MC_n_channel_phasespace.begin(), MC_n_channel_phasespace.end(), 0);
  //  MC_n_channel = MC_sum_channel_phasespace[n_dipoles - 1];
  for (int i_a = 0; i_a < n_dipoles; i_a++){
    logger << LOG_DEBUG << "MC_n_channel_phasespace[" << setw(2) << i_a << "] = " << setw(4) << MC_n_channel_phasespace[i_a] << "   MC_sum_channel_phasespace[" << setw(2) << i_a << "] = " << setw(4) << MC_sum_channel_phasespace[i_a] << endline_flush;
  }
  logger << LOG_DEBUG << "MC_n_channel = " << MC_n_channel << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "MC_IS_name_type.size() = " << MC_IS_name_type.size() << endline_flush;


  // first if-statement should be redundant !!!
  //  if (csi->class_contribution_CS_real){
  if (!switch_off_RS_mapping){
    for (int i_a = 1; i_a < n_dipoles; i_a++){ac_tau_psp_dipole(i_a, MC_x_dipole_mapping[i_a]);}
    if (csi->type_contribution == "RRA" || csi->type_contribution == "RRJ"){initialization_fake_dipole_mapping_RRA();}
    initialization_phasespace_MC_x_dipole_RA();
  }
  else if (switch_off_RS_mapping){
    logger << LOG_DEBUG << "csi->type_contribution = " << csi->type_contribution << endline_flush;
    logger << LOG_DEBUG << "csi->no_process_parton[" << 0 << "] = " << csi->no_process_parton[0] << endline_flush;
      ac_tau_psp(0, MC_x_dipole_mapping[0]);
    logger << LOG_DEBUG << "MC_x_dipole_mapping[0].size() = " << MC_x_dipole_mapping[0].size() << endline_flush;
    tau_MC_map = MC_x_dipole_mapping[0];
  }
  //  }

  logger << LOG_DEBUG << "switch_off_RS_mapping = " << switch_off_RS_mapping << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "MC_n_channel = " << MC_n_channel << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "tau_MC_map.size() = " << tau_MC_map.size() << endline_flush;

  logger.finished();
}


void phasespace_set::initialization_phasespace_subprocess(){
  static Logger logger("phasespace_set::initialization_phasespace_subprocess");
  logger.called();

  x_pdf.resize(3);

  start_xbp_all[0].resize(xb_max);
  start_xbs_all[0].resize(xb_max);
  start_xbsqrts_all[0].resize(xb_max);

  if (csi->coll_choice == 0){
    fourvector p1(E, 0, 0, E), p2(E, 0, 0, -E);
    start_xbp_all[0][0] = p1 + p2;
    start_xbp_all[0][1] = p1;
    start_xbp_all[0][2] = p2;
    for (int i = 0; i < 3; i++){x_pdf[i] = 1.;}
    start_xbs_all[0][0] = start_xbp_all[0][0].m2();
    start_xbsqrts_all[0][0] = sqrt(start_xbs_all[0][0]);
    start_xbp_all[0][xb_max - 4] = start_xbp_all[0][0];
    start_xbs_all[0][xb_max - 4] = start_xbs_all[0][0];
    start_xbsqrts_all[0][xb_max - 4] = start_xbsqrts_all[0][0];
    start_xbp_all[0][3] = start_xbp_all[0][0];
    start_xbs_all[0][3] = start_xbs_all[0][0];
    start_xbsqrts_all[0][3] = start_xbsqrts_all[0][0];
  }

  for (int xi = 1; xi <= csi->n_particle + 2; xi++){
    start_xbs_all[0][intpow(2, xi - 1)] = M2[abs(csi->type_parton[0][xi])];
    start_xbsqrts_all[0][intpow(2, xi - 1)] = M[abs(csi->type_parton[0][xi])];
  }

  sqrtsmin_opt = start_xbsqrts_all;
  smin_opt = start_xbs_all;

  xbp_all = start_xbp_all;
  xbs_all = start_xbs_all;
  xbsqrts_all = start_xbsqrts_all;

  logger.finished();
}


void phasespace_set::initialization_phasespace_subprocess_optimization(){
  static Logger logger("phasespace_set::initialization_phasespace_subprocess_optimization");
  logger.called();

  logger << LOG_DEBUG << "sqrtsmin_opt[0].size() = " << sqrtsmin_opt[0].size() << endline_flush;
  logger << LOG_DEBUG << "Before adaptation:  tau_0 = " << tau_0 << "   tau_0_s_had = " << tau_0_s_had << "   -> sqrt(s^_min) = " << 2 * sqrt(tau_0) * E << " (tau_0)" << " = " << sqrt(tau_0_s_had) << " (tau_0_s_had)" << endline_flush;

  for (int i_x = 4; i_x < sqrtsmin_opt[0].size(); i_x += 4){
    if (sqrtsmin_opt[0][i_x] == start_xbsqrts_all[0][i_x]){}
    else if (sqrtsmin_opt[0][i_x] < start_xbsqrts_all[0][i_x]){sqrtsmin_opt[0][i_x] = start_xbsqrts_all[0][i_x];}
    else if (sqrtsmin_opt[0][i_x] > start_xbsqrts_all[0][i_x] && vectorbinary_from_binary(i_x).size() == 1){
      logger << LOG_FATAL << "phase-space cut is in contradiction to on-shell conditions of intermediate particle " << i_x << "!" << endline_flush;
      exit(1);
    }
  }

  for (int i_x = 4; i_x < sqrtsmin_opt[0].size(); i_x += 4){
    if (vectorbinary_from_binary(i_x).size() > 1){
      double temp_sqrts_opt = max_subset_from_binary(sqrtsmin_opt[0], i_x);
      if (temp_sqrts_opt > sqrtsmin_opt[0][i_x]){sqrtsmin_opt[0][i_x] = temp_sqrts_opt;}
      smin_opt[0][i_x] = pow(sqrtsmin_opt[0][i_x], 2);
    }
  }
  if (tau_0_s_had < smin_opt[0][smin_opt[0].size() - 4]){
    tau_0_s_had = smin_opt[0][smin_opt[0].size() - 4];
    tau_0 = tau_0_s_had / s_had;
  }

  for (int i = 4; i < smin_opt[0].size(); i += 4){
    stringstream temp_ss;
    temp_ss << "sqrtsmin_opt[0][" << setw(3) << i << "] = " << setw(15) << setprecision(8) << sqrtsmin_opt[0][i] << "   smin_opt[0][" << setw(3) << i << "] = " << setw(15) << setprecision(8) << smin_opt[0][i];
    if (sqrtsmin_opt[0][i] != 0.){logger << LOG_DEBUG << temp_ss.str() << endline_flush;}
    else {logger << LOG_DEBUG_VERBOSE << temp_ss.str() << endline_flush;}
  }
  logger.newLine(LOG_DEBUG);
  for (int i = 0; i < start_xbp_all[0].size(); i++){
    if (start_xbs_all[0][i] != 0. || start_xbsqrts_all[0][i] != 0.){
      logger << LOG_DEBUG << "xbs[" << setw(3) << right << i << "] = " << start_xbs_all[0][i] << "   " << "xbsqrts[" << setw(3) << right << i << "] = " << start_xbsqrts_all[0][i] << "   " << endline_flush;
    }
  }
  logger.newLine(LOG_DEBUG);
  logger << LOG_DEBUG << "tau_0 = " << tau_0 << "   tau_0_s_had = " << tau_0_s_had << "   -> sqrt(s^_min) = " << 2 * sqrt(tau_0) * E << " (tau_0)" << " = " << sqrt(tau_0_s_had) << " (tau_0_s_had)" << endline_flush;
  logger.newLine(LOG_DEBUG);
  
  logger.finished();
}


void phasespace_set::initialization_phasespace_subprocess_dipole_IS_RA(){
  static Logger logger("phasespace_set::initialization_phasespace_subprocess_dipole_IS_RA");
  logger.called();

  for (int i_a = 1; i_a < n_dipoles; i_a++){
    logger << LOG_DEBUG_VERBOSE << "MC_IS_name_type[" << i_a << "].size() = " << MC_IS_name_type[i_a].size() << endline_flush;
    MC_IS_name_type[i_a][6].push_back("x of dipole " + csi->dipole[i_a].name());
    MC_IS_name_type[i_a][7].push_back("z of dipole " + csi->dipole[i_a].name());
    MC_IS_name_type[i_a][8].push_back("phi of dipole " + csi->dipole[i_a].name());
  }

  logger.finished();
}


void phasespace_set::initialization_phasespace_subprocess_dipole_RA(){
  static Logger logger("phasespace_set::initialization_phasespace_subprocess_dipole_RA");
  logger.called();

  x_pdf.resize(3);

  vector<vector<vector<double> > > dx_s_RA(n_dipoles);
  vector<vector<vector<double> > > dx_sqrts_RA(n_dipoles);
  for (int i_a = 0; i_a < dx_s_RA.size(); i_a++){
    dx_s_RA[i_a].resize(csi->dipole[i_a].type_parton().size(), vector<double> (1));
    dx_sqrts_RA[i_a].resize(csi->dipole[i_a].type_parton().size(), vector<double> (1));
  }

  start_xbp_all.resize(n_dipoles);
  start_xbs_all.resize(n_dipoles);
  start_xbsqrts_all.resize(n_dipoles);

  start_xbp_all[0].resize(xb_max);
  start_xbs_all[0].resize(xb_max);
  start_xbsqrts_all[0].resize(xb_max);
  for (int i_a = 1; i_a < n_dipoles; i_a++){
    start_xbp_all[i_a].resize(xb_max_dipoles);
    start_xbs_all[i_a].resize(xb_max_dipoles);
    start_xbsqrts_all[i_a].resize(xb_max_dipoles);
  }

  for (int i_a = 0; i_a < dx_s_RA.size(); i_a++){
    for (int i_b = 1; i_b < dx_s_RA[i_a].size(); i_b++){
      if (dx_s_RA[i_a][i_b].size() == 1){
	dx_s_RA[i_a][i_b][0] = M2[abs(csi->dipole[i_a].type_parton()[i_b])];
	dx_sqrts_RA[i_a][i_b][0] = M[abs(csi->dipole[i_a].type_parton()[i_b])];
      }
    }
  }

  logger << LOG_DEBUG << "definitions of particle momenta and masses" << endline_flush;

  for (int i_a = 0; i_a < n_dipoles; i_a++){
    for (int xbi = 0; xbi < start_xbp_all[i_a].size(); xbi++){
      start_xbp_all[i_a][xbi] = nullvector;
      start_xbs_all[i_a][xbi] = 0.;
      start_xbsqrts_all[i_a][xbi] = 0.;
    }
    int max_particles;
    if (i_a == 0){max_particles = csi->n_particle + 2 + 1;}
    else {max_particles = csi->n_particle + 2;}
    for (int xbi = 1; xbi < max_particles; xbi++){
      start_xbs_all[i_a][intpow(2, xbi - 1)] = dx_s_RA[i_a][xbi][0];
      start_xbsqrts_all[i_a][intpow(2, xbi - 1)] = dx_sqrts_RA[i_a][xbi][0];
    }
  }

  if (csi->coll_choice == 0){
    fourvector p1(E, 0, 0, E), p2(E, 0, 0, -E);
    for (int i = 0; i < 3; i++){x_pdf[i] = 1.;}

    for (int i_a = 0; i_a < n_dipoles; i_a++){
      logger << LOG_DEBUG << i_a << endline_flush;
      start_xbp_all[i_a][0] = p1 + p2;
      start_xbp_all[i_a][1] = p1;
      start_xbp_all[i_a][2] = p2;
      //      for (int i_p = 0; i_p < 3; i_p++){start_xbp_all[i_a][i_p] = start_xbp_all[0][i_p];}
      start_xbs_all[i_a][0] = start_xbp_all[0][0].m2();
      start_xbsqrts_all[i_a][0] = sqrt(start_xbs_all[i_a][0]);
      start_xbp_all[i_a][3] = start_xbp_all[i_a][0];
      start_xbs_all[i_a][3] = start_xbs_all[i_a][0];
      start_xbsqrts_all[i_a][3] = start_xbsqrts_all[i_a][0];
      if (i_a == 0){
	start_xbp_all[i_a][xb_max - 4] = start_xbp_all[i_a][0];
	start_xbs_all[i_a][xb_max - 4] = start_xbs_all[i_a][0];
	start_xbsqrts_all[i_a][xb_max - 4] = start_xbsqrts_all[i_a][0];
      }
      else {
	start_xbp_all[i_a][xb_max_dipoles - 4] = start_xbp_all[i_a][0];
	start_xbs_all[i_a][xb_max_dipoles - 4] = start_xbs_all[i_a][0];
	start_xbsqrts_all[i_a][xb_max_dipoles - 4] = start_xbsqrts_all[i_a][0];
      }
    }
  }

  for (int i_a = 0; i_a < n_dipoles; i_a++){
    for (int i_x = 4; i_x < start_xbs_all[i_a].size(); i_x += 4){
      vector<int> temp_vb = vectorbinary_from_binary(i_x);
      logger << LOG_DEBUG << "start_xbsqrts_all[" << setw(2) << i_a << "][" << setw(3) << i_x << "] = " << left << setw(23) << setprecision(15) << start_xbsqrts_all[i_a][i_x] << "   start_xbs_all[" << setw(2) << i_a << "][" << setw(3) << i_x << "] = " << left << setw(23) << setprecision(15) << start_xbs_all[i_a][i_x] << endline_flush;
      if (temp_vb.size() == 1 || start_xbs_all[i_a][i_x] != 0. || start_xbsqrts_all[i_a][i_x] != 0.){
	logger << LOG_DEBUG << "start_xbsqrts_all[" << setw(2) << i_a << "][" << setw(3) << i_x << "] = " << left << setw(23) << setprecision(15) << start_xbsqrts_all[i_a][i_x] << "   start_xbs_all[" << setw(2) << i_a << "][" << setw(3) << i_x << "] = " << left << setw(23) << setprecision(15) << start_xbs_all[i_a][i_x] << endline_flush;
      }
    }
  }

  sqrtsmin_opt = start_xbsqrts_all;
  smin_opt = start_xbs_all;

  logger.finished();
}


void phasespace_set::initialization_phasespace_subprocess_optimization_dipole_RA(){
  static Logger logger("phasespace_set::initialization_phasespace_subprocess_dipole_optimization_RA");
  logger.called();

  for (int i_a = 0; i_a < n_dipoles; i_a++){
    for (int i_x = 4; i_x < sqrtsmin_opt[i_a].size(); i_x += 4){
      if (sqrtsmin_opt[i_a][i_x] == start_xbsqrts_all[i_a][i_x]){}
      else if (sqrtsmin_opt[i_a][i_x] < start_xbsqrts_all[i_a][i_x]){sqrtsmin_opt[i_a][i_x] = start_xbsqrts_all[i_a][i_x];}
      else if (sqrtsmin_opt[i_a][i_x] > start_xbsqrts_all[i_a][i_x]){
	if (vectorbinary_from_binary(i_x).size() == 1){
	  logger << LOG_FATAL << "phase-space cut is in contradiction to on-shell conditions of intermediate particle " << i_x << "!" << endline_flush;
	  exit(1);
	}
      }
    }
  }

  for (int i_a = 0; i_a < n_dipoles; i_a++){
    for (int i_x = 4; i_x < sqrtsmin_opt[i_a].size(); i_x += 4){
      if (vectorbinary_from_binary(i_x).size() > 1){
	double temp_sqrts_opt = max_subset_from_binary(sqrtsmin_opt[i_a], i_x);
	if (temp_sqrts_opt > sqrtsmin_opt[i_a][i_x]){sqrtsmin_opt[i_a][i_x] = temp_sqrts_opt;}
	smin_opt[i_a][i_x] = pow(sqrtsmin_opt[i_a][i_x], 2);
      }
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG){
    for (int i_a = 0; i_a < n_dipoles; i_a++){
      for (int i = 4; i < smin_opt[i_a].size(); i += 4){
	stringstream temp_ss;
	temp_ss << "sqrtsmin_opt[" << i_a << "][" << setw(3) << i << "] = " << setw(15) << setprecision(8) << sqrtsmin_opt[i_a][i] << "   smin_opt[" << i_a << "][" << setw(3) << i << "] = " << setw(15) << setprecision(8) << smin_opt[i_a][i];
	if (sqrtsmin_opt[i_a][i] != 0.){logger << LOG_DEBUG << temp_ss.str() << endline_flush;}
	else {logger << LOG_DEBUG << temp_ss.str() << endline_flush;}
      }
    }
    logger.newLine(LOG_DEBUG);
    for (int i_a = 0; i_a < n_dipoles; i_a++){
      for (int i_x = 4; i_x < smin_opt[i_a].size(); i_x += 4){
	vector<int> temp_vb = vectorbinary_from_binary(i_x);
	if (temp_vb.size() == 1 || start_xbs_all[i_a][i_x] != 0. || start_xbsqrts_all[i_a][i_x] != 0.){
	  logger << LOG_DEBUG << "start_xbsqrts_all[" << setw(2) << i_a << "][" << setw(3) << i_x << "] = " << left << setw(23) << setprecision(15) << start_xbsqrts_all[i_a][i_x] << "   start_xbs_all[" << setw(2) << i_a << "][" << setw(3) << i_x << "] = " << left << setw(23) << setprecision(15) << start_xbs_all[i_a][i_x] << endline_flush;
	}
      }
    }
    logger.newLine(LOG_DEBUG);
  }
  
  vector<vector<vector<double> > > dx_s_RA(n_dipoles);
  vector<vector<vector<double> > > dx_sqrts_RA(n_dipoles);
  for (int i_a = 0; i_a < dx_s_RA.size(); i_a++){
    dx_s_RA[i_a].resize(csi->dipole[i_a].type_parton().size(), vector<double> (1));
    dx_sqrts_RA[i_a].resize(csi->dipole[i_a].type_parton().size(), vector<double> (1));
  }
  for (int i_a = 0; i_a < dx_s_RA.size(); i_a++){
    for (int i_b = 1; i_b < dx_s_RA[i_a].size(); i_b++){
      if (dx_s_RA[i_a][i_b].size() == 1){
	dx_s_RA[i_a][i_b][0] = M2[abs(csi->dipole[i_a].type_parton()[i_b])];
	dx_sqrts_RA[i_a][i_b][0] = M[abs(csi->dipole[i_a].type_parton()[i_b])];
      }
    }
  }

  dipole_sinx_min.resize(n_dipoles, 0.);
  for (int i_a = 0; i_a < dx_s_RA.size(); i_a++){
    for (int ib = 3; ib < dx_s_RA[i_a].size(); ib++){dipole_sinx_min[i_a] += M[abs(csi->dipole[i_a].type_parton()[ib])];}
    dipole_sinx_min[i_a] = pow(dipole_sinx_min[i_a], 2);
    logger << LOG_DEBUG << "smin_opt (only from masses):   dipole_sinx_min[" << setw(2) << i_a << "] = " << dipole_sinx_min[i_a] << endline_flush;
    dipole_sinx_min[i_a] = smin_opt[i_a][smin_opt[i_a].size() - 4];
    logger << LOG_DEBUG << "smin_opt (from smin_opt[" << i_a << "][" << smin_opt[i_a].size() - 4 << "]):    dipole_sinx_min[" << setw(2) << i_a << "] = " << dipole_sinx_min[i_a] << endline_flush;
  }

  logger << LOG_DEBUG << "tau_0 = " << tau_0 << "   tau_0_s_had = " << tau_0_s_had << "   -> sqrt(s^_min) = " << 2 * sqrt(tau_0) * E << " (tau_0)" << " = " << sqrt(tau_0_s_had) << " (tau_0_s_had)" << endline_flush;
  // Not only tau_0, but also tau_0_s_had need to be updated !!!
  if (tau_0_s_had < smin_opt[0][smin_opt[0].size() - 4]){
    tau_0_s_had = smin_opt[0][smin_opt[0].size() - 4];
    tau_0 = tau_0_s_had / s_had;
  }
  logger << LOG_DEBUG << "tau_0 = " << tau_0 << "   tau_0_s_had = " << tau_0_s_had << "   -> sqrt(s^_min) = " << 2 * sqrt(tau_0) * E << " (tau_0)" << " = " << sqrt(tau_0_s_had) << " (tau_0_s_had)" << endline_flush;
  logger.newLine(LOG_DEBUG);

  xbp_all = start_xbp_all;
  xbs_all = start_xbs_all;
  xbsqrts_all = start_xbsqrts_all;

  corrected_xbp_all = start_xbp_all;
  corrected_xbs_all = start_xbs_all;
  corrected_xbsqrts_all = start_xbsqrts_all;

  logger.finished();
}


// ???
void phasespace_set::initialization_phasespace_MC_tau(){
  static Logger logger("phasespace_set::initialization_phasespace_MC_tau");
  logger.called();

  // only non-output statements in initialization_phasespace_MC_tau (which are not repeated here):
  logger << LOG_DEBUG << "tau_MC_map.size() = " << tau_MC_map.size() << endline_flush;
  for (int i_m = tau_MC_map.size() - 1; i_m > 0; i_m--){
    logger << LOG_DEBUG << "map_Gamma[tau_MC_map[i_m = " << i_m << "] = " << tau_MC_map[i_m] << "] = " << map_Gamma[abs(tau_MC_map[i_m])] << endline_flush;
    if (map_Gamma[abs(tau_MC_map[i_m])] == 0.){tau_MC_map.erase(tau_MC_map.begin() + i_m);}
    else if (2 * sqrt(tau_0) * E > M[abs(tau_MC_map[i_m])]){
      logger << LOG_DEBUG << "2 * sqrt(tau_0) * E = " << 2 * sqrt(tau_0) * E << " > " << M[abs(tau_MC_map[i_m])] << " = M[abs(tau_MC_map[i_m = " << i_m << "] = " << tau_MC_map[i_m] << ")]" << endline_flush;
      tau_MC_map.erase(tau_MC_map.begin() + i_m);
    }
  }
  logger << LOG_DEBUG << "tau_MC_map.size() = " << tau_MC_map.size() << endline_flush;

  // to be shifted elsewhere - used so far somewhere...
  if (input_MC_tau.switch_optimization == -1){tau_MC_map.resize(1);}

  logger.finished();
}


void phasespace_set::initialization_phasespace_MC_x_dipole_RA(){
  static Logger logger("phasespace_set::initialization_phasespace_MC_x_dipole_RA");
  logger.called();

  // determine tau_MC_map from MC_x_dipole_mapping (including 0!)
  for (int i_a = 0; i_a < n_dipoles; i_a++){
    for (int i_m = MC_x_dipole_mapping[i_a].size() - 1; i_m > 0; i_m--){
      /*
      logger << LOG_DEBUG << "MC_x_dipole_mapping[" << i_a << "][" << i_m << "] = " << MC_x_dipole_mapping[i_a][i_m] << endline_flush;
      logger << LOG_DEBUG << "psi_M[" << abs(MC_x_dipole_mapping[i_a][i_m]) << "] = " << M[abs(MC_x_dipole_mapping[i_a][i_m])] << endline_flush;
      logger << LOG_DEBUG << "psi_Gamma[" << abs(MC_x_dipole_mapping[i_a][i_m]) << "] = " << Gamma[abs(MC_x_dipole_mapping[i_a][i_m])] << endline_flush;
      logger << LOG_DEBUG << "psi_map_Gamma[" << abs(MC_x_dipole_mapping[i_a][i_m]) << "] = " << map_Gamma[abs(MC_x_dipole_mapping[i_a][i_m])] << endline_flush;
      */
      if (map_Gamma[abs(MC_x_dipole_mapping[i_a][i_m])] == 0.){MC_x_dipole_mapping[i_a].erase(MC_x_dipole_mapping[i_a].begin() + i_m);}
      else if (2 * sqrt(tau_0) * E > M[abs(MC_x_dipole_mapping[i_a][i_m])]){MC_x_dipole_mapping[i_a].erase(MC_x_dipole_mapping[i_a].begin() + i_m);}
      // leads to problems if M_res < smin !!! needs to be investigated !!!
      //      else if (2 * sqrt(tau_0) * E > M[abs(MC_x_dipole_mapping[i_a][i_m])] + 5 * map_Gamma[abs(MC_x_dipole_mapping[i_a][i_m])]){MC_x_dipole_mapping[i_a].erase(MC_x_dipole_mapping[i_a].begin() + i_m);}
    }

    // to add MC_x_dipole mappings also to MC_tau
    for (int i_m = 0; i_m < MC_x_dipole_mapping[i_a].size(); i_m++){
      int flag = tau_MC_map.size();
      for (int j_m = 0; j_m < tau_MC_map.size(); j_m++){if (MC_x_dipole_mapping[i_a][i_m] == tau_MC_map[j_m]){flag = j_m; break;}}
      if (flag == tau_MC_map.size()){tau_MC_map.push_back(MC_x_dipole_mapping[i_a][i_m]);}
    }
  }

  MC_x_dipole.resize(n_dipoles);
  for (int i_a = 1; i_a < n_dipoles; i_a++){
    if (input_MC_x_dipole.switch_optimization == -1){MC_x_dipole_mapping[i_a].resize(1);}
    stringstream temp_ss;
    temp_ss << "MC_x_dipole_" << i_a;
    input_MC_x_dipole.name = temp_ss.str();
    input_MC_x_dipole.n_channel = MC_x_dipole_mapping[i_a].size();
    MC_x_dipole[i_a] = multichannel_set(input_MC_x_dipole, filename_weight_writeout, filename_weight_readin);
  }

  logger.finished();
}


void phasespace_set::initialization_phasespace_IS(){
  static Logger logger("phasespace_set::initialization_phasespace_IS");
  logger.called();

  g_z_coll.resize(3); // could be deleted later... !!!

  z1z2_opt_end = -1; // not needed here - only in CA (or qTsubtraction)

  input_MC_tau.n_channel = tau_MC_map.size();
  MC_tau = multichannel_set(input_MC_tau, filename_weight_writeout, filename_weight_readin);

  // Deactivate "IS_tau" if tau is fixed:
  if (start_xbs_all[0][xb_max - 4] != 0.){input_IS_tau.switch_optimization = -1;}

  if (csi->class_contribution_collinear){
    for (int i_c = 1; i_c < 3; i_c++){
      stringstream temp_name_ss;
      temp_name_ss << "IS_z1z2[" << i_c << "]";
      string temp_name = temp_name_ss.str();
      input_IS_z1z2.name = temp_name;
    }

    z_coll.resize(3);
    g_z_coll.resize(3);

    xz_pdf.resize(3, vector<double> (2, 0.));
  }
  else {xz_pdf.resize(3, vector<double> (1, 0.));}


  // determination of global variable 'active_optimization':
  // active_optimization:
  // active_optimization = -1: no optimization of any mapping (very unlikely)
  // active_optimization =  0: no active optimization (like in standard runs)
  // active_optimization =  1: at least one active optimization (like in grid runs)
  active_optimization = -1;
  if (MC_phasespace.active_optimization != -1){if (MC_phasespace.active_optimization > active_optimization){active_optimization = MC_phasespace.active_optimization;}}
  logger << LOG_DEBUG << "MC_phasespace.active_optimization = " << MC_phasespace.active_optimization << endline_flush;
  if (MC_tau.active_optimization != -1){if (MC_tau.active_optimization > active_optimization){active_optimization = MC_tau.active_optimization;}}
  logger << LOG_DEBUG << "MC_tau.active_optimization = " << MC_tau.active_optimization << endline_flush;
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      logger << LOG_DEBUG << "MC_x_dipole[" << i_a << "].active_optimization = " << MC_x_dipole[i_a].active_optimization << endline_flush;
      if (MC_x_dipole[i_a].active_optimization != -1){if (MC_x_dipole[i_a].active_optimization > active_optimization){active_optimization = MC_x_dipole[i_a].active_optimization;}}
    }
  }

  for (int i_r = 0; i_r < random_psp.size(); i_r++){
    if (random_psp[i_r]->IS_PS.active_optimization != -1){if (random_psp[i_r]->IS_PS.active_optimization > active_optimization){active_optimization = random_psp[i_r]->IS_PS.active_optimization;}}
  }

  logger.finished();
}


void phasespace_set::initialization_phasespace_IS_QT(){
  static Logger logger("phasespace_set::initialization_phasespace_IS_QT");
  logger.called();

  QT_eps = 0;
  QT_random_IS = 0.;

  QT_qt2 = 0.;
  QT_jacqt2 = 0.;
  QT_random_qt2 = 0.;

  logger.finished();
}


void phasespace_set::initialization_MC_rng(){
  static Logger logger("phasespace_set::initialization_MC_rng");
  logger.called();

  //  n_random_MC: 1 for the channel selection, (3n - 4) for the n-particle phasespace:
  n_random_MC = 1 + no_random;

  ///  int temp_switch_IS_MC = -1;
  ///  for (int i_s = 0; i_s < input_IS.size(); i_s++){if (input_IS[i_s]->name.substr(0, 5) == "IS_PS" && input_IS[i_s]->switch_optimization != -1){temp_switch_IS_MC = 0;}}
  ///  if (temp_switch_IS_MC != -1){n_random_MC += no_random;}
  logger << LOG_DEBUG << "input_MC_phasespace.switch_optimization      = " << setw(3) << input_MC_phasespace.switch_optimization << "   " << setw(20) << "n_random_MC" << " = " << n_random_MC << endline_flush;
  // Should be 0 if input_MC_phasespace.switch_optimization == -1 !!!
  ///  n_random_per_psp += n_random_MC;
  // Always request two random numbers per variable for the final-particle phase space!
  n_random_per_psp += 2 * n_random_MC;

  //  n_random_MC_tau: 1 for the MC_tau channel selection:
  //  n_random_MC_tau: 2 set instead in order to reproduce old results !!!
  n_random_MC_tau = 2;
  logger << LOG_DEBUG << "input_MC_tau.switch_optimization  = " << setw(3) << input_MC_tau.switch_optimization << "   " << setw(20) << "n_random_MC_tau" << " = " << n_random_MC_tau << endline_flush;
  // Should be 0 if input_MC_tau.switch_optimization == -1 !!!
  random_MC_tau.resize(n_random_MC_tau);
  n_random_per_psp += n_random_MC_tau;

  //  n_random_tau: 1 for the tau IS selection, 1 random number for the tau value:
  n_random_tau = 2;
  logger << LOG_DEBUG << "input_IS_tau.switch_optimization  = " << setw(3) << input_IS_tau.switch_optimization << "   " << setw(20) << "n_random_tau   " << " = " << n_random_tau << endline_flush;
  // Should be 1 if input_IS_tau.switch_optimization == -1 !!!
  random_tau.resize(n_random_tau);
  n_random_per_psp += n_random_tau;

  //  n_random_x1x2: 1 for the x1x2 IS selection, 1 random number for the x1x2 value:
  n_random_x12 = 2;
  logger << LOG_DEBUG << "input_IS_x1x2.switch_optimization = " << setw(3) << input_IS_x1x2.switch_optimization << "   " << setw(20) << "n_random_x12  " << " = " << n_random_x12 << endline_flush;
  // Should be 1 if input_IS_x1x2.switch_optimization == -1 !!!
  random_x12.resize(n_random_x12);
  // number of random numbers for x12 mapping:
  n_random_per_psp += n_random_x12;


  // new: Should be used for all collinear-emission contributions (i.e. with z1/z2 emissions):
  random_z.resize(3, vector<double> (2));
  if (csi->class_contribution_collinear){
    //  n_random_z: 1 for the z1/z2 IS selection, 1 random number for the z1/z2 value:
    n_random_z = 2;
    logger << LOG_DEBUG << "input_IS_z1z2.switch_optimization = " << setw(3) << input_IS_z1z2.switch_optimization << "   " << setw(20) << "n_random_z     " << " = " << n_random_z << " * 2 " << endline_flush;
    for (int i_z = 1; i_z < 3; i_z++){
      random_z[i_z].resize(n_random_z);
      // number of random numbers for z1/2 mapping:
      n_random_per_psp += n_random_z;
    }
  }

  if (csi->class_contribution_QT_powercorrection){
    n_random_qT = 3; // optimization for |qT|, flat for phi
    random_qT.resize(n_random_qT);
    // number of random numbers for 2-dimensional qT mapping (QT, phi):
    n_random_per_psp += n_random_qT;
  }

  // in case of resummation only !!!
  if (switch_resummation){
    n_random_qTres = 0; // default !!!
    logger << LOG_DEBUG << "input_IS_qTres.switch_optimization = " << setw(3) << input_IS_qTres.switch_optimization << "   " << setw(20) << "n_random_qTres  " << " = " << n_random_qTres << endline_flush;
    // Should be 1 if switch_IS_qTres == -1 !!!
    random_qTres.resize(n_random_qTres);
    // number of random numbers for qTres mapping:
    n_random_per_psp += n_random_qTres;
  }

  // to check if identical points are generated as in old run !!!
  n_random_per_psp += add_random_per_psp;

  logger << LOG_DEBUG << "n_random_per_psp = " << n_random_per_psp << endline_flush;

  rng.initialization(n_random_per_psp, n_shift_run, switch_off_random_generator);
  ///  rng.psi = this;

  logger.finished();
}


void phasespace_set::initialization_MC(){
  static Logger logger("phasespace_set::initialization_MC");
  logger.called();

  logger << LOG_DEBUG << "switch_IS_mode_phasespace = " << switch_IS_mode_phasespace << endline_flush;

  random_manager = randommanager(*this);

  initialization_phasespace_IS();

  g_tot = 0.;
  g_MC = 0.;
  if (csi->coll_choice){
    g_pdf = 0.;
    g_tau = 0.;
    g_x1x2 = 0.;
  }
  else {
    g_pdf = 1.;
    g_tau = 1.;
    g_x1x2 = 1.;
  }
  g_global_NWA = 1.;

  MC_g_IS_global = 1.;


  bool switch_readin_weights = false;
  if ((input_MC_phasespace.switch_optimization == 2 || input_MC_phasespace.switch_optimization == 3) ||
      (input_MC_tau.switch_optimization == 2 || input_MC_tau.switch_optimization == 3) ||
      (csi->class_contribution_collinear && (input_MC_x_dipole.switch_optimization == 2 || input_MC_x_dipole.switch_optimization == 3)) ||
      (input_IS_tau.switch_optimization == 2 || input_IS_tau.switch_optimization == 3) ||
      (input_IS_x1x2.switch_optimization == 2 || input_IS_x1x2.switch_optimization == 3) ||
      (csi->class_contribution_collinear && (input_IS_z1z2.switch_optimization == 2 || input_IS_z1z2.switch_optimization == 3)) ||
      (switch_resummation && (input_IS_qTres.switch_optimization == 2 || input_IS_qTres.switch_optimization == 3)) ||
      (csi->class_contribution_QT_powercorrection && (input_IS_qTres.switch_optimization == 2 || input_IS_qTres.switch_optimization == 3))){
    switch_readin_weights = true;
  }
  logger << LOG_DEBUG << "switch_readin_weights = " << switch_readin_weights << endline_flush;
  
  string type_data = "";
  map<string, size_t> position;
  vector<string> readin;
  vector<string> list_identifier;
  
  if (switch_readin_weights){
    get_data(filename_weight_readin, readin);
    if (readin.size() != 0){type_data = "hex";}
    else {
      string temp_filename_readin_dec = filename_weight_readin.insert(filename_weight_readin.size() - 4, "_dec");
      logger << LOG_DEBUG << "filename_readin (dec) = " << temp_filename_readin_dec << endline_flush;
      readin.clear();
      get_data(temp_filename_readin_dec, readin);
      if (readin.size() != 0){type_data = "dec";}
    }
    if (type_data != ""){
      logger << LOG_DEBUG << "type_data = " << type_data << endline_flush;
      for (size_t i_i = 0; i_i < readin.size(); i_i++){
	if (readin[i_i][0] == '%'){
	  string identifier = readin[i_i].substr(2, readin[i_i].size() - 2);
	  for (size_t i_s = 0; i_s < identifier.size(); i_s++){
	    if (identifier[identifier.size() - 1 - i_s] == ' '){identifier.erase(identifier.size() - 1);}
	    else {break;}
	  }
	  position[identifier] = i_i + 1;
	  list_identifier.push_back(identifier);
	}
      }
    }
  }
  
  for (int i_l = 0; i_l < list_identifier.size(); i_l++){
    logger << LOG_DEBUG << "list_identifier[" << setw(3) << i_l << "] = " << setw(50) << list_identifier[i_l] << " -> " << position[list_identifier[i_l]] << endline_flush;
  }
  

  int x_random = 0;

  n_IS_name_list = n_general_IS_name_list + n_MC_IS_name_list;
  random_manager.random_psp.resize(n_IS_name_list);
  random_psp.resize(n_IS_name_list);

  for (int i_r = 0; i_r < general_IS_name_list.size(); i_r++){
    //    logger << LOG_DEBUG << "general_IS_name_list[" << i_r << "] = " << general_IS_name_list[i_r] << "   "
    if (general_IS_name_list[i_r] == "IS_tau"){
      random_manager.random_psp[x_random + i_r] = randomvariable(input_IS_tau, general_IS_name_list[i_r], x_random + i_r, &random_manager);
      index_IS_tau = x_random + i_r;
      random_psp[index_IS_tau] = & random_manager.random_psp[index_IS_tau];
      IS_tau = & random_manager.random_psp[index_IS_tau].IS_PS;
    }
    else if (general_IS_name_list[i_r] == "IS_x1x2"){
      random_manager.random_psp[x_random + i_r] = randomvariable(input_IS_x1x2, general_IS_name_list[i_r], x_random + i_r, &random_manager);
      index_IS_x1x2 = x_random + i_r;
      random_psp[index_IS_x1x2] = & random_manager.random_psp[index_IS_x1x2];
      IS_x1x2 = & random_manager.random_psp[index_IS_x1x2].IS_PS;
    }
    else if (general_IS_name_list[i_r] == "IS_z1z2[1]" || general_IS_name_list[i_r] == "IS_z1z2[2]"){
      random_manager.random_psp[x_random + i_r] = randomvariable(input_IS_z1z2, general_IS_name_list[i_r], x_random + i_r, &random_manager);
      if (general_IS_name_list[i_r] == "IS_z1z2[1]"){
	index_IS_z1z2[1] = x_random + i_r;
	random_psp[index_IS_z1z2[1]] = & random_manager.random_psp[index_IS_z1z2[1]];
	IS_z1z2[1] = & random_manager.random_psp[index_IS_z1z2[1]].IS_PS;
      }
      if (general_IS_name_list[i_r] == "IS_z1z2[2]"){
	index_IS_z1z2[2] = x_random + i_r;
	random_psp[index_IS_z1z2[2]] = & random_manager.random_psp[index_IS_z1z2[2]];
	IS_z1z2[2] = & random_manager.random_psp[index_IS_z1z2[2]].IS_PS;
      }
    }
    else if (general_IS_name_list[i_r] == "IS_qTres"){
      random_manager.random_psp[x_random + i_r] = randomvariable(input_IS_qTres, general_IS_name_list[i_r], x_random + i_r, &random_manager);
      index_IS_qTres = x_random + i_r;
      random_psp[index_IS_qTres] = & random_manager.random_psp[index_IS_qTres];
      IS_qTres = & random_manager.random_psp[index_IS_qTres].IS_PS;
    }
    
    random_manager.initialization_random_psp(x_random + i_r);
    random_psp[x_random + i_r] = & random_manager.random_psp[x_random + i_r];
  }

  
  logger << LOG_DEBUG << "switch_IS_mode_phasespace = " << switch_IS_mode_phasespace << endline_flush;
  if (switch_IS_mode_phasespace == 1 || switch_IS_mode_phasespace == 2){
    /*// options should be removed !!!
    // create random variables for phase space integration with grids for each variable in each channel (MC_n_channel * (3n - 4))
    random_manager.random_psp.resize(no_random * MC_n_channel);
    random_psp.resize(no_random * MC_n_channel);
    ///    phasespace_randoms.resize(no_random * MC_n_channel);
    string tmp_name;
    ostringstream convert;
    for (int i = 0; i < no_random * MC_n_channel; i++) {
      convert.str(string());
      convert << "r" << i;
      tmp_name = convert.str();
      logger << LOG_DEBUG << "initialize randomvariable " << i << endline_flush;
      // needs to be replaced by new input !!!
      random_manager.random_psp[i] = randomvariable(n_IS_events, switch_IS_MC, n_IS_steps, n_IS_gridsize, tmp_name);
      ///      phasespace_randoms[i] = new randomvariable(n_IS_events, switch_IS_MC, n_IS_steps, n_IS_gridsize, tmp_name);
      ///      random_manager.register_variable(phasespace_randoms[i], true);
      random_manager.initialization_random_psp(i);
      random_psp[i] = & random_manager.random_psp[i];
      random_psp[i]->initialization();
      
      ///      random_manager.register_variable(phasespace_randoms[i]);
    }
    //    container_IS_switch.resize(container_IS_name.size(), 0);

    // Seems to be required, although it is set already in randommanager !!! ???
    ///    random_manager.psi =& *this;
    //    psi.phasespace_randoms=&phasespace_randoms;
    */
  }
  else if (switch_IS_mode_phasespace == 3 || switch_IS_mode_phasespace == 4){
    // create random variables for phase space integration with one grid only for each mapping (propagator, t-channel, etc.)

    map<size_t, size_t> map_MC_IS_type_to_input_IS;
    for (int i_s = 0; i_s < input_IS.size(); i_s++){
      if (input_IS[i_s]->name == input_IS_PS_p.name){map_MC_IS_type_to_input_IS[0] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_f.name){map_MC_IS_type_to_input_IS[1] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_t_t.name){map_MC_IS_type_to_input_IS[2] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_t_phi.name){map_MC_IS_type_to_input_IS[3] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_d_costheta.name){map_MC_IS_type_to_input_IS[4] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_d_phi.name){map_MC_IS_type_to_input_IS[5] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_dipole_xy.name){map_MC_IS_type_to_input_IS[6] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_dipole_uvz.name){map_MC_IS_type_to_input_IS[7] = i_s;}
      if (input_IS[i_s]->name == input_IS_PS_dipole_phi.name){map_MC_IS_type_to_input_IS[8] = i_s;}
    }

    for (int i_s = 0; i_s < 6; i_s++){logger << LOG_DEBUG << "map_MC_IS_type_to_input_IS[" << i_s << "] = " << map_MC_IS_type_to_input_IS[i_s] << endline_flush;}

    string tmp_name;
    ostringstream convert;

    /// new version which allows different grid sizes for different observables:
    //    int x_random = 0;
    for (int i_a = 0; i_a < MC_IS_name_type.size(); i_a++){
      for (int i_t = 0; i_t < MC_IS_name_type[i_a].size(); i_t++){
	for (int i_i = 0; i_i < MC_IS_name_type[i_a][i_t].size(); i_i++){
	  int y_r = n_general_IS_name_list + x_random;
	  if (y_r != map_type_to_list_MC_IS[vector<int> {i_a, i_t, i_i}]){logger << LOG_ERROR << "map_type_to_list_MC_IS[" << i_a << "][" << i_t << "][" << i_i << "] is not counted correctly!" << endline_flush;}
	  convert.str(string());
	  convert << "yr" << x_random;
	  tmp_name = convert.str() + "_" + MC_IS_name_type[i_a][i_t][i_i];
	  logger << LOG_DEBUG << "MC_IS_name_type[" << setw(2) << i_a << "][" << setw(2) << i_t << "][" << setw(4) << i_i << "] = " << MC_IS_name_type[i_a][i_t][i_i] << endline_flush;
	  logger << LOG_DEBUG << "map_MC_IS_type_to_input_IS.size() = " << map_MC_IS_type_to_input_IS.size() << "   i_t = " << i_t << endline_flush;
	  logger << LOG_DEBUG << "input_IS.size() = " << input_IS.size() << "   map_MC_IS_type_to_input_IS[" << i_t << "] = " << map_MC_IS_type_to_input_IS[i_t] << endline_flush;

	  random_manager.random_psp[y_r] = randomvariable(*input_IS[map_MC_IS_type_to_input_IS[i_t]], tmp_name, y_r, &random_manager);
	  random_manager.initialization_random_psp(y_r);
	  random_psp[y_r] = & random_manager.random_psp[y_r];

	  x_random++;
	}
      }
    }
  }
  else {r.resize(no_random + 1);}



  logger << LOG_DEBUG << "random_psp.size() = " << random_psp.size() << endline_flush;
  for (int i_r = 0; i_r < random_psp.size(); i_r++){
    logger << LOG_DEBUG << "random_psp[" << setw(4) << "] -> " << random_psp[i_r]->IS_PS.name << *random_psp[i_r] << endline_flush;
  }
  
 

  // Register all active multichannel optimizations:
  if (MC_phasespace.active_optimization != -1){MC_active.push_back(& MC_phasespace);}
  if (MC_tau.active_optimization != -1){MC_active.push_back(& MC_tau);}
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      if (MC_x_dipole[i_a].active_optimization != -1){MC_active.push_back(& MC_x_dipole[i_a]);}
    }
  }
  
  // Register all active importancesampling optimizations:
  for (int i_r = 0; i_r < random_psp.size(); i_r++){
    //    logger << LOG_DEBUG << "general_IS_name_list[" << i_r << "] = " << general_IS_name_list[i_r] << "   "
    if (random_psp[i_r]->imp_sampling){IS_active.push_back(& random_psp[i_r]->IS_PS);}
  }
  

  
  
  logger << LOG_DEBUG << "Readin of integration grids (MC = multichannel  and  IS = importancesampling)" << endline_flush;

  error_readin = false;
  if (switch_readin_weights){
    if (input_MC_phasespace.switch_optimization == 2 || input_MC_phasespace.switch_optimization == 3){
      MC_phasespace.readin_optimization_data(readin, position, type_data);
      if (MC_phasespace.error_readin){error_readin = true;}
    }
    if (input_MC_tau.switch_optimization == 2 || input_MC_tau.switch_optimization == 3){
      MC_tau.readin_optimization_data(readin, position, type_data);
      if (MC_tau.error_readin){error_readin = true;}
    }
    if (csi->class_contribution_CS_real){
      if (input_MC_x_dipole.switch_optimization == 2 || input_MC_x_dipole.switch_optimization == 3){
	for (int i_a = 1; i_a < n_dipoles; i_a++){
	  MC_x_dipole[i_a].readin_optimization_data(readin, position, type_data);
	  if (MC_x_dipole[i_a].error_readin){error_readin = true;}
	}
      }
    }
    for (int i_r = 0; i_r < random_psp.size(); i_r++){
      if (random_psp[i_r]->IS_PS.switch_optimization == 2 || random_psp[i_r]->IS_PS.switch_optimization == 3){
	random_psp[i_r]->IS_PS.readin_optimization_data(readin, position, type_data);
	if (random_psp[i_r]->IS_PS.error_readin){error_readin = true;}
	logger << LOG_DEBUG_VERBOSE << "random_psp[" << i_r << "] -> " << random_psp[i_r]->IS_PS.error_readin << endline_flush;
      }
    }

    if (error_readin){
      logger << LOG_ERROR << "error_readin -> " << error_readin << endline_flush;
      if (MC_phasespace.error_readin){logger << LOG_ERROR << "MC_phasespace.error_readin = " << MC_phasespace.error_readin << endline_flush;}
      if (MC_tau.error_readin){logger << LOG_ERROR << "MC_tau.error_readin = " << MC_tau.error_readin << endline_flush;}
      if (csi->class_contribution_CS_real){
	for (int i_a = 1; i_a < n_dipoles; i_a++){
	  if (MC_x_dipole[i_a].error_readin){logger << LOG_ERROR << "MC_x_dipole[" << i_a << "].error_readin = " << MC_x_dipole[i_a].error_readin << endline_flush;}
	}
      }
      for (int i_r = 0; i_r < random_psp.size(); i_r++){
	if (random_psp[i_r]->IS_PS.error_readin){logger << LOG_ERROR << "random_psp[" << i_r << "]->IS_PS.error_readin = " << random_psp[i_r]->IS_PS.error_readin << endline_flush;}
      }
      // Add treatment after error messages here ???
    }
  }

  logger.finished();
}



