#include "header.hpp"

// Why not references in many arguments - or inline functions ??? !!!

/////////////////////
//  angle mapping  //
/////////////////////

double phasespace_set::c_phi(double r1){
  static Logger logger("phasespace_set::c_phi");
  logger.called();

  double phi = 2 * pi * r1;

  logger.finished(phi);
  return phi;
}

double phasespace_set::inv_phi(double & phi){
  static Logger logger("phasespace_set::inv_phi");
  logger.called();

  double r_phi = phi / (2 * pi);

  logger.finished(r_phi);
  return r_phi;
}

double phasespace_set::c_costheta(double & r1){
  static Logger logger("phasespace_set::c_costheta");
  logger.called();

  double costheta = 2 * r1 - 1.;

  logger.finished(costheta);
  return costheta;
}

double phasespace_set::inv_costheta(double & costheta){
  static Logger logger("phasespace_set::inv_costheta");
  logger.called();

  double r_costheta = (costheta + 1.) / 2;

  logger.finished(r_costheta);
  return r_costheta;
}


double phasespace_set::c_costheta_pTcut(double & r1, double & costheta_pTcut, double n){
  static Logger logger("phasespace_set::c_costheta");
  logger.called();

  double costheta;
  if (r1 < 0.25){
    costheta = -costheta_pTcut - pow(1. - 4 * r1, n) * (1. - costheta_pTcut);
  }
  else if (r1 < 0.5){
    costheta = -costheta_pTcut + pow(4 * r1 - 1., n) * costheta_pTcut;
  }
  else if (r1 < 0.75){
    costheta = costheta_pTcut - pow(3. - 4 * r1, n) * costheta_pTcut;
  }
  else {
    costheta = costheta_pTcut + pow(4 * r1 - 3., n) * (1. - costheta_pTcut);
  }

  logger.finished(costheta);
  return costheta;
}


double phasespace_set::g_costheta_pTcut(double & costheta, double & costheta_pTcut, double n){
  static Logger logger("phasespace_set::c_costheta");
  logger.called();

  double g_costheta;
  if (costheta < -costheta_pTcut){
    g_costheta = 1. / (4 * n * pow(1. - costheta_pTcut, 1. / n) * pow(-(costheta + costheta_pTcut), 1. - 1. / n));
  }
  else if (costheta < 0.){
    g_costheta = 1. / (4 * n * pow(costheta_pTcut, 1. / n) * pow(costheta_pTcut + costheta, 1. - 1. / n));
  }
  else if (costheta < costheta_pTcut){
    g_costheta = 1. / (4 * n * pow(costheta_pTcut, 1. / n) * pow(costheta_pTcut - costheta, 1. - 1. / n));
  }
  else {
    g_costheta = 1. / (4 * n * pow(1. - costheta_pTcut, 1. / n) * pow(costheta - costheta_pTcut, 1. - 1. / n));
  }

  logger.finished(costheta);
  return g_costheta;
}


double phasespace_set::inv_costheta_pTcut(double & costheta, double & costheta_pTcut, double n){
  static Logger logger("phasespace_set::inv_costheta");
  logger.called();

  double r_costheta;
  if (costheta < -costheta_pTcut){
    r_costheta = 0.25 - 0.25 * pow(-(costheta + costheta_pTcut) / (1. - costheta_pTcut), 1. / n);
  }
  else if (costheta < 0.){
    r_costheta = 0.25 + 0.25 * pow((costheta + costheta_pTcut) / costheta_pTcut, 1. / n);
  }
  else if (costheta < costheta_pTcut){
    r_costheta = 0.75 - 0.25 * pow((costheta_pTcut - costheta) / costheta_pTcut, 1. / n);
  }
  else {
    r_costheta = 0.75 + 0.25 * pow((costheta - costheta_pTcut) / (1. - costheta_pTcut), 1. / n);
  }

  logger.finished(r_costheta);
  return r_costheta;
}


// **************************************************************************
// *                                                                        *
// *  various mappings for x in [0; 1] -> h in [smin; smax]                 *
// *                                                                        *
// **************************************************************************

//  result between smin and 1. - without "cut_technical" to avoid issues close to boundaries

double phasespace_set::h_propto_pot(double r, double smin, double n){
  static Logger logger("h_propto_pot (min to 1.)");
  logger.called();

  if (n == 1.){return smin * exp(-log(smin) * r);}
  else {return pow(pow(smin, 1. - n) + (1. -  pow(smin, 1. - n)) * r, 1. / (1. - n));}
}

double phasespace_set::g_propto_pot(double s, double smin, double n){
  static Logger logger("g_propto_pot (min to 1.)");
  logger.called();

  if (n == 1.){return 1. / (-log(smin) * s);}
  else {return 1. / (((1. - pow(smin, 1. - n)) / (1. - n)) * pow(s, n));}
}

double phasespace_set::inv_propto_pot(double s, double smin, double nu){
  //double inv_propto_pot(double s, double smin, double smax, double nu, double cut_technical){
  static Logger logger("inv_propto_pot (min to 1.)");
  logger.called();

  // temporary !!! switch_console_output_phasespace_issue  should be in psi !!!
  //  int switch_console_output_phasespace_issue = 0;

  double r;
  if (nu != 1.) {
    double sminm0 = pow(smin, 1. - nu);
    r = (pow(s, 1. - nu) - sminm0) / (pow(1., 1. - nu) - sminm0);
  }
  else {r = log(s / smin) / log(1. / smin);}
  return r;
}



double phasespace_set::h_costheta_collinear(double r, double nu){
  static Logger logger("h_costheta_collinear");
  logger.called();

  double s;
  //  s = -2 * pow(r, 2) + 1.;
  s = -2 * pow(r, nu) + 1.;
  
  return s;
}


double phasespace_set::g_costheta_collinear(double costheta, double nu){
  static Logger logger("g_propto_potcostheta_collinear");
  logger.called();

  double gcostheta;
  //  gcostheta = 1. / (4. * sqrt((1. - costheta) / 2.));
  gcostheta = 1. / (2. * nu * pow((1. - costheta) / 2., 1. - 1. / nu));
    
  return gcostheta;
}


double phasespace_set::inv_costheta_collinear(double costheta, double nu){
  static Logger logger("inv_costheta_collinear");
  logger.called();

  double r;
  //  r = sqrt((1. - costheta) / 2.);
  r = pow((1. - costheta) / 2., 1. / nu);

  return r;
}
  

double phasespace_set::h_costheta_anticollinear(double r, double nu){
  static Logger logger("h_costheta_anticollinear");
  logger.called();

  double s;
  //  s = 2 * pow(r, 2) - 1.;
  s = 2 * pow(r, nu) - 1.;
  
  return s;
}


double phasespace_set::g_costheta_anticollinear(double costheta, double nu){
  static Logger logger("g_propto_potcostheta_anticollinear");
  logger.called();

  double gcostheta;
  //  gcostheta = 1. / (4. * sqrt((1. + costheta) / 2.));
  gcostheta = 1. / (2. * nu * pow((1. + costheta) / 2., 1. - 1. / nu));

  return gcostheta;
}


double phasespace_set::inv_costheta_anticollinear(double costheta, double nu){
  static Logger logger("inv_costheta_anticollinear");
  logger.called();

  double r;
  //  r = sqrt((1. + costheta) / 2.);
  r = pow((1. + costheta) / 2., 1. / nu);

  return r;
}
  

//  result between smin and smax
double phasespace_set::h_propto_pot(double r, double smin, double smax, double nu, double cut_technical){
  static Logger logger("h_propto_pot");
  logger.called();

  double s;
  if (nu != 1.){ s = pow(r * pow(smax + cut_technical, 1. - nu) + (1. - r) * pow(smin + cut_technical, 1. - nu), (1. / (1. - nu))) - cut_technical; }
  else { s = exp(r * log(smax + cut_technical) + (1. - r) * log(smin + cut_technical)) - cut_technical; }

  return s;
}

double phasespace_set::g_propto_pot(double s, double smin, double smax, double nu, double cut_technical){
  static Logger logger("g_propto_pot");
  logger.called();

  double gs;
  if (nu != 1.){ gs = (1. - nu) / ((pow(smax + cut_technical, 1. - nu) - pow(smin + cut_technical, 1. - nu)) * pow(s + cut_technical, nu)); }
  else { gs = 1. / ((log(smax + cut_technical) - log(smin + cut_technical)) * (s + cut_technical)); }

  return gs;
}

double phasespace_set::inv_propto_pot(double s, double smin, double smax, double nu, double cut_technical){
  static Logger logger("inv_propto_pot");
  logger.called();

  double r;
  if (nu != 1.) {
    double sminm0 = pow(smin + cut_technical, 1. - nu);
    r = (pow(s + cut_technical, 1. - nu) - sminm0) / (pow(smax + cut_technical, 1. - nu) - sminm0);
  }
  else { r = log((s + cut_technical) / (smin + cut_technical)) / log((smax + cut_technical) / (smin + cut_technical)); }

  if (munich_isnan(r)) {
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "r = " << r <<"; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
      logger << LOG_ERROR << "r is NaN in inv_propto_pot!" << endline_flush;
    }
    //    r = 0.5;
    if (s <= smin){r = cut_technical;}
    else if (s >= smax){r = 1. - cut_technical;}
  }
  else if (r < 0.){
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "r = " << r <<"; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
      logger << LOG_ERROR << "r < 0 in inv_propto_pot!" << endline_flush;
    }
    r = cut_technical;
  }
  else if (r > 1.){
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "r = " << r <<"; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
      logger << LOG_ERROR << "r > 1 in inv_propto_pot!" << endline_flush;
    }
    r = 1. - cut_technical;
  }

  return r;
}


// check difference to h_propto_pot_mod !!! (in particular in nu == 1 case; nu != 1 case is identical) !!!

double phasespace_set::h_propto_pot_mod(double r, double smin, double smax, double nu, double cut_technical){
  static Logger logger("h_propto_pot_mod");
  logger.called();

  double s;
  if (nu != 1.){ s = pow(r * pow(smax + cut_technical, 1. - nu) + (1. - r) * pow(smin + cut_technical, 1. - nu), (1. / (1. - nu))) - cut_technical; }
  else { s = smin + (smax - smin) * (exp(r * log(1. + cut_technical) + (1. - r) * log(cut_technical)) - cut_technical); }
  return s;
}

double phasespace_set::g_propto_pot_mod(double s, double smin, double smax, double nu, double cut_technical){
  static Logger logger("g_propto_pot_mod");
  logger.called();

  double gs;
  if (nu != 1.){ gs = (1. - nu) / ((pow(smax + cut_technical, 1. - nu) - pow(smin + cut_technical, 1. - nu)) * pow(s + cut_technical, nu)); }
  else { gs = 1. / ((log(1. + cut_technical) - log(cut_technical)) * ((s - smin) + cut_technical * (smax - smin))); }
  return gs;
}

// to be adapted !!! in particular for nu == 1. !!!
double phasespace_set::inv_propto_pot_mod(double s, double smin, double smax, double nu, double cut_technical){
  static Logger logger("inv_propto_pot_mod");
  logger.called();

  double r;
  if (nu != 1.) {
    double sminm0 = pow(smin + cut_technical, 1. - nu);
    r = (pow(s + cut_technical, 1. - nu) - sminm0) / (pow(smax + cut_technical, 1. - nu) - sminm0);
  }
  else {r = log((s + cut_technical) / (smin + cut_technical)) / log((smax + cut_technical) / (smin + cut_technical));}

  if (munich_isnan(r)) {
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "r = " << r <<"; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
      logger << LOG_ERROR << "r is NaN in inv_propto_pot_mod!" << endline_flush;
    }
    if (s <= smin){r = cut_technical;}
    else if (s >= smax){r = 1. - cut_technical;}
  }
  else if (r < 0.){
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "r = " << r <<"; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
      logger << LOG_ERROR << "r < 0 in inv_propto_pot_mod!" << endline_flush;
    }
    r = cut_technical;
  }
  else if (r > 1.){
    if (switch_console_output_phasespace_issue){
      logger << LOG_WARN << "r = " << r <<"; smin, s, smax=" << smin << ", " << s << ", " << smax << endline_flush;
      logger << LOG_ERROR << "r > 1 in inv_propto_pot_mod!" << endline_flush;
    }
    r = 1. - cut_technical;
  }

  return r;
}


// unused at the moment !!!
/*
double phasespace_set::h_pot_min(double r, double smin, double smax, double n){
  if (n == 0.){return smax * exp(r * log(smin / smax));}
  else {return (smax - smin) * pow(r, n) + smin;}
}

double phasespace_set::g_pot_min(double s, double smin, double smax, double n){
  if (n == 0.){return 1. / (-log(smin / smax) * s);}
  else {return 1. / (pow(s - smin, (1. - 1. / n)) * pow(smax - smin, 1. / n) * n);}
}

double phasespace_set::h_pot_max(double r, double smin, double smax, double n){
  if (n == 0.){return smax + smin - smax * exp(r * log(smin / smax));}
  else {return (smax - smin) * (1. - pow(1. - r, n)) + smin;}
}

double phasespace_set::g_pot_max(double s, double smin, double smax, double n){
  if (n == 0.){return 1. / (log(smin / smax) * (s - (smin + smax)));}
  else {return 1. / (n * pow(smax - s, 1. - 1. / n) * pow(smax - smin, 1. / n));}
}

double phasespace_set::h_propto_pot_symm(double r, double smin, double smax, double nu, double cut_technical){
  static Logger logger("h_propto_pot_symm");
  logger.called();

  // temporary !!!
  int switch_console_output_phasespace_issue = 0;

  double s;
  double smin_smax_2 = .5 * (smax - smin);
  if (nu != 1.){
    if (r <= 0.5){
      s = smin + smin_smax_2 * pow(2 * r * pow(1. + cut_technical, 1. - nu) + (1. - 2 * r) * pow(cut_technical, 1. - nu), (1. / (1. - nu))) - cut_technical;
    }
    else {
      s = smax - smin_smax_2 * pow(2 * (1. - r) * pow(1. + cut_technical, 1. - nu) + (2 * r - 1.) * pow(cut_technical, 1. - nu), (1. / (1. - nu))) - cut_technical;
    }
  }
  else {
    if (r <= 0.5){
      s = smin + smin_smax_2 * (exp(2 * r * log(1. + cut_technical) + (1. - 2 * r) * log(cut_technical)) - cut_technical);
    }
    else {
      s = smax - smin_smax_2 * (exp(2 * (1. - r) * log(1. + cut_technical) + (2 * r - 1.) * log(cut_technical)) - cut_technical);
    }
    if (switch_console_output_phasespace_issue){
      if (s < 0. || s > 1.){logger << LOG_WARN << "s = " << setw(25) << setprecision(16) << s << "    sminsmax_2 = " << .5 * (smax + smin) << endline_flush;}
    }
  }
  return s;
}

double phasespace_set::g_propto_pot_symm(double s, double smin, double smax, double nu, double cut_technical){
  static Logger logger("g_propto_pot_symm");
  logger.called();

  double gs;
  double sminsmax_2 = .5 * (smax + smin);
  double smin_smax_2 = .5 * (smax - smin);
  if (nu != 1.){
    if (s <= sminsmax_2){
      gs = (1. - nu) / (2 * pow(smin_smax_2, 1. - nu) * (pow(1. + cut_technical, 1. - nu) - pow(cut_technical, 1. - nu)) * pow(s - smin + cut_technical * smin_smax_2, nu));
    }
    else {
      gs = (1. - nu) / (2 * pow(smin_smax_2, 1. - nu) * (pow(1. + cut_technical, 1. - nu) - pow(cut_technical, 1. - nu)) * pow(smax - s + cut_technical * smin_smax_2, nu));
    }
  }
  else {
    if (s <= sminsmax_2){
      gs = 1. / (2 * (log(1. + cut_technical) - log(cut_technical)) * ((s - smin) + cut_technical * smin_smax_2));
    }
    else {
      gs = 1. / (2 * (log(1. + cut_technical) - log(cut_technical)) * ((smax - s) + cut_technical * smin_smax_2));
    }
  }
  return gs;
}
*/
