#include "header.hpp"

void phasespace_set::psp_optimization_complete(double & integrand, double & this_psp_weight, double & this_psp_weight2, double & optimization_modifier){
  static Logger logger("phasespace_set::psp_optimization_complete");
  logger.called();

  // Points that show indications of being problematic due to miscancellations are excluded from phase space optimization:
  if (RA_techcut_optimization){ return; }

  double mod_integrand = integrand * optimization_modifier;
  double mod_this_psp_weight = this_psp_weight * optimization_modifier;
  double mod_this_psp_weight2 = this_psp_weight2 * pow(optimization_modifier, 2);

  MC_phasespace.psp_MCweight_optimization(mod_integrand, g_tot);
  MC_tau.psp_MCweight_optimization(mod_integrand, g_tot);
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      if (MC_x_dipole[i_a].active_optimization != -1){MC_x_dipole[i_a].psp_MCweight_optimization(mod_integrand, g_tot);}
    }
  }

  random_manager.psp_IS_optimization(mod_this_psp_weight, mod_this_psp_weight2);

  if (munich_isnan(mod_integrand)){logger << LOG_FATAL << "unidentified error - mod_integrand nan - run stopped!" << endline_flush; exit(1);}
  if (munich_isinf(mod_integrand)){logger << LOG_FATAL << "unidentified error - mod_integrand inf - run stopped!" << endline_flush; exit(1);}
  if (munich_isnan(g_tot)){logger << LOG_FATAL << "unidentified error - gtot nan - run stopped!" << endline_flush; exit(1);}

  logger.finished();
}


void phasespace_set::step_optimization_complete(){
  static Logger logger("phasespace_set::step_optimization_complete");
  logger.called();

  // enforce_step = true -> step performed immediately

  MC_phasespace.step_MCweight_optimization(*i_step_mode, enforce_step, i_step);
  MC_tau.step_MCweight_optimization(*i_step_mode, enforce_step, i_step);
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      MC_x_dipole[i_a].step_MCweight_optimization(*i_step_mode, enforce_step, i_step);
    }
  }

  // MC_opt_end should be replaced by something general like 'end_optimization'  !!!

  random_manager.step_optimization(*i_step_mode, enforce_step, i_step);

  // Refine meaning of switch_use_alpha_after_IS ???
  // Possibly needs to be updated after introduction of enforce_step !!!
  if (switch_use_alpha_after_IS){
    // to choose best alpha set only after previous IS optimization step - not used at present !!!
    int temp_switch_IS_MC = -1;
    int temp_n_IS_events = 1;
    for (int i_s = 0; i_s < input_IS.size(); i_s++){
      if (input_IS[i_s]->name.substr(0, 5) == "IS_PS"){
	if (input_IS[i_s]->switch_optimization != -1){temp_switch_IS_MC = 1;}
	if (input_IS[i_s]->n_event_per_step > temp_n_IS_events){temp_n_IS_events = input_IS[i_s]->n_event_per_step;}
      }
    }
    
    if (temp_switch_IS_MC && !MC_opt_end && *i_step_mode % temp_n_IS_events == 0){
      if (MC_phasespace.active_optimization && *i_step_mode < MC_phasespace.n_event_per_step * MC_phasespace.n_optimization_step){
	MC_phasespace.x_alpha_it_min = MC_phasespace.i_alpha_it;
      }

      if (MC_tau.active_optimization && *i_step_mode < MC_tau.n_event_per_step * MC_tau.n_optimization_step){
	MC_tau.x_alpha_it_min = MC_tau.i_alpha_it;
      }

      if (csi->class_contribution_CS_real){
	for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
	  if (MC_x_dipole[i_a].switch_optimization != -1 && MC_x_dipole[i_a].active_optimization && *i_step_mode < MC_x_dipole[i_a].n_event_per_step * MC_x_dipole[i_a].n_optimization_step){
	    MC_x_dipole[i_a].x_alpha_it_min = MC_x_dipole[i_a].i_alpha_it;
	  }
	}
      }
    }
  }

  logger.finished();
}


void phasespace_set::result_optimization_complete(){
  static Logger logger("phasespace_set::result_optimization_complete");
  logger.called();

  if (end_optimization){return;}

  MC_phasespace.result_MCweight_optimization(*i_step_mode, enforce_step, i_step);
  MC_tau.result_MCweight_optimization(*i_step_mode, enforce_step, i_step);
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      if (MC_x_dipole[i_a].active_optimization != -1){
	MC_x_dipole[i_a].result_MCweight_optimization(*i_step_mode, enforce_step, i_step);
      }
    }
  }
  // All multichannel optimizations are synchronized (nothing else tested...)

  random_manager.result_optimization();

  end_optimization = 1;
  if (MC_phasespace.active_optimization == 1 && MC_phasespace.end_optimization != 1){end_optimization = 0;}
  if (MC_tau.active_optimization == 1 && MC_tau.end_optimization != 1){end_optimization = 0;}
  if (csi->class_contribution_CS_real){for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){if (MC_x_dipole[i_a].active_optimization == 1 && MC_x_dipole[i_a].end_optimization != 1){end_optimization = 0;}}}

  if (random_manager.active_optimization == 1 && random_manager.end_optimization != 1){end_optimization = 0;}
  for (int i_r = 0; i_r < random_psp.size(); i_r++){if (random_psp[i_r]->IS_PS.active_optimization == 1 && random_psp[i_r]->IS_PS.end_optimization != 1){end_optimization = 0;}}

  if (Log::getLogThreshold() <= LOG_DEBUG){
    logger << LOG_DEBUG << setw(40) << "MC_phasespace.active_optimization" << " = " << MC_phasespace.active_optimization << "   "   << setw(40) << "MC_phasespace.end_optimization" << " = " << MC_phasespace.end_optimization << endline;
    logger << LOG_DEBUG << setw(40) << "MC_tau.active_optimization" << " = " << MC_tau.active_optimization << "   "   << setw(40) << "MC_tau.end_optimization" << " = " << MC_tau.end_optimization << endline;
    if (csi->class_contribution_CS_real){for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){logger << LOG_DEBUG << setw(40) << "MC_x_dipole[" << setw(4) << i_a << "].active_optimization" << " = " << MC_x_dipole[i_a].active_optimization << "   "   << setw(40) << "MC_x_dipole[" << setw(4) << i_a << "].end_optimization" << " = " << MC_x_dipole[i_a].end_optimization << endline;}}
    logger << LOG_DEBUG << setw(40) << "random_manager.active_optimization" << " = " << random_manager.active_optimization << "   "   << setw(40) << "random_manager.end_optimization" << " = " << random_manager.end_optimization << endline;
    for (int i_r = 0; i_r < random_psp.size(); i_r++){logger << LOG_DEBUG_VERBOSE << setw(40) << "random_psp[" << setw(4) << i_r << "]->IS_PS.active_optimization" << " = " << random_psp[i_r]->IS_PS.active_optimization << "   "   << setw(40) << "random_psp[" << setw(4) << i_r << "]->IS_PS.end_optimization" << " = " << random_psp[i_r]->IS_PS.end_optimization << endline;}
    logger << LOG_DEBUG << setw(40) << "end_optimization" << " = " << end_optimization << endline_flush;
  }
  
  logger.finished();
}


void phasespace_set::output_optimization_complete(){
  static Logger logger("phasespace_set::output_optimization_complete");
  logger.called();

  if (!end_optimization){return;}
  if (end_optimization == 2){return;}

  // Don't print weight files if ruled out by PDF selection in general, but print them for CT(2) since VT(2) could contribute although CT(2) does not !!!
  logger << LOG_DEBUG << "csi->int_end_pdf_selection = " << csi->int_end_pdf_selection << endline_flush;
  if (!csi->int_end_pdf_selection || csi->class_contribution_QT_collinear){
    ofstream out_weight;
    out_weight.open(filename_weight_writeout.c_str(), ofstream::out | ofstream::trunc);
    out_weight.close();

    string temp_filename_dec = filename_weight_writeout.insert(filename_weight_writeout.size() - 4, "_dec");
    logger << LOG_DEBUG << "filename_weight_writeout (dec) = " << temp_filename_dec << endline_flush;
    out_weight.open(temp_filename_dec.c_str(), ofstream::out | ofstream::trunc);
    out_weight.close();

    MC_phasespace.output_optimization();
    MC_tau.output_optimization();
    if (csi->class_contribution_CS_real){for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){if (MC_x_dipole[i_a].active_optimization != -1){MC_x_dipole[i_a].output_optimization();}}}

    random_manager.output_optimization();
  }
  
  end_optimization = 2;

  logger.finished();
}


void phasespace_set::enforce_end_optimization(){
  static Logger logger("phasespace_set::enforce_end_optimization");
  logger.called();

  if (Log::getLogThreshold() <= LOG_DEBUG){
    logger << LOG_DEBUG << "before enforce_end_optimization" << endline_flush;
    logger << LOG_DEBUG << "MC_opt_end = " << MC_opt_end << endline_flush;
    logger << LOG_DEBUG << "active_optimization = " << active_optimization << endline_flush;
    logger << LOG_DEBUG << "end_optimization    = " << end_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_phasespace.active_optimization = " << MC_phasespace.active_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_phasespace.end_optimization    = " << MC_phasespace.end_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_tau.active_optimization = " << MC_tau.active_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_tau.end_optimization    = " << MC_tau.end_optimization << endline_flush;
    if (csi->class_contribution_CS_real){
      for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
	logger << LOG_DEBUG << "MC_x_dipole[" << i_a << "].active_optimization = " << MC_x_dipole[i_a].active_optimization << endline_flush;
	logger << LOG_DEBUG << "MC_x_dipole[" << i_a << "].end_optimization    = " << MC_x_dipole[i_a].end_optimization << endline_flush;
      }
    }
  }

  MC_opt_end = 1;
  end_optimization = 1;
  if (MC_phasespace.active_optimization){MC_phasespace.end_optimization = 1;}
  if (MC_tau.active_optimization){MC_tau.end_optimization = 1;}
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      if (MC_x_dipole[i_a].active_optimization){MC_x_dipole[i_a].end_optimization = 1;}
    }
  }

  random_manager.end_optimization = 1;
  for (int i_r = 0; i_r < random_manager.random_psp.size(); i_r++){random_manager.random_psp[i_r].IS_PS.end_optimization = 1;}

  if (Log::getLogThreshold() <= LOG_DEBUG){
    logger << LOG_DEBUG << "before output_optimization_complete" << endline_flush;
    logger << LOG_DEBUG << "MC_opt_end = " << MC_opt_end << endline_flush;
    logger << LOG_DEBUG << "active_optimization = " << active_optimization << endline_flush;
    logger << LOG_DEBUG << "end_optimization    = " << end_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_phasespace.active_optimization = " << MC_phasespace.active_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_phasespace.end_optimization    = " << MC_phasespace.end_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_tau.active_optimization = " << MC_tau.active_optimization << endline_flush;
    logger << LOG_DEBUG << "MC_tau.end_optimization    = " << MC_tau.end_optimization << endline_flush;
    if (csi->class_contribution_CS_real){
      for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
	logger << LOG_DEBUG << "MC_x_dipole[" << i_a << "].active_optimization = " << MC_x_dipole[i_a].active_optimization << endline_flush;
	logger << LOG_DEBUG << "MC_x_dipole[" << i_a << "].end_optimization    = " << MC_x_dipole[i_a].end_optimization << endline_flush;
      }
    }
  }

  output_optimization_complete();

  logger.finished();
}
