#include "header.hpp"

void phasespace_set::initialization_fake_dipole_mapping_RT(){
  static Logger logger("phasespace_set::initialization_fake_dipole_mapping_RT");
  logger.called();

  csi->dipole.push_back(dipole_set("fakeR", csi, 0, 0));
  csi->determination_dipole_QCD();
  MC_x_dipole_mapping.resize(csi->dipole.size());
  for (int j_a = 1; j_a < csi->dipole.size(); j_a++){
    ac_tau_psp_dipole(j_a, MC_x_dipole_mapping[j_a]);
  }

  for (int j_a = 1; j_a < csi->dipole.size(); j_a++){
    for (int i_m = 0; i_m < MC_x_dipole_mapping[j_a].size(); i_m++){
      int flag = tau_MC_map.size();
      for (int j_m = 0; j_m < tau_MC_map.size(); j_m++){
	if (MC_x_dipole_mapping[j_a][i_m] == tau_MC_map[j_m]){flag = j_m; break;}
      }
      if (flag == tau_MC_map.size()){tau_MC_map.push_back(MC_x_dipole_mapping[j_a][i_m]);}
    }
  }

  logger << LOG_DEBUG_VERBOSE << "csi->dipole.size() = " << csi->dipole.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "csi->type_parton.size() = " << csi->type_parton.size() << endline_flush;
  csi->type_parton.erase(csi->type_parton.begin() + 1, csi->type_parton.end());
  csi->dipole.erase(csi->dipole.begin() + 1, csi->dipole.end());
  logger << LOG_DEBUG_VERBOSE << "csi->dipole.size() = " << csi->dipole.size() << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "csi->type_parton.size() = " << csi->type_parton.size() << endline_flush;

  logger.finished();
}


void phasespace_set::initialization_fake_dipole_mapping_RRA(){
  static Logger logger("phasespace_set::initialization_fake_dipole_mapping_RRA");
  logger.called();

  logger << LOG_DEBUG_VERBOSE << "fake_dipole to improve initial-state mappings started..." << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "REAL *RA_DIPOLE: csi->dipole.size() = " << csi->dipole.size() << endline_flush;

  fake_psi->msi = msi;
  fake_psi->csi->msi = msi;
  
  for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
    logger << LOG_DEBUG_VERBOSE << "csi->phasespace_order_alpha_s[" << i_a << "] = " << csi->phasespace_order_alpha_s[i_a] << "   csi->phasespace_order_alpha_e[" << i_a << "] = " << csi->phasespace_order_alpha_e[i_a] << endline_flush;
  }

  fake_psi->csi->type_contribution = "RA";

  for (int i_a = 1; i_a < csi->dipole.size(); i_a++){
    //  Setting  dipole i_a  as "basic process" of  fake_psi->csi :
    fake_psi->csi->no_process_parton.clear();
    fake_psi->csi->no_process_parton.resize(1, csi->no_process_parton[i_a]);
    fake_psi->csi->swap_parton.clear();
    fake_psi->csi->swap_parton.resize(1, csi->swap_parton[i_a]);
    fake_psi->csi->phasespace_order_alpha_s.clear();
    fake_psi->csi->phasespace_order_alpha_s.resize(1, csi->phasespace_order_alpha_s[i_a]);
    fake_psi->csi->phasespace_order_alpha_e.clear();
    fake_psi->csi->phasespace_order_alpha_e.resize(1, csi->phasespace_order_alpha_e[i_a]);
    fake_psi->csi->phasespace_order_interference.clear();
    fake_psi->csi->phasespace_order_interference.resize(1, csi->phasespace_order_interference[i_a]);
    fake_psi->csi->type_parton.clear();
    fake_psi->csi->type_parton.resize(1, csi->type_parton[i_a]);
    fake_psi->csi->basic_type_parton.clear();
    fake_psi->csi->basic_type_parton.resize(1, csi->basic_type_parton[i_a]);
    fake_psi->csi->symmetry_id_factor.clear();
    fake_psi->csi->symmetry_id_factor.resize(1, csi->symmetry_id_factor[i_a]);
    fake_psi->csi->dipole.clear();
    fake_psi->csi->dipole.push_back(dipole_set("fakeR", fake_psi->csi, 0, 0));
    logger << LOG_DEBUG_VERBOSE << "i_a = " << i_a << "   csi->dipole.size() = " << fake_psi->csi->dipole.size() << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "before: fake_psi->csi->type_parton.size() = " << fake_psi->csi->type_parton.size() << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "before: fake_psi->csi->dipole.size() = " << fake_psi->csi->dipole.size() << endline_flush;
    fake_psi->csi->determination_dipole_QCD();
    logger << LOG_DEBUG_VERBOSE << "after: fake_psi->csi->type_parton.size() = " << fake_psi->csi->type_parton.size() << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "after: fake_psi->csi->dipole.size() = " << fake_psi->csi->dipole.size() << endline_flush;

    stringstream temp;
    for (int i_p = 1; i_p < csi->type_parton[i_a].size(); i_p++){
      temp << setw(5) << csi->type_parton[i_a][i_p];
    }
    logger << LOG_DEBUG << setw(25) << right << "csi->dipole[" << setw(2) << i_a << "].name() = " << left << setw(15) << csi->dipole[i_a].name() << "   " << temp.str() << endline_flush;

    logger << LOG_DEBUG_VERBOSE << "fake_psi->csi->type_parton.size() = " << fake_psi->csi->type_parton.size() << endline_flush;
    for (int j_a = 0; j_a < fake_psi->csi->type_parton.size(); j_a++){
      stringstream temp;
      for (int i_p = 1; i_p < fake_psi->csi->type_parton[j_a].size(); i_p++){
	temp << setw(5) << fake_psi->csi->type_parton[j_a][i_p];
      }
      logger << LOG_DEBUG << setw(25) << right << "fake_psi->csi->dipole[" << setw(2) << j_a << "].name() = " << left << setw(15) << fake_psi->csi->dipole[j_a].name() << "   " << temp.str() << endline_flush;
    }

    //  Setting  dipole i_a  as "basic process" of  fake_psi :
    logger << LOG_DEBUG_VERBOSE << "MC_n_channel_phasespace.size() = " << MC_n_channel_phasespace.size() << endline_flush;
    fake_psi->MC_n_channel_phasespace.resize(1, MC_n_channel_phasespace[i_a]);
    fake_psi->MC_sum_channel_phasespace.resize(1, MC_sum_channel_phasespace[i_a]);
    fake_psi->M = M;
    logger << LOG_DEBUG_VERBOSE << "fake_psi->csi->type_parton.size() = " << fake_psi->csi->type_parton.size() << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "Before fake_MC_x_dipole_mapping" << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "fake_psi->csi->dipole.size() = " << fake_psi->csi->dipole.size() << endline_flush;
    fake_psi->MC_x_dipole_mapping.resize(fake_psi->csi->dipole.size());
    for (int j_a = 1; j_a < fake_psi->csi->dipole.size(); j_a++){
      logger << LOG_DEBUG_VERBOSE << "fake_psi->MC_x_dipole_mapping.size() = " << fake_psi->MC_x_dipole_mapping.size() << endline_flush;
      fake_psi->ac_tau_psp_dipole(j_a, fake_psi->MC_x_dipole_mapping[j_a]);
      logger << LOG_DEBUG_VERBOSE << "after:  fake_psi->MC_x_dipole_mapping[" << j_a << "].size() = " << fake_psi->MC_x_dipole_mapping[j_a].size() << endline_flush;
    }

    logger << LOG_DEBUG_VERBOSE << "fake_psi->csi->dipole[" << i_a << "].size() = " << fake_psi->csi->dipole.size() << endline_flush;
    for (int j_a = 1; j_a < fake_psi->csi->dipole.size(); j_a++){
      logger << LOG_DEBUG_VERBOSE << "fake_psi->MC_x_dipole_mapping[" << j_a << "].size() = " << fake_psi->MC_x_dipole_mapping[j_a].size() << endline_flush;
      for (int i_m = 0; i_m < fake_psi->MC_x_dipole_mapping[j_a].size(); i_m++){
	int flag = MC_x_dipole_mapping[i_a].size();
	for (int j_m = 0; j_m < MC_x_dipole_mapping[i_a].size(); j_m++){
	  logger << LOG_DEBUG_VERBOSE << "i_a = " << i_a << "   i_m = " << i_m << "   j_m = " << j_m << "   fake_psi->MC_x_dipole_mapping[" << j_a << "][" << i_m << "] = " << fake_psi->MC_x_dipole_mapping[j_a][i_m] << endline_flush;
	  if (fake_psi->MC_x_dipole_mapping[j_a][i_m] == MC_x_dipole_mapping[i_a][j_m]){flag = j_m; break;}
	}
	if (flag == MC_x_dipole_mapping[i_a].size()){MC_x_dipole_mapping[i_a].push_back(fake_psi->MC_x_dipole_mapping[j_a][i_m]);}
      }
    }
  }
  /*
    for (int i_a = 1; i_a < csi->dipole.size(); i_a++){
      for (int j_a = 1; j_a < fake_psi->csi->dipole.size(); j_a++){
	logger << LOG_DEBUG << "fake_psi->csi->dipole[" << i_a << "][" << j_a << "].name() = " << csi->dipole[i_a].name() << " - " << fake_psi->csi->dipole[j_a].name() << endline_flush;
      }
    }

    for (int i_a = 0; i_a < csi->dipole.size(); i_a++){
      for (int j_m = 0; j_m < MC_x_dipole_mapping[i_a].size(); j_m++){
	logger << LOG_DEBUG << "MC_x_dipole_mapping[" << i_a << "][" << j_m << "] = " << MC_x_dipole_mapping[i_a][j_m] << endline_flush;
      }
    }
  */
  
  /*
  logger << LOG_DEBUG << "Delete 'fake_psi' that was only required to construct IS mappings." << endline_flush;
  delete fake_psi->csi;
  delete fake_psi;
  logger << LOG_DEBUG << "Delete 'fake_psi' that was only required to construct IS mappings - done." << endline_flush;
  */
  
  logger << LOG_DEBUG << "fake_psi->csi->dipole to improve initial-state mappings finished..." << endline_flush;

  logger.finished();
}



