#include "header.hpp"

void phasespace_set::calculate_IS_QT_phi(){
  static Logger logger("phasespace_set::calculate_IS_QT_phi");
  logger.called();

  recoil_q = xbsqrts_all[0][0];
  recoil_q2 = xbs_all[0][0];
  // fill  n_random_qT (= 3) random numbers (two for |qT|, another one for phi)
  random_qT = rng.access_random_number(n_random_qT);
  for (int j = 0; j < IS_qTres->beta.size(); j++){if (random_qT[0] <= IS_qTres->beta[j]){IS_qTres->channel = j; break;}}
  double random_qT_abs = (double(IS_qTres->channel) + random_qT[1]) / double(IS_qTres->beta.size());
  // factor for Jacobian of the qT/phi mappings: needs to be added in weight combination
  g_qT = IS_qTres->alpha[IS_qTres->channel] * IS_qTres->alpha.size();
  logger << LOG_DEBUG_POINT << "g_qT = " << g_qT << endline_flush;

  double min_recoil_qT = 1.e-07 * recoil_q;
  //  double min_recoil_qT = esi->qT_basic->min_qTcut / 100. * recoil_q;
  double max_recoil_qT = esi->qT_basic->max_qTcut / 100. * recoil_q;
  if (user->switch_value[user->switch_map["CTfromPTmode"]]){
    min_recoil_qT = esi->qT_basic->min_qTcut / 100. * recoil_q;
    max_recoil_qT = E;  // to reproduce counterterm ???
  }
  double min_recoil_qT2 = pow(min_recoil_qT, 2);
  double max_recoil_qT2 = pow(max_recoil_qT, 2);

  // user-defined switch mapqT2
  // mapqT2 = 0: exponential mapping
  // mapqT2 = 1: linear mapping
  // mapqT2 = -1: exponential mapping, enhanced at max_recoil_qT2
  
  if (user->switch_value[user->switch_map["mapqT2"]] == 1){
    recoil_qT2 = min_recoil_qT2 + (max_recoil_qT2 - min_recoil_qT2) * random_qT_abs;
    g_qT /= (max_recoil_qT2 - min_recoil_qT2);
  }
  else if (user->switch_value[user->switch_map["mapqT2"]] == -1){
    recoil_qT2 = max_recoil_qT2 - min_recoil_qT2 * exp(log(max_recoil_qT2 / min_recoil_qT2) * random_qT_abs);
    g_qT /= log(max_recoil_qT2 / min_recoil_qT2) * (max_recoil_qT2 - recoil_qT2);
  }
  else {
    recoil_qT2 = min_recoil_qT2 * exp(log(max_recoil_qT2 / min_recoil_qT2) * random_qT_abs);
    g_qT /= log(max_recoil_qT2 / min_recoil_qT2) * recoil_qT2;
  }

  recoil_qT = sqrt(recoil_qT2);
  recoil_r = recoil_qT / recoil_q;
  
  logger << LOG_DEBUG_POINT << setw(25) << "min_recoil_qT" << " = " << min_recoil_qT << endline_flush;
  logger << LOG_DEBUG_POINT << setw(25) << "max_recoil_qT" << " = " << max_recoil_qT << endline_flush;
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_q" << " = " << recoil_q << endline_flush;
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_qT" << " = " << recoil_qT << endline_flush;
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_r" << " = " << recoil_r << endline_flush;
  

  esi->qT_basic->event_selection_recoil_qTcut();  // ->  recoil_rcut
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_rcut" << " = " << recoil_rcut << endline_flush;

  recoil_phi = c_phi(random_qT[2]);
  //  g_qT /= 2 * pi;
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_phi" << " = " << recoil_phi << endline_flush;

  //  recoil_system = fourvector(recoil_q, recoil_qT, 0., 0.);
  recoil_system = fourvector(sqrt(recoil_q2 + recoil_qT2), recoil_qT, 0., 0.);
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_system (before rot)" << " = " << recoil_system << endline_flush;
  recoil_system = recoil_system.rotation(recoil_phi, 1);
  logger << LOG_DEBUG_POINT << setw(25) << "recoil_system (after rot)" << " = " << recoil_system << endline_flush;

  logger.finished();
}


// used only for b-space resummation !!!
void phasespace_set::calculate_IS_QT(){
  static Logger logger("phasespace_set::calculate_IS_QT");
  logger.called();

  random_qTres = rng.access_random_number(n_random_qTres);
  for (int j = 0; j < IS_qTres->beta.size(); j++){if (random_qTres[0] <= IS_qTres->beta[j]){IS_qTres->channel = j; break;}}
  // Why is  random_x12[1]  used here ??? Should actually be random_qTres[1] !!!
  QT_random_qt2 = (double(IS_qTres->channel) + random_x12[1]) / double(IS_qTres->beta.size());
  g_qTres = 1. / (IS_qTres->alpha[IS_qTres->channel] * IS_qTres->alpha.size());
  g_pdf *= g_qTres;

  /* ///
  QT_random_qt->get_random(QT_random_qt2,QT_g_IS_);
  //  logger << LOG_DEBUG_VERBOSE << "qt: QT_random_qt2 = " << setw(23) << setprecision(15) << QT_random_qt2 << "   QT_g_IS_ = " << setw(23) << setprecision(15) << QT_g_IS_ << endline_flush;
  g_pdf *= QT_g_IS_;
  */

  if (csi->type_contribution == "CT" ||
      csi->type_contribution == "CT2"){
    // the (Bessel transformed) large logarithms vanish up to double precision accuracy for arguments larger than 40 -> use this as upper bound
    // 1.261 ~ b0*b0
    double maxqt2;
    double Qres_tmp=0.0, Qres_lower, Qres_upper;

    // Qres appears at two places here: as the reference value for the lower and for the upper qT cut
    // conceptually, these are partially unrelated
    // in fixed order computations, we have to use Qres as the lower cut off scale, because we precompute the qT integrals
    // in resummed computations, we explicitly integrate over qT, so this is unnecessary; in fact
    // it is more efficient to use mF as the lower reference scale, as qT/mF determine the universal limit
    // the upper cut off scale should alway be given by Qres, as otherwise qT/Qres can become huge, resulting in numerical
    // problems in the counterterm weight computation

    // Qres != mF does not work at the moment; has to be implemented in process specific cut files
    logger << LOG_DEBUG_VERBOSE << "switch_resummation = " << switch_resummation << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "Qres = " << Qres << endline_flush;

    assert(Qres == 0. || !switch_resummation);


    if (dynamical_Qres) {
      Qres_tmp = Qres_prefactor*xbsqrts_all[0][0];
    } else {
      Qres_tmp = Qres;
    }
    if (Qres_tmp == 0) {
      // default choice
      Qres_tmp = xbsqrts_all[0][0];
    }

    Qres_upper = Qres_tmp;

    if (switch_resummation){
      Qres_lower = xbsqrts_all[0][0];
    } else {
      Qres_lower = Qres_tmp;
    }

    if (csi->contribution_order_alpha_s == 1){maxqt2 = 1.261 * 30. * 20. * Qres_upper * Qres_upper;}
    else {maxqt2 = 1.261 * 40. * 50. * Qres_upper * Qres_upper;}

    if (switch_qTcut == 1){
      double r0 = min_qTcut / 100.;
      //  version with cut on pT/Qres_cut
      QT_qt2 = r0 * r0 * pow(Qres_lower,2) * exp(log(maxqt2 / r0 / r0 / Qres_lower / Qres_lower) * QT_random_qt2);
//       logger << LOG_DEBUG_VERBOSE << "qT=" << QT_qt2 << ", Qres=" << Qres_upper << ", " << Qres_lower << ", ratio=" << QT_qt2/Qres_upper << endline_flush;
      QT_jacqt2 = log(maxqt2 / r0 / r0 / Qres_lower / Qres_lower) * QT_qt2;
      //  version with cut on pT/m_inv
      //    QT_qt2 = r0 * r0 * xbs_all[0][0] * exp(log(maxqt2 / r0 / r0 / xbs_all[0][0]) * QT_random_qt2);
      //    QT_jacqt2 = log(maxqt2 / r0 / r0 / xbs_all[0][0]) * QT_qt2;
    }
    else if (switch_qTcut == 2){
      // old version with cut on pT only (not relative to m_inv)
      //   qt2=min_qTcut*min_qTcut*exp(log(maxqt2/min_qTcut/min_qTcut)*random_qt2);
      //   jacqt2=log(maxqt2/min_qTcut/min_qTcut)*qt2;
      logger << LOG_FATAL << "old qT variation is deprecated" << endline_flush;
      exit(1);
    }
  }

  else if (csi->type_contribution == "NLL_LO" || csi->type_contribution == "NLL_NLO" || csi->type_contribution == "NNLL_LO" || csi->type_contribution == "NNLL_NLO" || csi->type_contribution == "NNLL_NNLO"){
    if (switch_resummation) {
      min_qTcut = 0.1; // ??? why is that set manually (and overwrites the set value of min_qTcut) ???

      double maxqt2 = pow(2 * E, 2);
      maxqt2 = pow(E, 2); // ???

      if (lambda_qt2 == 0 || lambda_qt2 == 1) { // ??? meaning of lambda_qt2 ??? does it only choose the mapping of QT_qt2 ???
	QT_qt2 = min_qTcut * min_qTcut * exp(log(maxqt2 / min_qTcut / min_qTcut) * QT_random_qt2);
	QT_jacqt2 = log(maxqt2 / min_qTcut / min_qTcut) * QT_qt2;
      }

      //  random_qt2=0;

      //  double lambda=0.0355064;
      //  double C=-1.0/lambda*(exp(-lambda*sqrt(maxqt2))-exp(-lambda*min_qTcut));
      //  double y0=-1.0/lambda/C*exp(-lambda*min_qTcut);
      ////  logger << LOG_DEBUG_VERBOSE << exp(lambda*sqrt(maxqt2)) << ", " << exp(lambda*min_qTcut) << endline_flush;
      ////  logger << LOG_DEBUG_VERBOSE << C << ", " << y0 << endline_flush;
      //  double qt=-1.0/lambda*log(-lambda*C*(random_qt2+y0));
      //  double jacqt=C*exp(lambda*qt);
      //  qt2=qt*qt;
      //  jacqt2=jacqt*2*qt;

      else {
	double y0 = 1. / (pow(maxqt2 / min_qTcut / min_qTcut, 1 - lambda_qt2) - 1);
	double C = pow(min_qTcut, 2 * (1 - lambda_qt2)) / (1 - lambda_qt2) / y0;

	QT_qt2 = pow((1 - lambda_qt2) * C * (QT_random_qt2 + y0), 1. / (1 - lambda_qt2));
	QT_jacqt2 = C * pow(QT_qt2, lambda_qt2);

	//    logger << LOG_DEBUG_VERBOSE << C << ", " << y0 << ", " << (1-lambda_qt2)*C*(random_qt2+y0) << endline_flush;
	//    logger << LOG_DEBUG_VERBOSE << qt2 << ", " << maxqt2 << endline_flush;
      }

      //  maxqt2 = 1e6;
      //  qt2 = min_qTcut * min_qTcut + random_qt2 * (maxqt2 - min_qTcut * min_qTcut);
      //  jacqt2 = (maxqt2 - min_qTcut * min_qTcut);
    }
  }

  logger.finished();
}


