#include "header.hpp"

randomvariable::randomvariable(){
  static Logger logger("randomvariable::randomvariable()");
  logger.called();

  manager = NULL;

  logger.finished();
}

randomvariable::randomvariable(importancesampling_input _input, string _name, int _number, randommanager * _manager){
  static Logger logger("randomvariable::randomvariable (input)");
  logger.called();

  manager = _manager;
  number = _number;
  logger << LOG_DEBUG << _input << endline_flush;
  _input.name = _name;
  logger << LOG_DEBUG_VERBOSE << "name = " << _name << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "manager->psi->filename_weight_writeout = " << manager->psi->filename_weight_writeout << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "manager->psi->filename_weight_readin = " << manager->psi->filename_weight_readin << endline_flush;

  IS_PS = importancesampling_set(_input, manager->psi->filename_weight_writeout, manager->psi->filename_weight_readin);
  
  IS_PS.channel = 0;

  if (IS_PS.n_gridsize == 1){imp_sampling = false;}
  else {imp_sampling = true;}

  used = false;

  logger.finished();
}


void randomvariable::psp_IS_optimization(double & psp_weight, double & psp_weight2) {
  static Logger logger("randomvariable::psp_IS_optimization");
  logger.called();

  if (!imp_sampling){return;}
  if (!used && !manager->psi->switch_IS_optimization_each_variable){return;}
  //  logger << LOG_DEBUG << setw(5) << number << setw(50) << IS_PS.name << setw(5) << used << setw(5) << IS_PS.channel << endline_flush;
  IS_PS.psp_IS_optimization(psp_weight, psp_weight2);

  used = false;

  logger.finished();
}


void randomvariable::step_IS_optimization(int i_step_mode, bool enforce_step, int i_step){
  static Logger logger("randomvariable::step_IS_optimization");
  logger.called();

  logger << LOG_DEBUG << "IS_PS.name = " << IS_PS.name << endline_flush;
  
  IS_PS.step_IS_optimization(i_step_mode, enforce_step, i_step);
  // Why is  result_IS_optimization  called here already ???
  result_IS_optimization(i_step_mode, enforce_step, i_step);

  logger.finished();
}


void randomvariable::result_IS_optimization(int i_step_mode, bool enforce_step, int i_step) {
  static Logger logger("randomvariable::result_IS_optimization");
  logger.called();

  IS_PS.result_IS_optimization(i_step_mode, enforce_step, i_step);

  logger.finished();
}


double randomvariable::get_random(){
  static Logger logger("randomvariable::get_random()");
  logger.called();

  double x_random;
  // Problem in used variable ???
  /*
  if (used && !end_optimization) {
    logger << LOG_FATAL << "randomvariable::get_random  (" << name << " used && !end_optimization) should never be called with used == true !" << endline_flush;
    exit(1);
  }
  */


  if (imp_sampling) {
    logger << LOG_DEBUG_VERBOSE << IS_PS.name << "   called: used = true." << endline_flush;
    used = true;
    logger << LOG_DEBUG_VERBOSE << "manager = " << manager << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "manager->psi = " << manager->psi << endline_flush;
									//    logger << LOG_DEBUG_VERBOSE << "manager->psi->rng = " << manager->psi->rng << endline_flush;
    double random = manager->psi->rng.access_random_number();
    double random_channel = manager->psi->rng.access_random_number();
    logger << LOG_DEBUG_VERBOSE << IS_PS.name << "   called: used = true." << endline_flush;

    // to reproduce some old results in a pointwise way !!!
    // otherwise not required !!!
    if (IS_PS.name == "IS_tau"){swap(random, random_channel);}
    if (IS_PS.name == "IS_x1x2"){swap(random, random_channel);}

    // new implementation:
    IS_PS.channel = 0;
    if (IS_PS.switch_mode_optimization == 4){
      //      logger << LOG_INFO << "random_channel = " << random_channel << "   IS_PS.n_gridsize = " << IS_PS.n_gridsize << endline_flush;
      IS_PS.channel = random_channel * IS_PS.n_gridsize;
      if (IS_PS.channel == 0){x_random = random * IS_PS.alpha[IS_PS.channel];}
      else {x_random = IS_PS.beta[IS_PS.channel - 1] + random * IS_PS.alpha[IS_PS.channel];}
      //      logger << LOG_INFO << "IS_PS.channel = " << IS_PS.channel << "   IS_PS.alpha[IS_PS.channel] = " << IS_PS.alpha[IS_PS.channel] << endline_flush;
      //      logger << LOG_INFO << "random = " << random << endline_flush;
      //      logger << LOG_INFO << "x_random = " << x_random << endline_flush;
    }
    else {
      if (IS_PS.is_gridsize_power_of_two){
	for (int i_b = IS_PS.n_gridsize / 2; i_b > 0; i_b /= 2){
	  if (random_channel > IS_PS.beta[IS_PS.channel + i_b - 1]){IS_PS.channel += i_b;}
	}
      }
      else {
	int extra = 0;
	for (int i_b = IS_PS.n_gridsize / 2; i_b > 0; i_b /= 2){
	  if (random_channel > IS_PS.beta[IS_PS.channel + i_b - 1]){IS_PS.channel += i_b;}
	  if (i_b == 3){
	    for (int j_b = 0; j_b < 3 + extra; j_b++){
	      if (random_channel <= IS_PS.beta[IS_PS.channel + j_b]){
		IS_PS.channel += j_b;
		i_b = 0;
		break;
	      }
	    }
	  }
	  if (i_b % 2){extra++;}
	}
      }
      // 'x_random' replaces the flatly generated random number:
      x_random = (double(IS_PS.channel) + random) / IS_PS.n_gridsize;
    }

    if (manager->psi->switch_IS_mode_phasespace == 1 || manager->psi->switch_IS_mode_phasespace == 3){
      //      g_IS = IS_PS.alpha[IS_PS.channel] * IS_PS.n_gridsize;
      manager->psi->MC_g_IS_global *= IS_PS.alpha[IS_PS.channel] * IS_PS.n_gridsize;
    }
  }
  //  else {x_random = manager->psi->rng.access_random_number();}
  // Test if missing second random number for (!imp_sampling) causes problems !!!
  else {
    x_random = manager->psi->rng.access_random_number();
    manager->psi->rng.ignore_random_number();
  }

  if (!IS_PS.end_optimization){
    // It tells manager that this variable has been used at this phase-space point (can be used later)
    if (manager != NULL){manager->add_var_to_queue(this);}
  }

  //  logger << LOG_INFO << IS_PS.name << "   imp_sampling = " << imp_sampling << "   used = " << used << endline_flush;

  // Can be removed later:
  //  g_IS = 1.0;
  return x_random;

  logger.finished();
}


// Could also be directly accessed, without the detour via void randomvariable::get_random(double &x, double &g_IS) !!!

void randomvariable::get_g_IS(double r, double & g_IS) {
  static Logger logger("randomvariable::get_g_IS");
  logger.called();

  if (r < 0. || r > 1. || munich_isnan(r)){g_IS = sqrt(-1.);}
  else {
    if (IS_PS.switch_mode_optimization == 4){
      // could IS_PS.channel be used here ???
      // temp_channel -> IS_PS.channel 
      int temp_channel = 0;
      if (IS_PS.is_gridsize_power_of_two){
	for (int i_b = IS_PS.n_gridsize / 2; i_b > 0; i_b /= 2){
	  if (r > IS_PS.beta[temp_channel + i_b - 1]){temp_channel += i_b;}
	}
      }
      else {
	int extra = 0;
	for (int i_b = IS_PS.n_gridsize / 2; i_b > 0; i_b /= 2){
	  if (r > IS_PS.beta[temp_channel + i_b - 1]){temp_channel += i_b;}
	  if (i_b == 3){
	    for (int j_b = 0; j_b < 3 + extra; j_b++){
	      if (r <= IS_PS.beta[temp_channel + j_b]){
		temp_channel += j_b;
		i_b = 0;
		break;
	      }
	    }
	  }
	  if (i_b % 2){extra++;}
	}
      }
      IS_PS.channel = temp_channel;
      g_IS = 1. / (IS_PS.alpha[temp_channel] * IS_PS.n_gridsize);
    }
    else {
      //      int IS_PS.channel = (int)(r * IS_PS.n_gridsize);
      IS_PS.channel = (int)(r * IS_PS.n_gridsize);
      g_IS = IS_PS.alpha[IS_PS.channel] * IS_PS.n_gridsize;
    }
  }

  logger.finished();
}


void randomvariable::save_weights(ofstream & out_weights){
  static Logger logger("randomvariable::save_weights");
  logger.called();

  out_weights << "variable " << setw(50) << IS_PS.name << endline;
  // Why such low precision ???
  for (int i_b = 0; i_b < IS_PS.n_gridsize; i_b++){out_weights << setprecision(8) << IS_PS.alpha[i_b] << endline;}

  logger.finished();
}


void randomvariable::readin_weights(vector<string> & readin) {
  static Logger logger("randomvariable::readin_weights");
  logger.called();

  if (IS_PS.switch_optimization != 2 && IS_PS.switch_optimization != 3){return;}

  // check if correct !!!
  // find the correct section in the weights file
  int proc;
  for (proc=0; proc<readin.size(); proc++) {
    size_t pos=readin[proc].find("variable ");
    if (pos != string::npos) {
      pos += 9;
      string var_name=readin[proc].substr(pos,string::npos);
      int start_index = 0;
      for (int i = 0; i < var_name.size(); i++){
        if (var_name[i] != ' '){start_index = i; break;}
      }
      var_name=var_name.substr(start_index, var_name.size() - start_index);
      if (var_name == IS_PS.name) {
        logger << LOG_DEBUG << "weights found for " << var_name << endline_flush;
	//        logger << LOG_DEBUG_VERBOSE << "weights found for " << var_name << endline_flush;
        break;
      }
    }
  }

  if (proc == readin.size()){logger << LOG_FATAL << "could not read in weights for " << IS_PS.name << endline_flush; exit(1);}

  logger << LOG_DEBUG_VERBOSE << "reading weights for " << readin[proc] << endline_flush;
  proc++;

  for (int i_b = 0; i_b < IS_PS.n_gridsize; i_b++){
    IS_PS.alpha[i_b] = atof(readin[proc].c_str());
    proc++;
  }

  IS_PS.beta[0] = IS_PS.alpha[0];
  for (int i_b = 1; i_b < IS_PS.n_gridsize; i_b++){IS_PS.beta[i_b] = IS_PS.beta[i_b - 1] + IS_PS.alpha[i_b];}

  // Check for issues in read-in weights:
  if (fabs(IS_PS.beta[IS_PS.n_gridsize - 1] - 1) > 1e-6){
    for (int i_b = 0; i_b < IS_PS.n_gridsize; i_b++){
      logger << LOG_ERROR << "alpha[" << setw(4) << i_b << "] = " << IS_PS.alpha[i_b] << "   beta[" << setw(4) << i_b << "] = " << IS_PS.beta[i_b] << endline;
    }
    logger << LOG_FATAL << setw(50) << IS_PS.name << ": something went wrong in weight read-in, aborting.." << endline_flush; exit(1);
  }

  logger.finished();
}


void randomvariable::proceeding_out(ofstream &out_proceeding) {
  static Logger logger("randomvariable::proceeding_out");
  logger.called();

  IS_PS.proceeding_out(out_proceeding);

  logger.finished();
}


void randomvariable::proceeding_in(int & proc, vector<string> &readin) {
  static Logger logger("randomvariable::proceeding_in");
  logger.called();

  IS_PS.proceeding_in(proc, readin);

  logger.finished();
}


bool randomvariable::check_proceeding_in(int & int_end, int & temp_check_size, vector<string> & readin){
  static Logger logger("randomvariable::check_proceeding_in");
  logger.called();

  bool check_size = true;
  // new implementation:
  // check if usage of 'int_end' makes sense here !!!

  logger << LOG_DEBUG << "# " << IS_PS.name << "   temp_check_size (before) = " << temp_check_size << endline_flush;

  IS_PS.end_optimization = atoi(readin[temp_check_size].c_str());
  temp_check_size += 1;
  logger << LOG_DEBUG << "end_optimization = " << IS_PS.end_optimization << endline_flush;

  if (temp_check_size > readin.size()){
    int_end = 2;
    logger << LOG_DEBUG << "int_end = 2   temp_check_size = " << temp_check_size << " > " << readin.size() << " = readin.size()" << endline_flush;
    check_size = false;
  }
  else {
    logger << LOG_DEBUG << "temp_check_size = " << temp_check_size << " <= " << readin.size() << " = readin.size()" << endline_flush;
  }

  if (check_size && IS_PS.end_optimization != -1){
    //  Should be removed accordingly !!!
    //  s(3)
    ///    temp_check_size += 3;

    if (temp_check_size > readin.size()){
      int_end = 2;
      logger << LOG_DEBUG << "int_end = 2   temp_check_size = " << temp_check_size << " > " << readin.size() << " = readin.size()" << endline_flush;
      check_size = false;
      //      return;
    }
    else {
      logger << LOG_DEBUG << "temp_check_size = " << temp_check_size << " <= " << readin.size() << " = readin.size()" << endline_flush;
    }

    temp_check_size += IS_PS.n_gridsize;
    //  alpha(IS_PS.n_gridsize)
    if (temp_check_size > readin.size()){
      int_end = 2;
      logger << LOG_DEBUG << "int_end = 2   temp_check_size = " << temp_check_size << " > " << readin.size() << " = readin.size()" << endline_flush;
      check_size = false;
      //      return;
    }
    else {
      logger << LOG_DEBUG << "temp_check_size = " << temp_check_size << " <= " << readin.size() << " = readin.size()" << endline_flush;
    }
  }

  if (check_size && IS_PS.end_optimization == 0){
    temp_check_size += 2 + 4 * IS_PS.n_gridsize;
    //    temp_check_size += 3 * IS_PS.n_gridsize;
    if (temp_check_size > readin.size()){
      int_end = 2;
      logger << LOG_DEBUG << "int_end = 2   temp_check_size = " << temp_check_size << " > " << readin.size() << " = readin.size()" << endline_flush;
      check_size = false;
      //      return;
    }
    else {
      logger << LOG_DEBUG << "temp_check_size = " << temp_check_size << " <= " << readin.size() << " = readin.size()" << endline_flush;
    }

    if (IS_PS.switch_validation_optimization){
      temp_check_size += 2 + IS_PS.n_optimization_step + IS_PS.n_optimization_step * IS_PS.n_gridsize;
      if (temp_check_size > readin.size()){
	int_end = 2;
	logger << LOG_DEBUG << "int_end = 2   temp_check_size = " << temp_check_size << " > " << readin.size() << " = readin.size()" << endline_flush;
      check_size = false;
      //	return;
      }
      else {
	logger << LOG_DEBUG << "temp_check_size = " << temp_check_size << " <= " << readin.size() << " = readin.size()" << endline_flush;
      }
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << check_size << endline_flush; }
  logger.finished();
  return check_size;
}


void randomvariable::proceeding_group_by_type_out(ofstream & out_proceeding){
  static Logger logger("randomvariable::proceeding_out");
  logger.called();

  IS_PS.proceeding_group_by_type_out(out_proceeding);

  logger.finished();
}


void randomvariable::proceeding_group_by_type_in(int & proc, vector<string> & readin, bool switch_combination) {
  static Logger logger("randomvariable::proceeding_in");
  logger.called();

  IS_PS.proceeding_group_by_type_in(proc, readin, switch_combination);

  logger.finished();
}


ostream & operator << (ostream & s, const randomvariable & IS){
  s << IS.IS_PS;
  return s;
}

