#include "header.hpp"

multicollinear_set::multicollinear_set(){}

multicollinear_set::multicollinear_set(int _no_prc, vector<int> _type_parton, vector<vector<int> > _pdf){
  static Logger logger("multicollinear_set::multicollinear_set(no_prc, type_parton, pdf)");
  logger.called();

  no_prc = _no_prc;
  type_parton = _type_parton;
  pdf = _pdf;

  name = "C";
  all_name = vector<string> (pdf.size(), name);
  type_splitting.resize(1, 0);
  charge_factor.resize(1, 1.);
  in_collinear.resize(1, vector<int> (3, 0));
  no_emitter.resize(1, 0);
  type_correction.resize(1, 0);
  massive.resize(1, 0);
  pair.resize(1, vector<int> (2, 0));
  no_spectator.resize(1, 0);
  no_OL_entry.resize(1, 0);
  no_BLHA_entry.resize(1, 0);

  n_emission.resize(3, 0);
  emission.resize(3, 0);

  previous_splitting.resize(1, 0);
  no_endpoint.resize(4, -1);
  type_splitting_full.resize(3, 0);
  type_splitting_leg.resize(3);
  n_emission_leg_type.resize(3, vector<int> (3, 0));

  no_pdf = -1;
  
  logger.finished();
}

multicollinear_set::multicollinear_set(string _name, vector<string> _all_name, vector<int> _type_splitting, vector<vector<int> > _in_collinear, int _no_prc, vector<int> _type_parton, vector<vector<int> > _pdf, vector<double> _charge_factor, vector<int> _no_emitter, vector<int> _no_spectator, vector<vector<int> > _pair, vector<int> _type_correction, vector<int> _massive){
  static Logger logger("multicollinear_set::multicollinear_set(...)");
  logger.called();

  no_prc = _no_prc;
  type_parton = _type_parton;
  pdf = _pdf;

  name = _name;
  all_name = _all_name;
  type_splitting = _type_splitting;
  charge_factor = _charge_factor;
  in_collinear = _in_collinear;
  no_emitter = _no_emitter;
  type_correction = _type_correction;
  massive = _massive;

  // maybe modify this part: spectators are method-specific !!!
  pair = _pair;
  // needs to be modified - e.g. loop over spectators
  no_spectator = _no_spectator;

  for (int i_c = 0; i_c < no_spectator.size(); i_c++){
    if (no_spectator[i_c] == 0){no_OL_entry[i_c] = 0;}
    else if (no_emitter[i_c] < no_spectator[i_c]){no_OL_entry[i_c] = (no_spectator[i_c] * (no_spectator[i_c] - 1)) / 2 + no_emitter[i_c];}
    else if (no_emitter[i_c] > no_spectator[i_c]){no_OL_entry[i_c] = (no_emitter[i_c] * (no_emitter[i_c] - 1)) / 2 + no_spectator[i_c];}
    else {cout << "Should not happen !" << endl;}

    if (no_spectator[i_c] == 0){no_BLHA_entry[i_c] = 0;}
    else if (no_emitter[i_c] < no_spectator[i_c]){no_BLHA_entry[i_c] = ((no_spectator[i_c] - 1) * (no_spectator[i_c] - 2)) / 2 + (no_emitter[i_c] - 1);}
    else if (no_emitter[i_c] > no_spectator[i_c]){no_BLHA_entry[i_c] = ((no_emitter[i_c] - 1) * (no_emitter[i_c] - 2)) / 2 + (no_spectator[i_c] - 1);}
  }
  
  logger.finished();
}

