#include "header.hpp"

void observable_set::determine_integrand_CX_ncollinear_CT_QCD(){
  static Logger logger("observable_set::determine_integrand_CX_ncollinear_CT_QCD");
  logger.called();

  static int order = 1;

  // In resummed computations and for recoil-driven power corrections, the "artificial" qT is binned:
  if (switch_resummation){
    double qtoverq = sqrt(psi->QT_qt2 / QT_Qres / QT_Qres);
    calculate_coefficients_Itilde(qtoverq, order);
  }
  if (csi->class_contribution_QT_powercorrection){calculate_coefficients_Itilde(psi->recoil_r, order);}

  double global_factor = alpha_S / pi;

  determine_splitting_tH1F_QCD();

  double LQ = log(psi->xbs_all[0][0] / pow(QT_Qres, 2));
  if (QT_Qres == 0){LQ = 0.;}

  //  no power correction implementation for outdated CV variation !!!
  if (switch_CV){
    for (int i_vf = 0; i_vf < value_mu_fact.size(); i_vf++){
      for (int i_mf = 0; i_mf < value_mu_fact[i_vf].size(); i_mf++){
	for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
	  coll_tH1F_QCD_pdf[i_vf][i_mf][i_l] = value_list_pdf_factor[i_vf][i_mf][i_l][0] * coll_tH1F_QCD[i_l];
	}
	value_tH1F_QCD[i_vf][i_mf] = accumulate(coll_tH1F_QCD_pdf[i_vf][i_mf].begin(), coll_tH1F_QCD_pdf[i_vf][i_mf].end(), 0.);
	value_sig12_QCD[i_vf][i_mf] = calculate_sigma12_QCD(value_list_pdf_factor[i_vf][i_mf][0][0]);
	value_sig11_QCD[i_vf][i_mf] = calculate_sigma11_QCD(value_list_pdf_factor[i_vf][i_mf][0][0], value_tH1F_QCD[i_vf][i_mf], LQ);

	if (switch_distribution){
	  for (int i_y = 0; i_y < 3; i_y++){
	    for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
	      coll_tH1F_QCD_pdf_D[i_vf][i_mf][i_y][i_l] = value_list_pdf_factor[i_vf][i_mf][i_l][i_y] * coll_tH1F_QCD[i_l];
	    }
	    value_tH1F_QCD_D[i_vf][i_mf][i_y] = accumulate(coll_tH1F_QCD_pdf_D[i_vf][i_mf][i_y].begin(), coll_tH1F_QCD_pdf_D[i_vf][i_mf][i_y].end(), 0.);
	    value_sig12_QCD_D[i_vf][i_mf][i_y] = calculate_sigma12_QCD(value_list_pdf_factor[i_vf][i_mf][0][i_y]);
	    value_sig11_QCD_D[i_vf][i_mf][i_y] = calculate_sigma11_QCD(value_list_pdf_factor[i_vf][i_mf][0][i_y], value_tH1F_QCD_D[i_vf][i_mf][i_y], LQ);
	  }
	}
      }
    }

    if (switch_resummation){
      for (int i_vf = 0; i_vf < value_mu_fact.size(); i_vf++){
	for (int i_mf = 0; i_mf < value_mu_fact[i_vf].size(); i_mf++){
	  for (int i_q = 0; i_q < n_qTcut; i_q++){
	    if (i_q <= esi->cut_ps[0]){
	      CX_value_integrand_qTcut[i_q][i_vf][i_mf] = global_factor * (value_sig12_QCD[i_vf][i_mf] * LL[2] + value_sig11_QCD[i_vf][i_mf] * LL[1]) / pow(QT_Qres,2) * psi->QT_jacqt2;
	    }
	    else {CX_value_integrand_qTcut[i_q][i_vf][i_mf] = 0.;}
	    if (switch_distribution){
	      for (int i_y = 0; i_y < 3; i_y++){
		if (i_q <= esi->cut_ps[0]){
		  CX_value_integrand_qTcut_D[i_q][i_vf][i_mf][i_y] = global_factor * (value_sig12_QCD_D[i_vf][i_mf][i_y] * LL[2] + value_sig11_QCD_D[i_vf][i_mf][i_y] * LL[1]) / pow(QT_Qres,2) * psi->QT_jacqt2;
		}
		else {CX_value_integrand_qTcut_D[i_q][i_vf][i_mf][i_y] = 0.;}
	      }
	    }
	  }
	}
      }
    }
    else {
      for (int i_vf = 0; i_vf < value_mu_fact.size(); i_vf++){
	for (int i_mf = 0; i_mf < value_mu_fact[i_vf].size(); i_mf++){
	  for (int i_q = 0; i_q < n_qTcut; i_q++){
	    CX_value_integrand_qTcut[i_q][i_vf][i_mf] = global_factor * (value_sig12_QCD[i_vf][i_mf] * int_I[2][i_q] + value_sig11_QCD[i_vf][i_mf] * int_I[1][i_q]);
	    if (switch_distribution){
	      for (int i_y = 0; i_y < 3; i_y++){
		CX_value_integrand_qTcut_D[i_q][i_vf][i_mf][i_y] = global_factor * (value_sig12_QCD_D[i_vf][i_mf][i_y] * int_I[2][i_q] + value_sig11_QCD_D[i_vf][i_mf][i_y] * int_I[1][i_q]);
	      }
	    }
	  }
	}
      }
    }

    integrand = -var_rel_alpha_S * psi->ps_factor * rescaling_factor_alpha_e * ME2 * CX_value_integrand_qTcut[0][dynamic_scale][map_value_scale_fact];

    for (int i_s = 0; i_s < n_scales_CV; i_s++){
      integrand_CV[i_s] = -var_rel_alpha_S_CV[i_s] * psi->ps_factor * rescaling_factor_alpha_e * ME2 * CX_value_integrand_qTcut[0][dynamic_scale_CV][map_value_scale_fact_CV[i_s]];
      for (int i_q = 0; i_q < n_qTcut; i_q++){
	integrand_qTcut_CV[i_q][i_s] = -var_rel_alpha_S_CV[i_s] * psi->ps_factor * rescaling_factor_alpha_e * ME2 * CX_value_integrand_qTcut[i_q][dynamic_scale_CV][map_value_scale_fact_CV[i_s]];
      }
    }

    if (switch_distribution){
      int x_q = 0;
      for (int i_y = 0; i_y < 3; i_y++){
	integrand_D[i_y][0] = -var_rel_alpha_S * psi->ps_factor * rescaling_factor_alpha_e * ME2 * CX_value_integrand_qTcut_D[x_q][dynamic_scale][map_value_scale_fact][i_y];
	if (switch_CV){
	  for (int i_s = 0; i_s < n_scales_CV; i_s++){
	    integrand_D_CV[i_s][i_y][0] = -var_rel_alpha_S_CV[i_s] * psi->ps_factor * rescaling_factor_alpha_e * ME2 * CX_value_integrand_qTcut_D[x_q][dynamic_scale_CV][map_value_scale_fact_CV[i_s]][i_y];
	    for (int i_q = 0; i_q < n_qTcut; i_q++){
	      integrand_D_qTcut_CV[i_q][i_s][i_y][0] = -var_rel_alpha_S_CV[i_s] * psi->ps_factor * rescaling_factor_alpha_e * ME2 * CX_value_integrand_qTcut_D[i_q][dynamic_scale_CV][map_value_scale_fact_CV[i_s]][i_y];
	    }
	  }
	}
      }
    }
  }





  if (switch_TSV){
    for (int i_a = 0; i_a < csi->n_ps; i_a++){
      if (esi->cut_ps[i_a] == -1){continue;}
      for (int i_vf = 0; i_vf < max_dyn_fact + 1; i_vf++){
	for (int i_mf = 0; i_mf < n_scale_dyn_fact[i_vf]; i_mf++){
	  for (int i_y = 0; i_y < 3; i_y++){
	    for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
	      coll_tH1F_QCD_pdf_TSV[i_vf][i_mf][i_y][i_l] = value_list_pdf_factor_TSV[i_a][i_vf][i_mf][i_l][i_y] * coll_tH1F_QCD[i_l];
	    }
	    value_tH1F_QCD_TSV[i_vf][i_mf][i_y] = accumulate(coll_tH1F_QCD_pdf_TSV[i_vf][i_mf][i_y].begin(), coll_tH1F_QCD_pdf_TSV[i_vf][i_mf][i_y].end(), 0.);
	    value_sig12_QCD_TSV[i_vf][i_mf][i_y] = calculate_sigma12_QCD(value_list_pdf_factor_TSV[i_a][i_vf][i_mf][0][i_y]);
	    value_sig11_QCD_TSV[i_vf][i_mf][i_y] = calculate_sigma11_QCD(value_list_pdf_factor_TSV[i_a][i_vf][i_mf][0][i_y], value_tH1F_QCD_TSV[i_vf][i_mf][i_y], LQ);

	    if (switch_resummation){
	      for (int i_q = 0; i_q < n_qTcut; i_q++){
		if (i_q <= esi->cut_ps[0]){
		  value_fact_integrand_qTcut_TSV[i_q][i_vf][i_mf][i_y] = -ME2 * global_factor * (value_sig12_QCD_TSV[i_vf][i_mf][i_y] * LL[2] + value_sig11_QCD_TSV[i_vf][i_mf][i_y] * LL[1]) / pow(QT_Qres,2) * psi->QT_jacqt2;
		}
		else {value_fact_integrand_qTcut_TSV[i_q][i_vf][i_mf][i_y] = 0.;}
	      }
	    }
	    else if (csi->class_contribution_QT_powercorrection){
	      // Check if  value_fact_pc_integrand_qTcut_TSV[i_a][i_vf][i_mf][i_y]  doesn not contribute for rejected points if not re-set !!!
	      //	      if (esi->cut_ps[i_a] == -1){value_fact_pc_integrand_qTcut_TSV[i_a][i_vf][i_mf][i_y] = 0.;}
	      //	      else {
	      value_fact_pc_integrand_qTcut_TSV[i_a][i_vf][i_mf][i_y] = -ME2 * global_factor * (value_sig12_QCD_TSV[i_vf][i_mf][i_y] * LL[2] + value_sig11_QCD_TSV[i_vf][i_mf][i_y] * LL[1]) / psi->xbs_all[0][0] / psi->g_qT;
	      //	      }
	    }
	    else {
	      for (int i_q = 0; i_q < n_qTcut; i_q++){
		value_fact_integrand_qTcut_TSV[i_q][i_vf][i_mf][i_y] = -ME2 * global_factor * (value_sig12_QCD_TSV[i_vf][i_mf][i_y] * int_I[2][i_q] + value_sig11_QCD_TSV[i_vf][i_mf][i_y] * int_I[1][i_q]);
	      }
	    }
	  }
	}
      }
      if (Log::getLogThreshold() <= LOG_DEBUG_POINT){logger << LOG_DEBUG_POINT << "output_sigma_coefficients_CT_QCD: i_a = " << i_a << endline << output_sigma_coefficients_CT_QCD(i_a) << endline_flush;}
    }
  }

  logger.finished();
}


#define FrII right << setw(2)
#define FldXV setprecision(15) << setw(23) << left
#define Fcw left << setw(17)
#define FcpCVw left << setw(22)
#define FvCVw left << setw(18)
#define FcpTSVw left << setw(29)
#define FvTSVw left << setw(22)
string observable_set::output_ME2_CT_QCD(){
  static Logger logger("observable_set::output_ME2_CT_QCD");
  logger.called();

  stringstream out_comparison;
  out_comparison << "Coefficients in CT contribution (qT subtraction):" << endline;
  out_comparison << endline;
  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    if (list_combination_pdf[i_l].size() > 0){
      out_comparison << "Contribution from ncollinear[" << FrII << i_l << "] = " << ncollinear[i_l].name << endline;
      out_comparison << Fcw << "" << "[nc]" << endline;
      out_comparison << Fcw << "coll_tH1F_QCD" << "[" << FrII << i_l << "] = " << FldXV << coll_tH1F_QCD[i_l] << endline;
      out_comparison << endline;
    }
  }
  out_comparison << endline;
  if (switch_CV){
    out_comparison << "CV variation" << endl << endline;
    for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
      if (list_combination_pdf[i_l].size() > 0){
	out_comparison << "Contribution from ncollinear[" << FrII << i_l << "] = " << ncollinear[i_l].name << endline;
	for (int j_l = 0; j_l < list_combination_pdf[i_l].size(); j_l++){
	  stringstream temp_ss;
	  temp_ss << "PDF[" << FrII << j_l << "] (";
	  if (list_combination_pdf[i_l][j_l][0][0] == 1){temp_ss << "+";} else {temp_ss << "-";}
	  temp_ss << ") -> ";
	  for (int k_l = 0; k_l < list_combination_pdf[i_l][j_l].size(); k_l++){
	    if (k_l > 0 && k_l % 10 == 0){temp_ss << endl << setw(15) << "";}
	    stringstream temp_pair;
	    temp_pair << csi->name_particle[list_combination_pdf[i_l][j_l][k_l][1]] << csi->name_particle[list_combination_pdf[i_l][j_l][k_l][2]] << " ";
	    temp_ss << left << setw(5) << temp_pair.str();
	  }
	  out_comparison << temp_ss.str() << endline;
 	}
	out_comparison << endline;
	for (int i_vf = 0; i_vf < value_mu_fact.size(); i_vf++){
	  for (int i_mf = 0; i_mf < value_mu_fact[i_vf].size(); i_mf++){
	    out_comparison << "µ_F = " << value_mu_fact[i_vf][i_mf] << " = " << value_mu_fact_rel[i_vf][i_mf] << " x µ_" << i_vf << endline;
	    out_comparison << FcpCVw << "" << "[µF][xF][nc]" << endline;
	    out_comparison << FcpCVw << "value_list_pdf_factor" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][" << FrII << i_l << "][-] = " << FldXV << value_list_pdf_factor[i_vf][i_mf][i_l][0] << endline;
	    out_comparison << FcpCVw << "coll_tH1F_QCD_pdf" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][" << FrII << i_l << "]    = " << FldXV << coll_tH1F_QCD_pdf[i_vf][i_mf][i_l] << endline;
	    out_comparison << endline;
	  }
	}
      }
    }
    out_comparison << endline;
    for (int i_vf = 0; i_vf < value_mu_fact.size(); i_vf++){
      for (int i_mf = 0; i_mf < value_mu_fact[i_vf].size(); i_mf++){
	out_comparison << "µ_F = " << value_mu_fact[i_vf][i_mf] << " = " << value_mu_fact_rel[i_vf][i_mf] << " x µ_" << i_vf << endline;
	out_comparison << FvCVw << "" << "[µF][xF]" << endline;
	out_comparison << FvCVw << "value_sig11_QCD" << "[" << FrII << i_vf << "][" << FrII << i_mf << "] = " << FldXV << value_sig11_QCD[i_vf][i_mf] << endline;
	out_comparison << FvCVw << "value_sig12_QCD" << "[" << FrII << i_vf << "][" << FrII << i_mf << "] = " << FldXV << value_sig12_QCD[i_vf][i_mf] << endline;
	out_comparison << FvCVw << "value_tH1F_QCD" << "[" << FrII << i_vf << "][" << FrII << i_mf << "] = " << FldXV << value_tH1F_QCD[i_vf][i_mf] << endline;
	out_comparison << endline;
      }
    }
    out_comparison << endline;
  }
  if (switch_TSV){
    out_comparison << "TSV variation" << endl << endline;
    for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
      if (list_combination_pdf[i_l].size() > 0){
	out_comparison << "Contribution from ncollinear[" << FrII << i_l << "] = " << ncollinear[i_l].name << endline;
	for (int j_l = 0; j_l < list_combination_pdf[i_l].size(); j_l++){
	  stringstream temp_ss;
	  temp_ss << "PDF[" << FrII << j_l << "] (";
	  if (list_combination_pdf[i_l][j_l][0][0] == 1){temp_ss << "+";} else {temp_ss << "-";}
	  temp_ss << ") -> ";
	  for (int k_l = 0; k_l < list_combination_pdf[i_l][j_l].size(); k_l++){
	    if (k_l > 0 && k_l % 10 == 0){temp_ss << endl << setw(15) << "";}
	    stringstream temp_pair;
	    temp_pair << csi->name_particle[list_combination_pdf[i_l][j_l][k_l][1]] << csi->name_particle[list_combination_pdf[i_l][j_l][k_l][2]] << " ";
	    temp_ss << left << setw(5) << temp_pair.str();
	  }
	  out_comparison << temp_ss.str() << endline;
	}
	out_comparison << endline;
	for (int i_vf = 0; i_vf < max_dyn_fact + 1; i_vf++){
	  for (int i_mf = 0; i_mf < n_scale_dyn_fact[i_vf]; i_mf++){
	    out_comparison << "µ_F = " << value_scale_fact[0][i_vf][i_mf] << " = " << value_relative_scale_fact[i_vf][i_mf] << " x µ_" << i_vf << endline;
	    out_comparison << FcpTSVw << "" << "[µF][xF]   [nc]" << endline;
	    out_comparison << FcpTSVw << "value_list_pdf_factor_TSV [-]" << "[" << FrII << i_vf << "][" << FrII << i_mf << "]   [" << FrII << i_l << "][-] = " << FldXV << value_list_pdf_factor_TSV[0][i_vf][i_mf][i_l][0] << endline;
	    out_comparison << FcpTSVw << "coll_tH1F_QCD_pdf_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-][" << FrII << i_l << "]    = " << FldXV << coll_tH1F_QCD_pdf_TSV[i_vf][i_mf][0][i_l] << endline;
	    out_comparison << endline;
	  }
	}
      }
    }
    out_comparison << endline;
    out_comparison << "TSV variation - long format" << endl << endline;
    for (int i_vf = 0; i_vf < max_dyn_fact + 1; i_vf++){
      for (int i_mf = 0; i_mf < n_scale_dyn_fact[i_vf]; i_mf++){
	out_comparison << "µ_F = " << value_scale_fact[0][i_vf][i_mf] << " = " << value_relative_scale_fact[i_vf][i_mf] << " x µ_" << i_vf << endline;
	out_comparison << FvTSVw << "" << "[µF][xF]" << endline;
	out_comparison << FvTSVw << "value_tH1F_QCD_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-] = " << FldXV << value_tH1F_QCD_TSV[i_vf][i_mf][0] << endline;
	out_comparison << FvTSVw << "value_sig11_QCD_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-] = " << FldXV << value_sig11_QCD_TSV[i_vf][i_mf][0] << endline;
	out_comparison << FvTSVw << "value_sig12_QCD_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-] = " << FldXV << value_sig12_QCD_TSV[i_vf][i_mf][0] << endline;
	out_comparison << endline;
      }
    }
    out_comparison << endline;
    out_comparison << "TSV variation - compact format" << endl << endline;
    out_comparison << FvTSVw << "" << "[µF][xF]" << endline;
    for (int i_vf = 0; i_vf < max_dyn_fact + 1; i_vf++){
      for (int i_mf = 0; i_mf < n_scale_dyn_fact[i_vf]; i_mf++){
	out_comparison << FvTSVw << "µ_F" << " " << "[" << FrII << i_vf << "][" << FrII << i_mf << "]    = " << FldXV << value_scale_fact[0][i_vf][i_mf] << " = " << setw(5) << value_relative_scale_fact[i_vf][i_mf] << " x µ_" << i_vf << endline;
      }
    }
    out_comparison << endline;
    for (int i_vf = 0; i_vf < max_dyn_fact + 1; i_vf++){
      for (int i_mf = 0; i_mf < n_scale_dyn_fact[i_vf]; i_mf++){
	out_comparison << FvTSVw << "value_tH1F_QCD_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-] = " << FldXV << value_tH1F_QCD_TSV[i_vf][i_mf][0] << endline;
 	out_comparison << FvTSVw << "value_sig11_QCD_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-] = " << FldXV << value_sig11_QCD_TSV[i_vf][i_mf][0] << endline;
 	out_comparison << FvTSVw << "value_sig12_QCD_TSV" << "[" << FrII << i_vf << "][" << FrII << i_mf << "][-] = " << FldXV << value_sig12_QCD_TSV[i_vf][i_mf][0] << endline;
      }
    }
  }
  out_comparison << endline;

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << out_comparison.str() << endline_flush; }
  logger.finished();
  return out_comparison.str();
}


string observable_set::output_sigma_coefficients_CT_QCD(int i_a){
  static Logger logger("observable_set::output_sigma_coefficients_CT_QCD");
  logger.called();

  stringstream out_comparison;
  for (int i_vf = 0; i_vf < max_dyn_fact + 1; i_vf++){
    for (int i_mf = 0; i_mf < n_scale_dyn_fact[i_vf]; i_mf++){
      int i_y = 0;
      //for (int i_y = 0; i_y < 3; i_y++){
      if (i_mf != 1 || i_y != 0){continue;}
      if (csi->class_contribution_QT_powercorrection){
	if (i_a == 0){out_comparison << "Born phase space point" << endline;}
	else if (i_a == 1){out_comparison << "Phase space point with qT recoil" << endline;}
	out_comparison << "   LL[1] = " << setw(23) << setprecision(15) << LL[1] 
		       << "   LL[2] = " << setw(23) << setprecision(15) << LL[2]
		       << endline;
      }
      if (i_a == 0){
	out_comparison << endline;
	out_comparison << "   sig11 = " << setw(23) << setprecision(15) << value_sig11_QCD_TSV[i_vf][i_mf][i_y]
		       << "   sig12 = " << setw(23) << setprecision(15) << value_sig12_QCD_TSV[i_vf][i_mf][i_y]
		       << endline;
      }
      if (!csi->class_contribution_QT_powercorrection){
	for (int i_q = 0; i_q < n_qTcut; i_q++){
	  out_comparison << "   int_I[1][" << i_q << "] = " << setw(23) << setprecision(15) << int_I[1][i_q] 
			 << "   int_I[2][" << i_q << "] = " << setw(23) << setprecision(15) << int_I[2][i_q]
			 << endline;
	}
      }
      out_comparison << "value_fact_integrand_qTcut_TSV[0][" << i_vf << "][" << i_mf << "][0] = " << value_fact_integrand_qTcut_TSV[0][i_vf][i_mf][0] / (-ME2) << endline;
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "return " << endline << out_comparison.str() << endline_flush; }
  logger.finished();
  return out_comparison.str();
}


