#include "header.hpp"

//  important issues:
// - check which version in H2 is correct (potentially doubled term in VT2 contribution, which is already included via H1full -> relevant if LQ =/= 0) !!!
// - check if 'scale-dependent' parts are really scale dependent !!!
// - check missing LQ contributions in gg channel !!!

void observable_set::calculate_A_F(){
  static Logger logger("observable_set::calculate_A_F");
  logger.called();

  if (initial_gg){
    A_F = 2 * (QT_H1_delta_QCD - pi2_6 * C_A);
  }
  else if (initial_qqx){
    A_F = 2 * (QT_H1_delta_QCD - pi2_6 * C_F);
  }
  else {logger << LOG_FATAL << "Wrong process specified." << endline_flush; exit(1);}

  logger.finished();
}


void observable_set::calculate_B2_A_F(){
  static Logger logger("observable_set::calculate_B2_A_F");
  logger.called();

  if (initial_gg){
    B2_A_F = B2g + .5 * beta0 * A_F;
  }
  else if (initial_qqx){
    B2_A_F = B2q + .5 * beta0 * A_F;
  }
  else {logger << LOG_FATAL << "Wrong process specified." << endline_flush; exit(1);}

  logger.finished();
}





// All functions below are only relevant for CV variation:

void observable_set::determine_psp_weight_QT(){
  static Logger logger("observable_set::determine_psp_weight_QT");
  logger.called();

  // shifted to determine_psp_weight_TSV, to eventually remove this function completely with CV:
  if (!switch_TSV){
    this_psp_weight = integrand;
    this_psp_weight2 = pow(this_psp_weight, 2.);
    step_sum_weight += this_psp_weight;
    step_sum_weight2 += this_psp_weight2;
  }

  if (switch_CV){
    for (int i_s = 0; i_s < n_scales_CV; i_s++){
      for (int i_q = 0; i_q < output_n_qTcut; i_q++){
	this_psp_weight_CV[i_q][i_s] = integrand_qTcut_CV[i_q][i_s];
	this_psp_weight2_CV[i_q][i_s] = pow(this_psp_weight_CV[i_q][i_s], 2.);
	step_sum_weight_CV[i_q][i_s] += this_psp_weight_CV[i_q][i_s];
	step_sum_weight2_CV[i_q][i_s] += this_psp_weight2_CV[i_q][i_s];
      }
    }
  }

  logger.finished();
}

