#include "header.hpp"

void summary_list::extrapolation_linear(vector<double> & xvalue, double & min_extrapolation, double & max_extrapolation, vector<double> & data_result, vector<double> & data_deviation2, vector<double> & X, vector<double> & dX, double & chi2_dof){
  static Logger logger("summary_list::extrapolation_linear");
  logger.called();

  X.resize(2, 0.);
  dX.resize(2, 0.);

  vector<double> A(3, 0.);
  vector<double> B(2, 0.);
  vector<vector<double> > Ainv(3, vector<double> (3, 0.));
  double detA;
  vector<vector<double> > H(3, vector<double> (3, 0.));
  vector<vector<double> > Hinv(3, vector<double> (3, 0.));
  double detH;

  for (int i_q = 0; i_q < xvalue.size(); i_q++){
    if (xvalue[i_q] < min_extrapolation){continue;}
    if (xvalue[i_q] > max_extrapolation){break;}
    if (data_deviation2[i_q] != 0.){
      for (int i = 0; i < 3; i++){A[i] += pow(xvalue[i_q], i) / data_deviation2[i_q];}
      for (int i = 0; i < 2; i++){B[i] += pow(xvalue[i_q], i) * data_result[i_q] / data_deviation2[i_q];}
    }
  }

  Ainv[1][1] = A[0];
  Ainv[1][2] = -A[1];
  Ainv[2][2] = A[2];

  detA = A[0] * A[2] - A[1] * A[1];

  H[1][1] = 2 * A[2];
  H[1][2] = 2 * A[1];
  H[2][2] = 2 * A[0];

  detH = H[1][1] * H[2][2] - H[1][2] * H[1][2];

  Hinv[1][1] = H[2][2] / detH;
  Hinv[2][2] = H[1][1] / detH;
  Hinv[1][2] = -H[1][2] / detH;

  if (detA != 0.){
    X[1] = (Ainv[1][1] * B[1] + Ainv[1][2] * B[0]) / detA;
    X[0] = (Ainv[1][2] * B[1] + Ainv[2][2] * B[0]) / detA;

    dX[0] = sqrt(Hinv[1][1]);
    dX[1] = sqrt(Hinv[2][2]);
  }

  int n_data_point = 0;
  double chi2 = 0.;
  for (int i_q = 0; i_q < xvalue.size(); i_q++){
    if (xvalue[i_q] < min_extrapolation){continue;}
    if (xvalue[i_q] > max_extrapolation){break;}
    n_data_point++;
    //    chi2 += abs((data_result[i_q] - (X[0] + xvalue[i_q] * X[1])) / sqrt(data_deviation2[i_q]));
    chi2 += pow(data_result[i_q] - (X[0] + xvalue[i_q] * X[1]), 2) / data_deviation2[i_q];
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "xvalue[" << setw(3) << i_q << "] = " << setw(5) << xvalue[i_q] << "   chi2 = " << setw(15) << setprecision(8) << chi2 << "   chi2_dof = " << setw(15) << setprecision(8) << (chi2 / (n_data_point - 3)) << endline_flush; }
  }
  chi2_dof = chi2 / (n_data_point - 2);
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "min_extrapolation = " << setw(5) << min_extrapolation << "   max_extrapolation = " << setw(5) << max_extrapolation << "   chi2_dof = " << setw(15) << setprecision(8) << chi2_dof << endline_flush; }

  logger.finished();
}


void summary_list::extrapolation_quadratic(vector<double> & xvalue, double & min_extrapolation, double & max_extrapolation, vector<double> & data_result, vector<double> & data_deviation2, vector<double> & X, vector<double> & dX, double & chi2_dof){
  static Logger logger("summary_list::extrapolation_quadratic");
  logger.called();

  X.resize(3, 0.);
  dX.resize(3, 0.);

  vector<double> A(5, 0.);
  vector<double> B(3, 0.);
  vector<vector<double> > Ainv(4, vector<double> (4, 0.));
  double detA;
  vector<vector<double> > H(4, vector<double> (4, 0.));
  vector<vector<double> > Hinv(4, vector<double> (4, 0.));
  double detH;

  for (int i_q = 0; i_q < xvalue.size(); i_q++){
    if (xvalue[i_q] < min_extrapolation){continue;}
    if (xvalue[i_q] > max_extrapolation){break;}
    if (data_deviation2[i_q] != 0.){
      for (int i = 0; i < 5; i++){A[i] += pow(xvalue[i_q], i) / data_deviation2[i_q];}
      for (int i = 0; i < 3; i++){B[i] += pow(xvalue[i_q], i) * data_result[i_q] / data_deviation2[i_q];}
    }
  }

  Ainv[1][1] = A[0] * A[2] - A[1] * A[1];
  Ainv[1][2] = A[1] * A[2] - A[0] * A[3];
  Ainv[1][3] = A[1] * A[3] - A[2] * A[2];

  Ainv[2][2] = A[0] * A[4] - A[2] * A[2];
  Ainv[2][3] = A[2] * A[3] - A[1] * A[4];

  Ainv[3][3] = A[2] * A[4] - A[3] * A[3];

  detA = A[0] * A[2] * A[4] + 2 * A[1] * A[2] * A[3] - A[0] * A[3] * A[3] - A[4] * A[1] * A[1] - A[2] * A[2] * A[2];

  H[1][1] = 2 * A[4];
  H[1][2] = 2 * A[3];
  H[1][3] = 2 * A[2];
  H[2][2] = 2 * A[2];
  H[2][3] = 2 * A[1];
  H[3][3] = 2 * A[0];

  detH = H[1][1] * H[2][2] * H[3][3] + 2 * H[1][2] * H[2][3] * H[1][3] - H[2][2] * H[1][3] * H[1][3] - H[3][3] * H[1][2] * H[1][2] - H[1][1] * H[2][3] * H[2][3];

  Hinv[1][1] = (H[2][2] * H[3][3] - H[2][3] * H[2][3]) / detH;
  Hinv[1][2] = (H[1][3] * H[2][3] - H[1][2] * H[3][3]) / detH;
  Hinv[1][3] = (H[1][2] * H[2][3] - H[1][3] * H[2][2]) / detH;

  Hinv[2][2] = (H[1][1] * H[3][3] - H[1][3] * H[1][3]) / detH;
  Hinv[2][3] = (H[1][3] * H[1][2] - H[1][1] * H[2][3]) / detH;

  Hinv[3][3] = (H[1][1] * H[2][2] - H[1][2] * H[1][2]) / detH;

  if (detA != 0.){
    X[2] = (Ainv[1][1] * B[2] + Ainv[1][2] * B[1] + Ainv[1][3] * B[0]) / detA;
    X[1] = (Ainv[1][2] * B[2] + Ainv[2][2] * B[1] + Ainv[2][3] * B[0]) / detA;
    X[0] = (Ainv[1][3] * B[2] + Ainv[2][3] * B[1] + Ainv[3][3] * B[0]) / detA;

    dX[2] = Hinv[1][1];
    dX[1] = Hinv[2][2];
    dX[0] = Hinv[3][3];
  }

  int n_data_point = 0;
  double chi2 = 0.;
  for (int i_q = 0; i_q < xvalue.size(); i_q++){
    if (xvalue[i_q] < min_extrapolation){continue;}
    if (xvalue[i_q] > max_extrapolation){break;}
    n_data_point++;
    chi2 += pow(data_result[i_q] - (X[0] + xvalue[i_q] * X[1] + pow(xvalue[i_q], 2) * X[2]), 2) / data_deviation2[i_q];
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "xvalue[" << setw(3) << i_q << "] = " << setw(5) << xvalue[i_q] << "   chi2 = " << setw(15) << setprecision(8) << chi2 << "   chi2_dof = " << setw(15) << setprecision(8) << (chi2 / (n_data_point - 3)) << endline_flush; }
  }
  chi2_dof = chi2 / (n_data_point - 3);
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "min_extrapolation = " << setw(5) << min_extrapolation << "   max_extrapolation = " << setw(5) << max_extrapolation << "   chi2_dof = " << setw(15) << setprecision(8) << chi2_dof << endline_flush; }

  logger.finished();
}


