#include "header.hpp"

void summary_order::collect_distribution_TSV(){
  static Logger logger("summary_order::collect_distribution_TSV");
  logger.called();

  ///////////////////////////////////////////////////////////
  //  initialization of distribution_result/deviation_TSV  // -> extrapolated or rcut-independent
  ///////////////////////////////////////////////////////////

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "initialization of   distribution_result_TSV   started" << endline_flush; }

  distribution_result_TSV.resize(ygeneric->subgroup.size(), vector<vector<vector<vector<vector<double> > > > > (osi->extended_distribution.size()));
  distribution_deviation_TSV.resize(ygeneric->subgroup.size(), vector<vector<vector<vector<vector<double> > > > > (osi->extended_distribution.size()));
  for (int i_g = 0; i_g < ygeneric->subgroup.size(); i_g++){
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
      if (!ygeneric->switch_output_distribution[i_d]){continue;}
      distribution_result_TSV[i_g][i_d].resize(osi->extended_distribution[i_d].n_bins);
      distribution_deviation_TSV[i_g][i_d].resize(osi->extended_distribution[i_d].n_bins);
      for (int i_b = 0; i_b < osi->extended_distribution[i_d].n_bins; i_b++){
	distribution_result_TSV[i_g][i_d][i_b].resize(osi->n_extended_set_TSV);
	distribution_deviation_TSV[i_g][i_d][i_b].resize(osi->n_extended_set_TSV);
	for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	  if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	  if (!osi->switch_distribution_TSV[i_s]){continue;}
	  distribution_result_TSV[i_g][i_d][i_b][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	  distribution_deviation_TSV[i_g][i_d][i_b][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	}
      }
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "initialization of   distribution_result_TSV   finished" << endline_flush; }

  /////////////////////////////////////////////////////////////////
  //  initialization of distribution_result/deviation_qTcut_TSV  // -> rcut-dependent
  /////////////////////////////////////////////////////////////////

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "initialization of   distribution_result_qTcut_TSV   started" << endline_flush; }

  distribution_result_qTcut_TSV.resize(selection_n_qTcut, vector<vector<vector<vector<vector<vector<double> > > > > > (ygeneric->subgroup.size(), vector<vector<vector<vector<vector<double> > > > > (osi->extended_distribution.size())));
  distribution_deviation_qTcut_TSV.resize(selection_n_qTcut, vector<vector<vector<vector<vector<vector<double> > > > > > (ygeneric->subgroup.size(), vector<vector<vector<vector<vector<double> > > > > (osi->extended_distribution.size())));
  for (int i_q = 0; i_q < selection_n_qTcut; i_q++){
    for (int i_g = 0; i_g < ygeneric->subgroup.size(); i_g++){
      for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
	if (!ygeneric->switch_output_distribution[i_d]){continue;}
	distribution_result_qTcut_TSV[i_q][i_g][i_d].resize(osi->extended_distribution[i_d].n_bins);
	distribution_deviation_qTcut_TSV[i_q][i_g][i_d].resize(osi->extended_distribution[i_d].n_bins);
	for (int i_b = 0; i_b < osi->extended_distribution[i_d].n_bins; i_b++){
	  distribution_result_qTcut_TSV[i_q][i_g][i_d][i_b].resize(osi->n_extended_set_TSV);
	  distribution_deviation_qTcut_TSV[i_q][i_g][i_d][i_b].resize(osi->n_extended_set_TSV);
	  for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	    if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	    if (!osi->switch_distribution_TSV[i_s]){continue;}
	    distribution_result_qTcut_TSV[i_q][i_g][i_d][i_b][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	    distribution_deviation_qTcut_TSV[i_q][i_g][i_d][i_b][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	  }
	}
      }
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "initialization of   distribution_result_qTcut_TSV   finished" << endline_flush; }

  ///////////////////////////////////////////////////
  //  calculate distribution_result/deviation_TSV  //
  ///////////////////////////////////////////////////

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "general combination (qTcut-independent/extrapolated part) started" << endline_flush; }
  for (int i_g = 0; i_g < ygeneric->subgroup.size(); i_g++){ //  logger << LOG_DEBUG_VERBOSE << "i_g = " << i_g << endline_flush;
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){ //  logger << LOG_DEBUG_VERBOSE << "i_d = " << i_d << endline_flush;
      if (!ygeneric->switch_output_distribution[i_d]){continue;}
      for (int i_b = 0; i_b < osi->extended_distribution[i_d].n_bins; i_b++){ //  logger << LOG_DEBUG_VERBOSE << "i_b = " << i_b << endline_flush;
	for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){ //  logger << LOG_DEBUG_VERBOSE << "i_s = " << i_s << endline_flush;
	  if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	  if (!osi->switch_distribution_TSV[i_s]){continue;}
	  for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){ //  logger << LOG_DEBUG_VERBOSE << "i_r = " << i_r << endline_flush;
	    for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){ //  logger << LOG_DEBUG_VERBOSE << "i_f = " << i_f << endline_flush;
	      vector<double> combination_result(combination.size(), 0.);
	      vector<double> combination_deviation2(combination.size(), 0.);
	      for (int i_c = 0; i_c < combination.size(); i_c++){
		for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
		  int i_l = combination[i_c][j_c];
		  combination_result[i_c] += xlist[i_l]->distribution_result_TSV[i_g][i_d][i_b][i_s][i_r][i_f];
		  combination_deviation2[i_c] += pow(xlist[i_l]->distribution_deviation_TSV[i_g][i_d][i_b][i_s][i_r][i_f], 2);
		}
	      }
	      calculate_combination_result(combination_result, combination_deviation2, distribution_result_TSV[i_g][i_d][i_b][i_s][i_r][i_f], distribution_deviation_TSV[i_g][i_d][i_b][i_s][i_r][i_f]);
	    }
	  }
	}
      }
    }
  }
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "general combination (qTcut-independent/extrapolated part) finished" << endline_flush; }

  /////////////////////////////////////////////////////////
  //  calculate distribution_result/deviation_qTcut_TSV  //
  /////////////////////////////////////////////////////////

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "general combination (qTcut-dependent part) started" << endline_flush; }
  for (int i_g = 0; i_g < ygeneric->subgroup.size(); i_g++){ //  logger << LOG_DEBUG_VERBOSE << "i_g = " << i_g << endline_flush;
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){ //  logger << LOG_DEBUG_VERBOSE << "i_d = " << i_d << endline_flush;
      if (!ygeneric->switch_output_distribution[i_d]){continue;}
      for (int i_b = 0; i_b < osi->extended_distribution[i_d].n_bins; i_b++){ //  logger << LOG_DEBUG_VERBOSE << "i_b = " << i_b << endline_flush;
	for (int i_q = 0; i_q < selection_n_qTcut; i_q++){ //  logger << LOG_DEBUG_VERBOSE << "i_q = " << i_q << endline_flush;
	  for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){ //  logger << LOG_DEBUG_VERBOSE << "i_s = " << i_s << endline_flush;
	    if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	    if (!osi->switch_distribution_TSV[i_s]){continue;}
	    for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	      for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
		vector<double> combination_result(combination.size(), 0.);
		vector<double> combination_deviation2(combination.size(), 0.);
		for (int i_c = 0; i_c < combination.size(); i_c++){
		  for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
		    int i_l = combination[i_c][j_c];
		    int y_q = 0;
		    if (xlist[i_l]->xcontribution[0]->active_qTcut){y_q = i_q;}
		    combination_result[i_c] += xlist[i_l]->xcontribution[0]->distribution_result_TSV[i_g][i_d][i_b][y_q][i_s][i_r][i_f];
		    combination_deviation2[i_c] += pow(xlist[i_l]->xcontribution[0]->distribution_deviation_TSV[i_g][i_d][i_b][y_q][i_s][i_r][i_f], 2);
		  }
		}
		calculate_combination_result(combination_result, combination_deviation2, distribution_result_qTcut_TSV[i_q][i_g][i_d][i_b][i_s][i_r][i_f], distribution_deviation_qTcut_TSV[i_q][i_g][i_d][i_b][i_s][i_r][i_f]);
	      }
	    }
	  }
	}
      }
    }
  }
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "general combination (qTcut-dependent part) finished" << endline_flush; }

  if (ygeneric->switch_output_order){
    if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "output for TSV variation" << endline_flush; }
    if (ygeneric->switch_output_overview > 0){
      output_distribution_overview_TSV();
      output_dddistribution_overview_TSV();
    }
    if (ygeneric->switch_output_plot > 0){
      if (active_qTcut && ygeneric->switch_output_qTcut > 0){output_distribution_qTcut_TSV();}
      output_distribution_TSV();
    }
  }

  logger.finished();
}





// All functions below are only relevant for CV variation:

void summary_order::collect_distribution_CV(){
  static Logger logger("summary_order::collect_distribution_CV");
  logger.called();

  distribution_result_CV.resize(ygeneric->subgroup.size(), vector<vector<vector<double> > > (osi->extended_distribution.size()));
  distribution_deviation_CV.resize(ygeneric->subgroup.size(), vector<vector<vector<double> > > (osi->extended_distribution.size()));
  for (int i_g = 0; i_g < ygeneric->subgroup.size(); i_g++){
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
      if (!ygeneric->switch_output_distribution[i_d]){continue;}
      distribution_result_CV[i_g][i_d].resize(osi->extended_distribution[i_d].n_bins, vector<double> (osi->n_scales_CV, 0.));
      distribution_deviation_CV[i_g][i_d].resize(osi->extended_distribution[i_d].n_bins, vector<double> (osi->n_scales_CV, 0.));
    }
  }

  logger << LOG_DEBUG_VERBOSE << "general combination (only qTcut-independent part at value minqTcut available) started" << endline_flush;
  for (int i_g = 0; i_g < ygeneric->subgroup.size(); i_g++){logger << LOG_DEBUG_VERBOSE << "i_g = " << i_g << endline_flush;
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){logger << LOG_DEBUG_VERBOSE << "i_d = " << i_d << endline_flush;
      if (!ygeneric->switch_output_distribution[i_d]){continue;}
      for (int i_b = 0; i_b < osi->extended_distribution[i_d].n_bins; i_b++){logger << LOG_DEBUG_VERBOSE << "i_b = " << i_b << endline_flush;
	for (int i_s = 0; i_s < osi->n_scales_CV; i_s++){
	  vector<double> combination_result(combination.size(), 0.);
	  vector<double> combination_deviation2(combination.size(), 0.);
	  for (int i_c = 0; i_c < combination.size(); i_c++){
	    for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
	      int i_l = combination[i_c][j_c];
	      combination_result[i_c] += xlist[i_l]->xcontribution[0]->distribution_result_CV[i_g][i_d][i_b][i_s];
	      combination_deviation2[i_c] += pow(xlist[i_l]->xcontribution[0]->distribution_deviation_CV[i_g][i_d][i_b][i_s], 2);
	    }
	  }
	  calculate_combination_result(combination_result, combination_deviation2, distribution_result_CV[i_g][i_d][i_b][i_s], distribution_deviation_CV[i_g][i_d][i_b][i_s]);
	}
      }
    }
  }
  logger << LOG_DEBUG_VERBOSE << "general combination (only qTcut-independent part at value minqTcut available) finished" << endline_flush;


  if (ygeneric->switch_output_order){
    if (ygeneric->switch_output_overview > 0){output_distribution_overview_CV();}
    if (ygeneric->switch_output_plot > 0){output_distribution_CV();}
  }

  logger.finished();
}

