#include "header.hpp"

void summary_order::collect_result_TSV(vector<string> & subgroup){
  static Logger logger("summary_order::collect_result_TSV");
  logger.called();

  result_TSV.resize(osi->n_moments, vector<vector<vector<vector<double> > > > (subgroup.size(), vector<vector<vector<double> > > (osi->n_extended_set_TSV)));
  deviation_TSV.resize(osi->n_moments, vector<vector<vector<vector<double> > > > (subgroup.size(), vector<vector<vector<double> > > (osi->n_extended_set_TSV)));

  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 0; i_m < osi->n_moments; i_m++){
      for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	result_TSV[i_m][i_g][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	deviation_TSV[i_m][i_g][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
      }
    }
  }

  result_qTcut_TSV.resize(osi->n_moments, vector<vector<vector<vector<vector<double> > > > > (subgroup.size(), vector<vector<vector<vector<double> > > > (output_n_qTcut, vector<vector<vector<double> > > (osi->n_extended_set_TSV))));
  deviation_qTcut_TSV.resize(osi->n_moments, vector<vector<vector<vector<vector<double> > > > > (subgroup.size(), vector<vector<vector<vector<double> > > > (output_n_qTcut, vector<vector<vector<double> > > (osi->n_extended_set_TSV))));

  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 0; i_m < osi->n_moments; i_m++){
      for (int i_q = 0; i_q < output_n_qTcut; i_q++){
	for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	  if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	  if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	  result_qTcut_TSV[i_m][i_g][i_q][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	  deviation_qTcut_TSV[i_m][i_g][i_q][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	}
      }
    }
  }

  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "general combination (qTcut-dependent and qTcut-independent/extrapolated part) started" << endline_flush; }

  // qTcut-independent / extrapolated results: calculation of final results
  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 0; i_m < osi->n_moments; i_m++){
      for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
	    vector<double> combination_result(combination.size(), 0.);
	    vector<double> combination_deviation2(combination.size(), 0.);
	    for (int i_c = 0; i_c < combination.size(); i_c++){
	      for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
		int i_l = combination[i_c][j_c];
		combination_result[i_c] += xlist[i_l]->result_TSV[i_m][i_g][i_s][i_r][i_f];
		combination_deviation2[i_c] += pow(xlist[i_l]->deviation_TSV[i_m][i_g][i_s][i_r][i_f], 2);
	      }
	    }
	    calculate_combination_result(combination_result, combination_deviation2, result_TSV[i_m][i_g][i_s][i_r][i_f], deviation_TSV[i_m][i_g][i_s][i_r][i_f]);
	  }
	}
      }
    }
  }

  // qTcut-dependent results: calculation of final results
  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 0; i_m < osi->n_moments; i_m++){
      for (int i_q = 0; i_q < output_n_qTcut; i_q++){
	for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	  if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	  if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	  for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	    for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
	      vector<double> combination_result(combination.size(), 0.);
	      vector<double> combination_deviation2(combination.size(), 0.);
	      for (int i_c = 0; i_c < combination.size(); i_c++){
		for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
		  int i_l = combination[i_c][j_c];
		  combination_result[i_c] += xlist[i_l]->xcontribution[0]->result_TSV[i_m][i_g][i_q][i_s][i_r][i_f];
		  combination_deviation2[i_c] += pow(xlist[i_l]->xcontribution[0]->deviation_TSV[i_m][i_g][i_q][i_s][i_r][i_f], 2);
		}
	      }
	      calculate_combination_result(combination_result, combination_deviation2, result_qTcut_TSV[i_m][i_g][i_q][i_s][i_r][i_f], deviation_qTcut_TSV[i_m][i_g][i_q][i_s][i_r][i_f]);
	    }
	  }
	}
      }
    }
  }
  
  if (Log::getLogThreshold() <= LOG_DEBUG_VERBOSE){ logger << LOG_DEBUG_VERBOSE << "general combination (qTcut-dependent and qTcut-independent/extrapolated part) finished" << endline_flush; }

  // very non-general implementation so far !!!
  if (osi->n_moments > 1){calculate_moments_TSV(subgroup);}  //  > 0  or  > 1  ???

  
  if (ygeneric->switch_output_order){
    if (ygeneric->switch_output_overview > 0){ output_overview_TSV(); }
    if (ygeneric->switch_output_result > 0){ output_result_TSV(); }
    if (ygeneric->switch_output_plot > 0){
      output_plot_TSV();
      if (ygeneric->switch_output_band){ output_band_TSV(); }
    }
    if (ygeneric->switch_output_qTcut > 0){
      output_plot_qTcut_TSV();
      if (ygeneric->switch_output_band){ output_band_qTcut_TSV(); }
    }
    if (ygeneric->switch_output_script > 0){
      if (ygeneric->switch_output_plot > 0){
	output_plot_minqTcut_TSV();
	if (ygeneric->switch_output_band){ output_band_minqTcut_TSV(); }
	output_plot_extrapolated_TSV();
	if (ygeneric->switch_output_band){ output_band_extrapolated_TSV(); }
      }
    }
  }

  logger.finished();
}


void summary_order::calculate_moments_TSV(vector<string> & subgroup){
  static Logger logger("summary_order::calculate_moments_TSV");
  logger.called();

  vector<vector<vector<vector<vector<double> > > > > moment_result_TSV = result_TSV;
  vector<vector<vector<vector<vector<double> > > > > moment_deviation_TSV = deviation_TSV;

  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 1; i_m < osi->n_moments; i_m++){
      for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
	    // normalization: divide by cross section to get the moment
	    result_TSV[i_m][i_g][i_s][i_r][i_f] = result_TSV[i_m][i_g][i_s][i_r][i_f] / result_TSV[0][i_g][i_s][i_r][i_f];
	    deviation_TSV[i_m][i_g][i_s][i_r][i_f] = deviation_TSV[i_m][i_g][i_s][i_r][i_f] / result_TSV[0][i_g][i_s][i_r][i_f];
	  }
	}
      }
    }
  }

  vector<vector<vector<vector<vector<double> > > > > save_result_TSV = result_TSV;
  vector<vector<vector<vector<vector<double> > > > > save_deviation_TSV = deviation_TSV;

  if (ygeneric->switch_output_result > 0){output_moment_TSV("moment");}

  if (ygeneric->switch_output_result > 0){output_moment_TSV("ln_muR");}

  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 1; i_m < osi->n_moments; i_m++){
      for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
	    // normalization: divide by cross section to get moment
	    result_TSV[i_m][i_g][i_s][i_r][i_f] = exp(result_TSV[i_m][i_g][i_s][i_r][i_f]);
	    deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	  }
	}
      }
    }
  }

  if (ygeneric->switch_output_result > 0){output_moment_TSV("exp_ln_muR");}
  result_TSV = save_result_TSV;

  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 1; i_m < osi->n_moments; i_m++){
      for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
	    if (i_m == 1){
	      moment_result_TSV[i_m][i_g][i_s][i_r][i_f] = result_TSV[i_m][i_g][i_s][i_r][i_f] - result_TSV[1][i_g][i_s][i_r][i_f];
	      moment_deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	    }
	    if (i_m == 2){
	      moment_result_TSV[i_m][i_g][i_s][i_r][i_f] =
		result_TSV[i_m][i_g][i_s][i_r][i_f]
		- 2 * result_TSV[i_m - 1][i_g][i_s][i_r][i_f] * result_TSV[1][i_g][i_s][i_r][i_f]
		+ pow(result_TSV[1][i_g][i_s][i_r][i_f], 2);
	      moment_deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	    }
	    if (i_m == 3){
	      moment_result_TSV[i_m][i_g][i_s][i_r][i_f] =
		result_TSV[i_m][i_g][i_s][i_r][i_f]
		- 3 * result_TSV[i_m - 1][i_g][i_s][i_r][i_f] * result_TSV[1][i_g][i_s][i_r][i_f]
		+ 3 * result_TSV[i_m - 2][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 2)
		- pow(result_TSV[1][i_g][i_s][i_r][i_f], 3);
	      moment_deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	    }
	    if (i_m == 4){
	      moment_result_TSV[i_m][i_g][i_s][i_r][i_f] =
		result_TSV[i_m][i_g][i_s][i_r][i_f]
		- 4 * result_TSV[i_m - 1][i_g][i_s][i_r][i_f] * result_TSV[1][i_g][i_s][i_r][i_f]
		+ 6 * result_TSV[i_m - 2][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 2)
		- 4 * result_TSV[i_m - 3][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 3)
		+ pow(result_TSV[1][i_g][i_s][i_r][i_f], 4);
	      moment_deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	    }
	    if (i_m == 5){
	      moment_result_TSV[i_m][i_g][i_s][i_r][i_f] =
		result_TSV[i_m][i_g][i_s][i_r][i_f]
		- 5 * result_TSV[i_m - 1][i_g][i_s][i_r][i_f] * result_TSV[1][i_g][i_s][i_r][i_f]
		+ 10 * result_TSV[i_m - 2][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 2)
		- 10 * result_TSV[i_m - 3][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 3)
		+ 5 * result_TSV[i_m - 4][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 4)
		- pow(result_TSV[1][i_g][i_s][i_r][i_f], 5);
	      moment_deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	    }
	    if (i_m == 6){
	      moment_result_TSV[i_m][i_g][i_s][i_r][i_f] =
		result_TSV[i_m][i_g][i_s][i_r][i_f]
		- 6 * result_TSV[i_m - 1][i_g][i_s][i_r][i_f] * result_TSV[1][i_g][i_s][i_r][i_f]
		+ 15 * result_TSV[i_m - 2][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 2)
		- 20 * result_TSV[i_m - 3][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 3)
		+ 15 * result_TSV[i_m - 4][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 4)
		- 6 * result_TSV[i_m - 5][i_g][i_s][i_r][i_f] * pow(result_TSV[1][i_g][i_s][i_r][i_f], 5)
		+ pow(result_TSV[1][i_g][i_s][i_r][i_f], 6);
	      moment_deviation_TSV[i_m][i_g][i_s][i_r][i_f] = 0.;
	    }
	  }
	}
      }
    }
  }
  result_TSV = moment_result_TSV;
  if (ygeneric->switch_output_result > 0){output_moment_TSV("f_N");}
  result_TSV = save_result_TSV;

  for (int i_g = 0; i_g < subgroup.size(); i_g++){
    for (int i_m = 1; i_m < osi->n_moments; i_m++){
      for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	if (i_m > 0 && !osi->switch_moment_TSV[i_s]){continue;}
	for (int i_r = 0; i_r < osi->n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < osi->n_scale_fact_TSV[i_s]; i_f++){
	    moment_result_TSV[i_m][i_g][i_s][i_r][i_f] = moment_result_TSV[i_m][i_g][i_s][i_r][i_f] / pow(log(result_TSV[1][i_g][i_s][i_r][i_f]) - log(0.218), i_m);
	  }
	}
      }
    }
  }

  result_TSV = moment_result_TSV;
  if (ygeneric->switch_output_result > 0){output_moment_TSV("g_N");}
  result_TSV = save_result_TSV;

  logger.finished();
}





// All functions below are only relevant for CV variation:

void summary_order::collect_result_CV(vector<string> & subgroup){
  static Logger logger("summary_order::collect_result_CV");
  logger.called();

  // extrapolated/qTcut-independent result
  result_CV.resize(osi->n_moments, vector<vector<double> > (subgroup.size(), vector<double> (osi->n_scales_CV, 0.)));
  deviation_CV.resize(osi->n_moments, vector<vector<double> > (subgroup.size(), vector<double> (osi->n_scales_CV, 0.)));
  // qTcut-dependent result
  result_qTcut_CV.resize(osi->n_moments, vector<vector<vector<double> > > (subgroup.size(), vector<vector<double> > (osi->n_qTcut, vector<double> (osi->n_scales_CV, 0.))));
  deviation_qTcut_CV.resize(osi->n_moments, vector<vector<vector<double> > > (subgroup.size(), vector<vector<double> > (osi->n_qTcut, vector<double> (osi->n_scales_CV, 0.))));

  logger << LOG_DEBUG_VERBOSE << "multiplicative combination (qTcut-dependent and qTcut-independent/extrapolated part) started" << endline_flush;

  // implementation with correct error estimates for multiplicative combinations for Tcut-independent/extrapolated part
  for (int i_m = 0; i_m < osi->n_moments; i_m++){
    for (int i_s = 0; i_s < osi->n_scales_CV; i_s++){
      for (int i_g = 0; i_g < subgroup.size(); i_g++){
	vector<double> combination_result(combination.size(), 0.);
	vector<double> combination_deviation2(combination.size(), 0.);
	for (int i_c = 0; i_c < combination.size(); i_c++){
	  for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
	    int i_l = combination[i_c][j_c];
	    combination_result[i_c] += xlist[i_l]->result_CV[i_m][i_g][i_s];
	    combination_deviation2[i_c] += pow(xlist[i_l]->deviation_CV[i_m][i_g][i_s], 2);
	  }
	}
	calculate_combination_result(combination_result, combination_deviation2, result_CV[i_m][i_g][i_s], deviation_CV[i_m][i_g][i_s]);
      }
    }
  }

  // implementation with correct error estimates for multiplicative combinations for qTcut-dependent part
  for (int i_m = 0; i_m < osi->n_moments; i_m++){
    for (int i_s = 0; i_s < osi->n_scales_CV; i_s++){
      for (int i_q = 0; i_q < osi->n_qTcut; i_q++){
	for (int i_g = 0; i_g < subgroup.size(); i_g++){
	  vector<double> combination_result(combination.size(), 0.);
	  vector<double> combination_deviation2(combination.size(), 0.);
	  for (int i_c = 0; i_c < combination.size(); i_c++){
	    for (int j_c = 0; j_c < combination[i_c].size(); j_c++){
	      int i_l = combination[i_c][j_c];
	      combination_result[i_c] += xlist[i_l]->xcontribution[0]->result_CV[i_m][i_g][i_q][i_s];
	      combination_deviation2[i_c] += pow(xlist[i_l]->xcontribution[0]->deviation_CV[i_m][i_g][i_q][i_s], 2);
	    }
	  }
	  calculate_combination_result(combination_result, combination_deviation2, result_qTcut_CV[i_m][i_g][i_q][i_s], deviation_qTcut_CV[i_m][i_g][i_q][i_s]);
	}
      }
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "multiplicative combination (qTcut-dependent and qTcut-independent/extrapolated part) finished" << endline_flush;


  if (ygeneric->switch_output_order){
    if (ygeneric->switch_output_overview > 0){output_result_overview_CV();}
    if (ygeneric->switch_output_result > 0){output_result_CV();}
    if (ygeneric->switch_output_plot > 0){output_result_plot_CV();}
    if (ygeneric->switch_output_plot > 0){output_result_plot_qTcut_CV();}
  }

  logger.finished();
}

