// "The two-loop helicity amplitudes for gg -> V1V2 -> 4 leptons"
// A. von Manteuffel, L. Tancredi
//
// ggvvamp package


#include <iostream>
#include <iomanip>
#include <cmath>
#include <cstdio>
#include <ginac/ginac.h>
#include "ggvvamp.h"
#include "ggvvampprec.h"

using namespace std;
using namespace GiNaC;

void usage(const char* progname) {
  cout << "Usage:" << endl;
  cout << progname << " --benchmark" << endl;
  cout << progname << " s t ma2 mb2" << endl;
}

int main(int argc, char* argv[]) {

  int Nf = 5;
  ggvvampprec amp(Nf, IRschemeQt);
  amp.maxrel = 1e-6;

// alternatively use fixed precision evaluation:
//  typedef quadtype T;
//  typedef cln::cl_F T;
//  typedef double T;
//  GiNaC::Digits = 32;
//  ggvvamp<T> amp(Nf, IRschemeQt);

  // read from command line
  double s, t, ma2, mb2;
  if (argc == 2 && strcmp(argv[1],"--benchmark") == 0) {
    // benchmark point from 1408.6409
    s = 15625.;
    t = -2384.89727896833;
    ma2 = 6467.215561;
    mb2 = 625.;
  } else if (argc != 1 + 4 ||
      (sscanf(argv[1], "%lf", &s) != 1) || (sscanf(argv[2], "%lf", &t) != 1) ||
      (sscanf(argv[3], "%lf", &ma2) != 1) || (sscanf(argv[4], "%lf", &mb2) != 1)) {
    usage(argv[0]);
    return 1;
  }
  
  // for checks on crossings
  //swap(ma2,mb2); // crossing x12x34
  //t = ma2 + mb2 - s - t; // crossing x12

  amp.compute(s,t,ma2,mb2);

  cout << setprecision(15);
  for (int l = 0 ; l < nloop; ++l) {
    for (int i = 0 ; i < nEhel; ++i) {
      for (int j = 1 ; j <= nEcoeff; ++j) {
        cout << l << " \t" << i << " \t" << j;
        cout << " \t" << amp.E[i][j][l][0];
        cout << " \t" << amp.E[i][j][l][1];
        cout << "\n";
      }
    }
  }
  cout << endl;
  
  return 0;  
}
