FivePointAmplitudes
----------------

`FivePointAmplitudes` is a c++ library for fast numerical evaluation of analytic expressions
for finite remainders of two-loop scattering amplitudes
for triphoton production at hadron colliders.
The library is meant to provide double-virtual contributions for NNLO QCD cross section computations.

The library implements analytic expressions for the triphoton production which were obtained in [publication]


- S. Abreu, B. Page, E. Pascual, V. Sotnikov,
_Leading-Color Two-Loop QCD Corrections for Three-Photon Production at Hadron Colliders_
  
If you use this library in your studies, please cite this publication.



**Developers:**

Vasily Sotnikov


Dependencies
----------------

The library relies on [PentagonFunctions++] for numerical evaluation of pentagon functions.
If [PentagonFunctions++] is not available in the system, it will be configured and installed automatically.

For quadruple precision evaluations the [qd] library is required.



Installation
----------------

The library uses [meson] build system with [ninja] back-end.
Both can be obtained through the standard python package manager *pip*:

```
pip install --user ninja meson
```


To build the library in directory **build** and install it in the standard configuration into **$HOME/local**, do

```
meson setup build -D prefix=$HOME/local
cd build
ninja install
```

This will install to `$HOME/local`.

To enable [qd] support, add `-D high-precision=true` at the setup stage.
Make sure that [qd] installation is visible in your terminal environment
(environmental variables `$LD_LIBRARY_PATH`, `$PKG_CONFIG_PATH`).



Tests
-------------

A number of tests is provided to verify that the library works as intended.
It is recommended to always run the tests after installation. This can be done with

```
meson test
```

from the build directory.



Usage and examples
------------

The intended way to use the library is to write a C++ program which links to it.
The library is shipped with a *pkg-config* file **FivePointAmplitudes.pc**,
so the recommended approach is to use *pkg-config* facilities of your favorite build system.
Make sure that the library installation is visible in your terminal environment
(environmental variables `$LD_LIBRARY_PATH`, `$PKG_CONFIG_PATH`).

The interface of the library is presented by example programs in directory **examples/**.
We show examples of evaluation of the helicity finite remainders in **examples/example_helicity.cpp** 
and the squared remainders in **examples/example_H2.cpp**.

The compiled example programs can be found in **build/examples/**.


------------------

This work is licensed under **GNU GPLv3**

------------------


[publication]: https://arxiv.org/abs/2010.15834
[PentagonFunctions++]: https://gitlab.com/pentagon-functions/PentagonFunctions-cpp
[meson]: https://mesonbuild.com/
[ninja]: https://ninja-build.org/
[qd]: http://crd-legacy.lbl.gov/~dhbailey/mpdist/
