#include <iostream>
#include <iomanip>
#include <complex>

#include "RemaindersList.h"

/**
 * Here we demonstrate how to use the library to evaluate helicity finite remainders individually.
 */

int main () {

    using namespace FivePointAmplitudes;

    // Perform evaluations in double precision in this program.
    // Can be switched to dd_real if the support of higher precision is compiled
    using T = double;

    // The construction phase is performed once

    remainder_2q3y_1l_pmmpp_12345<T> r1_mpp;     // R-++(1)
    remainder_2q3y_1l_pmppp_12345<T> r1_ppp;     // R+++(1)

    remainder_2q3y_2l_pmmpp_12345<T> r_mpp;     // R-++(2,0)
    remainder_2q3y_2l_pmppp_12345<T> r_ppp;     // R+++(2,0)
    remainder_2q3y_2l_pmmpp_nf_12345<T> r_mpp_nf;  // R-++(2,Nf)
    remainder_2q3y_2l_pmppp_nf_12345<T> r_ppp_nf;  // R+++(2,Nf)

    // Now evaluate

    // Define kinematics by the values of {s12, s23, s34, s45, s15}

    std::array<T, 5> kin = {
        1.322499999999999795718963468971204450660825281407116617285652827,   -0.9941094982857642509598316394306294510649507329718547518082950804,
        0.2644715912591248749171852420867085059089492634565178442467326276,  0.2671260487335744070776761035463255162438335749571773920325166998,
        -0.8837952297598047718249741030261595723499834269647052179121872201,
    };

    std::cout << std::scientific << std::setprecision(17);

    // Evaluate R-++(2,0)
    // The first argument is kinematics, the second argument is the sign of Im[tr5], and the last is the renormalization scale.
    // The scale can be set to anything.
    auto r1_mpp_val = r1_mpp.eval(kin,-1,1);
    auto r1_ppp_val = r1_ppp.eval(kin,-1,1);

    auto r_mpp_val = r_mpp.eval(kin,-1,1);
    auto r_ppp_val = r_ppp.eval(kin,-1,1);
    auto r_mpp_nf_val = r_mpp_nf.eval(kin,-1,1);
    auto r_ppp_nf_val = r_ppp_nf.eval(kin,-1,1);

    std::cout << "R-++(1) : " << r1_mpp_val << std::endl;
    std::cout << "R-++(2,0) : " << r_mpp_val << std::endl;
    std::cout << "R-++(2,Nf) : " << r_mpp_nf_val << std::endl;

    std::cout << "\n";

    std::cout << "R+++(1) : " << r1_ppp_val << std::endl;
    std::cout << "R+++(2,0) : " << r_ppp_val << std::endl;
    std::cout << "R+++(2,Nf) : " << r_ppp_nf_val << std::endl;

    return 0;
}
